local pl = require "pl.pretty"
local _, features = pcall(require, "features")
features = features or {}
require "version"
local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
client.InitSocket(nil, adapter)
local api = {}

local function GetProxy(objPath, service, loc, timeout, socket_addr)
	local proxy
	local cli

	cli = client.New( service, loc, socket_addr, timeout or 5 ) --5 seconds timeout.
	if not cli then
		print( "CO", "Fail to create connection to service " .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		print( "CO", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

local function GetHostIP()
    local f = io.popen("/sbin/ifconfig")
    local ip = ""
    if f then
        local line = f:read("l")
        while line do
            if line:match("eth0 ") or line:match("wlan0 ") or line:match("wlan1 ") or line:match("wwan0 ") or line:match("wwan1 ") then
                line = f:read("l")
                ip = line:match("inet addr:(.-)%s")
                if ip and ip:match("%d+%.%d+%.%d+%.%d+") then
                    break
                end
            end
            line = f:read("l")
        end
        f:close()
    end
    return ip
end

local function GetDeviceName()
	local filename = "/etc/hostname"
    local f = io.open( filename, "r" )
	if f then
		local r = f:read("*l")
		f:close()
        if r then
		    return r
        end
    end
    return ""
end

local fixStreamType = {
    ["RTSP/SERVER"] = "rtsp",
    ["NDI_server"] = "ndi_hx",
    ["Rtmp_push"] = "rtmp",
    ["RTMP/PUSH"] = "rtmp",
    -- ["NET/PUSH"] = "net",
    ["Rtp_push"] = "rtp",
    ["RTP/PUSH"] = "rtp",
    ["KMP/PULLING_SERVER"] = "kmppulling",
    -- ["KMP/PUSH"] = "kmp",
    ["SPROXY/PUSH"] = "sproxy",
    ["TS-UDP/PUSH"] = "ts-udp",
    ["SRT-TS/PUSH"] = "srt",
    ["HTTP/HLS"] = "hls",
    ["INTERACT/JUNSHI"] = "junshi",
    ["OMNI/PUSH"] = "omni",
    ["SIP/CLIENT"] = "sip"
}

local function GetAmbaStreamStatus(idr, service, i, ip)
    local proxy = GetProxy("/DefaultApp/" .. idr .. "/" .. service, "streamer")
    if not proxy then
        return
    end
    local is_ok, skip, status = proxy:GetStatus()
    if is_ok and skip and skip >=0 and status then
        status.enabled, status.serviceUri, status.serviceType = proxy._P("Enabled","ServiceUri", "ServiceType")
        proxy:Destroy()
        if tostring(status.serviceType):upper() == "NIL" then
            return
        end
        if ip then
            status.serviceUri = status.serviceUri:gsub("%%HOST%%", ip, 1)
        end
        status.key = service
        status.stream = i
        status.status = status["stream.status"] == "online" and "online" or "offline"
        status["stream.status"] = nil
        status.serviceType = fixStreamType[status.serviceType] or status.serviceType
        return status
    end
    proxy:Destroy()
end

local function getWebPort(info)
    local proxy = GetProxy("/", "web_daemon")
    if proxy then
        local is_ok, skip, data = proxy.Get({
            ["webConfig.port"] = 80,
            ["webConfig.https_port"] = 443,
        })

        if is_ok and skip then
            info.web_port = data["webConfig.port"] or 80
            info.https_port = data["webConfig.https_port"] or 443
        end
    end
end

local function GetAmbaNDIHXStatus()
    if not features.FEATURE_NDI_HX then
        return
    end

    local streamProxy = GetProxy("/DefaultApp", "streamer")
    if streamProxy then
        local is_ok, skip, res = streamProxy:Get({
            ["streamer.stream[0].NDI_server.device_name"] = "",
            ["streamer.stream[0].NDI_server.group_name"] = "",
            ["streamer.stream[0].NDI_server.connection"] = "",
            ["streamer.stream[0].NDI_server.mcast_prefix"] = "",
            ["streamer.stream[0].NDI_server.mcast_mask"] = "",
            ["streamer.stream[0].NDI_server.server_addr"] = "",
            ["streamer.stream[0].NDI_server.enabled"] = "",
        })
        streamProxy:Destroy()
        if is_ok and skip then
            return {
                enabled = (tonumber(res["streamer.stream[0].NDI_server.enabled"]) or 0) ~= 0,
                serviceUri = "",
                serviceType = "ndi-hx",
                stream = "ndi-hx",
                key = "ndi-hx",
                status = "online",
            }
        end
    end

end

local function GetDeviceStatus()
    local tbl = {}
    local proxy = GetProxy( "/SYS_STATUS", "web_daemon")
    if proxy then
        local is_ok, cpuLoad = proxy.GetCpuLoad()
        tbl.cpu = is_ok and cpuLoad or 0

        local is_ok, memFree, memTotal = proxy.GetMemUse()
        tbl.mem = is_ok and math.floor((1 - (memFree / memTotal)) * 100) or 0

        local is_ok, ptime = proxy.GetPersisTime()
        tbl.ptime = is_ok and ptime or 0

        local is_ok, diskinfo = proxy.GetDiskSpaces()
        tbl.sysuse = is_ok and diskinfo.sys.percent or 0

        proxy:Destroy()
    end
    return tbl
end

local function addEncoderResolution(capture, encoder)
    if not capture or not encoder then
        return
    end
    local resolution = ""

    if encoder.scale_orig ~= "default" then
        resolution = encoder.scale_orig
    else
        resolution = string.format("%sx%s@", capture.width, capture.height)
    end

    if encoder.framerate_mode == "defualt" then
        resolution = resolution .. capture.frame_rate .. "Hz"
    elseif encoder.framerate_mode == "half" then
        if capture.frame_rate >= 50 then
            resolution = resolution .. math.floor(capture.frame_rate / 2) .. "Hz"
        else
            resolution = resolution .. capture.frame_rate .. "Hz"
        end
    else
        resolution = resolution .. encoder.framerate .. "Hz"
    end
    encoder.resolution = resolution
end

local hardwareinfo_version
function api.getEncoderInfo(info)
    if not hardwareinfo_version then
        local proxy = GetProxy("/SERVER", "web_daemon")
        if proxy then
            local is_ok, skip, versioninfo = proxy.CheckHardwareInfo()
            if is_ok and skip then
                hardwareinfo_version = versioninfo and versioninfo.hardwareVersion or skip and skip.hardwareVersion
            end
            proxy:Destroy()
        end
    end
    info.has_pic = true
    info.device_mode = "encoder"
    info.device_type = _G.OEM_TYPE
    info.software_version = _G.SOFTWARE_VERSION
    info.firmware_version = _G.FIRMWARE_VERSION or "2.5"
    info.hardware_version = hardwareinfo_version or _G.HARDWARE_VERSION or "1.4"
    info.show_streamer_button = true
    info.device_name  = GetDeviceName()
    if _G.FEATURE_VIDEO_SDI then
        info.Signal = "SDI"
    elseif _G.FEATURE_VIDEO_HDMI and not _G.FEATURE_VIDEO_VGA then
        info.Signal = "HDMI"
    elseif _G.FEATURE_VIDEO_VGA then
        info.Signal = "HDMI"
    else
        info.Signal = "UNKNOW"
    end

    -- 序列号
    local klnkSerial
    local proxy = GetProxy("/SERVER", "web_daemon")
    if proxy then
        _, klnkSerial = proxy.GetSerialNo()
        proxy:Destroy()
    end

    info.capture = {}
    local afeProxy = GetProxy( "/AVIO/00/AFE", "codec" )
    if afeProxy then
        local r, x, res = afeProxy:Control( "detect-source", {source="any"} )
        afeProxy:Destroy()
        if r and x and x >= 0 then
            info.capture.video = info.capture.video or {}
            info.capture.audio = info.capture.audio or {}
            for k, v in pairs(res) do
                local t, key = k:match("^(.-)%.(.+)")
                info.capture[t][key] = v
            end
        end
    end

    local vfeProxy = GetProxy( "/AVIO/00/VFE", "codec" )
    if vfeProxy then
        local is_ok, skip, res = vfeProxy:Control( "detect-source", {source="any"} )
        vfeProxy:Destroy()
        if is_ok and skip and skip >=0 then
            info.capture.video = info.capture.video or {}
            info.capture.audio = info.capture.audio or {}
            for k, v in pairs(res) do
                local t, key = k:match("^(.-)%.(.+)")
                info.capture[t][key] = v
            end
            local as = info.capture.audio.source
            if as == "DIGITAL" or as == "digital" or as == "DEFAULT" or as == "default" then
                info.capture.audio.source = info.Signal
            elseif as == "MIC" or as == "mic" then
                info.capture.audio.source = "MIC"
            elseif as == "AES" or as == "aes" then
                info.capture.audio.source = "AES"
            else
                info.capture.audio.source = "Line-IN"
            end
            info.capture.audio.audio_source = info.capture.audio.source
            info.capture.video.video_source = info.Signal
        end
    end

    info.encoder = {}
    local ip = GetHostIP()
    for i = 0, 1 do
        local video_params = {
            ["encoder.stream[" .. i .. "].audio.bitrate"] = 64000,
            ["encoder.stream[" .. i .. "].audio.br_ctrl"] = "cbr",
            ["encoder.stream[" .. i .. "].audio.codec"] = "AAC",
            ["encoder.stream[" .. i .. "].video.avclevel"] = 40,
            ["encoder.stream[" .. i .. "].video.bitrate"] = 4000000.00,
            ["encoder.stream[" .. i .. "].video.br_ctrl"] = "cbr",
            ["encoder.stream[" .. i .. "].video.chrome"] = 0,
            ["encoder.stream[" .. i .. "].video.codec"] = "H264",
            ["encoder.stream[" .. i .. "].video.dynamic_bitrate"] = 0,
            ["encoder.stream[" .. i .. "].video.framerate"] = 0.00,
            ["encoder.stream[" .. i .. "].video.framerate_mode"] = "default",
            ["encoder.stream[" .. i .. "].video.gop"] = "IP",
            ["encoder.stream[" .. i .. "].video.gop_size"] = 60,
            ["encoder.stream[" .. i .. "].video.profile"] = "high",
            ["encoder.stream[" .. i .. "].video.qp_max"] = 51,
            ["encoder.stream[" .. i .. "].video.qp_min"] = 12.00,
            ["encoder.stream[" .. i .. "].video.quality"] = 80,
            ["encoder.stream[" .. i .. "].video.reduce_framerate"] = 0,
            ["encoder.stream[" .. i .. "].video.ref_frames"] = 1,
            ["encoder.stream[" .. i .. "].video.scale_height"] = 0,
            ["encoder.stream[" .. i .. "].video.scale_orig"] = "default",
        }

        local vencProxy = GetProxy("/Codec/Encoders", "codec")
        if vencProxy then
            local is_ok, skip, res = vencProxy:Get({
                ["encoder.encode_mode"] = "main"
            })
            if is_ok and skip then
                local s = res["encoder.encode_mode"]
                if s == "main" and i == 1 then
                    break
                end
            end

            info.encoder[i] = {}
            local is_ok, skip, res = vencProxy:Get(video_params)
            vencProxy:Destroy()
            if is_ok and skip then
                info.encoder[i].video = info.encoder[i].video or {}
                for kk, vv in pairs(res) do
                    local key = kk:match("encoder%.stream%[" .. i .. "%]%.video%.(.+)")
                    if key then
                        info.encoder[i].video[key] = vv
                    end
                end
            end
        end

        local aChannel = -1
        local streamer_params = {
            ["streamer.stream[" .. i .. "].audioChannel"] = -1
        }
        local streamProxy = GetProxy("/DefaultApp", "streamer")
        if streamProxy then
            local is_ok, skip, res = streamProxy:Get(streamer_params)
            streamProxy:Destroy()
            if is_ok and skip then
                aChannel = res["streamer.stream[" .. i .. "].audioChannel"] or -1
            end
        end

        if aChannel >= 0 then
            local audio_params = {
                ["audioEngine.encoders[1].enable"] = 1,
                ["audioEngine.encoders[1].encoder.aac.type"] = "AAC-LC",
                ["audioEngine.encoders[1].encoder.bitrate"] = 64000,
                ["audioEngine.encoders[1].encoder.channels"] = 2,
                ["audioEngine.encoders[1].encoder.codec"] = "AAC",
                ["audioEngine.encoders[1].encoder.g711.law"] = "ULAW",
                ["audioEngine.encoders[1].encoder.sampleSize"] = 160,
                ["audioEngine.encoders[1].encoder.sampling"] = 48000,
                ["audioEngine.encoders[1].source.device"] = "default",
                ["audioEngine.encoders[1].source.resample"] = "fastest",
                ["audioEngine.encoders[1].source.type"] = "ALSA",
            }
            local aencProxy = GetProxy("/AudioEngine", "streamer")
            if aencProxy then
                local is_ok, skip, res = aencProxy:Get(audio_params)
                aencProxy:Destroy()
                if is_ok and skip then
                    info.encoder[i].audio = {
                        {
                            codec = res["audioEngine.encoders[1].encoder.codec"],
                            sampling = res["audioEngine.encoders[1].encoder.sampling"],
                            channels = res["audioEngine.encoders[1].encoder.channels"],
                            bitrate = res["audioEngine.encoders[1].encoder.bitrate"],
                            g711_law = res["audioEngine.encoders[1].encoder.g711.law"]
                        }
                        -- enable = res["audioEngine.encoders[1].enable"] == 1,
                        -- aac_type = res["audioEngine.encoders[1].encoder.aac.type"],
                        -- law = res["audioEngine.encoders[1].encoder.g711.law"],
                        -- sampleSize = res["audioEngine.encoders[1].encoder.sampleSize"],
                        -- device = res["audioEngine.encoders[1].source.device"],
                        -- resample = res["audioEngine.encoders[1].source.resample"],
                        -- source_type = res["audioEngine.encoders[1].source.type"]
                    }
                end
            end
        else
            info.encoder[i].audio = {
                {
                    codec = "NONE",
                    sampling = 0,
                    channels = 0,
                    bitrate = 0
                }
            }
        end

        info.stream = info.stream or {}
        if i == 0 then
            local _fix = string.format("%02d", i)
            local status = GetAmbaStreamStatus(_fix, "Rtsp_server", i, ip)
            if status then
                table.insert(info.stream, status)
            end
            for idr = 0, 7 do
                local suffix = string.format("Dynamic_%02d", idr)
                local status = GetAmbaStreamStatus(_fix, suffix, i, ip)
                if status then
                    table.insert(info.stream, status)
                end
            end
        end
    end

    local srcProxy = GetProxy( "/DefaultApp/00/Source", "streamer")
    if srcProxy then
        local is_ok, skip, ret = srcProxy:GetStatus()
        if is_ok and skip and skip >= 0 and ret then
            if info.encoder and info.encoder[0] and info.encoder[0].video then
                info.encoder[0].video.real_bitrate = ret["video.bitrate"]
            end
        end
        srcProxy:Destroy()
    end

    local ndi_stream = GetAmbaNDIHXStatus()
    if ndi_stream then
        table.insert(info.stream, ndi_stream)
    end

    info.encoder[2] = info.encoder[1]
    info.encoder[1] = info.encoder[0]
    info.encoder[0] = nil

    addEncoderResolution(info.capture.video, info.encoder[1].video)

    info.device_status = GetDeviceStatus()
    info.disk_status = {}
    info.record_status = {}
    local recordProxy = GetProxy( "/", "record" )
    if recordProxy then
        local is_ok, status = recordProxy:GetStatus("main")
        if is_ok then
            info.record_status[1] = status
        end
        local is_ok, status = recordProxy:GetStatus("sub")
        if is_ok then
            info.record_status[2] = status
        end
        local is_ok, skip, disks = recordProxy:DiskList()
        if is_ok and skip and type(disks) == "table" then
            for k, v in pairs(disks) do
                local r, st, detail = recordProxy:DiskDetail(k)
                info.disk_status[k] = detail
            end
        end
        recordProxy:Destroy()
    end

    info.serial_number = klnkSerial
    info.bonding_status = api.GetBondingStatus()
    -- getWebPort(info)
    return info, klnkSerial
end

function api.Reboot()
    local proxy = GetProxy( "/SYS_STATUS", "web_daemon")
	if proxy then
		proxy.Reboot()
		proxy:Destroy()
    end
end

function api.updatePicture()
    local proxy = GetProxy("/DefaultApp/02", "streamer")
    if not proxy then
        return
    end
    proxy:Control( "snap", {
		["path"] = "/tmp",
		["file"] = "WEB-SNAP",
		["suffix"] = ".jpg",
		["count"] = 1
	} )
    proxy:Destroy()
    local path = "/tmp/WEB-SNAP.jpg"
    local f = io.open(path, "rb")
    if f then
        local jpg = f:read("*a")
        f:close()
        return jpg
    end
end

function api.Reset()
    local proxy = GetProxy( "/SYS_STATUS", "web_daemon", nil, 5 )
    if proxy then
        proxy.Reset()
        proxy:Destroy()
    end
end

function api.ControlStream(cfg)
    local key = cfg.key
    local stream = cfg.stream
    local enabled = cfg.enabled

    if key == "ndi-hx" and stream == "ndi-hx" then
        if features.FEATURE_NDI_HX then
            local streamProxy = GetProxy("/DefaultApp", "streamer")
            if streamProxy then
                local is_ok, skip, res = streamProxy:Set({
                    ["streamer.stream[0].NDI_server.enabled"] = enabled == true and 1 or 0
                })
                streamProxy:Destroy()
            end
        end
        return
    end

    local _fix = string.format("%02d", stream)
    local proxy = GetProxy("/DefaultApp/" .. _fix .. "/" .. key, "streamer")
    if proxy then
        proxy:Enable(enabled == true)
        proxy:Destroy()
    end
end

function api.ControlAllStream(cfg)
    local enable = cfg.enabled == true
    for _, idr in ipairs{"00", "01"} do
        for idx = 0, 7 do
            local service = string.format("Dynamic_%02d", idx)
            local proxy = GetProxy("/DefaultApp/" .. idr .. "/" .. service, "streamer")
            if proxy then
                local x, skip, err = proxy:Enable(enable == true)
                proxy:Destroy()
            end
        end
    end

    if features.FEATURE_NDI_HX then
        local streamProxy = GetProxy("/DefaultApp", "streamer")
        if streamProxy then
            local is_ok, skip, res = streamProxy:Set({
                ["streamer.stream[0].NDI_server.enabled"] = enable == true and 1 or 0
            })
            streamProxy:Destroy()
        end
    end
end

function api.UpgradeDevice(file)
    local proxy = GetProxy("/UPGRADE", "web_daemon", nil, 60)
    if not proxy then
        print("Can't Get the UPGRADE proxy !!!")
        return
    end
    local is_ok, skip = proxy.ProcessUpgradeFirmware(file)
    proxy:Destroy()
    return is_ok and skip
end

function api.VideoSet(map)
    --TODO :  map 解析
    local tbl = {
        ["encoder.stream[0].video.scale"] = map.scale_orig or nil,--"1920x1080"
        ["encoder.stream[0].video.bitrate"] = map.bitrate or nil,
        ["encoder.stream[0].video.framerate"] = tonumber(map.framerate) or nil,
        ["encoder.stream[0].video.framerate_mode"] = map.framerate_mode, -- half(减半), custom(自定义), default(与源一致)
    }

    local vencProxy = GetProxy("/Codec/Encoders", "codec")
    if vencProxy then
        local is_ok, skip = vencProxy:Set(tbl)
        vencProxy:Destroy()
    end
end

function api.AudioSet(map)
    local tbl = {
        ["audioEngine.encoders[1].encoder.codec"] = map.codec or nil,
        ["audioEngine.encoders[1].encoder.channels"] = map.channels or nil,
        ["audioEngine.encoders[1].encoder.sampling"] = map.sampling or nil,
        ["audioEngine.encoders[1].encoder.g711.law"] = map.g711_law or nil
    }
    if next(tbl) then
        local aencProxy = GetProxy("/AudioEngine", "streamer")
        if aencProxy then
            local is_ok, skip, res = aencProxy:Set(tbl)
            aencProxy:Destroy()
        end
        api.Reset()
    end

    if map.audio_source or map.source then
        local asourceProxy = GetProxy("/AVIO", "codec")
        if asourceProxy then
            local source = map.audio_source or map.source
            if source ~= "HDMI" and source ~= "SDI" then
                source = "LINE"
            end
            asourceProxy:Set({
                ["avio.audio.source"] = source
            })
            asourceProxy:Destroy()
        end
    end
end

local authcode = "kiloview"
function api.AddStreamToService(map, ifnames)
    local ip, port = map.ip, map.port
    local polyProxy = GetProxy("/klnk", "streamer")
    if not polyProxy then
        return
    end
    local is_ok, old = polyProxy.GetNetLink()
    if is_ok and old then
        local new_devs = {}
        for i, v in ipairs(ifnames) do
            table.insert(new_devs, { id = i, ifname = v})
        end
        local is_ok, skip = polyProxy.SetNetLink(ip, port, authcode, old.timeout, new_devs, 1)
    end
    polyProxy:Destroy()
end

function api.DeviceDelStreamService()
    local polyProxy = GetProxy("/klnk", "streamer")
    if not polyProxy then
        return
    end
    local is_ok, old = polyProxy.GetNetLink()
    if is_ok and old then
        local is_ok, skip = polyProxy.SetNetLink(ip, port, authcode, old.timeout, old.polydevs, 0)
    end
    polyProxy:Destroy()
end

function api.GetBondingStatus()
    local polyProxy = GetProxy("/klnk", "streamer")
    if not polyProxy then
        return
    end
    local is_ok, ret = polyProxy.GetBondingStatus()
    polyProxy:Destroy()
    if is_ok and ret then
        local result = {}
        for _, v in ipairs(ret.links or {}) do
            table.insert(result, {name = v.link, sendrate = v.sendrate, recvrate = v.recvrate})
        end
        return result
    end
end

return api