
local pl = require "pl.pretty"
require "version"
local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
client.InitSocket(nil, adapter)
local api = {
	device_type = "",
	driving_mode = "",
	network = {}
}

local function isN6Device()
	if not api.device_type then
		return
	end
	return api.device_type:match("^N6") or api.device_type:match("^N5")
end

local function GetProxy(objPath, service, loc, timeout, socket_addr)
	local proxy
	local cli

	cli = client.New( service, loc, socket_addr, timeout or 5 ) --5 seconds timeout.
	if not cli then
		print( "CO", "Fail to create connection to service " .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		print( "CO", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

local function GetDrivingMode()
    local proxy = GetProxy( "/", "codec" )
	if proxy then
		local r, mode = proxy.GetMode()
        proxy:Destroy()
		if r and mode then
			if mode == "decoding" or mode == "decoder" then
				api.driving_mode = "decoder"
			else
				api.driving_mode = "encoder"
			end
            return api.driving_mode
		end
	end
    return ""
end

local function SerialNumber()
	local serial = "000000000"
    local f = io.open( "/boot/SERIAL_NO", "r" )
    if f then
        serial = f:read("*l")
        f:close()
	else
		local t = io.open( "/etc/SERIAL_NO", "r" )
		if t then
			serial = t:read("*l")
			t:close()
		end
    end
    return serial
	--for test
	-- return "20220333"
end

local function GetDeviceName()
	local filename = "/etc/hostname"
    local f = io.open( filename, "r" )
	if f then
		local r = f:read("*l")
		f:close()
        if r then
		    return r
        end
    end
    return ""
end

function api.Init( device_type )
	api.device_type = device_type
end

function api.SerialNumber()
    return SerialNumber()
end

function api.DeviceName()
    return GetDeviceName()
end

function api.Reboot()
	local proxy = GetProxy( "/", "systemctrl" )
	if proxy then
		proxy.Reboot(2)
		proxy:Destroy()
	end
end

function api.Reset()
	local proxy = GetProxy( "/", "systemctrl" )
	if proxy then
		proxy.Reset()
		proxy:Destroy()
	end
end

function api.GetBasicInformation(info)
    info.device_name  = GetDeviceName()
    info.device_mode  = "IO"
	info.device_type  = api.device_type
    info.driving_mode = GetDrivingMode()
	info.serial_number    = SerialNumber()
	info.software_version = _G.SOFTWARE_VERSION
	info.hardware_version = _G.HARDWARE_VERSION
	info.web_index = info.driving_mode
	if isN6Device() then
		info.web_index = "dashboard"
	end
    return true, info
end

local function RequestEncoderInfo(proxy, types)
	local r, status = proxy.GetStatus(types)
	local audio_meta = {}
	local video_meta = {}
	if r and status then
		local audio_signal
		if not status.audio_signal then
			audio_signal = "none"
		elseif status.audio_signal == "hdmi" or status.audio_signal == "sdi" or status.audio_signal == "digital" or status.audio_signal == "embedded" then
			if api.device_type == "N3" or api.device_type == "N30" or api.device_type == "N3-s" then
				audio_signal = "SDI"
			else
				audio_signal = "HDMI"
			end
		elseif status.audio_signal == "line" or status.audio_signal == "linein" or status.audio_signal == "line-in" or status.audio_signal == "mic" or status.audio_signal == "micin" or status.audio_signal == "mic-in" or status.audio_signal == "analog" then
			audio_signal = "analog"
		else
			audio_signal = status.audio_signal
		end
		local audio_chn = tonumber(status.channels) or 2
		local audio_channel_names = {
			[0] = "Auto",
			[1] = "Mono",
			[2] = "Stereo"
		}
		
		audio_meta.audio_codec = status.audio_format or "PCM"
		audio_meta.audio_format   = (status.sampling or 48000) .. "Hz/" .. (audio_channel_names[audio_chn] or ("" .. audio_chn .. " CH"))
		audio_meta.audio_channels = audio_chn
		-- if isN6Device() then
		-- 	audio_meta.audio_channels = 0
		-- end
		audio_meta.audio_sampling = status.sampling
		audio_meta.audio_signal   = audio_signal
		if status.video_signal then
			if api.device_type == "N3" or api.device_type == "N30" or api.device_type == "N3-s" then
				video_meta.video_signal = "SDI"
			else
				video_meta.video_signal = "HDMI"
			end
			video_meta.video_encode = types == "ndifull" and "NDI-FULL" or "NDI-HX"
		else
			video_meta.video_signal = "NULL"
		end
		video_meta.interlaced     = status.interlaced
		video_meta.resolution = status.resolution
		video_meta.frame_rate = status.frame_rate
		video_meta.bitrate    = status.bitrate
		video_meta.xRes = status.xRes
		video_meta.yRes = status.yRes
	end
	return audio_meta, video_meta
end

local function RequestGroupChannelAndEncQuality(proxy, types)
	local r1, group, channel = proxy.GetNameWithGroup(types)
	local r2, quality = proxy.GetEncQuality(types)
	local r3, ndi_conn, mcast_pfx, mcast_mask, ttl = proxy.GetNDIConnection(types)
	return group, channel, quality, ndi_conn, mcast_pfx, mcast_mask, ttl
end

function api.GetEncoderInfo(info)
	--判断运行模式
	if api.driving_mode == "" or api.driving_mode == "decoder" then 
		return false, info 
	end

	local proxy = GetProxy( "/", "codec" )
	if proxy then
		-- audio_channels: 2
		-- audio_format: "48000Hz/Stereo"
		-- audio_sampling: 48000
		-- audio_signal: "embedded"
		-- bitrate: 130329
		-- frame_rate: 59.94
		-- interlaced: false
		-- resolution: "1920x1080@59.94 Hz"
		-- video_signal: true
		-- xRes: 1920
		-- yRes: 1080
		info.audio = {}
		info.video = {}

		local r, gain = proxy.GetAudioGain()
		
		local audio_meta, video_meta = RequestEncoderInfo(proxy, "ndifull")
		local group, channel, quality, ndi_conn, mcast_pfx, mcast_mask, ttl = RequestGroupChannelAndEncQuality(proxy, "ndifull")
		video_meta.group_name = group
		video_meta.channel_name = channel
		video_meta.level = quality >= 200 and "Superior" or quality >= 150 and "High" or quality >= 100 and "Default" or quality >= 75 and "Low" or "Poor"
		video_meta.ndi_connection = ndi_conn or "tcp"
		video_meta.netprefix = mcast_pfx or "239.255.0.0"
		video_meta.netmask = mcast_mask or "255.255.0.0"
		video_meta.ttl = tonumber(ttl) or 127
		audio_meta.volume = gain or 100
		-- video_meta.quality = quality
		table.insert(info.audio, audio_meta)
		table.insert(info.video, video_meta)
		if isN6Device() then
			local _, source_meta = RequestEncoderInfo(proxy)
			info.video[1].resolution = source_meta.resolution
			info.video[1].xRes = source_meta.xRes
			info.video[1].yRes = source_meta.yRes
			audio_meta, video_meta = RequestEncoderInfo(proxy, "ndihx")
			group, channel, quality, ndi_conn, mcast_pfx, mcast_mask, ttl = RequestGroupChannelAndEncQuality(proxy, "ndihx")
			video_meta.group_name = group
			video_meta.channel_name = channel
			video_meta.level = quality >= 20 and "High Quality" or quality >= 14 and "High" or quality >= 8 and "Default" or quality >= 2 and "Low" or "Poor"
			-- video_meta.set_bitrate = quality --* 1000000
			video_meta.ndi_connection = ndi_conn or "tcp"
			video_meta.netprefix = mcast_pfx or "239.255.0.0"
			video_meta.netmask = mcast_mask or "255.255.0.0"
			video_meta.ttl = tonumber(ttl) or 127
			audio_meta.volume = gain or 100
			table.insert(info.audio, audio_meta)
			table.insert(info.video, video_meta)
		end
		proxy:Destroy()
	end
	return false, info 
end

function api.GetCaptureInfo(info)
	return false, info
end

function api.GetDecoderInfo(info)
	if api.driving_mode == "" or api.driving_mode == "encoder" then 
		return false, info 
	end
	local proxy = GetProxy( "/", "codec" )
	if proxy then
		local r, status = proxy.GetStatus()
		local r2, more = proxy.GetPresets()
		local r3, current = proxy.GetCurrentSettings()
		proxy:Destroy()
		-- audio: "48.0KHz / 2CH"
		-- audio_channels: 2
		-- audio_sampling: 48000
		-- bitrate: 132375
		-- codec: "SHQ2"
		-- deInterlace: false
		-- frame_rate: 60.03
		-- interlaced: false
		-- ip: "192.168.35.57"
		-- name: "DESKTOP-0ERC3PJ (Intel UHD Graphics 630 2)"
		-- online: true
		-- ptz: 1
		-- resolution: "1920x1080p 60Hz"
		-- url: "192.168.35.57:6041"
		-- video_mode: "hdmi"
		-- warning: ""
		-- web: ""
		-- xRes: 1920
		-- yRes: 1080
		if r and status then
			info.audio = {}
			info.video = {}
			info.audio.audio 		  = status.audio
			info.audio.audio_channels = status.audio_channels
			info.audio.audio_sampling = status.audio_sampling
			if api.device_type == "N3" or api.device_type == "N30" or api.device_type == "N3-s" then
				info.audio.audio_mode = "SDI"
			else
				info.audio.audio_mode = "HDMI"
			end
			info.bitrate = status.bitrate
			
			info.video.deInterlace = status.deInterlace
			info.video.frame_rate  = status.frame_rate
			info.video.resolution = status.resolution
			info.video.video_mode = status.video_mode
			info.video.xRes = status.xRes
			info.video.yRes = status.yRes
			info.video.bitrate = status.bitrate

			if status.codec:match("^H") then
				info.codec   = "NDI-HX"
			else
				info.codec   = "NDI-FULL"
			end
			info.interlaced  = status.interlaced
			info.ip 	= status.ip
			info.name 	= status.name
			info.online = status.online
			info.ptz 	= status.ptz
			info.url 		= status.url
			info.warning 	= status.warning
			info.web  = status.web

			if r2 and more then
				if not r3 or type(current) ~= "table" then
					current = nil
				end
				local preset = {}
				for id = 1, 9 do
					local fnd
					for k = 1, #more do
						if more[k].id == id or more[k].id == tostring(id) then
							fnd = more[k]
							break
						end
					end
					if not fnd then
						fnd = {
							id = tostring(id), name = "", group = "", group_name = "", device_name = "",
							channel_name = "", ip = "", url = "", online = "", enable = 0
						}
					else
						fnd.id = tostring(id)
						fnd.group_name = fnd.group
						fnd.enable = fnd.name and fnd.name ~= "" and 1 or 0
						fnd.online = fnd.online and "on" or "off"
					end
					if current and current.preset_id == tostring(id) then
						fnd.current = true
					end
					table.insert(preset, fnd)
				end
				table.insert(preset, {
					id = "10",
					current = current and current.preset_id == "0",
					BlankColor = blank_color,
				})
				info.preset = preset
			end
			
			return true, info
		end
	end
	return false, info
end

function api.MediaStream( info )
	return false, info
end

function api.GetSnapshoot()
	return false, ""
end

function api.UpgradeDevice(file)
    local proxy = GetProxy("/", "systemctrl", "local", 30)
    if not proxy then
        print("Can't Get the systemctrl proxy !!!")
        return
    end

    local is_ok, skip = proxy.ProcessFirmware(file)
    proxy:Destroy()
    return is_ok and skip
end

local cpu_saved = {}
local function GetCpuLoad()
	local file = io.open( "/proc/stat", "r" )
	if not file then
		return 0, 99	--load: 0, idle: 99
	end

	local line = file:read("l")
	file:close()

	local cpu,u,n,s,i,w,x,y = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)" )
	u = u and tonumber(u) or 0
	n = n and tonumber(n) or 0
	s = s and tonumber(s) or 0
	i = i and tonumber(i) or 0
	w = w and tonumber(w) or 0
	x = x and tonumber(x) or 0
	y = y and tonumber(y) or 0

	local du = u - (cpu_saved.u or 0)
	local dn = n - (cpu_saved.n or 0)
	local ds = s - (cpu_saved.s or 0)
	local di = i - (cpu_saved.i or 0)
	local dw = w - (cpu_saved.w or 0)
	local dx = x - (cpu_saved.x or 0)
	local dy = y - (cpu_saved.y or 0)

	cpu_saved.u = u
	cpu_saved.n = n
	cpu_saved.s = s
	cpu_saved.i = i
	cpu_saved.w = w
	cpu_saved.x = x
	cpu_saved.y = y

	local total = du + dn + ds + di + dw + dx + dy
	if total < 1 then
		total = 1
	end

	local scale = 100/total
	local idle = di*scale
	return 100-idle, idle
end

local function GetMemUse()
	local file = io.open( "/proc/meminfo", "r" )
	if not file then
		return nil, nil
	end

	local free = nil
	local total = nil
	local v
	local line = file:read("l")
	while line and ( not free or not total ) do
		v = line:match( "MemTotal:%s+(%d+)" )
		if v then
			total = tonumber(v)
		else
			v = line:match( "MemFree:%s+(%d+)" )
			if v then
				free = tonumber(v)
			end
		end
		line = file:read("l")
	end

	file:close()
	return free, total
end

local function GetPersisTime()
	local file=io.open("/proc/uptime", "r")
	if not file then
		return 0, {year=1970,month=1,day=1,hour=8,min=0,sec=0}
	end

	local line = file:read("l")
	file:close()
	local pers = line:match("(%d+)")
	if not pers then
		return 0, {year=1970,month=1,day=1,hour=8,min=0,sec=0}
	else
		return tonumber(pers), os.date("*t",os.time() - tonumber(pers))
	end
end

local function GetDiskSpaces()
	local result = {
		sys = { total=0, used=0, free=0, percent=100 },
		data = { total=0, used=0, free=0, percent=100 }
	}

	local file=io.popen("/bin/df -m","r")
	if not file then
		return result
	end

	local line = file:read("l")
	while line and (result.sys.total <= 0 or result.data.total <=0 ) do
		local flag,t,u,f,p = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%%%s+/$" )
		if flag then
			result.sys.total = tonumber(t)
			result.sys.used = tonumber(u)
			result.sys.free = tonumber(f)
			result.sys.percent = tonumber(p)
		else
			flag,t,u,f,p = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%%%s+/data$" )
			if flag then
				result.data.total = tonumber(t)
				result.data.used = tonumber(u)
				result.data.free = tonumber(f)
				result.data.percent = tonumber(p)
			end
		end

		line=file:read("l")
	end

	file:close()
	return result
end

function api.GetDeviceStatus()
	local memFree, memTotal = GetMemUse()
	return {
		cpu = GetCpuLoad(),
		mem = math.floor((1 - (memFree / memTotal)) * 100) or 0, 
		ptime = GetPersisTime(), 
		sysuse = GetDiskSpaces().sys.percent,
	}
end

function api.VideoSet(map)

end

function api.AudioSet(map)
    local proxy = GetProxy("/", "codec")
	if proxy then
		local source = map.audio_signal or map.source
		if source then
			if source ~= "HDMI" and source ~= "SDI" then
				source = "linein"
			end
			proxy.SelectAudioSource(map.audio_signal or map.source)
		end
		if map.volume then
			proxy.SetAudioGain(tonumber(map.volume) or 100)
		end
		if map.channels then
			proxy.SetAudioChannel(tonumber(map.channels) or -1, {})
		end
		proxy:Destroy()
	end
end

function api.NDISet(map)
	if not map.types or (map.types ~= "ndifull" and map.types ~= "ndihx") then
		print("Can not do NDISet, no map.types or map.types is invalid. map.types = ", map.types)
		return
	end

	if not (map.ndi_connection and map.netprefix and map.netmask and map.ttl) then
		print("Can not do NDISet, args is invalid ", map.ndi_connection, map.netprefix, map.netmask, map.ttl)
		return
	end
	local proxy = GetProxy("/", "codec")
	if proxy then
		if isN6Device() then
			proxy.SetNDIConnection(map.types, map.ndi_connection, map.netprefix, map.netmask, map.ttl)
		else
			proxy.SetNDIConnection(map.ndi_connection, map.netprefix, map.netmask, map.ttl)
		end
		proxy:Destroy()
	end
end

function api.OutputSet(map)
	if not map.id then
		print("Can not set output id, not have map.id.")
		return
	end

	local proxy = GetProxy("/", "codec")
	if proxy then
		local id = tostring(map.id)
		if id == "0" or id == "10" then
			proxy.SelectBlank()
		else
			proxy.SelectPreset(id)
		end
		proxy:Destroy()
	end
end

return api