local pl = require "pl.pretty"
require "version"
local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
client.InitSocket(nil, adapter)
local api = {}

local function GetProxy(objPath, service, loc, timeout, socket_addr)
	local proxy
	local cli

	cli = client.New( service, loc, socket_addr, timeout or 5 ) --5 seconds timeout.
	if not cli then
		_dbg.Error( "CO", "Fail to create connection to service " .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		_dbg.Error( "CO", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

local function GetDeviceName()
	local filename = "/etc/hostname"
    local f = io.open( filename, "r" )
	if f then
		local r = f:read("*l")
		f:close()
        if r then
		    return r
        end
    end
    return ""
end

local function GetHostIP()
    local f = io.popen("/sbin/ifconfig")
    local ip = ""
    if f then
        local line = f:read("l")
        while line do
            if line:match("eth0 ") or line:match("wlan0 ") or line:match("wlan1 ") or line:match("wwan0 ") or line:match("wwan1 ") then
                line = f:read("l")
                ip = line:match("inet addr:(.-)%s")
                if ip and ip:match("%d+%.%d+%.%d+%.%d+") then
                    break
                end
            end
            line = f:read("l")
        end
    end
    f:close()
    return ip
end

local function fixS2EncKeyToServer(tbl)
    local width, height = tbl.ve.u32PicWidth, tbl.ve.u32PicHeight
    if width == 1 and height == 1 then
        width = tbl.capSize[1]
        height = tbl.capSize[2]
    end
    return {
        bitrate = (tbl.rc.mode == "CBR" and tbl.rc.u32BitRate or tbl.rc.u32MaxBitRate) * 1000,
        qp_max = tbl.rc.u32MaxQp,
        dynamic_bitrate = (tbl.rc.u32MaxBitRate) * 1000,
        codec = tbl.type,
        gop_size = tbl.rc.u32Gop,
        --avclevel = 40,
        --gop = "IP",
        profile = tbl.ve.u32Profile,
        --reduce_framerate = 0,
        framerate = tbl.rc.u32SrcFrmRate,
        --quality = 80,
        br_ctrl = tbl.rc.mode,
        chrome = tbl.grey and 1 or 0,
        qp_min = tbl.rc.u32MinQp,
        --scale_height = 0,
        --ref_frames = 1,
        scale_orig = width .. "x" .. height
    }
end

local function fixS2StreamKeyToServer(tbl, ip, stream)
    local new = {}
    for i, v in ipairs(tbl) do
        table.insert(new, {
            serviceUri = ip and v.addressUrl:gsub("%%HOST%%", ip, 1) or v.addressUrl,
            serviceType = v.type,
            enabled = v.enable,
            status = v.status and v.status == "start" and "online" or "offline",
            key = i,
            stream = stream - 1,  -- 这里是为了与安霸保持一致
            stream_name = v.name,
            bind_audio = v.bindAudio,
        })
    end
    return new
end

function api.getEncoderInfo(info, features)
    info.Signal = "HDMI"
    info.device_mode = "encoder"
    info.device_type = features.PRODUCT_ID
    info.software_version = _G.SOFTWARE_VERSION
    info.hardware_version = _G.HARDWARE_VERSION
    info.device_name  = GetDeviceName()

    local klnkSerial
    local proxy = GetProxy("/", "systemctrl")
    if proxy then
        local is_ok, sn = proxy.GetSerialNo()
        proxy:Destroy()
        if is_ok then
            klnkSerial = sn or "00000000"
        end
    end

    info.encoder = {
        [1] = {}, [2] = {}
    }
    info.capture = {}
    local codecProxy = GetProxy("/codec", "codec")
    if codecProxy then
        local is_ok, res = codecProxy.GetInfo()
        codecProxy:Destroy()
        if is_ok then
            info.encoder[1].video = fixS2EncKeyToServer(res.main)
            info.encoder[2].video = fixS2EncKeyToServer(res.sub)
        end
    end

    local audioProxy = GetProxy("/audio", "codec")
    if audioProxy then
        local is_ok, res = audioProxy.GetInfo()
        audioProxy:Destroy()
        if is_ok then
            info.capture.audio = {
                source = res.source.audioSource,
                audio_source = res.source.audioSource,
                locked = 0,
                channels = res.source.u32ChnCnt,
                sampling = res.source.enSamplerate,
                sample_width = res.source.enBitwidth,
                sample_size = 0
            }

            info.encoder[1].audio = {}
            for _, v in ipairs(res.audioEncode) do
                table.insert(info.encoder[1].audio, {
                    bitrate = v.bitrate,
                    channels = v.channels,
                    sampling = v.sampling,
                    codec = v.aacType
                })
            end

            info.encoder[2].audio = info.encoder[1].audio
        end
    end

    local captureProxy = GetProxy("/capture", "codec")
    if captureProxy then
        local is_ok, res = captureProxy.GetInfo(1)

        if is_ok then
            info.capture.video = {
                -- format_code = ,
                format = res.sourceSize .. "@" .. res.sourceFrameRate .. "Hz",
                frame_rate = res.sourceFrameRate,
                source = res.videoSource,
                video_source= res.videoSource,
                -- height = res.stDestSize[2],
                -- width = res.stDestSize[1],
                locked = 1,
                -- standard = ,
                -- interlaced = ,
                -- source_detected = ,
                -- format_name = ,
                -- format_seq = ,
                -- freerun = ,
            }
        end
    end

    info.stream = {}
    local streamProxy = GetProxy("/streamer", "codec")
    if streamProxy then
        local is_ok, res = streamProxy.GetInfo()
        streamProxy:Destroy()
        if is_ok then
            local ip = GetHostIP()
            for _, v in ipairs(fixS2StreamKeyToServer(res.main, ip, 1)) do
                table.insert(info.stream, v)
            end
            for _, v in ipairs(fixS2StreamKeyToServer(res.sub, ip, 2)) do
                table.insert(info.stream, v)
            end

            -- info.stream[1] = fixS2StreamKeyToServer(res.main, ip, 1)
            -- info.stream[2] = fixS2StreamKeyToServer(res.sub, ip, 2)
        end
    end
    info.serial_number = klnkSerial
    info.web_index = "dashboard"
    return info, klnkSerial
end

function api.Reboot()
    local proxy = GetProxy( "/", "systemctrl" )
	if proxy then
		proxy.Reboot(2)
		proxy:Destroy()
	end
end

function api.Reset()
    -- 暂无此方法
end

function api.ControlStream(cfg)
    local key = cfg.key
    local stream = cfg.stream + 1 -- 这里是为了修复上面的 - 1
    local enabled = cfg.enabled

    local proxy = GetProxy("/streamer", "codec")
    if proxy then
        proxy.UpdateStreamer(stream == 1 and "main" or "sub", key, {
            enable = enabled == true
        })
        proxy:Destroy()
    end
end

function api.ControlAllStream(cfg)
    local enable = cfg.enabled == true
    local proxy = GetProxy("/streamer", "codec")
    if proxy then
        local is_ok, info = proxy.GetInfo()
        if is_ok then
            for stream, obj in pairs(info) do
                for i in ipairs(obj) do
                    proxy.UpdateStreamer(stream, i, {
                        enable = enable
                    })
                end
            end
        end
        proxy:Destroy()
    end
end

function api.UpgradeDevice(file)
    local proxy = GetProxy("/", "systemctrl", "local", 30)
    if not proxy then
        print("Can't Get the systemctrl proxy !!!")
        return
    end

    local is_ok, skip = proxy.ProcessFirmware(file)
    proxy:Destroy()
    return is_ok and skip
end

return api