-- KILOVIEW M2 products Audio/Video front end.
local ADV7441A = require("chips.adv7441a")
local periphery = require( "periphery" )
local hwChk = require( "kiloview.M2.hardwareCheck" )
local vga_formats = require("vga_formats")

----------------------------------------------------------------

local _M = {
	deviceType = "M2",
	deviceName = "HDMI/VGA H.264 ENCODER",
	vendor = "KILOVIEW Electronics Co., Ltd.",
	chip = nil,
	currentVideoSrc = "UNKNOWN",
	currentAudioSrc = "DIGITAL",
	audSelIO = nil,
	autoAudioSrc = true,
	forceReselectAudio = false,
	lastVideoFormatVersion = 0,
	lastVideoFormat = {},
	lastAudioFormatVersion = 0,
	lastAudioFormat = {},
	analogAudioFormat = {
		_valid = true,
		audio_src = "LINE",
		sampling = 48000, --TODO: How can dynamic support?
		audio_signal_type = 3, --AKA analog
		channels = 2,
		bit_width = 16
	},
}

----------------------------------------------------------------
local function AlsaSelect( src )
	if src == "MIC" then
		os.execute( [[/usr/bin/amixer cset name='Capture Source' 1]] )
		--os.execute( [[/usr/bin/amixer cset name='Line Input Switch' off]] )
		--os.execute( [[/usr/bin/amixer cset name='Mic Input Switch' on]] )
	else
		os.execute( [[/usr/bin/amixer cset name='Capture Source' 0]] )
		--os.execute( [[/usr/bin/amixer cset name='Line Input Switch' on]] )
		--os.execute( [[/usr/bin/amixer cset name='Mic Input Switch' off]] )
	end

end
----------------------------------------------------------------
local function CompareFormatInfo( a, b )
	if a == b then
		return false
	end
	
	local k,v
	for k,v in pairs(a) do
		if v ~= b[k] then
			return true
		end
	end
	
	for k,v in pairs(b) do
		if v ~= a[k] then
			return true
		end
	end
	
	return false
end
----------------------------------------------------------------
function _M:Init()
	--1st, Check the hardware version, if >=2.0, that uses a special audio driver.
	hwChk.Check()
	self.audSelIO = periphery.GPIO( 13, "out" )
	if not self.audSelIO then
		print( "WARNING: Could not open GPIO13 controller for audio digital/analog selection!" )
	else
		self.audSelIO:write( false )	-- 0: Digital(Default); 1: Analog
	end
	
	self.chip = ADV7441A( { "HDMI", "VGA" }, "/dev/i2c-0", 0x42, "AUTO", 6 --[[HPD]] )
	if not self.chip then
		print( "ERROR: Initialize M2 AVFE: Could not create ADV7441A controller!" )
		return false
	end
	self.chip:Reset()
	return true
end
----------------------------------------------------------------
function _M:StepCheck( enableSwitchNextSource )
	if self.chip then
		self.chip:Step( enableSwitchNextSource )
		local changed = false
		if self.currentVideoSrc ~= self.chip.currentSource then
			print( "***M2 source port changed:", self.currentVideoSrc, self.chip.currentSource )
			self.currentVideoSrc = self.chip.currentSource
			changed = true
		end
		
		if self.autoAudioSrc and (self.forceReselectAudio or changed) then
			if self.currentVideoSrc == "VGA" then
				self:SelectAudioSource( "ANALOG", true )
			else
				self:SelectAudioSource( "HDMI", true )
			end
		end
		
		self.forceReselectAudio = false
	
		if CompareFormatInfo( self.chip.videoFormat, self.lastVideoFormat ) then	--Different
			self.lastVideoFormatVersion = self.lastVideoFormatVersion + 1
			self.lastVideoFormat = {}
			local k,v
			for k,v in pairs(self.chip.videoFormat) do
				self.lastVideoFormat[k] = v
			end
		end
		
		if CompareFormatInfo( self.chip.audioFormat, self.lastAudioFormat) then -- Different audio
			self.lastAudioFormatVersion = self.lastAudioFormatVersion + 1
			self.lastAudioFormat = {}
			local k,v
			for k, v in pairs(self.chip.audioFormat) do
				self.lastAudioFormat[k] = v
			end
		end
	end
end
----------------------------------------------------------------
function _M:ReportDetectingStatus( srcType )
	if srcType == "video" then
		if self.currentVideoSrc == "UNKNOWN" then
			return false, "HDMI/VGA detecting"
		else
			return true, self.currentVideoSrc
		end
	elseif srcType == "audio" then
		return true, self.currentAudioSrc
	end
end
----------------------------------------------------------------
function _M:ReportVideoFormat( curVersion )
	--if not curVersion or curVersion ~= self.lastVideoFormatVersion then
		return self.lastVideoFormatVersion, self.lastVideoFormat
	--else
	--	return self.lastVideoFormatVersion, nil	--Don't return format
	--end
end
----------------------------------------------------------------
function _M:ReportAudioFormat( curVersion )
	if self.savedAudioSrc ~= self.currentAudioSrc then
		self.savedAudioSrc = self.currentAudioSrc
		self.lastAudioFormatVersion = self.lastAudioFormatVersion + 1
		self.analogAudioFormat.audio_src = self.currentAudioSrc
	end

	if self.currentAudioSrc == "DIGITAL" then
		return self.lastAudioFormatVersion, self.lastAudioFormat
	else
		return self.lastAudioFormatVersion, self.analogAudioFormat
	end
end
----------------------------------------------------------------
function _M:SelectVideoSource( src )
	src = src:upper()
	print( "*** Changed video source port to: ", src )
	self.forceReselectAudio = true	--Try re-select audio source
	if src == "DEFAULT" or src == "AUTO" then
		self.chip:ChangeSource( "AUTO" )
		return "AUTO"
	elseif src == "ONESHOT" then
		self.chip:ChangeSource( "ONESHOT" )
		return "ONESHOT"
	elseif src == "HDMI" or src == "DIGITAL" then
		self.chip:ChangeSource( "HDMI" )
		return "HDMI"
	else
		self.chip:ChangeSource( "VGA" )
		return "VGA"
	end
end
----------------------------------------------------------------
local function PatchI2SWordLen(len)
	if hwChk.hardwareVersionSwCode < 0x20 then
		local file = io.open( "/proc/k2e_audio/wlen", "w" )
		if file then
			file:write( "" .. len )
			file:close()
		end
	end
end

function _M:SelectAudioSource( src, nochangeAuto )
	--TODO: Select audio source (SDI digital/Analog line-in/Analog Mic-in)
	src = src:upper()
	if src == "DIGITAL" or src == "HDMI" then
		self.currentAudioSrc = "DIGITAL"
		if not nochangeAuto then
			self.autoAudioSrc = false
		end
	
		if self.audSelIO then
			self.audSelIO:write( true )		-- 1: Digital
		end

		PatchI2SWordLen(32)
	elseif src == "ANALOG" or src == "LINE" then
		self.currentAudioSrc = "LINE"
		if not nochangeAuto then
			self.autoAudioSrc = false
		end
		if self.audSelIO then
			self.audSelIO:write( false )		-- 0: Analog
		end
		AlsaSelect( "LINE" )
		PatchI2SWordLen(64)
	elseif src == "MIC" then
		self.currentAudioSrc = "MIC"
		if not nochangeAuto then
			self.autoAudioSrc = false
		end
		if self.audSelIO then
			self.audSelIO:write( false )		-- 0: Analog
		end
		AlsaSelect( "MIC" )
		PatchI2SWordLen(64)
	elseif src == "AUTO" or src == "DEFAULT" then
		self.autoAudioSrc = true
		self.forceReselectAudio = true
	end
	
	return self.currentAudioSrc
end
----------------------------------------------------------------
function _M:SetVideoParameter( pmName, pmValue, src )
	--Nothing
end
----------------------------------------------------------------
function _M:SetAudioParameter( pmName, pmValue, src )
	if not src then
		src = self.currentAudioSrc
	end
	
	if src == "DIGITAL" or src == "HDMI" then
		return --Nothing
	end
	
	if pmName == "gain" then	--Adjust analog audio gain
		if src == "ANALOG" or src == "LINE" then
			pmValue = tonumber(pmValue)	--As DB
			local dbSteps = math.floor(pmValue/0.75)	--For AIC23 chip, value 1 means 0.75db
			if dbSteps > 8 then
				dbSteps = 8
			elseif dbSteps < -23 then
				dbSteps = -23
			end
			
			local dbVal = 23 + dbSteps
			os.execute( [[/usr/bin/amixer cset name='Line Input Volume' ]] .. dbVal )
		elseif src == "MIC" then
			pmValue = tonumber(pmValue)
			if pmValue < 3 then	-- <3db, as default mic gain (0db)
				os.execute( [[/usr/bin/amixer cset name='Mic Booster Switch' off]] )
			else	-- >=3db, Mic boost (so with 20db gain)
				os.execute( [[/usr/bin/amixer cset name='Mic Booster Switch' on]] )
			end
		end
	elseif pmName == "line_bypass" then
		if pmValue == 0 then
			os.execute( [[/usr/bin/amixer cset name='Output Mixer Line Bypass Switch' off]] )
		else
			os.execute( [[/usr/bin/amixer cset name='Output Mixer Line Bypass Switch' on]] )
		end
	end
end
----------------------------------------------------------------
function _M:GetSupportVideoSources()
	return { ["AUTO"] = "Auto Detect Video Input", ["ONESHOT"] = "Check Which Source With Signal and Lock It", ["HDMI"] = "HDMI Video Input", ["VGA"] = "VGA Video Input" }
end
----------------------------------------------------------------
function _M:GetSupportAudioSources()
	return { ["AUTO"] = "Auto Select Audio from HDMI/VGA Video Selection", ["HDMI"] = "HDMI Digital Audio", ["LINE"] = "Analog Line Input" --[[, ["MIC"] = "Analog MIC Input"]] }
end
----------------------------------------------------------------

------------------------------------------------------------------------
--All those special VGA operations
------------------------------------------------------------------------
function _M:VGAGetOptionFormats( formatCode )
	if not formatCode then
		return nil
	end
	
	local LCF, LCVS, FCL, BL, IorP = formatCode:match( "(%d+)/(%d+)/(%d+)/(%d+)/(%w)" )
	if not LCF or not LCF or not FCL or not BL then
		return nil
	end
	
	local allMatchs = vga_formats.STDI_AllMatches( tonumber(LCF), tonumber(LCVS), tonumber(FCL), tonumber(BL), IorP == "I" or Ior == "i" )
	if not allMatchs then 
		return nil
	else
		local i
		local r = {}
		for i=1, #allMatchs do
			table.insert( r, {
				_SEQ = allMatchs[i]._SEQ,
				formatName = allMatchs[i].formatName,
				modeline = allMatchs[i].modeline
			} )
		end
		
		return r
	end
end
------------------------------------------------------------------------
--For VGA applications: Set current special format
function _M:VGASetSpecFormat(formatCode, modeline)
	if vga_formats.SetSpecialFormat( formatCode, modeline ) then
		self.chip:ForceUpdate()
		return true
	else
		return false
	end
end
------------------------------------------------------------------------
--For VGA applications: Adjust the X/Y shift for picture
function _M:VGAPicAdjustment( formatCode, shiftX, shiftY, phase )
	if vga_formats.SetSpecialAdjustment( formatCode, shiftX, shiftY, phase ) then
		self.chip:ForceUpdate()
		return true
	else
		return false
	end
end
------------------------------------------------------------------------
function _M:VGAGetPicAdjustment( formatCode )
	return vga_formats.GetSpecialAdjustment( formatCode )
end
------------------------------------------------------------------------
--Set VGA user modelines
function _M:VGASetModelines( modelines )
	if vga_formats.AddModelines( modelines ) then
		self.chip:ForceUpdate()
		return true
	else
		return false
	end
end
------------------------------------------------------------------------
--Get VGA user modelines
function _M:VGAGetModelines()
	local lines = vga_formats.GetUserModelines()
	if type(lines) == "table" then
		local r = ""
		local i
		for i = 1, #lines do
			r = r .. lines[i] .. "\r\n"
		end
		return r
	else
		return nil
	end
end
------------------------------------------------------------------------

return _M

