-----------------------------------------------------------------------------------------------------
-- ADV7441A chip controller.
-- Copyright (C) 2016, KILOVIEW Elecontrics Co., Ltd.
--
-- Author: 	Jacob Zuo, 2016/6/4
-----------------------------------------------------------------------------------------------------

local periphery = require "periphery"
local edid = require "edid_7441a"	--EDID
local I2C = periphery.I2C
local GPIO = periphery.GPIO
--For debug only.
--local pretty = require "pl.pretty"

local MAIN_FREQ = 28636360

local hdtv_formats = require "hdtv_formats"
local vga_formats = require "vga_formats"

hdtv_formats.SetMainFreq( MAIN_FREQ )
vga_formats.SetMainFreq( MAIN_FREQ )

--Init both HDTV and VGA format standards
hdtv_formats.Init( "STDI" )
vga_formats.Init()

------------------------------------------------------
-- The factory function to create a ADV7441A handle object
------------------------------------------------------
local function CreateADV7441A( enablePorts, i2c_dev, baseAddr, defaultSrc, hpdGpio )
	if not enablePorts or type(enablePorts) ~= "table" or #enablePorts < 1 then
		enablePorts = { "HDMI", "VGA" }
	end
	hpdGpio = hpdGpio or 7
	i2c_dev = i2c_dev or "/dev/i2c-0"
	defaultSrc = defaultSrc or "AUTO"

	local USER_MAP = baseAddr or 0x42
	local USER_MAP1=0x46
	local USER_MAP2=0x62
	local VDP_MAP=0x4a
	local RSV_MAP=0x4e
	local HDMI_MAP=0x6a
	local KSV_MAP=0x66
	local EDID_MAP=0x6e

	local HPD_HIGH=false
	local HPD_LOW=true

	if USER_MAP == 0x42 then
		print( "Creating ADV7441A control object at 0x42@" .. i2c_dev .. "..." )
	elseif USER_MAP == 0x40 then
		print( "Creating ADV7441A control object at 0x40@" .. i2c_dev .. "..." )
		USER_MAP1 = USER_MAP1 - 2
		USER_MAP2 = USER_MAP2 - 2
		VDP_MAP = VDP_MAP - 2
		RSV_MAP = RSV_MAP - 2
		HDMI_MAP = HDMI_MAP - 2
		KSV_MAP = KSV_MAP - 2
		EDID_MAP = EDID_MAP - 2
	else
		return nil, "Invalid ADV7441A I2C address specified"
	end

	------------------------------------------------------
	-- Object
	------------------------------------------------------
	local ADV7441A = {
		enablePorts = enablePorts,
		defaultSource = defaultSrc:upper(),
		currentSource = "UNKNOWN",
		currentSignal = false,
		--currentFormat = nil,
		forceReassignFormat = false,
		lastFormatRecord = {},
		videoFormat = { _valid = false },
		audioFormat = { _valid = false },
		nostandardFormatInfo = {},
		detectingPort = -1,
		i2c_handle = nil,
		hpd_handle = nil,
		
		formatCode = "",	--The format code, e.g. LCF/LCVS/FCL/BL/P, unique for a format.
		optionFormats = nil, --The optional matched formats
	}

	-------------------------------------------------------
	-- I2C and GPIO openning
	-------------------------------------------------------
	local done,err = pcall( function() ADV7441A.i2c_handle = I2C( i2c_dev ) end )
	if not done then
		if type(err) == "table" then
			print( "ERROR: Could not open I2C device for ADV7441A controlling. ErrMsg:", err.message or "---" )
		else
			print( "ERROR: Could not open I2C device for ADV7441A controlling. ErrMsg:", err )
		end
		return nil, err
	end

	done,err = pcall( function() ADV7441A.hpd_handle = GPIO( hpdGpio, "out" ) end )
	if not done then
		if type(err) == "table" then
			print( "WARNING: Could not open HPD GPIO controller for pin [" .. hpdGpio .. "] ErrMsg:", err.message or "---" )
		else
			print( "WARNING: Could not open HPD GPIO controller for pin [" .. hpdGpio .. "] ErrMsg:", err )
		end
	else
		ADV7441A.hpd_handle:write( HPD_LOW )
	end

	-------------------------------------------------------
	-- Object method: Register read
	-------------------------------------------------------
	function ADV7441A:ReadReg( map, addr, itemArr )
		if not self.i2c_handle then
			return nil, "Invalid I2C handle"
		end
		if not itemArr or type(itemArr) ~= "table" then
			itemArr = { 0xff, flags = I2C.I2C_M_RD }
		else
			if #itemArr < 1 then
				itemArr[1] = 0xff
			end
			if not itemArr.flags then
				itemArr.flags = I2C.I2C_M_RD
			else
				itemArr.flags = itemArr.flags | I2C.I2C_M_RD
			end
		end

		local msg = { {addr}, itemArr }
		local done,err = pcall( function() self.i2c_handle:transfer( map>>1, msg ) end )
		if not done then
			if type(err) == "table" then
				--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", map, addr), err.message or "---" )
				return nil, err.message or "---"
			else
				--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", map, addr), err )
				return nil, err
			end
		else
			return itemArr
		end
	end

	------------------------------------------------------
	-- Register write
	------------------------------------------------------
	function ADV7441A:WriteReg( map, addr, itemArr )
		if not self.i2c_handle then
			return false, "Invalid I2C handle"
		end
		if not itemArr or type(itemArr) ~= "table" or #itemArr < 1 then
			return false, "Invalid data"
		end

		table.insert( itemArr, 1, addr )
		local msg = { itemArr }
		local done,err = pcall( function() self.i2c_handle:transfer( map>>1, msg ) end )
		if not done then
			if type(err) == "table" then
				--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", map, addr), err.message or "---" )
				return nil, err.message or "---"
			else
				--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", map, addr), err )
				return nil, err
			end
		else
			return true
		end
	end

	------------------------------------------------------
	-- Control chip to select corresponding port as current input
	------------------------------------------------------
	function ADV7441A:_SelectCurrentPort(src)
		if not src or src == "" or src == "CURRENT" then
			src = self.currentSource
		end

		if src ~= "COMPONENT" then		--Clear the ADC maps
			self:WriteReg( USER_MAP, 0xc3, {0x00, 0x00} )
		end

		if src == "HDMI" or src == "HDMI2" or src == "DVI" or src == "DVI2" then
			self:WriteReg( USER_MAP, 0x05, {0x06} )		--Select HDMI as source
			--[[
			if src == "HDMI" or src == "DVI" then
				self:WriteReg( HDMI_MAP, 0x00, {0x00} )	--Select port A
			else
				self:WriteReg( HDMI_MAP, 0x00, {0x01} )	--Select port B
			end
			]]
		elseif src == "VGA" then
			self:WriteReg( USER_MAP, 0x05, {0x02} )					--Select VGA as source

		elseif src == "COMPONENT" then
			--MAP Y,Pb,Pr
			self:WriteReg( USER_MAP, 0xc3, {0x31, 0xc2} )		--AIN1 - ADC0, AIN3 - ADC1
																									--AIN2 - ADC2, ADC_SW_MAIN_EN=1, SOG.
			self:WriteReg( USER_MAP, 0x05, {0x01} )					--Select Component as source

		elseif src == "CVBS" or src == "AV" then
			--self:WriteReg( USER_MAP, 0x00, { 0x0e } )	--Select AIN10 as CVBS input
			self:WriteReg( USER_MAP, 0x05, {0x00} )					--Select CVBS as source
		--else unknown, so ignore this
		end
	end

	------------------------------------------------------
	-- Soft reset the chip!
	------------------------------------------------------
	function	ADV7441A:Reset()	--Soft reset
		--GPIO HPD control (deassert)
		if self.hpd_handle then
			self.hpd_handle:write( HPD_LOW )
		end
		periphery.sleep_ms(10);

		self:WriteReg( KSV_MAP, 0x73, {0} );	--Control all disable first.

		self:WriteReg( USER_MAP, 0x0f, {0x80, flags=I2C.I2C_M_IGNORE_NAK} )
		local i, powerup
		powerup = false
		for i=1,100 do
			periphery.sleep_ms(10)
			local rslt = self:ReadReg( USER_MAP, 0x0f, { 0xff } )
			if rslt and (rslt[1] & 0x80) == 0 then
				powerup = true
				print( "Chip reset DONE." )
				break
			end
		end

		if not powerup then
			print( "***ERROR*** This chip looks could not power up after reset!" )
		end

--		self:WriteReg( USER_MAP, 0x0e, {USER_MAP2} )	--Configure USER_MAP2 address
--		periphery.sleep_ms( 10 )	--Wait 10ms
--		--Then configure other address
--		self:WriteReg( USER_MAP2, 0xeb, { USER_MAP1} )
--		self:WriteReg( USER_MAP2, 0xec, { VDP_MAP } )
--		self:WriteReg( USER_MAP2, 0xed, { KSV_MAP } )
--		self:WriteReg( USER_MAP2, 0xee, { EDID_MAP } )
--		self:WriteReg( USER_MAP2, 0xef, { HDMI_MAP } )
--		self:WriteReg( USER_MAP2, 0xea, { RSV_MAP } )
--		periphery.sleep_ms( 10 )	--Wait 10ms

		--local test = ADV7441A:ReadReg( USER_MAP2, 0xea, { 0, 0, 0, 0, 0, 0 } )
		--print( string.format( "%02x %02x %02x %02x %02x %02x", test[1], test[2], test[3], test[4], test[5], test[6] ) )

		periphery.sleep_ms(5);

		self:_InitInterrupts()
		self:_InitEDID()	--Init EDID

		self:WriteReg( USER_MAP, 0x00, { 0x0e } )	--Select AIN10 as CVBS input

		if self.defaultSource ~= "AUTO" and self.defaultSource ~= "ONESHOT" then
			--Select HDMI/DVI ports
			self.currentSource = self.defaultSource
			self:_SelectCurrentPort()
		else --As "AUTO"/"ONESHOT"
			self.currentSource = "UNKNOWN"
			self.detectingPort = -1
		end

		self.currentSignal = false
		--self.currentFormat = nil
		self.forceReassignFormat = false
		self.lastFormatRecord = {}		--Clear recorded format detection informations.
		self.videoFormat = {_valid = false}

		--TODO: Now fixed it.
		self.audioFormat = { 
			_valid = true,
			audio_src = "HDMI",
			sampling = 48000, 
			audio_signal_type = 0, --AKA HDMI
			channels = 2,
			bit_width = 16
		}

		--Assert HPD high
		if self.hpd_handle then
			self.hpd_handle:write( HPD_HIGH )
		end

	end

	------------------------------------------------------
	-- Init the ADV7441A interrupts
	------------------------------------------------------

	function ADV7441A:_InitInterrupts()
		self:WriteReg( USER_MAP1, 0x44, { 0xff } )	--INT1MSK1
		self:WriteReg( USER_MAP1, 0x42, { 0xff } )	--INT1CLR1
		self:WriteReg( USER_MAP1, 0x42, { 0x00 } )	--INT1CLR1

		self:WriteReg( USER_MAP1, 0x48, { 0xff } )	--INT1MSK2
		self:WriteReg( USER_MAP1, 0x46, { 0xff } )	--INT1CLR2
		self:WriteReg( USER_MAP1, 0x46, { 0x00 } )	--INT1CLR2

		self:WriteReg( USER_MAP1, 0x4c, { 0xff } )	--INT1MSK3
		self:WriteReg( USER_MAP1, 0x4a, { 0xff } )	--INT1CLR3
		self:WriteReg( USER_MAP1, 0x4a, { 0x00 } )	--INT1CLR3

		self:WriteReg( USER_MAP1, 0x66, { 0x10 } )	--HDMIINT2MSK2
		self:WriteReg( USER_MAP1, 0x62, { 0x10 } )	--HDMIINT2MSK1
		self:WriteReg( USER_MAP1, 0x65, { 0xff } )	--HDMIINTSTUS2
	end

	------------------------------------------------------
	--EDID configuration
	------------------------------------------------------
	function ADV7441A:_InitEDID()
		--GPIO HPD Control (deassert)
		if self.hpd_handle then
			self.hpd_handle:write( HPD_LOW )
		end

		self:WriteReg( KSV_MAP, 0x73, {0} );	--Control all disable first.
		--self:WriteReg( KSV_MAP, 0x40, { 0x81 } ); --BCAPS (???0x83)
		local i
		for i=1, #edid, 16 do
			self:WriteReg( EDID_MAP, i-1, { edid[i], edid[i+1], edid[i+2], edid[i+3], edid[i+4], edid[i+5], edid[i+6], edid[i+7] ,
																edid[i+8], edid[i+9], edid[i+10], edid[i+11], edid[i+12], edid[i+13], edid[i+14], edid[i+15]
															  } )
		end

		self:WriteReg( KSV_MAP, 0x72, { 0x8c } )	--CustomEDID_SPA_Start
		self:WriteReg( KSV_MAP, 0x70, { 0x00 } )	--CustomEDID_SPA_VAL1
		self:WriteReg( KSV_MAP, 0x71, { 0x00 } )	--CustomEDID_SPA_VAL2

		self:WriteReg( KSV_MAP, 0x73, { 0x07 } )	--Enable PortA and PortB EDID

		periphery.sleep_ms(5)		--Wait 5ms for Calc CRC

		--[[
		for i=1, 256, 8 do
			local v = self:ReadReg( EDID_MAP, i-1, { 0, 0, 0, 0, 0, 0, 0, 0 } )
			print( string.format( "%02x %02x %02x %02x %02x %02x %02x %02x", table.unpack(v) ) )
		end
		]]

		--GPIO HPD control (assert)
		if self.hpd_handle then
			self.hpd_handle:write( HPD_HIGH )
		end

	end

	------------------------------------------------------
	-- Basic check the signal status of current input
	------------------------------------------------------

	function ADV7441A:_CheckSignal(src)
		if not src then
			src = self.currentSource
		end

		if src == "CVBS" or src == "AV" then
			local rslt = self:ReadReg( USER_MAP, 0x13, {0x00} )
			if rslt then
				return (rslt[1] & 0x01) ~= 0
			else
				return false
			end
		else
			local rslt = self:ReadReg( USER_MAP, 0xb1, {0x00} )
			if rslt then
				return (rslt[1] & 0x80 ) ~= 0
			else
				return false
			end
		end
	end

	------------------------------------------------------
	-- A step handler for ADV7441A processing.
	-- To fit the LUA working mode, you need call the
	-- 'Step' periodically ( suggest per 200~500ms).
	-- This routine should do :
	--	* Detect which source has signal and auto select it as current input.
	--        (Set default source to "AUTO" or "ONESHOT")
	--	* Signal status checking
	--	*	Auto format detecting
	--	*	Configue the chip to make it output BT.656/BT.601 signals
	--
	-- The argument 'switchSrcEnabled' controls that when default source be "AUTO"
	-- or "ONESHOT" and signal detecting is on the way, this 'Step' calling need or need not
	-- switch source to next trying port [ If signal detected and port ensured, nothing should
	-- happen ]. So this is no specials, so you can enable it each calling time, but if you do like
	-- this, you need think whether or not switch detecting ports to fast, espeically the 'Step'
	-- calling period is little (such as < 200ms).
	-- So, a suggestion is setting this be 'true' periodically when call 'Step', like:
	--		Step(false) - Step(false) - Step(false) - Step(TRUE) - Step(false) - Step(false) - Step(false) - Step(TRUE) - ...
	------------------------------------------------------

	function ADV7441A:Step(switchSrcEnabled)
		if self.currentSource == "UNKNOWN" then	--Auto detection required
			if self.detectingPort < 0 then --First start
				self.detectingPort = 1
				self:_SelectCurrentPort( self.enablePorts[ self.detectingPort ] )
			end

			local thisPort = self.enablePorts[ self.detectingPort ]
			local signal = self:_CheckSignal( thisPort )

			if signal then
				self.forceReassignFormat = true	-- First detect the port, re-assign format arguments.
				self.currentSource = thisPort
				self.currentSignal = true
				
				if self:_DetectCurrentFormat( thisPort ) then
					self:_ConfigureCurrentFormat()
				end
				--Otherwise, though has no correct format, but assume it detected.
				return
			else
				self.videoFormat.signal_locked = false
			end

			if switchSrcEnabled then
				self.detectingPort = (self.detectingPort + 1)
				if self.detectingPort > #self.enablePorts then
					self.detectingPort = 1
				end
				self:_SelectCurrentPort( self.enablePorts[ self.detectingPort ] )
			end
			return	--Changed port, so next time, it can do detecting.
		end

		-- Speicified or detected a source port
		local signal = self:_CheckSignal()	--Check current signal
		if not signal then	--Lost signal
			self.currentSignal = false
			if self.defaultSource == "AUTO" and switchSrcEnabled then --Auto detect again
				self.currentSource = "UNKNOWN"
			end	--Otherwise, specified or "ONESHOT", always use current source
		elseif not self.currentSignal and signal then	--Hot plugged
			self.currentSignal = true
			self.forceReassignFormat = true	--Force re-assign format arguments
		end

		if self.currentSignal then
			if self:_DetectCurrentFormat() then
				self:_ConfigureCurrentFormat()
			end
		else
			self.videoFormat.signal_locked = false
		end
	end
	------------------------------------------------------
	------------------------------------------------------

	function ADV7441A:ForceUpdate()
		self.lastFormatRecord = {}		--Clear the recorded format information.
	end

	------------------------------------------------------
	-- Change the source port by special name: "HDMI"/"HDMI2", "DVI"/"DVI2", "VGA", "COMPONENT", "CVBS" ...
	-- Also you can set 'src' be "AUTO", "ONESHOT", if that, it should detect an input which with signal and
	-- select it as current source.
	------------------------------------------------------
	function ADV7441A:ChangeSource(src)
		if not src then
			return false
		end

		if src == "AUTO" or src == "ONESHOT" then
			self.currentSource = "UNKNOWN"
			self.defaultSource = src
		elseif src == "UNKNOWN" or src == "DEFAULT" then
			self.currentSource = "UNKNOWN"
			self.defaultSource = "AUTO"
		else
			self.currentSource = src
			self.defaultSource = src
			self:_SelectCurrentPort( src )
		end

		self:Step(true)	--Immediately checking.

		return true
	end

	------------------------------------------------------
	-- Detect current signal format
	------------------------------------------------------

	function ADV7441A:_DetectCurrentFormat(src)
		if not src then
			src = self.currentSource
		end

		if src == "HDMI" or src == "HDMI2" or src == "DVI" or src == "DVI2" then
			return self:_DetectHDMIFormat(src)
		elseif src == "COMPONENT" then
			return self:_DetectComponentFormat(src)
		elseif src == "VGA" then
			return self:_DetectVGAFormat(src)
		elseif src == "CVBS" or src == "AV" then
			return self:_DetectCVBSFormat(src)
		else
			return false
		end
	end

	------------------------------------------------------
	-- Configure signal format
	------------------------------------------------------

	function ADV7441A:_ConfigureCurrentFormat(src)
		if not self.videoFormat._lastFormatChanged and not self.forceReassignFormat then
			return true
		end

		if not src then
			src = self.currentSource
		end

		self.forceReassignFormat = false	--Clear this flag.

		if src == "HDMI" or src == "HDMI2" or src == "DVI" or src == "DVI2" then
			return self:_ConfigureHDMIFormat(src)
		elseif src == "COMPONENT" then
			return self:_ConfigureComponentFormat(src)
		elseif src == "VGA" then
			return self:_ConfigureVGAFormat(src)
		elseif src == "CVBS" or src == "AV" then
			return self:_ConfigureCVBSFormat(src)
		else
			return false
		end
	end

	------------------------------------------------------
	-- According detected format information, record it to Amba based formation context.
	------------------------------------------------------

	function ADV7441A:_RecordVideoFormatInfo(src, formatInfo)
		if type(formatInfo) ~= "table" then
			return false
		end

		--pretty.dump( formatInfo )

		if self.videoFormat._formatInfo ~= formatInfo or (src and src ~= self.videoFormat.video_src) then
			self.videoFormat._lastFormatChanged = true
			self.videoFormat._formatInfo = formatInfo

			if src then
				self.videoFormat.video_src = src
			end

			--TODO:
			self.videoFormat.cap_start_x = 0
			self.videoFormat.cap_start_y = 0

			self.videoFormat.cap_cap_w = formatInfo.hAct
			self.videoFormat.line_width = formatInfo.hAct

			self.videoFormat.total_width = formatInfo.hTotal

			if formatInfo.vAct then
				self.videoFormat.cap_cap_h = formatInfo.vAct
				self.videoFormat.total_height = formatInfo.vTotal

				if formatInfo.interlaced then
					self.videoFormat.f0_height = formatInfo.vAct/2
					self.videoFormat.f1_height = formatInfo.vAct/2
				else
					self.videoFormat.f0_height = formatInfo.vAct
					self.videoFormat.f1_height = formatInfo.vAct
				end
			elseif formatInfo.vAct0 then
				self.videoFormat.f0_height = formatInfo.vAct0
				if formatInfo.interlaced then
					if formatInfo.vAct1 then
						self.videoFormat.f1_height = formatInfo.vAct1
						self.videoFormat.cap_cap_h = formatInfo.vAct0 + formatInfo.vAct1
						if self.videoFormat.cap_cap_h >= 486 and self.videoFormat.cap_cap_h <= 488 then
							self.videoFormat.cap_cap_h = 480	--This is a special format
						else
							self.videoFormat.cap_cap_h = self.videoFormat.cap_cap_h & 0xfffffffe
						end
						self.videoFormat.total_height = self.videoFormat.cap_cap_h	--??? total_height for interlace?
					else
						self.videoFormat.f1_height = formatInfo.vAct0
						self.videoFormat.cap_cap_h = formatInfo.vAct0 * 2
						self.videoFormat.total_height = formatInfo.vAct0 * 2
					end
				else
					self.videoFormat.f1_height = formatInfo.vAct0
					self.videoFormat.cap_cap_h = formatInfo.vAct0
					self.videoFormat.total_height = formatInfo.vTotal0
				end
			else
				self.videoFormat.f0_height = 0
				self.videoFormat.f1_height = 0
				self.videoFormat.cap_cap_h = 0	--This is wrong.
				self.videoFormat.total_height  = 0
			end

			--TODO:
			self.videoFormat.sync_start = 0

			if formatInfo.video_system then
				self.videoFormat.video_system = formatInfo.video_system
			else
				self.videoFormat.video_system = _G.AMBA_VIDEO_SYSTEM_AUTO or 0x0
			end

			if AMBA_VIDEO_FPS then
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS(formatInfo.frameRate)
			else
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS_60 or formatInfo.frameRate
			end
			self.videoFormat.frame_rate_value = formatInfo.frameRate

			if formatInfo.aspect_ratio then
				self.videoFormat.aspect_ratio = formatInfo.aspect_ratio
			else
				self.videoFormat.aspect_ratio = _G.AMBA_VIDEO_RATIO_AUTO or 0
			end

			--TODO: Here, need fix this to support 8 bits/ BT.601 options.
			self.videoFormat.input_type = _G.AMBA_VIDEO_TYPE_YUV_656 or 2
			self.videoFormat.bit_resolution = 16
			if formatInfo.interlaced then
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_INTERLACE or 1
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_INTLC or 1
				self.videoFormat.interlaced = 1
			else
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_PROGRESSIVE or 2
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_PROG or 2
				self.videoFormat.interlaced = 0
			end

			--TODO:
			self.videoFormat.hs_front_porch = 0
			self.videoFormat.hs_pulse_width = 0
			self.videoFormat.hs_back_porch = 0
			--TODO:
			self.videoFormat.f0_vs_front_porch = 0
			self.videoFormat.f0_vs_pulse_width = 0
			self.videoFormat.f0_vs_back_porch = 0
			--TODO:
			self.videoFormat.f1_vs_front_porch = 0
			self.videoFormat.f1_vs_pulse_width = 0
			self.videoFormat.f1_vs_back_porch = 0

			if formatInfo.tmdsFreq and formatInfo.tmdsFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.tmdsFreq
			elseif formatInfo.pixelFreq and formatInfo.pixelFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.pixelFreq
			else
				self.videoFormat.pixel_freq = 0
			end

			self.videoFormat.format_code = formatInfo.formatId	-- Make the formatId as format-code.
			self.videoFormat.format_name = formatInfo.formatName or (self.videoFormat.cap_cap_w .. "x" .. self.videoFormat.cap_cap_h)
			self.videoFormat.format_seq = formatInfo._SEQ or 0
			--self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_DIGITAL or 3
			return true
		else
			return true	--No changed.
		end
	end

	------------------------------------------------------
	-- Detect HDMI Audio
	------------------------------------------------------
	function ADV7441A:_DetectHDMIAudio(src)
	end

	------------------------------------------------------
	-- Detect HDMI AVI info frame
	------------------------------------------------------

	function ADV7441A:_DetectHDMIInfoFrame(src)
		local rslt
		rslt = self:ReadReg( USER_MAP1, 0x60, {0xff} )	--HDMI Raw status
		if rslt and (rslt[1] & 0x01) ~= 0 then	--AVI Info detected
			rslt = self:ReadReg( USER_MAP1, 0x6c, {0xff} )	--Try read HDMI INT Status4
			if not rslt then
				return false
			end

			local i
			local veri_ok = false
			for i=1,3 do
				rslt = self:ReadReg( HDMI_MAP, 0x80, {0x00} )	--Read info-frame version
				if rslt and rslt[1] >= 2 then
					veri_ok = true
					break
				end
			end

			if not veri_ok then
				return false
			end

			rslt = self:ReadReg( HDMI_MAP, 0x84, {0x00} )	--Read AVIINFODATA4 (id code)
			if rslt then
				local idCode = rslt[1] & 0x7f
				if src ~= self.lastFormatRecord.source or idCode ~= self.lastFormatRecord.infoFrameId then
					local formatInfo = hdtv_formats.Standard_Match( idCode )
					if formatInfo then
						if self:_RecordVideoFormatInfo( src, formatInfo ) then
							self.lastFormatRecord.source = src
							self.lastFormatRecord.infoFrameId = idCode
							return true
						else
							return false
						end
					else --of matched.
						return false
					end
				else --Last record is this one, so no do again.
					return true
				end
			else
				return false
			end
		else
			return false
		end
	end

	------------------------------------------------------
	-- STDI checking
	------------------------------------------------------

	function ADV7441A:_STDICheck()
		local interlaced = false
		local rslt = self:ReadReg( USER_MAP, 0xb1, {0x00} )
		if not rslt or (rslt[1]&0x80) == 0 then
			--self.videoFormat._DebugInfo = "E:-/-/-/-"
			return false, nil, nil, nil, nil, nil
		end
		interlaced = ((rslt[1]&0x40) ~= 0)

		local BL = rslt[1] & 0x3f
		rslt = self:ReadReg( USER_MAP, 0xb2, {0x00} )
		BL = (BL << 8) + rslt[1]

		rslt = self:ReadReg( USER_MAP, 0xb3, {0x00} )
		local LCVS = (rslt[1] & 0xf8) >> 3

		local LCF = rslt[1] & 0x07
		LCF = LCF << 8

		rslt = self:ReadReg( USER_MAP, 0xb4, {0x00} )
		LCF = LCF + rslt[1]

		rslt = self:ReadReg( USER_MAP, 0xca, {0x00, 0x00} )
		local FCL = (rslt[1] & 0x1f) << 8
		FCL =  FCL +rslt[2]

		--self.videoFormat._DebugInfo = "E:" .. BL .. "/" .. LCVS .. "/" .. LCF .. "/" .. FCL ..  (interlaced and "/I" or "/P")
		--print( self.videoFormat._DebugInfo )
		return true, BL, LCF, FCL, LCVS, interlaced
	end

	------------------------------------------------------
	-- Assume HDMI customized format
	------------------------------------------------------
	function ADV7441A:_AssumeHDMICustomFormat(src, BL, LCF, FCL, LCVS, interlaced)
		local refreshFreq = 0
		if FCL and FCL > 0 then
			refreshFreq = MAIN_FREQ/(FCL*256)
			if refreshFreq <= 24 then
				refreshFreq = 24
			elseif refreshFreq <= 26 then
				refreshFreq = 25
			elseif refreshFreq <= 31 then
				refreshFreq = 30
			elseif refreshFreq <=51 then
				refreshFreq = 50
			elseif refreshFreq <=61 then
				refreshFreq = 60
			elseif refreshFreq <=71 then
				refreshFreq = 70
			elseif refreshFreq <= 76 then
				refreshFreq = 75
			elseif refreshFreq <= 85 then
				refreshFreq = 85
			end	--Else ignore restrict.
		end

		if refreshFreq <= 0 then
			return false
		end

		local rslt = self:ReadReg( HDMI_MAP, 0x06, {0x00} )
		local tmdsFreq = rslt[1]
		tmdsFreq = tmdsFreq * 1000000

		local actW = 0
		local totalW = 0
		local actH = 0
		local totalH = 0
		local actH2 = 0
		local totalH2 = 0

		rslt = self:ReadReg( HDMI_MAP, 0x07, {0x00,0x00} )
		if rslt then
			actW = ((rslt[1]&0x0f) << 8) + rslt[2]
		end

		rslt = self:ReadReg( HDMI_MAP, 0x1e, {0x00, 0x00} )
		if rslt then
			totalW = ((rslt[1]&0x0f) << 8) + rslt[2]
		end

		rslt = self:ReadReg( HDMI_MAP, 0x26, {0x00, 0x00} )
		if rslt then
			totalH = ((rslt[1]&0x1f) << 8) + rslt[2]
		end

		rslt = self:ReadReg( HDMI_MAP, 0x28, {0x00, 0x00} )
		if rslt then
			totalH2 = ((rslt[1]&0x1f) << 8) + rslt[2]
		end

		rslt = self:ReadReg( HDMI_MAP, 0x09, {0x00, 0x00} )
		if rslt then
			actH = ((rslt[1]&0x0f) << 8) + rslt[2]
		end

		rslt = self:ReadReg( HDMI_MAP, 0x0b, {0x00, 0x00} )
		if rslt then
			actH2 = ((rslt[1]&0x0f) << 8) + rslt[2]
		end

		if actW < 160 or actH < 90 or refreshFreq < 23 then
			return false
		end

		local formatName
		if interlaced then
			formatName = actW .. "x" .. (actH + actH2) .. "@" .. refreshFreq .. "i"
		else
			formatName = actW .. "x" .. actH .. "@" .. refreshFreq .. "Hz"
		end

		local v = self.nostandardFormatInfo[formatName]
		if not v then	--Create this format information
			v = {
				formatId = LCF .. "/" .. LCVS .. "/" .. FCL .. "/" .. BL .. "/" .. (interlaced and "I" or "P"),
				formatName = formatName,
				frameRate = refreshFreq,
				hTotal = totalW,
				hAct = actW,
				vTotal0 = totalH,
				vAct0 = actH,
				vTotal1 = totalH2,
				vAct1 = actH2,
				interlaced = interlaced,
				rateMode = 1,	--That is VIRATE_HD
				tmdsFreq = tmdsFreq,
				spec_amba_mode = 0,	--That is AMBA_VIDEO_MODE_AUTO
				pixelRepeat = false,	--TODO:
				aspect_ratio = 0, --That is ASPECT_RATIO_AUTO
				video_system = 0, --That is AMBA_VIDEO_SYSTEM_AUTO
			}
			self.nostandardFormatInfo[formatName] = v
		end

		return self:_RecordVideoFormatInfo( src, v )
	end

	------------------------------------------------------
	-- Detect HDMI input format
	------------------------------------------------------

	function ADV7441A:_DetectHDMIFormat(src)
		local rslt = self:ReadReg( HDMI_MAP, 0x05, {0x00} )		--Check HDCP
		local hdcp = false
		if rslt and (rslt[1]&0x40) ~= 0 then
			hdcp = true
			self.videoFormat.hdcp_content = 1
		else
			self.videoFormat.hdcp_content = 0	--It may contains an error status, but, no matter.
		end

		self.videoFormat._valid = false

		rslt = self:ReadReg( USER_MAP1, 0x68, {0x00} )	--Check TMDS lock status
		if rslt then
			if (rslt[1]&0x01) ~= 0 then	--TMDS locked
				if (rslt[1]&0x02) ~= 0 then --Is HDMI mode
					self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_HDMI or 1
				else
					self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_DVI or 0
				end

				rslt = self:ReadReg( HDMI_MAP, 0x04, {0x00} ) --Check Video PLL locked
				if not rslt or (rslt[1]&0x02) == 0 then
					self.videoFormat.signal_locked = false
					return false
				end

				--DE_REGEN_FILTER_LOCKED && VERTICAL_FILTER_LOCKED
				rslt = self:ReadReg( HDMI_MAP, 0x07, {0x00} )
				if not rslt or (rslt[1]&0x10) == 0 or (rslt[1]&0x40) == 0 then
					self.videoFormat.signal_locked = false
					return false
				end

				--OK, all ready, check standards (Bellow)
			else
				self.videoFormat.signal_locked = false
				return false
			end
		else
			self.videoFormat.signal_locked = false
			return false
		end

		self.videoFormat.signal_locked = true

		--NOTE: some HDMI source does not provided valid format-id.
		--[[
		--Check info frame first.
		if self:_DetectHDMIInfoFrame(src) then
			--Clear all recorded BL/LCF/FCL/LCVS/interlaced
			self.videoFormat._valid = true
			self.lastFormatRecord.BL = nil
			self.lastFormatRecord.LCF = nil
			self.lastFormatRecord.FCL = nil
			self.lastFormatRecord.LCVS = nil
			self.lastFormatRecord.interlaced = nil
			return true
		end
		]]

		self.lastFormatRecord.infoFrameId = -1		--Invalid recorded infoFrameId

		--So sorry, I need check special formats
		local done, BL, LCF, FCL, LCVS, interlaced = self:_STDICheck()

		if done and ( BL ~= self.lastFormatRecord.BL
								or LCF ~= self.lastFormatRecord.LCF
								or FCL ~= self.lastFormatRecord.FCL
								or LCVS ~= self.lastFormatRecord.LCVS
								or interlaced ~= self.lastFormatRecord.interlaced
								or src ~= self.lastFormatRecord.source
							)
		then
			rslt = hdtv_formats.STDI_Match( LCF, LCVS, FCL, BL, interlaced )
			if rslt then
				if self:_RecordVideoFormatInfo( src, rslt ) then
					self.videoFormat._valid = true
					self.lastFormatRecord.source = src
					self.lastFormatRecord.BL = BL
					self.lastFormatRecord.LCF = LCF
					self.lastFormatRecord.FCL = FCL
					self.lastFormatRecord.LCVS = LCVS
					self.lastFormatRecord.interlaced = interlaced
					return true
				end
			end

--[[
			rslt = vga_formats.STDI_Match( LCF, LCVS, FCL, BL, interlaced )	--Check VGA standards
			if rslt then
				if self:_RecordVideoFormatInfo( src, rslt ) then
					self.videoFormat._valid = true
					self.lastFormatRecord.source = src
					self.lastFormatRecord.BL = BL
					self.lastFormatRecord.LCF = LCF
					self.lastFormatRecord.FCL = FCL
					self.lastFormatRecord.LCVS = LCVS
					self.lastFormatRecord.interlaced = interlaced
					return true
				end
			end
]]

			if self:_AssumeHDMICustomFormat( src, BL, LCF, FCL, LCVS, interlaced ) then
					self.videoFormat._valid = true
					self.lastFormatRecord.source = src
					self.lastFormatRecord.BL = BL
					self.lastFormatRecord.LCF = LCF
					self.lastFormatRecord.FCL = FCL
					self.lastFormatRecord.LCVS = LCVS
					self.lastFormatRecord.interlaced = interlaced
					return true
			else
				return false	--Not detect
			end
		elseif done then
			self.videoFormat._valid = true
			return true	--Already recorded
		end

		return false

	end

	------------------------------------------------------
	-- Detect VGA format
	------------------------------------------------------

	function ADV7441A:_DetectVGAFormat(src)
		self.lastFormatRecord.infoFrameId = -1
		local rslt
		local done, BL, LCF, FCL, LCVS, interlaced = self:_STDICheck()

		self.videoFormat._valid = false
		self.videoFormat.hdcp_content = 0
		self.videoFormat.signal_locked = done	--Signal is locked? STDI get, then make it be.
		
		if done and ( BL ~= self.lastFormatRecord.BL
								or LCF ~= self.lastFormatRecord.LCF
								or FCL ~= self.lastFormatRecord.FCL
								or LCVS ~= self.lastFormatRecord.LCVS
								or interlaced ~= self.lastFormatRecord.interlaced
								or src ~= self.lastFormatRecord.source
							)
		then
			-- Don't check the HDTV formats, because:
			-- * All my builtin modelines include HDTV format declaration.
			-- * If we check HDTV formats, but the standards does not include SYNC informations
			--	   so I can't configure registers correctly!
			--[[
			rslt = hdtv_formats.STDI_Match( LCF, LCVS, FCL, BL, interlaced )
			... others
			]]

			rslt = vga_formats.STDI_Match( LCF, LCVS, FCL, BL, interlaced )	--Check VGA standards
		
			if rslt then --Matched.
				if self:_RecordVideoFormatInfo( src, rslt ) then
					self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_ANALOG or 4
					self.videoFormat._valid = true
					self.lastFormatRecord.source = src
					self.lastFormatRecord.BL = BL
					self.lastFormatRecord.LCF = LCF
					self.lastFormatRecord.FCL = FCL
					self.lastFormatRecord.LCVS = LCVS
					self.lastFormatRecord.interlaced = interlaced
					return true
				end
			end
			
			--Otherwise: (1) not matched (2) record format fail (that may never occus)
			self.videoFormat._valid = false
			self.videoFormat.formatId = LCF .. "/" .. LCVS .. "/" .. FCL .. "/" .. BL .. "/" .. (interlaced and "I" or "P" )
			self.videoFormat.format_name = "E: " ..  self.videoFormat.formatId
			return false -- Note false.
		elseif done then
			self.videoFormat._valid = true
			return true	--Already recorded
		else --Can not get STDI
			return false
		end
		
		return false
	end

	------------------------------------------------------
	-- Detect component (YPbPr) format
	------------------------------------------------------
	function ADV7441A:_DetectComponentFormat(src)
		--TODO:
		return false
	end

	------------------------------------------------------
	-- Detect CVBS format
	------------------------------------------------------

	function ADV7441A:_DetectCVBSFormat(src)
		--TODO:
		return false
	end

	------------------------------------------------------
	-- Configure HDMI format
	------------------------------------------------------

	function ADV7441A:_ConfigureHDMIFormat(src)

		self.videoFormat._lastFormatChanged = false

		self:WriteReg( USER_MAP, 0x03, { 0x05 } )	-- 0c: 20bit 4:2:2, DUP AV code.
		self:WriteReg( USER_MAP, 0x05, { 0x06 } )	-- ?05: Prim_Mode =0110b HDMI
		self:WriteReg( USER_MAP, 0x1D, { 0x40 } )	-- Disable TRI_LLC
		self:WriteReg( USER_MAP, 0x37, { 0x01 } )	-- Enable PCLK
		self:WriteReg( USER_MAP, 0x6B, { 0xc1 } )	-- Setting cp_op_sel = 1 (20bit) & Enable DE(range 0-255)
		self:WriteReg( USER_MAP, 0x68, { 0xF0 } )	-- f2: Auto CSC , YPbPr Out
		self:WriteReg( USER_MAP, 0xBA, { 0xA0 } )	-- Enable HDMI and Analog in
		self:WriteReg( USER_MAP, 0xC8, { 0x08 } )	-- Digital Fine Clamp Start position
		self:WriteReg( USER_MAP, 0x7B, { 0x0f, 0xc0, 0xff, 0xff } )	-- Enable AV code Insertion
		--self:WriteReg( USER_MAP, 0x7C, { 0xc0 } )	-- Pos H/V Sync out
		--self:WriteReg( USER_MAP, 0x7D, { 0xFF } )	-- Start_hs inactive CLK
		--self:WriteReg( USER_MAP, 0x7E, { 0xFF } )	-- End_hs inactive CLK

		--MARK: Adjust the SAV/EAV as -1, for de-process a bad line at left.
		self:WriteReg( USER_MAP, 0x8b, { 0x4f, 0xff, 0xff } )
		--self:WriteReg( USER_MAP, 0x8c, { 0xff } )
		--self:WriteReg( USER_MAP, 0x8d, { 0xff } )

		self:WriteReg( USER_MAP, 0xb7, { 0x13 } )	-- Recomend, for 720p30 switching

		self:WriteReg( USER_MAP, 0xF4, { 0x3F } )		-- Max Drive Strength
		self:WriteReg( USER_MAP2, 0xF0, { 0x10, 0x0f } )	-- ADI Recommended Write
		--self:WriteReg( USER_MAP2, 0xF1, { 0x0F } )	-- ADI Recommended Write
		self:WriteReg( USER_MAP2, 0xF4, { 0x20 } )	-- ADI Recommended Write
		self:WriteReg( HDMI_MAP, 0x15, { 0xEC } )		-- Disable these mute mask bits

		self:WriteReg( HDMI_MAP, 0x1C, { 0x49, 0x04 } )		-- Set MCLKOUT to output 128fs
		--self:WriteReg( HDMI_MAP, 0x1D, { 0x04 } )		--  & Set Unmute Delay to 1_5 sec.

		--**self:WriteReg( HDMI_MAP, 0x3C, { 0x42 } )	-- Set MCLKOUT to output 128fs
		--**self:WriteReg( HDMI_MAP, 0x3D, { 0x40 } )	-- Change the VCO range for the audio PLL
		self:WriteReg( HDMI_MAP, 0x5A, { 0x01 } )		-- Reset Audio Pll

		return true
	end

	------------------------------------------------------
	-- Configure VGA format
	------------------------------------------------------
	local SpecialVGAPhaseAjusts = {
		{ w = 1440, h = 900, f = 60, phase = 0x40 },
		{ w = 1680, h = 1050, f = 60, phase = 0x40 },
		{ w = 1280, h = 768, f = 60, phase = 0x50 },
		{ w = 1280, h = 768, f = 75, phase = 0x50 },
		{ w = 1360, h = 768, f = 60, phase = 0x50 },
		{ w = 1368, h = 768, f = 60, phase = 0x50 },
		{ w = 1600, h = 900, f = 60, phase = 0x50 },
		{ w = 1024, h = 768, f = 60, phase = 0x50 }
	}

	function ADV7441A:_ConfigureVGAFormat(src)
		self:WriteReg( USER_MAP, 0x03, { 0x05 })	--0c: 20bit 4:2:2, DUP AV code.
		self:WriteReg( USER_MAP, 0x05, { 0x02 })	--Prim_Mode =010b for automatic graphics mode

		--**self:WriteReg( USER_MAP, 0x06, { 0x07 })	--@@ Manual set bellow.

		self:WriteReg( USER_MAP, 0x1D, { 0x40 })		--Disable TRI_LLC
		self:WriteReg( USER_MAP, 0x3c, { 0x58 })		--SOG Sync level for atenuated sync, PLL Qpump to default

		--**self:WriteReg( USER_MAP, 0x37, 0x01 })	--@@01? PCLK Polarity

		self:WriteReg( USER_MAP, 0x47, { 0x0a })			--0a: Enable Automatic PLL_Qpump and VCO Range
		self:WriteReg( USER_MAP, 0x68, { 0xf0 })			--Auto CSC, YPbPr Out

		--@@aaaa MARK: Change the sample phase
		self:WriteReg( USER_MAP, 0x6a, { 0x50 })			-- 40 is default, low 5bits are change the phase.
		--@@Mark by Jacob:	Change 0x6a to 0x48( Fix the sample phase, is help for fix the [shake] problem )

		--@@By ME.
		self:WriteReg( USER_MAP, 0x6b, { 0xd1 })			-- (c1 default) CPOP_SEL = 1(mode 7), 656 range = 0-255
		self:WriteReg( USER_MAP, 0x7B, { 0x0f })			--(0x0f) Enable AV code Insertion

		--**self:WriteReg( USER_MAP, 0x7C, { 0xc0 })	--Pos H/V Sync out
		--**self:WriteReg( USER_MAP, 0x7D, { 0xFF })	--Start_hs inactive CLK
		--**self:WriteReg( USER_MAP, 0x7E, { 0xFF })		--End_hs inactive CLK

		--MARK: Adjust the SAV/EAV as -1, for de-process a bad line at left.
		--**self:WriteReg( USER_MAP, 0x8b, { 0x4f })
		--**self:WriteReg( USER_MAP, 0x8c, { 0xff })
		--**self:WriteReg( USER_MAP, 0x8d, { 0xff })

		self:WriteReg( USER_MAP, 0x81, { 0xda })		--Enable AV code insert
		self:WriteReg( USER_MAP, 0x91, { 0x10 })		--10: Progressive.
		self:WriteReg( USER_MAP, 0xb7, { 0x13 })		--Recomend, for 720p30 switching

		self:WriteReg( USER_MAP, 0xF4, { 0x3f })	--F Max Drive Strength
		--**self:WriteReg( USER_MAP, 0xf4, 0x15 })	--Medium Low Drive Strength

		--Correctup some resolution's DLL phase.
		if self.videoFormat._formatInfo and self.videoFormat._formatInfo.hSyncStart then	--Check 'hSyncStart' to ensure it's a VGA style format information

			self.videoFormat._lastFormatChanged = false

			local v = self.videoFormat._formatInfo
			--[[
			local i
			for i=1, #SpecialVGAPhaseAjusts do
				if SpecialVGAPhaseAjusts[i].w == v.hAct and
					SpecialVGAPhaseAjusts[i].h == v.vAct and
					math.abs( SpecialVGAPhaseAjusts[i].f - v.frameRate ) <= 1 then
						self:WriteReg( USER_MAP, 0x6a, { SpecialVGAPhaseAjusts[i].phase } )
						break
				end
			end
			]]
			if v.samplePhase then
				local phase = math.floor( tonumber(v.samplePhase) ) & 0x1f
				phase = phase | 0x40
				self:WriteReg( USER_MAP, 0x6a, { phase } )
			end
			
			if v._ADV7441_MODE then --Builtin supports
				self:WriteReg( USER_MAP, 0x06, {v._ADV7441_MODE} )	--Set MODE.
				self:WriteReg( USER_MAP, 0x87, { 0x63, 0x5a } ) --Disable PLL ratio
				self:WriteReg( USER_MAP, 0x8f, {0} )
				self:WriteReg( USER_MAP, 0x90, {0} )
				self:WriteReg( USER_MAP, 0xab, {0} )
				self:WriteReg( USER_MAP, 0xac, {0} )

				self:WriteReg( USER_MAP, 0xa2, {0,0,0,0,0,0} )
			else -- Other resolution which ADV7441 doesn't known
				self:WriteReg( USER_MAP, 0x06, { 0x07 } )		--A customized resolution.
				--TODO: Now, we only consider progressive formats, VGA interlaced format be ignored now.

				-- PLL div.
				local plldiv = v.hTotal
				self:WriteReg( USER_MAP, 0x87, { 0xe0 | ((plldiv>>8)&0x0f), (plldiv&0xff) } )

				-- Freerun length
				local lineCnt = v.vTotal or 0
				local frunLen
				if not v.hTotal or not v.pixelFreq or  v.hTotal <= 0 then
					frunLen = 0
				else
					frunLen = math.floor( MAIN_FREQ / ( v.pixelFreq / v.hTotal ) )
				end
				self:WriteReg( USER_MAP, 0x8f, { ((frunLen>>8)&0x0f), (frunLen&0xff) } )

				-- Total line count
				self:WriteReg( USER_MAP, 0xab, { ((lineCnt>>4)&0xff), ((lineCnt<<4)&0xf0) } )

				-- CP_START_SAV / CP_START_EAV
				local nv1 = v.hTotal - v.hSyncStart - 4  --May be 4, but 8 is good for position shift!
				local nv2 = v.hTotal - ( v.hSyncStart - v.hAct )

				self:WriteReg( USER_MAP, 0xa2, { ((nv1>>4)&0xff), ((nv1&0x0f)<<4) + ((nv2>>8)&0x0f), (nv2&0xff) }  )

				--CP_START_VBI/CP_END_VBI
				nv1 = v.vTotal - (v.vSyncStart - v.vAct) + 1
				nv2 = v.vTotal - v.vSyncStart + 1
				self:WriteReg( USER_MAP, 0xa5, { ((nv1>>4)&0xff), ((nv1&0x0f)<<4) + ((nv2>>8)&0x0f), (nv2&0xff) }  )

				--Progressive mode **TODO here.
				self:WriteReg( USER_MAP, 0x91, {0x10} )
			end
		end

		return true
	end

	------------------------------------------------------
	-- Configure Component format
	------------------------------------------------------

	function ADV7441A:_ConfigureComponentFormat(src)
		--TODO:
		return false
	end

	------------------------------------------------------
	-- Configure CVBS format
	------------------------------------------------------

	function ADV7441A:_ConfigureCVBSFormat(src)
		--TODO:
		return false
	end


	------------------------------------------------------

	--Now, initialize the I2C addresses
	local rslt,err = ADV7441A:ReadReg( USER_MAP, 0x11, {0xff} )
	if not rslt then
		print( "ERROR: Could not detect ADV7441A chip type, this may not work correctly! (" .. err .. ")" )
	else
		if rslt[1] == 0x01 then
			print( "ADV7441A: Detected chip ES1." )
		elseif rslt[1] == 0x02 then
			print( "ADV7441A: Detected chip ES2." )
		elseif rslt[1] == 0x04 then
			print( "ADV7441A: Detected chip ES3." )
		else
			print( "ADV7441A: Unknown chip type (" .. rslt[1] .. ")" )
		end

--		ADV7441A:WriteReg( USER_MAP, 0x0e, {USER_MAP2} )	--Configure USER_MAP2 address
--		periphery.sleep_ms( 10 )	--Wait 10ms
--		--Then configure other address

--		ADV7441A:WriteReg( USER_MAP2, 0xeb, { USER_MAP1} )
--		ADV7441A:WriteReg( USER_MAP2, 0xec, { VDP_MAP } )
--		ADV7441A:WriteReg( USER_MAP2, 0xed, { KSV_MAP } )
--		ADV7441A:WriteReg( USER_MAP2, 0xee, { EDID_MAP } )
--		ADV7441A:WriteReg( USER_MAP2, 0xef, { HDMI_MAP } )
--		ADV7441A:WriteReg( USER_MAP2, 0xea, { RSV_MAP } )

	end

	return ADV7441A
end --END of CreateADV7441A()

return CreateADV7441A

