-----------------------------------------------------------------------------------------------------
-- GV7601 chip controller.
-- Copyright (C) 2016, KILOVIEW Elecontrics Co., Ltd.
--
-- Author: 	Jacob Zuo, 2016/6/4
-----------------------------------------------------------------------------------------------------

local periphery = require "periphery"
local GPIO = periphery.GPIO
--For debug only.
--local pretty = require "pl.pretty"

require "gspi"

local sdi_formats = require "sdi_formats"
sdi_formats.Init()

------------------------------------------------------
-- The factory function to create a GV7601 handle object
------------------------------------------------------
local function CreateGV7601( gspiOptions )
	gspiOptions = gspiOptions or { cs=80, clk=77, dout=79, din=78, io_10b20b = 76, io_sdhd=74 }
	print( "Creating GV7601 control object..." )

	------------------------------------------------------
	-- Object
	------------------------------------------------------
	local GV7601 = {
		currentSource = "SDI",
		currentSignal = false,
		freeRun = false,
		currentFormat = nil,
		forceReassignFormat = false,
		lastFormatRecord = {},
		videoFormat = { _valid = false },
		audioFormat = { _valid = false },
		nostandardFormatInfo = {},
		gspi_handle = nil,
		io_10b20b = nil,
		io_sdhd = nil
	}

	-------------------------------------------------------
	-- GSPI openning
	-------------------------------------------------------
	GV7601.gspi_handle = GSPI.createNew( gspiOptions )
	if not GV7601.gspi_handle then
		print( "ERROR: Could not open GSPI handler for GV7601!" )
		return nil, "Fail open GSPI bus"
	end
	
	if gspiOptions.io_10b20b and gspiOptions.io_10b20b >= 0 then
		GV7601.io_10b20b = GPIO( gspiOptions.io_10b20b, "out" )
		if not GV7601.io_10b20b then
			print( "WARNING: Could not open IO control for 10bit/20bit" )
		end
	end
	
	if gspiOptions.io_sdhd and gspiOptions.io_sdhd >= 0 then
		GV7601.io_sdhd = GPIO( gspiOptions.io_sdhd, "out" )
		if not GV7601.io_sdhd then
			print( "WARNING: Could not open IO control for sd/hd loopback" )
		end
	end
	
	-------------------------------------------------------
	-- Object method: Register read
	-------------------------------------------------------
	function GV7601:ReadReg( addr, itemArr )
		if not self.gspi_handle then
			return nil, "Invalid GSPI handle"
		end
		return self.gspi_handle:Read( addr, itemArr )
	end

	------------------------------------------------------
	-- Register write
	------------------------------------------------------
	function GV7601:WriteReg( addr, itemArr )
		if not self.gspi_handle then
			return false, "Invalid GSPI handle"
		end
		return self.gspi_handle:Write( addr, itemArr )
	end

	------------------------------------------------------
	-- Soft reset the chip!
	------------------------------------------------------
	function GV7601:Reset()
		--Nothing
		self.currentSignal = false
		self.freeRun = false
		self.currentFormat = nil
		self.forceReassignFormat = false
		self.lastFormatRecord = {}		--Clear recorded format detection informations.
		self.videoFormat = {_valid = false}
		
		--TODO: Now fixed it.
		self.audioFormat = { 
			_valid = true,
			audio_src = "SDI",
			sampling = 48000, 
			audio_signal_type = 1, --AKA SDI
			channels = 2,
			bit_width = 16
		}

		self:WriteReg( 0x39, {0x0001} ) --Configure audio fs to 256fs.
		self:WriteReg( 0x408, {0x00ff} ) --SD audio core: 24bits/i2s
		self:WriteReg( 0x200, {0x00f4} ) --HD audio core
	end

	------------------------------------------------------
	-- Init the GV7601 interrupts 
	------------------------------------------------------
	
	function GV7601:_InitInterrupts()
		--TODO: Nothing yet.
	end
	
	------------------------------------------------------
	-- Check SDI signal and type 
	------------------------------------------------------
	
	function GV7601:_CheckSignalAndType()
		local rslt = self:ReadReg( 0x022, { 0 } )
		if not rslt then
			self.currentSignal = false
			return false, nil
		else
			local signal = ((rslt[1]&0x1000)~=0)
			if not signal then
				return false, nil
			end
			
			local vMode = (rslt[1]>>14) & 0x3
			local viRate = sdi_formats.VIRATE_SD
			if vMode == 0 then
				viRate = sdi_formats.VIRATE_HD
			elseif vMode == 1 or vMode == 3 then
				viRate = sdi_formats.VIRATE_SD
			else
				viRate = sdi_formats.VIRATE_3G
			end
			
			if viRate == sdi_formats.VIRATE_SD then
				if self.io_10b20b then
					self.io_10b20b:write( false )		--Low for 10bit
				end
			
				if self.io_sdhd then
					self.io_sdhd:write( true )		--HIGH for SD
				end
			else --HD or 3G
				if self.io_10b20b then
					self.io_10b20b:write( true )		--High for 20bit
				end
			
				if self.io_sdhd then
					self.io_sdhd:write( false )		--Low for HD
				end
			end
			return true, viRate
		end
	end
	
	------------------------------------------------------
	
	function GV7601:ForceUpdate()
		self.lastFormatRecord = {}		--Clear the recorded format information.
	end

	------------------------------------------------------
	-- A step handler for GV7601 processing.
	-- To fit the LUA working mode, you need call the 
	-- 'Step' periodically ( suggest per 200~500ms).
	-- This routine should do :
	--	* Signal status checking
	--	*	Auto format detecting
	--	*	Configue the chip to make it output BT.656/BT.601 signals
	--
	------------------------------------------------------
	
	function GV7601:Step()
		local oldSignal, viRate
		oldSignal = self.currentSignal
		self.currentSignal, viRate = self:_CheckSignalAndType()
		
		if oldSignal ~= self.currentSignal then
			print( "SDI INFO: SIGNAL=" .. (self.currentSignal and "<locked>" or "<unlocked>") .. ", Rate=" .. (viRate or "<unknown>") )
		end

		if self.currentSignal then
			self.videoFormat.signal_locked = true	--Assume it be, but when detecting, it may changed
			self:_DetectCurrentFormat()
		else
			self.videoFormat.signal_locked = false
		end
	end


	------------------------------------------------------
	-- Detect current signal format
	------------------------------------------------------
	
	function GV7601:_DetectCurrentFormat(src)
		local rslt = self:ReadReg( 0x006, {0} )
		if not rslt then
			return false
		end
		
		local std_id = (rslt[1] >> 8) & 0x3f
		rslt = self:ReadReg( 0x1f, {0, 0, 0, 0} )
		local ntsc = ((rslt[4]&0x2000) ~= 0)
		local interlace = ((rslt[4]&0x800) ~= 0)
		local vAct = rslt[4]&0x7ff
		local hAct = rslt[1]&0x3fff
		local hTotal = rslt[2]&0x3fff
		local vTotal = rslt[3]&0x7ff
		
		if std_id ~= self.lastFormatRecord.std_id or
			ntsc ~= self.lastFormatRecord.ntsc or
			interlace ~= self.lastFormatRecord.interlace or
			vAct ~= self.lastFormatRecord.vAct or
			hAct ~= self.lastFormatRecord.hAct or
			vTotal ~= self.lastFormatRecord.vTotal or
			hTotal ~= self.lastFormatRecord.hTotal then
			
			print( "SDI INFO: Format STD-ID = " .. std_id )
			print( "Details:", hTotal, hAct, vTotal, vAct, interlace, ntsc )

			local formatInfo = sdi_formats.Standard_Match( std_id, hTotal, hAct, vTotal, vAct, interlace, ntsc )
			if formatInfo then
				self.lastFormatRecord.std_id = std_id
				self.lastFormatRecord.ntsc = ntsc
				self.lastFormatRecord.interlace = interlace
				self.lastFormatRecord.vAct = vAct
				self.lastFormatRecord.vTotal = vTotal
				self.lastFormatRecord.hAct = hAct
				self.lastFormatRecord.hTotal = hTotal
				
				print( "Matched format: ", formatInfo.formatId )
				self.videoFormat._valid = true
				self:_RecordVideoFormatInfo(src, formatInfo)
				return true
			else
				print( "*** Oh! No matched format for it, what's wrong?" )
			end
		else
			self.videoFormat._valid = true
			return true
		end
		
		self.videoFormat._valid = false
		--TODO: Other format checking but now not support
		return false
	end
	
	------------------------------------------------------
	-- According detected format information, record it to Amba based formation context.
	------------------------------------------------------
	
	function GV7601:_RecordVideoFormatInfo(src, formatInfo)
		if type(formatInfo) ~= "table" then
			return false
		end

		--pretty.dump( formatInfo )
		
		if self.videoFormat._formatInfo ~= formatInfo or (src and src ~= self.videoFormat.video_src) then
			self.videoFormat._lastFormatChanged = true
			self.videoFormat._formatInfo = formatInfo
			
			if src then
				self.videoFormat.video_src = src
			end
			
			--TODO:
			self.videoFormat.cap_start_x = 0
			self.videoFormat.cap_start_y = 0
			
			self.videoFormat.cap_cap_w = formatInfo.hAct
			self.videoFormat.line_width = formatInfo.hAct
			
			self.videoFormat.total_width = formatInfo.hTotal
			
			if formatInfo.vAct then
				self.videoFormat.cap_cap_h = formatInfo.vAct
				self.videoFormat.total_height = formatInfo.vTotal
				
				if formatInfo.interlaced then
					self.videoFormat.f0_height = formatInfo.vAct/2
					self.videoFormat.f1_height = formatInfo.vAct/2
				else
					self.videoFormat.f0_height = formatInfo.vAct
					self.videoFormat.f1_height = formatInfo.vAct
				end
			elseif formatInfo.vAct0 then
				self.videoFormat.f0_height = formatInfo.vAct0
				if formatInfo.interlaced then
					if formatInfo.vAct1 then
						self.videoFormat.f1_height = formatInfo.vAct1
						self.videoFormat.cap_cap_h = formatInfo.vAct0 + formatInfo.vAct1
						if self.videoFormat.cap_cap_h >= 486 and self.videoFormat.cap_cap_h <= 488 then
							self.videoFormat.cap_cap_h = 480	--This is a special format
						else
							self.videoFormat.cap_cap_h = self.videoFormat.cap_cap_h & 0xfffffffe
						end
						self.videoFormat.total_height = self.videoFormat.cap_cap_h	--??? total_height for interlace?
					else
						self.videoFormat.f1_height = formatInfo.vAct0
						self.videoFormat.cap_cap_h = formatInfo.vAct0 * 2
						self.videoFormat.total_height = formatInfo.vAct0 * 2
					end
				else
					self.videoFormat.f1_height = formatInfo.vAct0
					self.videoFormat.cap_cap_h = formatInfo.vAct0
					self.videoFormat.total_height = formatInfo.vTotal0
				end
			else
				self.videoFormat.f0_height = 0
				self.videoFormat.f1_height = 0
				self.videoFormat.cap_cap_h = 0	--This is wrong.
				self.videoFormat.total_height  = 0
			end
			
			--TODO:
			self.videoFormat.sync_start = 0
			
			if formatInfo.video_system then
				self.videoFormat.video_system = formatInfo.video_system
			else
				self.videoFormat.video_system = _G.AMBA_VIDEO_SYSTEM_AUTO or 0x0
			end
			
			if AMBA_VIDEO_FPS then
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS(formatInfo.frameRate)
			else
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS_60 or formatInfo.frameRate
			end
			
			self.videoFormat.frame_rate_value = formatInfo.frameRate
			
			if formatInfo.aspect_ratio then
				self.videoFormat.aspect_ratio = formatInfo.aspect_ratio
			else
				self.videoFormat.aspect_ratio = _G.AMBA_VIDEO_RATIO_AUTO or 0
			end
			
			--TODO: Here, need fix this to support 8 bits/ BT.601 options.
			self.videoFormat.input_type = _G.AMBA_VIDEO_TYPE_YUV_656 or 2
			self.videoFormat.bit_resolution = 16
			if formatInfo.interlaced then
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_INTERLACE or 1
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_INTLC or 1
				self.videoFormat.interlaced = 1
			else
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_PROGRESSIVE or 2
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_PROG or 2
				self.videoFormat.interlaced = 0
			end
			
			--TODO:
			self.videoFormat.hs_front_porch = 0
			self.videoFormat.hs_pulse_width = 0
			self.videoFormat.hs_back_porch = 0
			--TODO:
			self.videoFormat.f0_vs_front_porch = 0
			self.videoFormat.f0_vs_pulse_width = 0
			self.videoFormat.f0_vs_back_porch = 0
			--TODO:
			self.videoFormat.f1_vs_front_porch = 0
			self.videoFormat.f1_vs_pulse_width = 0
			self.videoFormat.f1_vs_back_porch = 0
		
			if formatInfo.tmdsFreq and formatInfo.tmdsFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.tmdsFreq 
			elseif formatInfo.pixelFreq and formatInfo.pixelFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.pixelFreq
			else
				self.videoFormat.pixel_freq = 0
			end
			
			self.videoFormat.format_code = formatInfo.formatId	-- Make the formatId as format-code.
			self.videoFormat.format_name = formatInfo.formatName or (self.videoFormat.cap_cap_w .. "x" .. self.videoFormat.cap_cap_h)
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_SDI or 2
			return true
		else
			return true	--No changed.
		end
	end

	------------------------------------------------------
	-- Detect SDI Audio
	------------------------------------------------------
	function GV7601:_DetectAudio(src)
		--TODO:
	end

	------------------------------------------------------
	-- Assume SDI customized format
	------------------------------------------------------
	function GV7601:_AssumeCustomFormat(src, BL, LCF, FCL, LCVS, interlaced)
		return false	--TODO
	end

	--GV7601:_Reset()
	
	return GV7601
end --END of CreateGV7601()

return CreateGV7601
