-----------------------------------------------------------------------------------------------------
-- ITE66021 chip controller.
-- Copyright (C) 2016, KILOVIEW Elecontrics Co., Ltd.
--
-- Author: 	Jacob Zuo, 2019/3/18
-----------------------------------------------------------------------------------------------------

local coaf=require "coaf.core"
local _dbg = require "coaf.debug"

local periphery = require "periphery"
local I2C_REG = require "chips.i2c_reg"
--For debug only.
--local pretty = require "pl.pretty"

local MAIN_FREQ = 27000000
local CP_MEAN_VALUE = 48696
local CP_MAX_DEVIATION = (CP_MEAN_VALUE*15/100)
local m_ROSCCLK = CP_MEAN_VALUE/4
local hdtv_formats = require "chips.hdtv_formats"

hdtv_formats.SetMainFreq( MAIN_FREQ )

--Init both HDTV and VGA format standards
hdtv_formats.Init( "STDI" )

------------------------------------------------------
-- The factory function to create a ITE66021 handle object
------------------------------------------------------
local function CreateITE66021( i2c_dev, rxAddr )
	------------------------------------------------------
	-- Object
	------------------------------------------------------
	local ITE66021 = {
		currentSource = "HDMI",
		currentSignal = false,
		freeRun = false,
		currentFormat = nil,
		forceReassignFormat = false,
		lastFormatRecord = {},
		videoinfo={pix_saved=0,pixclk_cnt=0,pixclk_last=0},
		videoFormat = { _valid = false },
		audioFormat = { _valid = false, audio_src = "HDMI", audio_signal_type = 0 }, --type 0 aka. HDMI
		orgAudioFormat = { _valid = false },
		nostandardFormatInfo = {},
		i2c_reg = nil,
	}

	------------------------------------------------------
	i2c_dev = i2c_dev or "/dev/i2c-0"
	
	local BASE_MAP = rxAddr or 0x48
	local EDID_MAP = 0x54
	local MHL_ADDR = 0X60

	_dbg.Info("ITE66021", "Creating ITE66021 control object at 0x" .. string.format( "%02x", BASE_MAP) .. "@" .. i2c_dev .. "..." )

	ITE66021.i2c_reg = I2C_REG.Create()
	if not ITE66021.i2c_reg:Open( i2c_dev ) then
		return nil, "Fail to open I2C device"
	end

	-------------------------------------------------------
	-- Object method: Register read
	-------------------------------------------------------
	function ITE66021:ReadReg( map, addr, itemArr )
		return self.i2c_reg:ReadReg(map, addr, itemArr)
	end

	------------------------------------------------------
	-- Register write
	------------------------------------------------------
	function ITE66021:WriteReg( map, addr, itemArr )
		return self.i2c_reg:WriteReg(map, addr, itemArr)
	end

	------------------------------------------------------
	
	function ITE66021:ForceUpdate()
		self.lastFormatRecord = {}		--Clear the recorded format information.
	end


	------------------------------------------------------
	-- Soft reset the chip!
	------------------------------------------------------
	function ITE66021:Reset()
		self:WriteReg( BASE_MAP, 0x67, {0x81} )
		self:WriteReg( BASE_MAP, 0x58, {0xaa} )
		self:WriteReg( BASE_MAP, 0x65, {0x50} )
		self:WriteReg( BASE_MAP, 0x77, {0x80} )
		self:WriteReg( BASE_MAP, 0x7D, {0x00} )
		self:WriteReg( BASE_MAP, 0x0F, {0x01} )
		self:WriteReg( BASE_MAP, 0x70, {0x00} )
		self:WriteReg( BASE_MAP, 0x71, {0x80} )
		self:WriteReg( BASE_MAP, 0x72, {0x10} )
		self:WriteReg( BASE_MAP, 0x73, {0xB2} )
		self:WriteReg( BASE_MAP, 0x74, {0x04} )
		self:WriteReg( BASE_MAP, 0x75, {0x65} )
		self:WriteReg( BASE_MAP, 0x76, {0x02} )
		self:WriteReg( BASE_MAP, 0x77, {0xE9} )
		self:WriteReg( BASE_MAP, 0x78, {0x00} )
		self:WriteReg( BASE_MAP, 0x79, {0x93} )
		self:WriteReg( BASE_MAP, 0x7A, {0x3C} )
		self:WriteReg( BASE_MAP, 0x7B, {0x18} )
		self:WriteReg( BASE_MAP, 0x7C, {0x04} )
		self:WriteReg( BASE_MAP, 0x7D, {0x55} )
		self:WriteReg( BASE_MAP, 0x7E, {0x3F} )
		self:WriteReg( BASE_MAP, 0x7F, {0x49} )
		self:WriteReg( BASE_MAP, 0x80, {0x3D} )
		self:WriteReg( BASE_MAP, 0x81, {0x9F} )
		self:WriteReg( BASE_MAP, 0x82, {0x3E} )
		self:WriteReg( BASE_MAP, 0x83, {0x18} )
		self:WriteReg( BASE_MAP, 0x84, {0x04} )
		self:WriteReg( BASE_MAP, 0xB0, {0x03} ) -- config hpd out
		self:WriteReg( BASE_MAP, 0x0F, {0x00} )
		self:WriteReg( BASE_MAP, 0x87, {0xA9} ) -- config edid addr 0x54
		self:WriteReg( BASE_MAP, 0x34, {0xC1} ) -- config MHL_ADDR 0X60
		self:WriteReg( BASE_MAP, 0x54, {0x11} ) -- config RCLKFreqSel=01 and audio fs = 256
		self:WriteReg( BASE_MAP, 0xC0, {0x04} )
		self:WriteReg( BASE_MAP, 0x10, {0x30} ) --add Enable auto video reset
		self:WriteReg( BASE_MAP, 0x11, {0xa0} )	--add Enable auto video reset	
			
		
		self:_Cal_oclk()
		self:_InitInterrupts()
		self:_InitEDID()	--Init EDID
		
		self:_ConfigureHDMIFormat()	--Force default blue
		
		self.currentSignal = false
		self.freeRun = false
		self.currentFormat = nil
		self.forceReassignFormat = false

		self.lastFormatRecord = {}		--Clear recorded format detection informations.
		self.audioFormat = { _valid = false, audio_src = "HDMI", audio_signal_type = 0 }
		self.orgAudioFormat = { _valid = false }
		self.videoFormat = {_valid = false}
	end
	------------------------------------------------------
	-- ITE66021 Calibration Ring Oscillator
	------------------------------------------------------	
	function ITE66021:_Cal_oclk()
		local CP_100ms_00B
		local CP_100ms_01B
		local CP_100ms_02B
		local OSCCLK
		local RCLKFreqSel = 1
		self:WriteReg( BASE_MAP, 0x0F, {0x00} )
		self:WriteReg( BASE_MAP, 0x73, {0x05} )
		self:WriteReg( BASE_MAP, 0x0F, {0x01} )
		self:WriteReg( BASE_MAP, 0x57, {0x12} )
		self:WriteReg( BASE_MAP, 0x57, {0x02} )
		self:WriteReg( BASE_MAP, 0x56, {0x01} )

		-- set Read Address and enable
		self:WriteReg( BASE_MAP, 0x50, {0x00} )
		self:WriteReg( BASE_MAP, 0x51, {0x00} )
		self:WriteReg( BASE_MAP, 0x53, {0x04} )	

		
		local SIP_00B = self:ReadReg( BASE_MAP, 0x61, {0x00} )
		local SIP_01B = self:ReadReg( BASE_MAP, 0x61, {0x00} )
		local SIP_02B = self:ReadReg( BASE_MAP, 0x61, {0x00} )

		if SIP_00B[1]==0x02 and SIP_01B[1]==0x02 and SIP_02B[1] ==0x02 	then
			self:WriteReg( BASE_MAP, 0x50, {0x01} )
			self:WriteReg( BASE_MAP, 0x51, {0x60} )
			self:WriteReg( BASE_MAP, 0x53, {0x04} )
			CP_100ms_00B=self:ReadReg( BASE_MAP, 0x61, {0x00} )
			CP_100ms_01B=self:ReadReg( BASE_MAP, 0x61, {0x00} )
			CP_100ms_02B=self:ReadReg( BASE_MAP, 0x61, {0x00} )
			OSCCLK =CP_100ms_00B[1] + (CP_100ms_01B[1]<<8) + (CP_100ms_02B[1]<<16)
		elseif SIP_00B[1]==0xFF and SIP_01B[1] ==0x00 and SIP_02B[1]==0xFF then
			self:WriteReg( BASE_MAP, 0x50, {0x03} )
			self:WriteReg( BASE_MAP, 0x51, {0x60} )
			self:WriteReg( BASE_MAP, 0x53, {0x04} )
			CP_100ms_00B=self:ReadReg( BASE_MAP, 0x61, {0x00} )
			CP_100ms_01B=self:ReadReg( BASE_MAP, 0x61, {0x00} )
			CP_100ms_02B=self:ReadReg( BASE_MAP, 0x61, {0x00} )
			OSCCLK =CP_100ms_00B[1] + (CP_100ms_01B[1]<<8) + (CP_100ms_02B[1]<<16)
		else
			OSCCLK = 0;
		end
		self:WriteReg( BASE_MAP, 0x0F, {0x00} )
		self:WriteReg( BASE_MAP, 0x73, {0x00} )
		
		local diff_cp
		local cp_cal_mean = CP_MEAN_VALUE

		if OSCCLK ~= 0 then
			OSCCLK=OSCCLK/100
		end
		if OSCCLK >= cp_cal_mean then
			diff_cp = OSCCLK-cp_cal_mean
		else
			diff_cp = cp_cal_mean - OSCCLK
		end
		if diff_cp > CP_MAX_DEVIATION then
			OSCCLK = CP_MEAN_VALUE
		end		

		local oscdiv = (OSCCLK/1000/10)

		if ((OSCCLK/1000/oscdiv)-10)>(10-(OSCCLK/1000/(oscdiv+1))) then
			oscdiv= oscdiv + 1
		end
		oscdiv = math.ceil(oscdiv)  --math.ceil math.floor
		oscdiv	= (oscdiv<<4)&0x0F0
		
		local temp = self:ReadReg(MHL_ADDR,0X01)
		temp[1] = temp[1]&0x8f + oscdiv&0x70	
		self:WriteReg( MHL_ADDR, 0x01, {temp[1]} )	

		temp = self:ReadReg(MHL_ADDR,0X29)
		temp[1] = temp[1]&0x7c + 0x83
		self:WriteReg( MHL_ADDR, 0x29, {temp[1]} )

		if RCLKFreqSel==0 then 
			m_ROSCCLK = math.ceil(OSCCLK/2)   --// 20 MHz
		elseif RCLKFreqSel==1	then 
			m_ROSCCLK = math.ceil(OSCCLK/4)   --// 10 MHz
		elseif( RCLKFreqSel==2 ) then 
			m_ROSCCLK = math.ceil(OSCCLK/8)
		else
			m_ROSCCLK = math.ceil(OSCCLK/16)
		end
		local t10usint = math.ceil(m_ROSCCLK/100)
		local t10usflt = math.ceil((m_ROSCCLK/100 - t10usint)*128)
		self:WriteReg( MHL_ADDR, 0x02, {t10usint&0xFF} )
		self:WriteReg( MHL_ADDR, 0x03, {((t10usint&0x100)>>1)+t10usflt} )
	end	
	------------------------------------------------------
	-- Init the ITE66021 interrupts 
	------------------------------------------------------
	
	function ITE66021:_InitInterrupts()
		--TODO: Nothing yet.
	end
	
	------------------------------------------------------
	--EDID configuration
	------------------------------------------------------
	function ITE66021:_InitEDID()
--		local edid 
--		if _G.DISABLE_4Kp60 then
--			edid = require "chips.edid"	--EDID
--		else
--			edid = require "chips.edid_uhd60"
--		end
--		self:WriteReg( BASE_MAP, 0x0F, {0x01} )
--		self:WriteReg( BASE_MAP, 0xB0, {0x01} ) 
--		self:WriteReg( BASE_MAP, 0x0F, {0x00} )	
--
--		local i
--		for i=1, #edid, 8 do
--			self:WriteReg( EDID_MAP, i-1, { edid[i], edid[i+1], edid[i+2], edid[i+3], edid[i+4], edid[i+5], edid[i+6], edid[i+7]} )
--		end
--		self:WriteReg( BASE_MAP, 0xc1, {0x9e} )
--		self:WriteReg( BASE_MAP, 0xc2, {0x10} )
--		self:WriteReg( BASE_MAP, 0xC3, {0x00} )
--		self:WriteReg( BASE_MAP, 0xC4, {edid[128]} )
--		self:WriteReg( BASE_MAP, 0xC5, {edid[256]} )	
--		self:WriteReg( BASE_MAP, 0x0F, {0x01} )
--		self:WriteReg( BASE_MAP, 0xB0, {0x03} ) 
--		self:WriteReg( BASE_MAP, 0x0F, {0x00} )	
--		periphery.sleep_ms(5) --Wait 5ms for Calc CRC
	end
	
	function ITE66021:_CheckSignal(src)
		local freeRun = false
		local rslt = self:ReadReg( BASE_MAP, 0x0a, {0x00} )
		if rslt then
			if (rslt[1] & 0xad) == 0xad then
				rslt = self:ReadReg( BASE_MAP, 0x99, {0x00} )
				if rslt and (rslt[1]&0x08) ~= 0 then
					return true, freeRun
				else
					return false, freeRun
				end
			else
				return false, freeRun
			end
		else
			return false, freeRun
		end
	end

	------------------------------------------------------
	-- A step handler for ITE66021 processing.
	-- To fit the LUA working mode, you need call the 
	-- 'Step' periodically ( suggest per 200~500ms).
	-- This routine should do :
	--	* Signal status checking
	--	*	Auto format detecting
	--	*	Configue the chip to make it output BT.656/BT.601 signals
	--
	------------------------------------------------------
	
	function ITE66021:Step()
		self.currentSignal, self.freeRun = self:_CheckSignal()	--Check current signal

		--TODO: consider free-run but now not support.
		if self.currentSignal ~= self.videoFormat.signal_locked then
			if not self.videoFormat.version then --Keep track the updated version
				self.videoFormat.version = 0
			else
				self.videoFormat.version = self.videoFormat.version + 1
			end
		end

		if not self.audioFormat.version then
			self.audioFormat.version = 0
		end

		if self.currentSignal then
			self.videoFormat.signal_locked = true	--Assume it be, but when detecting, it may changed
			self.videoFormat.free_run = false
			self:_DetectCurrentFormat()
			self:_ConfigureCurrentFormat()
		elseif self.freeRun then
			self.videoFormat.signal_locked = true
			self.videoFormat.free_run = true
		else
			self.videoFormat.signal_locked = false
			self.videoFormat.free_run = false
		end
	end


	------------------------------------------------------
	-- Detect current signal format
	------------------------------------------------------
	
	function ITE66021:_DetectCurrentFormat(src)
		self:_DetectHDMIAudio(src)
		self:_UpdateAudioFormatInfo()
		return self:_DetectHDMIFormat(src)
	end
	
	function ITE66021:_UpdateAudioFormatInfo()
		if self.audioFormat._valid ~= self.orgAudioFormat.withAudio or
			self.audioFormat.sampling ~= self.orgAudioFormat.sample then

			self.audioFormat.version = self.audioFormat.version + 1

			self.audioFormat._valid = self.orgAudioFormat.withAudio
			self.audioFormat.fs = self.orgAudioFormat.fs
			self.audioFormat.sampling = self.orgAudioFormat.sample
			self.audioFormat.channels = 2
			self.audioFormat.bit_width = 16
		end
	end

	------------------------------------------------------
	-- Configure signal format
	------------------------------------------------------
	
	function ITE66021:_ConfigureCurrentFormat(src)
		if not self.videoFormat._lastFormatChanged and not self.forceReassignFormat then
			return true
		end
		
		return self:_ConfigureHDMIFormat(src)
	end
	
	------------------------------------------------------
	-- According detected format information, record it to Amba based formation context.
	------------------------------------------------------
	
	function ITE66021:_RecordVideoFormatInfo(src, formatInfo)
		if type(formatInfo) ~= "table" then
			return false
		end
		--pretty.dump( formatInfo )
		
		if self.videoFormat._formatInfo ~= formatInfo or (src and src ~= self.videoFormat.video_src) then
			if not self.videoFormat.version then --Keep track the updated version
				self.videoFormat.version = 0
			else
				self.videoFormat.version = self.videoFormat.version + 1
			end

			self.videoFormat._lastFormatChanged = true
			self.videoFormat._formatInfo = formatInfo
			
			if src then
				self.videoFormat.video_src = src
			end
			self.videoFormat.cap_start_x = 0
			self.videoFormat.cap_start_y = 0
			
			self.videoFormat.cap_cap_w = formatInfo.hAct
			self.videoFormat.line_width = formatInfo.hAct
			
			self.videoFormat.total_width = formatInfo.hTotal
			
			if formatInfo.vAct then
				self.videoFormat.cap_cap_h = formatInfo.vAct
				self.videoFormat.total_height = formatInfo.vTotal
				
				if formatInfo.interlaced then
					self.videoFormat.f0_height = formatInfo.vAct/2
					self.videoFormat.f1_height = formatInfo.vAct/2
				else
					self.videoFormat.f0_height = formatInfo.vAct
					self.videoFormat.f1_height = formatInfo.vAct
				end
			elseif formatInfo.vAct0 then
				self.videoFormat.f0_height = formatInfo.vAct0
				if formatInfo.interlaced then
					if formatInfo.vAct1 then
						self.videoFormat.f1_height = formatInfo.vAct1
						self.videoFormat.cap_cap_h = formatInfo.vAct0 + formatInfo.vAct1
						if self.videoFormat.cap_cap_h >= 486 and self.videoFormat.cap_cap_h <= 488 then
							self.videoFormat.cap_cap_h = 480	--This is a special format
						else
							self.videoFormat.cap_cap_h = self.videoFormat.cap_cap_h & 0xfffffffe
						end
						self.videoFormat.total_height = self.videoFormat.cap_cap_h	--??? total_height for interlace?
					else
						self.videoFormat.f1_height = formatInfo.vAct0
						self.videoFormat.cap_cap_h = formatInfo.vAct0 * 2
						self.videoFormat.total_height = formatInfo.vAct0 * 2
					end
				else
					self.videoFormat.f1_height = formatInfo.vAct0
					self.videoFormat.cap_cap_h = formatInfo.vAct0
					self.videoFormat.total_height = formatInfo.vTotal0
				end
			else
				self.videoFormat.f0_height = 0
				self.videoFormat.f1_height = 0
				self.videoFormat.cap_cap_h = 0	--This is wrong.
				self.videoFormat.total_height  = 0
			end
			
			--TODO:
			self.videoFormat.sync_start = 0
			
			if formatInfo.video_system then
				self.videoFormat.video_system = formatInfo.video_system
			else
				self.videoFormat.video_system = _G.AMBA_VIDEO_SYSTEM_AUTO or 0x0
			end
			
			if AMBA_VIDEO_FPS then
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS(formatInfo.frameRate)
			else
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS_60 or formatInfo.frameRate
			end
			
			self.videoFormat.frame_rate_value = formatInfo.frameRate
			
			if formatInfo.aspect_ratio then
				self.videoFormat.aspect_ratio = formatInfo.aspect_ratio
			else
				self.videoFormat.aspect_ratio = _G.AMBA_VIDEO_RATIO_AUTO or 0
			end
			
			--TODO: Here, need fix this to support 8 bits/ BT.601 options.
			self.videoFormat.input_type = _G.AMBA_VIDEO_TYPE_YUV_656 or 2
			self.videoFormat.bit_resolution = 16
			if formatInfo.interlaced then
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_INTERLACE or 1
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_INTLC or 1
				self.videoFormat.interlaced = 1
			else
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_PROGRESSIVE or 2
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_PROG or 2
				self.videoFormat.interlaced = 0
			end
			
			--TODO:
			self.videoFormat.hs_front_porch = 0
			self.videoFormat.hs_pulse_width = 0
			self.videoFormat.hs_back_porch = 0
			--TODO:
			self.videoFormat.f0_vs_front_porch = 0
			self.videoFormat.f0_vs_pulse_width = 0
			self.videoFormat.f0_vs_back_porch = 0
			--TODO:
			self.videoFormat.f1_vs_front_porch = 0
			self.videoFormat.f1_vs_pulse_width = 0
			self.videoFormat.f1_vs_back_porch = 0
		
			if formatInfo.tmdsFreq and formatInfo.tmdsFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.tmdsFreq 
			elseif formatInfo.pixelFreq and formatInfo.pixelFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.pixelFreq
			else
				self.videoFormat.pixel_freq = 0
			end
			
			self.videoFormat.format_code = formatInfo.formatId	-- Make the formatId as format-code.
			self.videoFormat.format_name = formatInfo.formatName or (formatInfo.hAct .. "x" .. formatInfo.vAct0)
			if formatInfo.rateMode == hdtv_formats.VIRATE_4K then
				self.videoFormat.video_mode = "4k"
			elseif formatInfo.rateMode == hdtv_formats.VIRATE_4K60 then
				self.videoFormat.video_mode = "4k60"
			elseif formatInfo.rateMode == hdtv_formats.VIRATE_SD then
				self.videoFormat.video_mode = "sd"
			else
				self.videoFormat.video_mode = "hd"
			end

			return true
		else
			return true	--No changed.
		end
	end

	------------------------------------------------------
	-- Detect HDMI Audio
	------------------------------------------------------
	function ITE66021:_DetectHDMIAudio(src)
		self.orgAudioFormat.withAudio = false
		local rslt
		
		local rslt = self:ReadReg( BASE_MAP, 0x0a, {0x00} )
		if rslt then
			if (rslt[1] & 0x2c) == 0x2c then
				self.orgAudioFormat.pll_locked = true
			else
				self.orgAudioFormat.pll_locked = false
				return false
			end
		else
			self.orgAudioFormat.pll_locked = false
			return false
		end

		if (rslt[1] & 0x02) == 0 then --DVI mode
			return false
		end

		rslt = self:ReadReg( BASE_MAP, 0xaa, {0x00} )
		if not rslt or (rslt[1]&0x80) == 0 then --Audio ON flag is not set.
			return false
		end

		rslt = self:ReadReg( BASE_MAP, 0xae, {0x00,0x00} )
		if not rslt then
			return false
		end

		local sampleFlag = rslt[1] & 0x0f
		if sampleFlag == 0 then
			self.orgAudioFormat.sample = 44100
		elseif sampleFlag == 2 then
			self.orgAudioFormat.sample = 48000
		elseif sampleFlag == 3 then
			self.orgAudioFormat.sample = 32000
		elseif sampleFlag == 8 then
			self.orgAudioFormat.sample = 88200
		elseif sampleFlag == 9 then
			self.orgAudioFormat.sample = 768000
		elseif sampleFlag == 10 then
			self.orgAudioFormat.sample = 96000
		elseif sampleFlag == 12 then
			self.orgAudioFormat.sample = 176000
		elseif sampleFlag == 14 then
			self.orgAudioFormat.sample = 192000
		else
			self.orgAudioFormat.sample = 48000
		end	

		self.orgAudioFormat.withAudio = true
		--TODO: Check the fs type from register.
		self.orgAudioFormat.fs = 256 --(rslt[1]+1)*128
		return true
	end

	------------------------------------------------------
	-- Detect HDMI AVI info frame
	------------------------------------------------------
	
	function ITE66021:_DetectHDMIInfoFrame(src)
		--TODO: 从66021的寄存器中读取 AVI InfoFrame 中视频格式的CEA_ID
		--self.videoFormat.cea_id = 0 --TODO:
		--Check the color-space:
		--self:WriteReg( BASE_MAP, 0x0f, {0x02} ) --Read 2xx regs.
		--self.videoFormat.avi_infor = self:ReadReg( BASE_MAP, 0x12, {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0} ) --reg 0x215
		--self:WriteReg( BASE_MAP, 0x0f, {0x00} ) --Reset to 0xx regs.
		--for i=1, 18, 1 do
		--	print(self.videoFormat.avi_infor[i])
		--end			
		--return true
		local cea_id = 255
		--Check the cea_id:
		self:WriteReg( BASE_MAP, 0x0f, {0x02} ) --Read 2xx regs.
		local rslt = self:ReadReg( BASE_MAP, 0x18, {0x00} ) --reg 0x218
		if rslt and rslt[1] then
			cea_id = rslt[1] & 0x7f
		end
		self:WriteReg( BASE_MAP, 0x0f, {0x00} ) --Reset to 0xx regs.

		if cea_id ~= self.videoFormat.cea_id then
			self.videoFormat.cea_id = cea_id
			self.videoFormat._lastFormatChanged = true
		end	
		return true	
	end
	
	------------------------------------------------------
	-- Get format
	------------------------------------------------------
	
	function ITE66021:_GetVideoSyncDetail()
		local interlaced = false	
		local diff_pixclk
		local pxclk = 0
		local rslt = self:ReadReg( BASE_MAP, 0x99, {0x00} )
		if rslt and (rslt[1]&0x02) ~= 0 then
			interlaced = true
		end

		rslt = self:ReadReg( BASE_MAP, 0x9a, {0x00} )
		if not rslt or rslt[1] == 0 then
			return nil
		end
		--local pxclk = math.floor(MAIN_FREQ/rslt[1] * 255)
		--local pixclk_temp = math.ceil(m_ROSCCLK/rslt[1] * 255 * 1000)
		local pixclk_temp = math.ceil(m_ROSCCLK/rslt[1] * 255 * 9975 /10)
		rslt = self:ReadReg( BASE_MAP, 0x9c, {0,0,0,0,0,0,0,0,0,0,0,0,0,0} )
		if not rslt then
			return nil
		end

		local HTotal = ((rslt[2] & 0x3F) << 8) + rslt[1]
		local HActive = ((rslt[4] & 0x3F) << 8) + rslt[3]
		--local HSyncWidth = ((rslt[6] & 0x01) << 8) + rslt[5]
		--local HFrontPorch = ((rslt[6] & 0xF0) << 4) + rslt[7]
		--local HBackPorch = HTotal - HActive - HFrontPorch - HSyncWidth
		local VTotal = ((rslt[9] & 0x0F) << 8) + rslt[8]
		local VActive = ((rslt[9] & 0xF0) << 4) + rslt[10]
		--local VSyncWidth = rslt[11] & 0x1F
		--local VFrontPorch = rslt[12] & 0x3F
		--local VBackPorch = VTotal - VActive - VFrontPorch - VSyncWidth
		--local VPolarity = (rslt[13] & 0x08) >> 3
		--local HPolarity = (rslt[13] & 0x04) >> 2

		if self.videoinfo.pixclk_last == 0 and self.videoinfo.pixclk_saved == 0 then 
			self.videoinfo.pixclk_last = pixclk_temp
			self.videoinfo.pixclk_saved = pixclk_temp
		end

		if self.videoinfo.HTotal ~= HTotal or self.videoinfo.HActive ~= HActive or self.videoinfo.VTotal ~= VTotal
			or self.videoinfo.VActive ~= VActive or self.videoinfo.interlaced ~= interlaced then 
		 	self.videoinfo.pixclk_last = pixclk_temp
		 	self.videoinfo.pixclk_saved = pixclk_temp
	 	end		

		if self.videoinfo.pixclk_last >= pixclk_temp then
			diff_pixclk = self.videoinfo.pixclk_last - pixclk_temp
		else
			diff_pixclk = pixclk_temp - self.videoinfo.pixclk_last
		end
		if diff_pixclk >= 1000000 then 
			self.videoinfo.pixclk_cnt = 0
		elseif diff_pixclk < 1000000  and self.videoinfo.pixclk_cnt < 10 then   -- < 1mhz
			self.videoinfo.pixclk_cnt = self.videoinfo.pixclk_cnt + 1
		else
			self.videoinfo.pixclk_saved = pixclk_temp		
		end

		self.videoinfo.pixclk_last = pixclk_temp	
		pxclk = self.videoinfo.pixclk_saved			
		self.videoinfo.HTotal = HTotal
		self.videoinfo.HActive = HActive
		self.videoinfo.VTotal = VTotal
		self.videoinfo.VActive = VActive
		self.videoinfo.interlaced = interlaced
		--print("_GetVideoSyncDetail")
		--print(pxclk, HTotal, HActive, VTotal, VActive, interlaced)
		return pxclk, HTotal, HActive, VTotal, VActive, interlaced
	end
	
	------------------------------------------------------
	-- Assume HDMI customized format
	------------------------------------------------------
	function ITE66021:_AssumeHDMICustomFormat(src, pxclk, HTotal, actW, VTotal, actH, interlaced)
		local frame_pxs = HTotal * VTotal
		if frame_pxs <= 0 then
			return false
		end

		local refreshFreq = math.floor(pxclk/frame_pxs)
		if refreshFreq <= 24 then
			refreshFreq = 24
		elseif refreshFreq <= 27 then
			refreshFreq = 25
		elseif refreshFreq <= 35 then
			refreshFreq = 30
		elseif refreshFreq <=52 then
			refreshFreq = 50
		elseif refreshFreq <=65 then
			refreshFreq = 60
		elseif refreshFreq <=72 then
			refreshFreq = 70
		elseif refreshFreq <= 77 then
			refreshFreq = 75
		elseif refreshFreq <= 88 then
			refreshFreq = 85
		end	--Else ignore restrict.
		
		if refreshFreq <= 0 then
			return false
		end
		
		if actW < 160 or actW > 4096 or actH < 90 or actH > 2200 or refreshFreq < 23 then
			return false
		end

		local vMode = hdtv_formats.VIRATE_HD

		if pxclk > 250000000 then
			if refreshFreq >= 45 then --Assume it's 4K50 or 4K60
				HTotal = HTotal*2
				actW = actW*2
				vMode = hdtv_formats.VIRATE_4K60
			else
				vMode = hdtv_formats.VIRATE_4K
			end
		elseif pxclk < 30000000 then
			vMode = hdtv_formats.VIRATE_SD
		end

		actH2 = actH
		totalH2 = VTotal
		
		local formatName
		if interlaced then
			formatName = actW .. "x" .. (actH + actH2) .. "@" .. refreshFreq .. "i"
		else
			formatName = actW .. "x" .. actH .. "@" .. refreshFreq .. " Hz"
		end
		
		local v = self.nostandardFormatInfo[formatName]
		if not v then	--Create this format information
			v = {
				formatId = formatName,
				formatName = formatName,
				frameRate = refreshFreq,
				hTotal = HTotal,
				hAct = actW,
				vTotal0 = VTotal,
				vAct0 = actH,
				vTotal1 = totalH2,
				vAct1 = actH2,
				interlaced = interlaced,
				rateMode = vMode,
				tmdsFreq = pxclk,
				pixelRepeat = false,--TODO:
				aspect_ratio = 0, --That is ASPECT_RATIO_AUTO
				video_system = 0, --That is AMBA_VIDEO_SYSTEM_AUTO
			}
			self.nostandardFormatInfo[formatName] = v
		end
		
		return self:_RecordVideoFormatInfo( src, v )
	end

	------------------------------------------------------
	-- Detect HDMI input format
	------------------------------------------------------
	
	function ITE66021:_DetectHDMIFormat(src)
		local hdcp = false --TODO:
		self.videoFormat.hdcp_content = 0	--It may contains an error status, but, no matter.

		local rslt = self:ReadReg( BASE_MAP, 0x0a, {0x00} )
		if rslt and (rslt[1]&0x02) ~= 0 then
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_HDMI or 1
		else
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_DVI or 0
		end

		self.videoFormat._valid = false
		self.videoFormat.signal_locked = true
		
		local colorSpace = "default"

		--Check the color-space:
		self:WriteReg( BASE_MAP, 0x0f, {0x02} ) --Read 2xx regs.
		local rslt = self:ReadReg( BASE_MAP, 0x15, {0x00} ) --reg 0x215
		if rslt and rslt[1] then
			local flg = rslt[1] & 0x60
			if flg == 0x20 then
				colorSpace = "yuv422"
			elseif flg == 0x40 then
				colorSpace = "yuv444"
			elseif flg == 0x00 then
				colorSpace = "rgb444"
			else
				colorSpace = "yuv420"
			end
		end
		self:WriteReg( BASE_MAP, 0x0f, {0x00} ) --Reset to 0xx regs.

		if colorSpace ~= self.videoFormat.colorSpace then
			self.videoFormat.colorSpace = colorSpace
			self.videoFormat._lastFormatChanged = true
		end

		self:_DetectHDMIInfoFrame(src)

		--So sorry, I need check special formats
		local pxclk, HTotal, HActive, VTotal, VActive, interlaced = self:_GetVideoSyncDetail()
		if pxclk then
			if not self.lastFormatRecord.pxClk or math.abs(pxclk - self.lastFormatRecord.pxClk) >= 20000000 
				or not self.lastFormatRecord.HTotal or math.abs(HTotal - self.lastFormatRecord.HTotal) > 4 
				or not self.lastFormatRecord.HActive or math.abs(HActive - self.lastFormatRecord.HActive) > 4
				or not self.lastFormatRecord.VTotal or math.abs(VTotal - self.lastFormatRecord.VTotal) > 2
				or not self.lastFormatRecord.VActive or math.abs(VActive - self.lastFormatRecord.VActive) > 2
				or interlaced ~= self.lastFormatRecord.interlaced then

				--[[ TODO: We needn't this step.
				rslt = hdtv_formats.Format_Match( pxclk/1000000, HTotal, HActive, VTotal, VActive, interlaced )
				print( "***********", rslt, rslt and rslt.formatName or "***" )
				if rslt then
					if self:_RecordVideoFormatInfo( src, rslt ) then
						self.videoFormat._valid = true
						self.lastFormatRecord.pxClk = pxclk
						self.lastFormatRecord.HTotal = HTotal
						self.lastFormatRecord.HActive = HActive
						self.lastFormatRecord.VTotal = VTotal
						self.lastFormatRecord.VActive = VActive
						self.lastFormatRecord.interlaced = interlaced
						return true
					end
				end
				]]

				if self:_AssumeHDMICustomFormat( src, pxclk, HTotal, HActive, VTotal, VActive, interlaced ) then
						self.videoFormat._valid = true
						self.lastFormatRecord.pxClk = pxclk
						self.lastFormatRecord.HTotal = HTotal
						self.lastFormatRecord.HActive = HActive
						self.lastFormatRecord.VTotal = VTotal
						self.lastFormatRecord.VActive = VActive
						self.lastFormatRecord.interlaced = interlaced
						return true
				else
					return false	--Not detect
				end
			else
				self.videoFormat._valid = true
				return true
			end
		else
			return false
		end
	end

	------------------------------------------------------
	-- Configure HDMI format
	------------------------------------------------------
	function ITE66021:_ConfigureHDMIFormat(src)
		self.videoFormat._lastFormatChanged = false
		self:SetVideoMode( self.videoFormat.signal_locked, self.videoFormat.video_mode, self.videoFormat.colorSpace )
		return true
	end

	------------------------------------------------------
	function ITE66021:SetVideoMode(sig,vidMod,colorSpace)
		if not sig then
			self:WriteReg( BASE_MAP, 0x50, {0x12} )
			self:WriteReg( BASE_MAP, 0x51, {0x00} )
		elseif vidMod == "4k60" then
			self:WriteReg( BASE_MAP, 0x50, {0x01} )
			self:WriteReg( BASE_MAP, 0x51, {0xc0} )
		elseif vidMod == "4k" then
			if colorSpace == "rgb444" then --Clock phase adjustment
				self:WriteReg( BASE_MAP, 0x51, {0x40} )
				self:WriteReg( BASE_MAP, 0x50, {0x02} ) --0x10
			elseif colorSpace == "yuv444" then
				self:WriteReg( BASE_MAP, 0x51, {0x40} )
				self:WriteReg( BASE_MAP, 0x50, {0x02} )	--0x10		
			else
				self:WriteReg( BASE_MAP, 0x51, {0xc0} )
				--self:WriteReg( BASE_MAP, 0x50, {0x01} )
				self:WriteReg( BASE_MAP, 0x50, {0x01} ) --NOTE: changed here but not ensured.
			end		
		elseif vidMod == "sd" then
			self:WriteReg( BASE_MAP, 0x50, {0x12} )
			self:WriteReg( BASE_MAP, 0x51, {0x00} ) -- 0x04 output is CCIR656 mode 对amba不适用
		else --Assume all others be 'hd'
			self:WriteReg( BASE_MAP, 0x50, {0x12} ) --or 0x11 0x13
			self:WriteReg( BASE_MAP, 0x51, {0x00} )
		end
	end
	
	------------------------------------------------------
	
	ITE66021:Reset()
	return ITE66021
end --END of CreateITE66021()

return CreateITE66021

