-----------------------------------------------------------------------------------------------------
-- ITE6661 chip controller.
-- Copyright (C) 2016, KILOVIEW Elecontrics Co., Ltd.
--
-- Author: 	Jacob Zuo, 2019/3/18
-----------------------------------------------------------------------------------------------------

local coaf=require "coaf.core"
local _dbg = require "coaf.debug"

local periphery = require "periphery"
local I2C_REG = require "chips.i2c_reg"
--local pretty = require "pl.pretty"

local MAIN_FREQ = 20000000

local hdtv_formats = require "chips.hdtv_formats"
--hdtv_formats.SetMainFreq( MAIN_FREQ )

--Init both HDTV and VGA format standards
--hdtv_formats.Init( "STDI" )

------------------------------------------------------
-- The factory function to create a ITE6661 handle object
------------------------------------------------------
local function CreateITE6661( i2c_dev, rxAddr )
	------------------------------------------------------
	-- Object
	------------------------------------------------------
	local ITE6661 = {
		currentSource = "HDMI",
		currentSignal = false,
		freeRun = false,
		currentFormat = nil,
		forceReassignFormat = false,
		lastFormatRecord = {},
		videoFormat = { _valid = false },
		audioFormat = { _valid = false, audio_src = "HDMI", audio_signal_type = 0 }, --type 0 aka. HDMI
		orgAudioFormat = { _valid = false },
		nostandardFormatInfo = {},
		i2c_reg = nil,
	}

	------------------------------------------------------
	i2c_dev = i2c_dev or "/dev/i2c-0"
	
	local BASE_MAP = rxAddr or 0x49
	local TX_MAP = 0x4a
	local EDID_MAP = 0x54

	_dbg.Info("ITE6661", "Creating ITE6661 control object at 0x" .. string.format( "%02x", BASE_MAP) .. "@" .. i2c_dev .. "..." )

	ITE6661.i2c_reg = I2C_REG.Create()
	if not ITE6661.i2c_reg:Open( i2c_dev ) then
		return nil, "Fail to open I2C device"
	end

	-------------------------------------------------------
	-- Object method: Register read
	-------------------------------------------------------
	function ITE6661:ReadReg( map, addr, itemArr )
		return self.i2c_reg:ReadReg(map, addr, itemArr)
	end

	------------------------------------------------------
	-- Register write
	------------------------------------------------------
	function ITE6661:WriteReg( map, addr, itemArr )
		return self.i2c_reg:WriteReg(map, addr, itemArr)
	end

	------------------------------------------------------
	
	function ITE6661:ForceUpdate()
		self.lastFormatRecord = {}		--Clear the recorded format information.
	end

	------------------------------------------------------
	-- Soft reset the chip!
	------------------------------------------------------

	function ITE6661:Reset()	--Soft reset
		self:WriteReg( BASE_MAP, 0x11, {0x01} ) -- 软复位
		self:WriteReg( BASE_MAP, 0x11, {0x00} )

		self:WriteReg( BASE_MAP, 0x65, {0x10} )
		self:WriteReg( BASE_MAP, 0x8d, {0x95} )
		self:WriteReg( BASE_MAP, 0xc0, {0x04} )
		self:WriteReg( BASE_MAP, 0x0f, {0x01} )
		self:WriteReg( BASE_MAP, 0xb0, {0x03} )
		self:WriteReg( BASE_MAP, 0x0f, {0x00} )
		self:WriteReg( BASE_MAP, 0x87, {0xa9} )
		self:WriteReg( BASE_MAP, 0x10, {0x10} ) --add Enable auto video reset
		self:WriteReg( BASE_MAP, 0x11, {0xa0} ) --add Enable auto video reset
		
		self:_InitInterrupts()
		self:_InitEDID()	--Init EDID
		self:_InitTx()
		
		self:_ConfigureHDMIFormat()
		
		self.currentSignal = false
		self.freeRun = false
		self.currentFormat = nil
		self.forceReassignFormat = false
		self.lastFormatRecord = {}		--Clear recorded format detection informations.

		self.audioFormat = { _valid = false, audio_src = "HDMI", audio_signal_type = 0 }
		self.orgAudioFormat = { _valid = false }
		self.videoFormat = {_valid = false}

	end

	------------------------------------------------------
	-- Init the ITE6661 interrupts 
	------------------------------------------------------
	
	function ITE6661:_InitInterrupts()
		--TODO: Nothing yet.
	end
	
	function ITE6661:_InitTx()
		self:WriteReg( TX_MAP, 0x05, {0x00} )
		self:WriteReg( TX_MAP, 0x04, {0x00} )
		self:WriteReg( TX_MAP, 0x19, {0x06} )
		self:WriteReg( TX_MAP, 0x22, {0x0f} )
		self:WriteReg( TX_MAP, 0x23, {0x00} )
		self:WriteReg( TX_MAP, 0x26, {0x03} )
		self:WriteReg( TX_MAP, 0x32, {0x00} )
		self:WriteReg( TX_MAP, 0x33, {0x0f} )
		self:WriteReg( TX_MAP, 0xc1, {0x00} )
	end

	------------------------------------------------------
	--EDID configuration
	------------------------------------------------------
	function ITE6661:_InitEDID()
		local edid 
		_G.DISABLE_4Kp60 = true
		if _G.DISABLE_4Kp60 then
			edid = require "chips.edid"	--EDID
		else
			edid = require "chips.edid_uhd60"
		end

		local i
		for i=1, #edid, 8 do
			self:WriteReg( EDID_MAP, i-1, { edid[i], edid[i+1], edid[i+2], edid[i+3], edid[i+4], edid[i+5], edid[i+6], edid[i+7]} )
		end
		periphery.sleep_ms(5) --Wait 5ms for Calc CRC
	end
	
	-----------------------------------------------------
	function ITE6661:_CheckSignal(src)
		local freeRun = false
		--TODO: Not support freeRun for ITE6661
		local rslt = self:ReadReg( BASE_MAP, 0x0a, {0x00} )
		if rslt then
			if (rslt[1] & 0xad) == 0xad then
				rslt = self:ReadReg( BASE_MAP, 0x99, {0x00} )
				if rslt and (rslt[1]&0x08) ~= 0 then
					return true, freeRun
				else
					return false, freeRun
				end
			else
				return false, freeRun
			end
		else
			return false, freeRun
		end
	end

	------------------------------------------------------
	-- A step handler for ITE6661 processing.
	-- To fit the LUA working mode, you need call the 
	-- 'Step' periodically ( suggest per 200~500ms).
	-- This routine should do :
	--	* Signal status checking
	--	*	Auto format detecting
	--	*	Configue the chip to make it output BT.656/BT.601 signals
	--
	------------------------------------------------------
	
	function ITE6661:Step()
		self.currentSignal, self.freeRun = self:_CheckSignal()	--Check current signal

		--TODO: consider free-run but now not support.
		if self.currentSignal ~= self.videoFormat.signal_locked then
			if not self.videoFormat.version then --Keep track the updated version
				self.videoFormat.version = 0
			else
				self.videoFormat.version = self.videoFormat.version + 1
			end
		end

		if not self.audioFormat.version then
			self.audioFormat.version = 0
		end

		if self.currentSignal then
			self.videoFormat.signal_locked = self.currentSignal	--Assume it be, but when detecting, it may changed
			self.videoFormat.free_run = false
			self:_DetectCurrentFormat()
			self:_ConfigureCurrentFormat()
		elseif self.freeRun then
			self.videoFormat.signal_locked = true
			self.videoFormat.free_run = true
		else
			self.videoFormat.signal_locked = false
			self.videoFormat.free_run = false
		end
	end


	------------------------------------------------------
	-- Detect current signal format
	------------------------------------------------------
	
	function ITE6661:_DetectCurrentFormat(src)
		self:_DetectHDMIAudio(src)
		self:_UpdateAudioFormatInfo()
		return self:_DetectHDMIFormat(src)
	end
	
	function ITE6661:_UpdateAudioFormatInfo()
		if self.audioFormat._valid ~= self.orgAudioFormat.withAudio or
			self.audioFormat.sampling ~= self.orgAudioFormat.sample then

			self.audioFormat.version = self.audioFormat.version + 1

			self.audioFormat._valid = self.orgAudioFormat.withAudio
			self.audioFormat.fs = self.orgAudioFormat.fs
			self.audioFormat.sampling = self.orgAudioFormat.sample
			self.audioFormat.channels = 2
			self.audioFormat.bit_width = 16
		end
	end
	------------------------------------------------------
	-- Configure signal format
	------------------------------------------------------
	
	function ITE6661:_ConfigureCurrentFormat(src)
		if not self.videoFormat._lastFormatChanged and not self.forceReassignFormat then
			return true
		end
		
		return self:_ConfigureHDMIFormat(src)
	end
	
	------------------------------------------------------
	-- According detected format information, record it to Amba based formation context.
	------------------------------------------------------
	
	function ITE6661:_RecordVideoFormatInfo(src, formatInfo)
		if type(formatInfo) ~= "table" then
			return false
		end
		--pretty.dump( formatInfo )
		if self.videoFormat._formatInfo ~= formatInfo or (src and src ~= self.videoFormat.video_src) then
			if not self.videoFormat.version then --Keep track the updated version
				self.videoFormat.version = 0
			else
				self.videoFormat.version = self.videoFormat.version + 1
			end

			self.videoFormat._lastFormatChanged = true
			self.videoFormat._formatInfo = formatInfo
			
			if src then
				self.videoFormat.video_src = src
			end
			
			self.videoFormat.format_code = formatInfo.formatId	-- Make the formatId as format-code.
			self.videoFormat.format_name = formatInfo.formatName or (self.videoFormat.cap_cap_w .. "x" .. self.videoFormat.cap_cap_h)

			if formatInfo.rateMode == hdtv_formats.VIRATE_4K then
				self.videoFormat.video_mode = "4k"
			elseif formatInfo.rateMode == hdtv_formats.VIRATE_4K60 then
				self.videoFormat.video_mode = "4k60"
			elseif formatInfo.rateMode == hdtv_formats.VIRATE_SD then
				self.videoFormat.video_mode = "sd"
			else
				self.videoFormat.video_mode = "hd"
			end

			return true
		else
			return true	--No changed.
		end
	end

	------------------------------------------------------
	-- Detect HDMI Audio
	------------------------------------------------------
	function ITE6661:_DetectHDMIAudio(src)
		self.orgAudioFormat.withAudio = false
		local rslt
		
		local rslt = self:ReadReg( BASE_MAP, 0x0a, {0x00} )
		if rslt then
			if (rslt[1] & 0x2c) == 0x2c then
				self.orgAudioFormat.pll_locked = true
			else
				self.orgAudioFormat.pll_locked = false
				return false
			end
		else
			self.orgAudioFormat.pll_locked = false
			return false
		end

		if (rslt[1] & 0x02) == 0 then --DVI mode
			return false
		end

		rslt = self:ReadReg( BASE_MAP, 0xaa, {0x00} )
		if not rslt or (rslt[1]&0x80) == 0 then --Audio ON flag is not set.
			return false
		end

		rslt = self:ReadReg( BASE_MAP, 0xae, {0x00,0x00} )
		if not rslt then
			return false
		end

		local sampleFlag = rslt[1] & 0x0f
		if sampleFlag == 0 then
			self.orgAudioFormat.sample = 44100
		elseif sampleFlag == 2 then
			self.orgAudioFormat.sample = 48000
		elseif sampleFlag == 3 then
			self.orgAudioFormat.sample = 32000
		elseif sampleFlag == 8 then
			self.orgAudioFormat.sample = 88200
		elseif sampleFlag == 9 then
			self.orgAudioFormat.sample = 768000
		elseif sampleFlag == 10 then
			self.orgAudioFormat.sample = 96000
		elseif sampleFlag == 12 then
			self.orgAudioFormat.sample = 176000
		elseif sampleFlag == 14 then
			self.orgAudioFormat.sample = 192000
		else
			self.orgAudioFormat.sample = 48000
		end	

		self.orgAudioFormat.withAudio = true
		--TODO:
		self.orgAudioFormat.fs = 256 --(rslt[1]+1)*128

		return true
	end

	------------------------------------------------------
	-- Detect HDMI AVI info frame
	------------------------------------------------------
	
	function ITE6661:_DetectHDMIInfoFrame(src)
		--TODO: 
		return false
	end
	
	------------------------------------------------------
	-- Get format
	------------------------------------------------------
	
	function ITE6661:_GetVideoSyncDetail()
		local interlaced = false	
		local rslt = self:ReadReg( BASE_MAP, 0x99, {0x00} )
		if rslt and (rslt[1]&0x02) ~= 0 then
			interlaced = true
		end

		rslt = self:ReadReg( BASE_MAP, 0x9a, {0x00} )
		if not rslt or rslt[1] == 0 then
			return nil
		end

		local pxclk = math.floor(MAIN_FREQ/rslt[1] * 255)
		rslt = self:ReadReg( BASE_MAP, 0x9c, {0,0,0,0,0,0,0,0,0,0,0,0,0,0} )
		if not rslt then
			return nil
		end

		local HTotal = ((rslt[2] & 0x3F) << 8) + rslt[1]
		local HActive = ((rslt[4] & 0x3F) << 8) + rslt[3]
		--local HSyncWidth = ((rslt[6] & 0x01) << 8) + rslt[5]
		--local HFrontPorch = ((rslt[6] & 0xF0) << 4) + rslt[7]
		--local HBackPorch = HTotal - HActive - HFrontPorch - HSyncWidth
		local VTotal = ((rslt[9] & 0x0F) << 8) + rslt[8]
		local VActive = ((rslt[9] & 0xF0) << 4) + rslt[10]
		--local VSyncWidth = rslt[11] & 0x1F
		--local VFrontPorch = rslt[12] & 0x3F
		--local VBackPorch = VTotal - VActive - VFrontPorch - VSyncWidth
		--local VPolarity = (rslt[13] & 0x08) >> 3
		--local HPolarity = (rslt[13] & 0x04) >> 2
		
		return pxclk, HTotal, HActive, VTotal, VActive, interlaced
	end
	
	------------------------------------------------------
	-- Assume HDMI customized format
	------------------------------------------------------
	function ITE6661:_AssumeHDMICustomFormat(src, pxclk, HTotal, actW, VTotal, actH, interlaced)
		local frame_pxs = HTotal * VTotal
		if frame_pxs <= 0 then
			return false
		end

		local refreshFreq = math.floor(pxclk/frame_pxs)
		if refreshFreq <= 24 then
			refreshFreq = 24
		elseif refreshFreq <= 27 then
			refreshFreq = 25
		elseif refreshFreq <= 35 then
			refreshFreq = 30
		elseif refreshFreq <=52 then
			refreshFreq = 50
		elseif refreshFreq <=65 then
			refreshFreq = 60
		elseif refreshFreq <=72 then
			refreshFreq = 70
		elseif refreshFreq <= 77 then
			refreshFreq = 75
		elseif refreshFreq <= 88 then
			refreshFreq = 85
		end	--Else ignore restrict.
		
		if refreshFreq <= 0 then
			return false
		end
		
		if actW < 160 or actW > 4096 or actH < 90 or actH > 2200 or refreshFreq < 23 then
			return false
		end

		local vMode = hdtv_formats.VIRATE_HD

		if pxclk > 250000000 then
			if refreshFreq >= 45 then --Assume it's 4K50 or 4K60
				HTotal = HTotal*2
				actW = actW*2
				vMode = hdtv_formats.VIRATE_4K60
			else
				vMode = hdtv_formats.VIRATE_4K
			end
		elseif pxclk < 30000000 then
			vMode = hdtv_formats.VIRATE_SD
		end

		actH2 = actH
		totalH2 = VTotal
		
		local formatName
		if interlaced then
			formatName = actW .. "x" .. (actH + actH2) .. "@" .. refreshFreq .. "i"
		else
			formatName = actW .. "x" .. actH .. "@" .. refreshFreq .. " Hz"
		end
		
		local v = self.nostandardFormatInfo[formatName]
		if not v then	--Create this format information
			v = {
				formatId = formatName,
				formatName = formatName,
				frameRate = refreshFreq,
				hTotal = HTotal,
				hAct = actW,
				vTotal0 = VTotal,
				vAct0 = actH,
				vTotal1 = totalH2,
				vAct1 = actH2,
				interlaced = interlaced,
				rateMode = vMode,
				tmdsFreq = pxclk,
				pixelRepeat = false,--TODO:
				aspect_ratio = 0, --That is ASPECT_RATIO_AUTO
				video_system = 0, --That is AMBA_VIDEO_SYSTEM_AUTO
			}
			self.nostandardFormatInfo[formatName] = v
		end
		
		return self:_RecordVideoFormatInfo( src, v )
	end

	------------------------------------------------------
	-- Detect HDMI input format
	------------------------------------------------------
	
	function ITE6661:_DetectHDMIFormat(src)
		local hdcp = false --TODO:
		self.videoFormat.hdcp_content = 0	--It may contains an error status, but, no matter.

		local rslt = self:ReadReg( BASE_MAP, 0x0a, {0x00} )
		if rslt and (rslt[1]&0x02) ~= 0 then
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_HDMI or 1
		else
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_DVI or 0
		end

		self.videoFormat._valid = false
		
		--Check info frame first. NOTE: Some HDMI source does not provide a valid 
		--format Id.
		--[[ TODO: This is not implemented yet, but also I will not consider it.
		if self:_DetectHDMIInfoFrame(src) then
			--Clear all recorded HTotal/HActive/VTotal/VActive/interlaced
			self.videoFormat._valid = true
			self.lastFormatRecord.pxClk = nil
			self.lastFormatRecord.HTotal = nil
			self.lastFormatRecord.HActive = nil
			self.lastFormatRecord.VTotal = nil
			self.lastFormatRecord.VActive = nil
			self.lastFormatRecord.interlaced = nil
			return true
		end
		]]

		self.lastFormatRecord.infoFrameId = -1		--Invalid recorded infoFrameId
		
		--So sorry, I need check special formats
		local pxclk, HTotal, HActive, VTotal, VActive, interlaced = self:_GetVideoSyncDetail()
		if pxclk then
			if not self.lastFormatRecord.pxClk or math.abs(pxclk - self.lastFormatRecord.pxClk) >= 20000000 
				or not self.lastFormatRecord.HTotal or math.abs(HTotal - self.lastFormatRecord.HTotal) > 4 
				or not self.lastFormatRecord.HActive or math.abs(HActive - self.lastFormatRecord.HActive) > 4
				or not self.lastFormatRecord.VTotal or math.abs(VTotal - self.lastFormatRecord.VTotal) > 2
				or not self.lastFormatRecord.VActive or math.abs(VActive - self.lastFormatRecord.VActive) > 2
				or interlaced ~= self.lastFormatRecord.interlaced then

				--[[ TODO: We needn't this step.
				rslt = hdtv_formats.Format_Match( pxclk/1000000, HTotal, HActive, VTotal, VActive, interlaced )
				print( "***********", rslt, rslt and rslt.formatName or "***" )
				if rslt then
					if self:_RecordVideoFormatInfo( src, rslt ) then
						self.videoFormat._valid = true
						self.lastFormatRecord.pxClk = pxclk
						self.lastFormatRecord.HTotal = HTotal
						self.lastFormatRecord.HActive = HActive
						self.lastFormatRecord.VTotal = VTotal
						self.lastFormatRecord.VActive = VActive
						self.lastFormatRecord.interlaced = interlaced
						return true
					end
				end
				]]

				if self:_AssumeHDMICustomFormat( src, pxclk, HTotal, HActive, VTotal, VActive, interlaced ) then
						self.videoFormat._valid = true
						self.lastFormatRecord.pxClk = pxclk
						self.lastFormatRecord.HTotal = HTotal
						self.lastFormatRecord.HActive = HActive
						self.lastFormatRecord.VTotal = VTotal
						self.lastFormatRecord.VActive = VActive
						self.lastFormatRecord.interlaced = interlaced
						return true
				else
					return false	--Not detect
				end
			else
				self.videoFormat._valid = true
				return true
			end
		else
			return false
		end
	end

	------------------------------------------------------
	-- Configure HDMI format
	------------------------------------------------------
	function ITE6661:_ConfigureHDMIFormat(src)
		self.videoFormat._lastFormatChanged = false
		self:SetVideoMode( self.videoFormat.signal_locked, self.videoFormat.video_mode )
		return true
	end

	------------------------------------------------------
	function ITE6661:SetVideoMode(sig,vidMod,flag)
		if flag then
			if not self.lasttime then
				self.lasttime = os.time()
			end
			if not self.count then
				self.count = 0
			end
			if (os.time()-self.lasttime)<7 and self.count>=2 then
				self:Reset()
				self.lasttime = os.time()
				self.count = 1
			elseif (os.time()-self.lasttime)<7 then
				self.count = self.count + 1
			else
				self.lasttime = os.time()
				self.count = 1
			end
		end
		--[[
		if not sig then
			self:WriteReg( TX_MAP, 0x19, {0x02} )
		elseif vidMod == "4k60" then
			self:WriteReg( TX_MAP, 0x19, {0x02} )
		elseif vidMod == "4k" then
			self:WriteReg( TX_MAP, 0x19, {0x22} )
		elseif vidMod == "sd" then
			self:WriteReg( TX_MAP, 0x19, {0x02} )
		else --Assume all others be 'hd'
			self:WriteReg( TX_MAP, 0x19, {0x02} )
		end
		]]
		--The new policy:
		self:WriteReg( TX_MAP, 0x19, {0x06} )
	end
	
	------------------------------------------------------
	
	ITE6661:Reset()
	return ITE6661

end --END of CreateITE6661()

return CreateITE6661

