local _M = {
	user_special_parsed_formats = {},
	user_modelines = nil,
	user_special_formats = nil,
	user_special_adjusts = nil,
}

local MAIN_FREQ = _G.MAIN_FREQ or 28636360
local format_map = require "builtin_modelines"
local CONFIG_LOC = "/data/configs"

if not format_map then
	format_map = {}
end

--------------------------------------------

function _M.ParseModeline( k, v )
	local hPulse = 1
	local vPulse = 1
	local interlaced = false
	local doublescan = false
	local uniKey = nil
	
	local name, pixelFreq, actH, hSyncStart, hSyncEnd, totalH, actV, vSyncStart, vSyncEnd, totalV =
			k:match('"(.*)"%s+([%d%.]+)%s+(%d-)%s+(%d-)%s+(%d-)%s+(%d-)%s+(%d-)%s+(%d-)%s+(%d-)%s+(%d+).*$')
	
	if nil ~= name then	--Parse OK
		if not v or type(v) ~= "table" then
			v = {}
		end
		
		k = string.lower(k)
		if k:match("-hsync") then
			hPulse = 0
		end
		
		if k:match("-vsync") then
			vPulse = 0
		end
		
		if k:match("interlace") then
			interlaced = true
		end
		
		if k:match("doublescan") then
			doublescan = true
		end
		
		v.pixelFreq = tonumber( pixelFreq ) * 1000000
		v.hTotal = tonumber(totalH)
		v.hAct = tonumber(actH)
		v.hSyncStart = tonumber(hSyncStart)
		v.hSyncEnd = tonumber(hSyncEnd)
		v.vTotal = tonumber(totalV)
		v.vAct = tonumber(actV)
		v.vSyncStart = tonumber(vSyncStart)
		v.vSyncEnd = tonumber(vSyncEnd)
		
		v.hPulse = hPulse
		v.vPulse = vPulse
		v.doubleScan = doublescan
		
		v.interlaced = interlaced
		
		if v.pixelFreq > 0 and v.hTotal > 0 and v.vTotal > 0 then
			--Restrict it in our video standards
			v.frameRate = v.pixelFreq / (v.hTotal * v.vTotal)
			if v.doubleScan then
				v.frameRate = v.frameRate /2 
			end
			
			local base = math.floor(v.frameRate)
			if v.frameRate - base < 0.5 then
				v.frameRate = base
			else
				v.frameRate = base + 1
			end
			
			v.BL = math.floor( MAIN_FREQ / ((v.pixelFreq/v.hTotal)/8) )
			v.LCF = math.floor( v.vTotal )
			v.FCL = math.floor( MAIN_FREQ / (v.pixelFreq / (v.hTotal * v.vTotal)) / 256 )
			v.LCVS = math.abs( math.floor( v.vSyncEnd - v.vSyncStart ) )
		else
			v.frameRate = 60	--Assume default
			v.BL = math.floor( MAIN_FREQ/ ((v.frameRate*v.vTotal)/8) )
			v.LCF = math.floor( v.vTotal )
			v.FCL = math.floor( MAIN_FREQ / v.frameRate / 256 )
			v.LCVS = math.abs( math.floor( v.vSyncEnd - v.vSyncStart ) )
		end
		
		if name:match("@") then
			v.formatName = name
		else
			v.formatName = name .. "@" .. v.frameRate .. (interlaced and "i" or "Hz") 
		end
	else
		return nil
	end
	
	if v._VESA then
		v._MATCH_LEVEL = (v._MATCH_LEVEL or 60) + (v._MFLAG or 0)
	elseif v._HDTV then
		v._MATCH_LEVEL = (v._MATCH_LEVEL or 50) + (v._MFLAG or 0)
	elseif v._NTSC or v._PAL then
		v._MATCH_LEVEL = (v._MATCH_LEVEL or 40) + (v._MFLAG or 0)
	elseif v._CVT or v._GTF then
		v._MATCH_LEVEL = (v._MATCH_LEVEL or 30) + (v._MFLAG or 0)
	elseif v._NVIDIA then
		v._MATCH_LEVEL = (v._MATCH_LEVEL or 20) + (v._MFLAG or 0)
	else
		v._MATCH_LEVEL = (v._MATCH_LEVEL or 10) + (v._MFLAG or 0)
	end

	return v
end

--------------------------------------------

function _M.Init()
	local r
	r, _M.user_modelines = pcall( dofile, CONFIG_LOC .. "/vga_user_modelines.conf" )
	if not r or not _M.user_modelines then
		_M.user_modelines = {}
		print( "*** No user customized VGA modelines." )
	else
		print( "*** Found #" .. #_M.user_modelines .. " modelines of user customized." )
	end
	
	r, _M.user_special_formats = pcall( dofile, CONFIG_LOC .. "/vga_specials.conf" )
	if not r or not _M.user_special_formats then
		_M.user_special_formats = {}
		print( "*** No user specialized VGA formats found." )
	else
		print( "*** Found user specialized VGA format options." )
	end
	
	r, _M.user_special_adjusts = pcall( dofile, CONFIG_LOC .. "/vga_adjusts.conf" )
	if not r or not _M.user_special_adjusts then
		_M.user_special_adjusts = {}
		print( "*** No VGA adjustment options." )
	else
		print( "*** Found user adjustment options of VGA format." )
	end

	local k, v --To parse the user modelines. { "line1", "line2", ... }
	for k=1, #_M.user_modelines do
		if type(_M.user_modelines[k]) == "string" then
			v = _M.ParseModeline( _M.user_modelines[k] )
			if type(v) == "table" and v.formatName then
				v._SEQ = 1000 + k
				_M.user_special_parsed_formats[ _M.user_modelines[k] ] = v
			end
		end
	end
end

--------------------------------------------

function _M.ClearAddedModelines()
	print( "*** All VGA user specialized modelines are cleared." )
	_M.user_modelines = {}
	_M.user_special_parsed_formats = {}
	local file = io.open( CONFIG_LOC .. "/vga_user_modelines.conf", "w" )
	if file then
		file:write( "return {}" )
		file:close()
	end
end

--------------------------------------------

function _M.AddModelines( modelines )
	print( "*** Added user specialized VGA modelines:" )
	print( modelines )
	if not modelines or type(modelines) ~= "string" then
		return false
	end
	
	_M.user_modelines = {}
	_M.user_special_parsed_formats = {}
	
	modelines = modelines .. "\r\n"
	local line
	for line in modelines:gmatch( "(.-)[\r\n]+" ) do
		if line ~= "" then
			table.insert( _M.user_modelines, line )
		end
	end
	
	local file = io.open( CONFIG_LOC .. "/vga_user_modelines.conf", "w" )
	if file then
		if #_M.user_modelines == 0 then
			file:write( "return {}" )
		else
			local i
			file:write( "return {\n" )
			for i=1, #_M.user_modelines do
				file:write( "[[" .. _M.user_modelines[i] .. "]], \n" )
			end
			file:write( "}\n" )
		end
		file:close()
	end
	
	local k, v --To parse the user modelines. { "line1", "line2", ... }
	for k=1, #_M.user_modelines do
		v = _M.ParseModeline( _M.user_modelines[k] )
		if type(v) == "table" and v.formatName then
			v._SEQ = 1000 + k
			_M.user_special_parsed_formats[ _M.user_modelines[k] ] = v
		end
	end
	
	return true
end
--------------------------------------------

function _M.GetUserModelines()
	return _M.user_modelines
end

--------------------------------------------
function _M.SetSpecialFormat( code, modeline )
	print( "*** User specialize the VGA format: ", code, modeline )
	if not modeline or modeline == "" then
		_M.user_special_formats[ code ] = nil
	else
		_M.user_special_formats[ code ] = { ML = modeline }
	end
	local file = io.open( CONFIG_LOC .. "/vga_specials.conf", "w" )
	if file then
		file:write( "return {\n" )
		local k, v
		for k, v in pairs( _M.user_special_formats ) do
			if type(v) == "table" and not v.tmp then
				file:write( "[ [[" .. k .. "]] ] = { ML=[[" .. (v.ML or "") .. "]] }, \n" )
			end
		end
		file:write( "}\n" )
		file:close()
	end
	return true
end

--------------------------------------------
function _M.RemoveSpecialFormat( code )
	print( "*** User removed a special VGA format: ", code )
	if not code then
		_M.user_special_formats = {}
	else
		_M.user_special_formats[ code ] = nil
	end
	
	local file = io.open( CONFIG_LOC .. "/vga_specials.conf", "w" )
	if file then
		file:write( "return {\n" )
		local k, v
		for k, v in pairs( _M.user_special_formats ) do
			if type(v) == "table" and not v.tmp then
				file:write( "[ [[" .. k .. "]] ] = { ML=[[" .. (v.ML or "") .. "]] }, \n" )
			end
		end
		file:write( "}\n" )
		file:close()
	end
	return true
end

--------------------------------------------
function _M.SetSpecialAdjustment( code, shiftX, shiftY, phase )
	print( "*** User adjust the VGA picture: ", code, shiftX, shiftY, phase )
	if code then
		if shiftX or shiftY then
			shiftX = shiftX or 0
			shiftY = shiftY or 0
			phase = phase or 16
			_M.user_special_adjusts[ code ] = { X = shiftX, Y = shiftY, P = phase }
		else
			_M.user_special_adjusts[ code ] = nil
		end
	else
		return false
	end
	
	local file = io.open( CONFIG_LOC .. "/vga_adjusts.conf", "w" )
	if file then
		file:write( "return {\n" )
		local k, v
		for k, v in pairs( _M.user_special_adjusts ) do
			if type(v) == "table" then
				file:write( "[ [[" .. k .. "]] ] = { X=" .. (v.X or 0) .. ", Y=" .. (v.Y or 0) .. ", P=" .. (v.P or "nil") .. " }, \n" )
			end
		end
		file:write( "}\n" )
		file:close()
	end
	
	return true
end
--------------------------------------------
function _M.GetSpecialAdjustment( code )
	if not code then
		return nil
	else
		return _M.user_special_adjusts[code]	--May be nil
	end
end
--------------------------------------------
function _M.STDI_AllMatches( LCF, LCVS, FCL, BL, interlaced )
	local k,v
	local diff_BL, diff_LCF, diff_FCL, diff_LCVS
	local score
	
	local seqs = { _M.user_special_parsed_formats, format_map }
	local r = {}
	
	--NOTE:
	-- BL:  (MAIN_FREQ/1sec-total-lines)*8 ~~= 8 lines Crytical clocks
	-- LCF: Total Lines per frame.
	-- FCL: (each-frame-with-crytical-clocks)/256
	-- LCVS: Vertical-Sync pulse width (lines)
	local i
	for i=1, #seqs do
		for k,v in pairs(seqs[i]) do
			if v.BL and v.LCF and v.FCL and v.LCVS then
				diff_BL = math.abs( BL - v.BL )
				diff_LCF = math.abs( LCF - v.LCF )
				diff_FCL = math.abs( FCL - v.FCL )
				diff_LCVS = math.abs(LCVS - v.LCVS)
				
				if diff_LCF <= 1 and diff_LCVS <=3 and diff_FCL <=10 and ( nil == interlaced or interlaced == v.interlaced ) then
					score = diff_LCF * 1000000 + diff_LCVS * 10000 + diff_FCL * 100 + diff_BL
					table.insert( r, {
						score = score,
						formatName = v.formatName,
						_SEQ = v._SEQ or 0,
						modeline = k,
						options = v
					} )
				end -- If matched
			end --If valid item
		end --For match
	end--For each maps
	
	table.sort( r, function(a,b)
		if a.score < b.score then
			return true
		else
			return false
		end
	end )

	return r
end

--------------------------------------------
local function FormatAdjustment( code, bestItem )
	if bestItem then	--Directly shot!
		if _M.user_special_adjusts[code] then --User special adjustment
			local k,v
			local dummy = {}
			for k,v in pairs(bestItem) do
				dummy[k] = v
			end
			
			local shiftX = _M.user_special_adjusts[code].X or 0
			local shiftY = _M.user_special_adjusts[code].Y or 0
			if dummy.hSyncStart and dummy.hSyncEnd then
				dummy.hSyncStart = dummy.hSyncStart + shiftX
				dummy.hSyncEnd = dummy.hSyncEnd + shiftX
			end
			
			if dummy.vSyncStart and dummy.vSyncEnd then
				dummy.vSyncStart = dummy.vSyncStart + shiftY
				dummy.vSyncEnd = dummy.vSyncEnd + shiftY
			end
			
			dummy.samplePhase = _M.user_special_adjusts[code].P
			return dummy
		else
			return bestItem
		end
	else
		return nil
	end
end
--------------------------------------------

function _M.STDI_Match( LCF, LCVS, FCL, BL, interlaced )
	local bestItem = nil
	local code = LCF .. "/" .. LCVS .. "/" .. FCL .. "/" .. BL .. "/" .. (interlaced and "I" or "P")
	
	if type(_M.user_special_formats[ code ]) == "table" and type(_M.user_special_formats[ code ].ML) == "string" then --Fast matched!
		local ML = _M.user_special_formats[ code ].ML
		bestItem = _M.user_special_parsed_formats[ ML ]  --First, try match user definitions
		if not bestItem then
			bestItem = format_map[ ML ] --Then try builtin
		end
		if bestItem then
			bestItem.formatId = code		--Update the formatId ( that should assume as format_code )
			return FormatAdjustment( code, bestItem )
		end
	end
	
	local allMatchs = _M.STDI_AllMatches( LCF, LCVS, FCL, BL, interlaced )
	if #allMatchs < 1 then
		return nil, "No matched"
	else
		--Debug only.
		local xx 
		for xx=1, #allMatchs do
			print( "Matched: Score=" .. allMatchs[xx].score .. ", format=" .. allMatchs[xx].formatName .. "[#" .. allMatchs[xx]._SEQ .. "]" )
		end
	
		bestItem = allMatchs[1].options
		bestItem.formatId = code --Update the formatId ( that should assume as format_code )
		_M.user_special_formats[ code ] = { ML = allMatchs[1].modeline, tmp = true }	--Record it, so can fast detect next time!
	end
	
	return FormatAdjustment( code, bestItem )
end

--------------------------------------------

function _M.Format_Match( tmdsFreqMHz , total_w, act_w, total_v0, act_v0, interlaced, total_v1, act_v1 )
	local diff_tmds, diff_tw, diff_w, diff_tv0, diff_v0
	local k,v
	local bestItem, bestScore
	local score
	
	local matched_vesa = {}
	
	for k,v in pairs(format_map) do
		if not tmdsFreqMHz or tmdsFreqMHz <= 0 then
			diff_tmds = 0
		else
			diff_tmds = math.abs( tmdsFreqMHz - v.pixelFreq/1000000 )
		end
		diff_tw = math.abs( total_w - v.hTotal )
		diff_w = math.abs( act_w - v.hAct )
		diff_tv0 = math.abs( total_v0 - v.vTotal0 )
		diff_v0 = math.abs( act_v0 - v.vAct0 )
		
		if diff_tmds <=2 and
			diff_tw <= 8 and
			diff_w <= 4 and
			diff_tv0 <= 8 and
			diff_v0 <= 4 and
			( nil == interlaced or interlaced == v.interlaced ) 
			then
			
			if v._VESA then
				table.insert( matched_vesa, v )
			end
				
			--score = diff_tmds*1000000 + diff_w*10000 + diff_tw*100 + diff_v0*10 + diff_tv0
			score = diff_w*1000000 + diff_tw*10000 + diff_v0*100 + diff_tv0
			if not bestItem then
				bestItem = v
				bestScore = score
			elseif score < bestScore then
				bestItem = v
				bestScore = score
			end
		end
	end
	
	if bestItem then
		for k=1,#matched_vesa do	--If same resolution matched in VESA, so choice VESA standard
			if bestItem.hAct == matched_vesa[k].hAct and
				bestItem.vAct == matched_vesa[k].vAct then
				bestItem = matched_vesa[k]
				break
			end
		end
		return bestItem
	else
		return nil, "No matched"
	end
	
end

--------------------------------------------

function _M.SetMainFreq( freq )
	MAIN_FREQ = freq
end

--------------------------------------------

return _M
