local dbus = require "kiloview.dbusClients"
local l2dbus = require "l2dbus"
local nu = require "kiloview.netUtil"
local _dbg = require "coaf.debug"
local pt = require "pl.pretty"
local wwan_config = require "mod.wwan_config"
local modem_config = require "mod.modem_config"
local ethernet_config = require "mod.ethernet_config"
local funcs = require "mod.nm_funcs"
local coaf = require "coaf.core"
local thread = require "coaf.thread"
local stringx = require "pl.stringx"
local utils = require "coaf.utils"
require "features"

local static = require "nm_static"

local _M = {
	config = nil,
	modemcfg = nil,
	ethcfg = nil,
	checkThreads = {},
	wwan_autoconnect = true,
	devices = {},
	modemdevices = {},
	simcardstat = {},
}

local NM_BUS_NAME = "org.freedesktop.NetworkManager"

local NM_MANAGER_OBJNAME = "/org/freedesktop/NetworkManager"
local NM_SETTINGS_OBJNAME = "/org/freedesktop/NetworkManager/Settings"

local NM_MANAGER_INTERFACE = "org.freedesktop.NetworkManager"
local NM_DEVICE_INTERFACE = "org.freedesktop.NetworkManager.Device"
local NM_IP4CONFIG_INTERFACE = "org.freedesktop.NetworkManager.IP4Config"
local NM_DHCP4CONFIG_INTERFACE = "org.freedesktop.NetworkManager.DHCP4Config"

local NM_IP6CONFIG_INTERFACE = "org.freedesktop.NetworkManager.IP6Config"
local NM_DHCP6CONFIG_INTERFACE = "org.freedesktop.NetworkManager.DHCP6Config"

local NM_GENERIC_INTERFACE = "org.freedesktop.NetworkManager.Device.Generic"
local NM_WIRED_INTERFACE = "org.freedesktop.NetworkManager.Device.Wired"
local NM_WIRELESS_INTERFACE = "org.freedesktop.NetworkManager.Device.Wireless"
local NM_MODEM_INTERFACE = "org.freedesktop.NetworkManager.Device.Modem"
local NM_WIFI_AP_INTERFACE = "org.freedesktop.NetworkManager.AccessPoint"

local NM_SETTINGS_INTERFACE = "org.freedesktop.NetworkManager.Settings"
local NM_CONNECTION_INTERFACE = "org.freedesktop.NetworkManager.Settings.Connection"
local NM_ACTIVE_CONNECTION_INTERFACE = "org.freedesktop.NetworkManager.Connection.Active"

local MM_MANAGER_OBJNAME = "/org/freedesktop/ModemManager1"
local MM_BUS_NAME = "org.freedesktop.ModemManager1"
local DBUS_OBJECT_INTERFACE = "org.freedesktop.DBus.ObjectManager"
local MM_MODEM_INTERFACE = "org.freedesktop.ModemManager1.Modem"
local MM_MODEM_SIMPLE_INTERFACE = "org.freedesktop.ModemManager1.Modem.Simple"
local MM_SIM_INTERFACE = "org.freedesktop.ModemManager1.Sim"
local MM_BEARER_INTERFACE = "org.freedesktop.ModemManager1.Bearer"

-----------

local NM_DEVICE_STATE_UNKNOWN = 0  -- The device is in an unknown state. 
local NM_DEVICE_STATE_UNMANAGED = 10 -- The device is recognized but not managed by NetworkManager. 
local NM_DEVICE_STATE_UNAVAILABLE = 20 -- The device cannot be used (carrier off, rfkill, etc). 
local NM_DEVICE_STATE_DISCONNECTED = 30 --The device is not connected. 
local NM_DEVICE_STATE_PREPARE = 40 --The device is preparing to connect. 
local NM_DEVICE_STATE_CONFIG = 50 --The device is being configured. 
local NM_DEVICE_STATE_NEED_AUTH = 60 --The device is awaiting secrets necessary to continue connection. 
local NM_DEVICE_STATE_IP_CONFIG = 70 --The IP settings of the device are being requested and configured. 
local NM_DEVICE_STATE_IP_CHECK = 80 --The device's IP connectivity ability is being determined. 
local NM_DEVICE_STATE_SECONDARIES = 90 --The device is waiting for secondary connections to be activated. 
local NM_DEVICE_STATE_ACTIVATED = 100 --The device is active. 
local NM_DEVICE_STATE_DEACTIVATING = 110 --The device's network connection is being torn down. 
local NM_DEVICE_STATE_FAILED = 120 --The device is in a failure state following an attempt to activate it. 

------------

_M.NM_DEVICE_STATE_NAMES = {
	[NM_DEVICE_STATE_UNKNOWN] = "UNKNOWN",
	[NM_DEVICE_STATE_UNMANAGED] = "UNMANAGED",
	[NM_DEVICE_STATE_UNAVAILABLE] = "UNAVAILABLE",
	[NM_DEVICE_STATE_DISCONNECTED] = "DISCONNECTED",
	[NM_DEVICE_STATE_PREPARE] = "PREPARE",
	[NM_DEVICE_STATE_CONFIG] = "CONFIG",
	[NM_DEVICE_STATE_NEED_AUTH] = "NEED_AUTH",
	[NM_DEVICE_STATE_IP_CONFIG] = "IP_CONFIG",
	[NM_DEVICE_STATE_IP_CHECK] = "IP_CHECK",
	[NM_DEVICE_STATE_SECONDARIES] = "SECONDARIES",
	[NM_DEVICE_STATE_ACTIVATED] = "ACTIVATED",
	[NM_DEVICE_STATE_DEACTIVATING] = "DEACTIVATING",
	[NM_DEVICE_STATE_FAILED] = "FAILED"
}

------------

local NM_DEVICE_TYPE_UNKNOWN = 0
local NM_DEVICE_TYPE_ETHERNET = 1
local NM_DEVICE_TYPE_WIFI = 2
local NM_DEVICE_TYPE_UNUSED1 = 3
local NM_DEVICE_TYPE_UNUSED2 = 4
local NM_DEVICE_TYPE_BT = 5
local NM_DEVICE_TYPE_OLPC_MESH = 6
local NM_DEVICE_TYPE_WIMAX = 7
local NM_DEVICE_TYPE_MODEM = 8
local NM_DEVICE_TYPE_INFINIBAND = 9
local NM_DEVICE_TYPE_BOND = 10
local NM_DEVICE_TYPE_VLAN = 11
local NM_DEVICE_TYPE_ADSL = 12
local NM_DEVICE_TYPE_BRIDGE = 13
local NM_DEVICE_TYPE_GENERIC = 14
local NM_DEVICE_TYPE_TEAM = 15

------------

_M.NM_DEVICE_TYPE_NAMES = {
	[NM_DEVICE_TYPE_UNKNOWN] = "UNKNOWN",
	[NM_DEVICE_TYPE_ETHERNET] = "ETHERNET",
	[NM_DEVICE_TYPE_WIFI] = "WIFI",
	[NM_DEVICE_TYPE_UNUSED1] = "UNUSED1",
	[NM_DEVICE_TYPE_UNUSED2] = "UNUSED2",
	[NM_DEVICE_TYPE_BT] = "BT",
	[NM_DEVICE_TYPE_OLPC_MESH] = "OLPC_MESH",
	[NM_DEVICE_TYPE_WIMAX] = "WIMAX",
	[NM_DEVICE_TYPE_MODEM] = "MODEM",
	[NM_DEVICE_TYPE_INFINIBAND] = "INFINIBAND",
	[NM_DEVICE_TYPE_BOND] = "BOND",
	[NM_DEVICE_TYPE_VLAN] = "VLAN",
	[NM_DEVICE_TYPE_ADSL] = "ADSL",
	[NM_DEVICE_TYPE_BRIDGE] = "BRIDGE",
	[NM_DEVICE_TYPE_GENERIC] = "GENERIC",
	[NM_DEVICE_TYPE_TEAM] = "TEAM"
}

------------

local NM_802_11_MODE_UNKNOWN = 0
local NM_802_11_MODE_ADHOC = 1
local NM_802_11_MODE_INFRA = 2
local NM_802_11_MODE_AP = 3

------------

_M.NM_802_11_MODE_NAMES = {
	[NM_802_11_MODE_UNKNOWN] = "UNKNOWN",
	[NM_802_11_MODE_ADHOC] = "ADHOC",
	[NM_802_11_MODE_INFRA] = "INFRA",
	[NM_802_11_MODE_AP] = "AP"
}

------------

local NM_802_11_AP_SEC_NONE = 0x0
local NM_802_11_AP_SEC_PAIR_WEP40 = 0x1
local NM_802_11_AP_SEC_PAIR_WEP104 = 0x2
local NM_802_11_AP_SEC_PAIR_TKIP = 0x4
local NM_802_11_AP_SEC_PAIR_CCMP = 0x8
local NM_802_11_AP_SEC_GROUP_WEP40 = 0x10
local NM_802_11_AP_SEC_GROUP_WEP104 = 0x20
local NM_802_11_AP_SEC_GROUP_TKIP = 0x40
local NM_802_11_AP_SEC_GROUP_CCMP = 0x80
local NM_802_11_AP_SEC_KEY_MGMT_PSK = 0x100
local NM_802_11_AP_SEC_KEY_MGMT_802_1X = 0x200

------------

local NM_DEVICE_STATE_REASON_UNKNOWN = 0
local NM_DEVICE_STATE_REASON_NONE = 1
local NM_DEVICE_STATE_REASON_NOW_MANAGED = 2
local NM_DEVICE_STATE_REASON_NOW_UNMANAGED = 3
local NM_DEVICE_STATE_REASON_CONFIG_FAILED = 4
local NM_DEVICE_STATE_REASON_CONFIG_UNAVAILABLE = 5
local NM_DEVICE_STATE_REASON_CONFIG_EXPIRED = 6
local NM_DEVICE_STATE_REASON_NO_SECRETS = 7
local NM_DEVICE_STATE_REASON_SUPPLICANT_DISCONNECT = 8
local NM_DEVICE_STATE_REASON_SUPPLICANT_CONFIG_FAILED = 9
local NM_DEVICE_STATE_REASON_SUPPLICANT_FAILED = 10
local NM_DEVICE_STATE_REASON_SUPPLICANT_TIMEOUT = 11
local NM_DEVICE_STATE_REASON_PPP_START_FAILED = 12
local NM_DEVICE_STATE_REASON_PPP_DISCONNECT = 13
local NM_DEVICE_STATE_REASON_PPP_FAILED = 14
local NM_DEVICE_STATE_REASON_DHCP_START_FAILED = 15
local NM_DEVICE_STATE_REASON_DHCP_ERROR = 16
local NM_DEVICE_STATE_REASON_DHCP_FAILED = 17
local NM_DEVICE_STATE_REASON_SHARED_START_FAILED = 18
local NM_DEVICE_STATE_REASON_SHARED_FAILED = 19
local NM_DEVICE_STATE_REASON_AUTOIP_START_FAILED = 20
local NM_DEVICE_STATE_REASON_AUTOIP_ERROR = 21
local NM_DEVICE_STATE_REASON_AUTOIP_FAILED = 22
local NM_DEVICE_STATE_REASON_MODEM_BUSY = 23
local NM_DEVICE_STATE_REASON_MODEM_NO_DIAL_TONE = 24
local NM_DEVICE_STATE_REASON_MODEM_NO_CARRIER = 25
local NM_DEVICE_STATE_REASON_MODEM_DIAL_TIMEOUT = 26
local NM_DEVICE_STATE_REASON_MODEM_DIAL_FAILED = 27
local NM_DEVICE_STATE_REASON_MODEM_INIT_FAILED = 28
local NM_DEVICE_STATE_REASON_GSM_APN_FAILED = 29
local NM_DEVICE_STATE_REASON_GSM_REGISTRATION_NOT_SEARCHING = 30
local NM_DEVICE_STATE_REASON_GSM_REGISTRATION_DENIED = 31
local NM_DEVICE_STATE_REASON_GSM_REGISTRATION_TIMEOUT = 32
local NM_DEVICE_STATE_REASON_GSM_REGISTRATION_FAILED = 33
local NM_DEVICE_STATE_REASON_GSM_PIN_CHECK_FAILED = 34
local NM_DEVICE_STATE_REASON_FIRMWARE_MISSING = 35
local NM_DEVICE_STATE_REASON_REMOVED = 36
local NM_DEVICE_STATE_REASON_SLEEPING = 37
local NM_DEVICE_STATE_REASON_CONNECTION_REMOVED = 38
local NM_DEVICE_STATE_REASON_USER_REQUESTED = 39
local NM_DEVICE_STATE_REASON_CARRIER = 40
local NM_DEVICE_STATE_REASON_CONNECTION_ASSUMED = 41
local NM_DEVICE_STATE_REASON_SUPPLICANT_AVAILABLE = 42
local NM_DEVICE_STATE_REASON_MODEM_NOT_FOUND = 43
local NM_DEVICE_STATE_REASON_BT_FAILED = 44
local NM_DEVICE_STATE_REASON_GSM_SIM_NOT_INSERTED = 45
local NM_DEVICE_STATE_REASON_GSM_SIM_PIN_REQUIRED = 46
local NM_DEVICE_STATE_REASON_GSM_SIM_PUK_REQUIRED = 47
local NM_DEVICE_STATE_REASON_GSM_SIM_WRONG = 48
local NM_DEVICE_STATE_REASON_INFINIBAND_MODE = 49
local NM_DEVICE_STATE_REASON_DEPENDENCY_FAILED = 50
local NM_DEVICE_STATE_REASON_BR2684_FAILED = 51
local NM_DEVICE_STATE_REASON_MODEM_MANAGER_UNAVAILABLE = 52
local NM_DEVICE_STATE_REASON_SSID_NOT_FOUND = 53
local NM_DEVICE_STATE_REASON_SECONDARY_CONNECTION_FAILED = 54
local NM_DEVICE_STATE_REASON_DCB_FCOE_FAILED = 55
local NM_DEVICE_STATE_REASON_TEAMD_CONTROL_FAILED = 56
local NM_DEVICE_STATE_REASON_MODEM_FAILED = 57
local NM_DEVICE_STATE_REASON_MODEM_AVAILABLE = 58
local NM_DEVICE_STATE_REASON_SIM_PIN_INCORRECT = 59
local NM_DEVICE_STATE_REASON_NEW_ACTIVATION = 60
local NM_DEVICE_STATE_REASON_PARENT_CHANGED = 61
local NM_DEVICE_STATE_REASON_PARENT_MANAGED_CHANGED = 62

------------

local deepcopy = function(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end

    return _copy(object)
end

local function workThread(thrObj, devname)
	_dbg.Info( "NM_DAEMON", devname, "workThread is started!" )
	_M.checkThreads[devname].inited = false
    while not thrObj.terminated do
		local modem = _M.GetModemInfo(devname,false)
		if modem and modem.ModemInfo and modem.ModemInfo.PrimaryPort then
			if not _M.checkThreads[devname].inited then
				if _M.config[devname] then 
					if not _M.config[devname].connection["interface-name"] then
						_M.config[devname].connection["interface-name"] = modem.ModemInfo.PrimaryPort
					end

					if _M.AddConnection(_M.config[devname]) then
						_M.checkThreads[devname].inited = true
					end
				else
					_dbg.Info( "NM_DAEMON", devname, "no config, init ok" )
					_M.checkThreads[devname].inited = true
				end
			end

			--check connect and auto reconnect
			local deviceinfo = _M.GetDeviceBy(devname)
			if deviceinfo and deviceinfo.State == NM_DEVICE_STATE_DISCONNECTED and _M.wwan_autoconnect then
				local existconsetting = _M.ListConnectionAndGetDetails( { interface = modem.ModemInfo.PrimaryPort } )
				if existconsetting then
					local k, v
					for k, v in pairs(existconsetting) do
						_dbg.Info( "NM_DAEMON", "Autoconnect:",k)
						_M.ConnectDevice( deviceinfo.Object, k)
					end
				end
			end
		else
			_M.checkThreads[devname].inited = false
			-- print("modem not init!!!",devname)
			--remove exist connection
			--[[
			if _M.config[devname] and _M.config[devname].connection and _M.config[devname].connection["interface-name"] then
				local existconsetting = _M.ListConnectionAndGetDetails( { interface = _M.config[devname].connection["interface-name"] } )
				if existconsetting then
					local k, v
					for k, v in pairs(existconsetting) do
						if v.connection and v.connection["interface-name"] 
							and v.connection["interface-name"] == _M.config[devname].connection["interface-name"] then
							if not _M.DeleteConnection(k,false) then
								_dbg.Error( "NM_DAEMON", "DeleteConnection failed!",k )
							end
							--移除之前的端口记录
							_M.config[devname].connection["interface-name"] = nil
							_M.config:__TRIGGER(true)
						end
					end
				end
			end
			--]]
		end
        coaf.Sleep(5)
    end
    _dbg.Info( "NM_DAEMON", devname, "workThread is terminated." )
end

local function config_ethernet(devicename, cfg)
	if devicename and type(cfg) == "table" then
		local cmd = "/usr/sbin/ethtool -s " .. devicename
		local autoneg = false
		if cfg.auto_neg ~= "off" then
			autoneg = true
			local advertise = "0x00f"
			cmd = cmd .. " autoneg on"
			if cfg.auto_neg == "10m_100m" then
				advertise = "0x00f"
			elseif cfg.auto_neg == "100m" then
				advertise = "0x00c"
			elseif cfg.auto_neg == "10m" then
				advertise = "0x003"
			else
				advertise = "0x00f"
			end
			cmd = cmd .. " advertise " .. advertise
		else
			cmd = cmd .. " autoneg off"
		end

		if not autoneg then
			if cfg.speed and cfg.speed ~= "" then
				cmd = cmd .. " speed " .. cfg.speed
			else
				cmd = cmd .. " speed 100"
			end


			if cfg.duplex and cfg.duplex ~= "half" then
				cmd = cmd .. " duplex full"
			else
				cmd = cmd .. " duplex half"
			end
		end

		_dbg.Info( "NM_DAEMON", "Configure ethernet with command:", cmd)
		os.execute(cmd)
	end
end

_M.Init = function()
	if not _M.config then
		_M.config = wwan_config.Load()
	end

	if not _M.modemcfg then
		_M.modemcfg = modem_config.Load()
	end

	if not _M.ethcfg then
		_M.ethcfg = ethernet_config.Load()
	end

	_M.dbus_session = dbus.CreateClientSession()
	if not _M.dbus_session then
		return false, "Could not create DBUS client session for network manager"
	end

	_M.nm_manager_proxy = _M.dbus_session:GetProxy( NM_MANAGER_OBJNAME, NM_BUS_NAME, NM_MANAGER_INTERFACE )
	if not _M.nm_manager_proxy then
		_M.dbus_session:Free()
		_M.dbus_session = nil
		return false, "Could not get network manager interface proxy"
	end

	_M.nm_settings_proxy = _M.dbus_session:GetProxy( NM_SETTINGS_OBJNAME, NM_BUS_NAME, NM_SETTINGS_INTERFACE )
	if not _M.nm_settings_proxy then
		_M.dbus_session:Free()
		_M.dbus_session = nil
		return false, "Could not get network manager settings interface proxy"
	end

	_M.dbus_object_proxy = _M.dbus_session:GetProxy( MM_MANAGER_OBJNAME, MM_BUS_NAME, DBUS_OBJECT_INTERFACE )
	if not _M.dbus_object_proxy then
		_dbg.Error( "NM_DAEMON", "Could not get dbus object proxy" )
		return false, "Could not get dbus object proxy"
	end

	--根据配置需要创建连接
	if _M.config then
		local i
		for i=1, 4 do
			local devname = "modem"..tostring(i)
			_M.checkThreads[devname] = {
				thread = thread.New( workThread )
			}
			_M.checkThreads[devname].thread:Start(devname)
		end
	end

	_M.Steps()

	--TODO: 重置MAC
	coaf.NewTimer( _M.resetMac, 1):start()

	-- 湖南卫视定制模块连接
	if _G.FEATURE_HUNANWEISHI_MODEM then
		coaf.NewTimer(function ()
			for k, v in pairs(_M.modemdevices) do
				local AT_PORT = _M.GetATPATH(k)
				local SIM = v["org.freedesktop.ModemManager1.Modem"]["Sim"] or "/"
				if AT_PORT and SIM ~= "/" then
					local AT_MODEL = v["org.freedesktop.ModemManager1.Modem"]["Model"] or ""
					os.execute("/usr/bin/start_393_386.sh " .. AT_PORT .. " " .. AT_MODEL .. " &")
				end
			end
		end,60):start()
	end

	return true
end

-----------------------------------------------

_M.GetNetworkGlobalStatus = function()
	if not _M.nm_manager_proxy then
		return nil, "Network manager module is not initialized yet"
	end

	local results = {}

	results.NetworkingEnabled = _M.nm_manager_proxy.p.NetworkingEnabled
	results.WirelessEnabled = _M.nm_manager_proxy.p.WirelessEnabled
	results.WirelessHardwareEnabled = _M.nm_manager_proxy.p.WirelessHardwareEnabled
	results.WwanEnabled = _M.nm_manager_proxy.p.WwanEnabled
	results.WwanHardwareEnabled = _M.nm_manager_proxy.p.WwanHardwareEnabled
	results.WimaxEnabled = _M.nm_manager_proxy.p.WimaxEnabled
	results.WimaxHardwareEnabled = _M.nm_manager_proxy.p.WimaxHardwareEnabled
	results.ActiveConnections = _M.nm_manager_proxy.p.ActiveConnections
	results.PrimaryConnection = _M.nm_manager_proxy.p.PrimaryConnection
	results.PrimaryConnectionType = _M.nm_manager_proxy.p.PrimaryConnectionType
	results.State = _M.nm_manager_proxy.p.State
	results.Connectivity = _M.nm_manager_proxy.p.Connectivity

	return results
end

-----------------------------------------------
-- Spec the device name and/or type name to find special device(s)
-- (ref NM_DEVICE_TYPE_NAMES).
-- If 'name' is nil/""/"*", means any devices
-- If 'typefilter' is nil/"*"/"", means all types
-- Returns a list of device basic informations.
---------------------------------------------
_M.GetDevicesBy = function( name, typefilter )
	if not _M.nm_manager_proxy then
		return nil, "Network manager module is not initialized yet"
	end

	local r, objs = _M.nm_manager_proxy:GetDevices()
	if not r or type(objs) ~= "table" then
		return nil, objs -- The objs as error message returns.
	end

	if not typefilter or typefilter == "" then
		typefilter = "*"
	end

	if not name or name == "" then
		name = "*"
	end

	local results = {}
	local k,v
	for k,v in pairs(objs) do
		local devProxy = _M.dbus_session:GetProxy( v, NM_BUS_NAME, NM_DEVICE_INTERFACE )
		if devProxy then
			local devType = devProxy.p.DeviceType or 0
			local typeStr = _M.NM_DEVICE_TYPE_NAMES[devType] or "UNKNOWN"
			local devName = devProxy.p.Interface or ""
			local PrimaryPort = devName

			--patch Modem!
			if devType == NM_DEVICE_TYPE_MODEM then
				local modemObjs = _M.modemdevices
				if modemObjs then
					local key,value
					for key,value in pairs(modemObjs) do
						if value["org.freedesktop.ModemManager1.Modem"] and 
							value["org.freedesktop.ModemManager1.Modem"].PrimaryPort and 
							value["org.freedesktop.ModemManager1.Modem"].PrimaryPort == devName then
							devName = key
						end
					end
				end
			end

			if (name == "*" or name == devName) and (typefilter == "*" or typefilter == typeStr) then
				local stateReason = devProxy.p.StateReason
				local devbaseinfo = { 	Name=devName, 
										Object=v, -- NOTE it. This is the reference id of device 
										State=devProxy.p.State,
										StateReason = stateReason and stateReason.reason or 0, 
										Type=devType,
										TypeName = typeStr,
										Driver = devProxy.p.Driver,
										DriverVersion = devProxy.p.DriverVersion,
										ActiveConnection = devProxy.p.ActiveConnection, -- To get connection
										AvailableConnections = devProxy.p.AvailableConnections, -- All connections
										PrimaryPort = PrimaryPort
									}		

				table.insert( results, devbaseinfo )
			end
		end
	end
	table.sort( results, function(a,b)
		return a.Name < b.Name
	end)
	return results
end

--------------------------

local function ip4_prefixToMask(pfx)
	--The general using.
	if pfx == 0 then
		return "0.0.0.0"
	elseif pfx == 8 then
		return "255.0.0.0"
	elseif pfx == 16 then
		return "255.255.0.0"
	elseif pfx == 24 then
		return "255.255.255.0"
	elseif pfx == 32 then
		return "255.255.255.255"
	else
		local val = 0xffffffff
		val = (val << (32-pfx)) & 0xffffffff
		local rslt = tostring( (val >> 24) & 0xff )
		rslt = rslt .. "." .. tostring( (val >> 16) & 0xff )
		rslt = rslt .. "." .. tostring( (val >> 8 ) & 0xff )
		rslt = rslt .. "." .. tostring( val & 0xff )
		return rslt
	end
end

---------------------
_M.GetWifiProxy = function(obj)
	return _M.dbus_session:GetProxy( obj, NM_BUS_NAME, NM_WIRELESS_INTERFACE )
end

---------------------

_M.GetWifiAPInfo  = function ( o )
	if not o or o == "" or o == "/" then
		return nil
	end

	local apProxy = _M.dbus_session:GetProxy( o, NM_BUS_NAME, NM_WIFI_AP_INTERFACE )
	if apProxy then
		local results = {}
		results.object = o	--Save the object id.

		local wpaFlags = apProxy.p.WpaFlags
		local rsnFlags = apProxy.p.RsnFlags

		wpaFlags = wpaFlags | rsnFlags;	--NOTE: Need merge these two flags together.
		results.security = {}

		if wpaFlags == 0 then
			results.security["OPEN"] = true
		else
			if 	(wpaFlags&NM_802_11_AP_SEC_PAIR_WEP40) > 0 or
				(wpaFlags&NM_802_11_AP_SEC_PAIR_WEP104) > 0 or
				(wpaFlags&NM_802_11_AP_SEC_GROUP_WEP40) > 0 or
				(wpaFlags&NM_802_11_AP_SEC_GROUP_WEP104) > 0 
			then
				results.security["WEP"] = true
			end

			if (wpaFlags&NM_802_11_AP_SEC_PAIR_TKIP) > 0 or
				(wpaFlags&NM_802_11_AP_SEC_PAIR_CCMP) > 0 or
				(wpaFlags&NM_802_11_AP_SEC_GROUP_TKIP) > 0 or
				(wpaFlags&NM_802_11_AP_SEC_GROUP_CCMP) > 0 or 
				(wpaFlags&NM_802_11_AP_SEC_KEY_MGMT_PSK) > 0 
			then
				results.security["PSK"] = true
			end

			if (wpaFlags&NM_802_11_AP_SEC_KEY_MGMT_802_1X) > 0 then
				results.security["8021X"] = true
			end
		end

		results.Ssid = apProxy.p.Ssid 	--TODO: Extract the SSID need
		results.Ssid_Array = results.Ssid

		if type(results.Ssid) == "table" then
			results.Ssid = string.char( table.unpack(results.Ssid) )
		end

		results.Frequency = apProxy.p.Frequency
		results.HwAddress = apProxy.p.HwAddress
		results.Mode = _M.NM_802_11_MODE_NAMES[ apProxy.p.Mode or 0 ]
		results.MaxBitrate = apProxy.p.MaxBitrate
		results.Strength = apProxy.p.Strength
		results.LastSeen = apProxy.p.LastSeen

		return results
	else
		return nil, "Could not get WIFI AP proxy"
	end
end

---------------------

local function getModemCap(cap)
	local r
	if (cap & 0x01) > 0 then
		r = "DIAL"
	end
	
	if (cap & 0x02) > 0 then
		if r then
			r = r .. ",CDMA/EVDO"
		else
			r = "CDMA/EVDO"
		end
	end

	if (cap & 0x04) > 0 then
		if r then
			r = r .. ",GSM/UMTS/HSPA"
		else
			r = "GSM/UMTS/HSPA"
		end
	end

	if (cap & 0x08) > 0 then
		if r then
			r = r .. ",LTE"
		else
			r = "LTE"
		end
	end

	if not r then
		return ""
	else
		return r
	end
end

----------------------------------------------------
--Get the details about the device.
-- Specify the device 'obj' object id [ what returned
-- by GetDevicesand 'opts' options can
-- be nil or "*" if request all informations, or an 
-- array contains these requires:
--
-- IP4 = true : IP4 configures
-- DHCP4 = true: DHCP4 configures
-- IP6 = true : IP6 configures (not implement yet)
-- DHCP6 = true: DHCP6 configures(not implement yet)

-- GENERIC = true: Generic network informations (hardware address only)
-- WIRE = true: Wired connection informations
-- MODEM = true: Modem informations
-- WIFI = true, Wifi informations

-- NOTE: WIFI scan list and do scan should in seperated
--       functions.
----------------------------------------------------

_M.GetDeviceInfo = function( obj, opts )
	-- _dbg.Error( "NM_DAEMON", "GetDeviceInfo", obj )

	if not _M.nm_manager_proxy then
		return nil, "Network manager module is not initialized yet"
	end
	
	local devProxy, msg = _M.dbus_session:GetProxy( obj, NM_BUS_NAME, NM_DEVICE_INTERFACE )
	if not devProxy then
		return nil, msg
	end

	local checkAll = false
	if type(opts) ~= "table" then
		checkAll = true
	end

	local state = devProxy.p.State or 0
	local devType = devProxy.p.DeviceType or 0
	local stateReason = devProxy.p.StateReason

	--patch modem
	local devName = devProxy.p.Interface or ""
	local PrimaryPort = devName
	if devType == NM_DEVICE_TYPE_MODEM then
		local modemObjs = _M.modemdevices
		local key,value
		for key,value in pairs(modemObjs) do
			if value["org.freedesktop.ModemManager1.Modem"].PrimaryPort == devName then
				devName = key
			end
		end
	end

	local results = {
		Name = devName,
		PrimaryPort = PrimaryPort,
		State = state,
		StateReason = stateReason and stateReason.reason or 0,
		Type = devType,
		ActiveConnection = devProxy.p.ActiveConnection, -- To get connection
		AvailableConnections = devProxy.p.AvailableConnections -- All connections
	}

	-- Process UNManager connections
	if PrimaryPort then
		 local existconsettings = _M.ListConnectionAndGetDetails( { interface = PrimaryPort } )
		 local exist = {}                                                                      
		 local k, v                                                                            
		 for k, v in pairs(results.AvailableConnections) do
			 exist[v] = true
		 end
	     for k, v in pairs(existconsettings) do
			 if not exist[k] then
				 table.insert( results.AvailableConnections, k )
			 end
		 end
	 end

	-- Check IP4
	if state == NM_DEVICE_STATE_ACTIVATED and (checkAll or opts.IP4) then -- Get IP4 informations
		local IP4 = {}
		local o = devProxy.p.Ip4Config
		if o then
			local ip4Proxy = _M.dbus_session:GetProxy( o, NM_BUS_NAME, NM_IP4CONFIG_INTERFACE )
			if ip4Proxy then
				IP4.Addresses = {}
				--Get ip addresses
				local k,v 
				for k,v in pairs(ip4Proxy.p.AddressData) do
					if type(v) == "table" then
						local addr = {}
						addr.address = v["address"]
						addr.prefix = v["prefix"]
						addr.mask = ip4_prefixToMask( addr.prefix )
						table.insert( IP4.Addresses, addr )
					end
				end

				IP4.Gateway = ip4Proxy.p.Gateway or ""	--Default gateway
				IP4.Dns = ip4Proxy.p.Nameservers
				local kk,vv 
				for kk,vv in pairs(IP4.Dns) do
					--print( "DNS:", vv )
					IP4.Dns[kk] = nu.u32ToIp(vv)
				end

				results.IP4 = IP4
			end
		end
	end

	--Check DHCP4
	if state == NM_DEVICE_STATE_ACTIVATED and (checkAll or opts.DHCP4) then -- Get DHCP4 informations.
		--TODO: No used, so ignore it.
		results.DHCP4 = {}
		results.DHCP4.Options = {}
	end

	--TODO: Check IP6/DHCP6 but ignored here.

	if checkAll or opts.GENERIC then -- Check generic informations
		--local genericProxy = _M.dbus_session:GetProxy( obj, NM_BUS_NAME, NM_GENERIC_INTERFACE )
		--if genericProxy then
		--	results.GENERIC = {}
		--	results.GENERIC.hwAddress = genericProxy.p.HwAddress
		--end
	end

	if devType == NM_DEVICE_TYPE_ETHERNET and (checkAll or opts.WIRE) then --Check wired informations
		local wiredProxy = _M.dbus_session:GetProxy( obj, NM_BUS_NAME, NM_WIRED_INTERFACE )
		if wiredProxy then
			results.WIRE = {}
			results.WIRE.HwAddress = wiredProxy.p.HwAddress
			results.WIRE.PermHwAddress = wiredProxy.p.PermHwAddress
			results.WIRE.Speed = wiredProxy.p.Speed
			results.WIRE.Carrier = wiredProxy.p.Carrier
		end
	end

	if devType == NM_DEVICE_TYPE_WIFI and (checkAll or opts.WIFI ) then --Check wifi status
		local wifiProxy = _M.dbus_session:GetProxy( obj, NM_BUS_NAME, NM_WIRELESS_INTERFACE )
		if wifiProxy then
			results.WIFI = {}
			results.WIFI.HwAddress = wifiProxy.p.HwAddress
			results.WIFI.PermHwAddress = wifiProxy.p.PermHwAddress
			results.WIFI.Mode = wifiProxy.p.Mode or 0
			results.WIFI.ModeName = _M.NM_802_11_MODE_NAMES[results.WIFI.Mode]
			results.WIFI.Bitrate = wifiProxy.p.Bitrate or 0
			results.WIFI.ActiveAccessPoint = _M.GetWifiAPInfo( wifiProxy.p.ActiveAccessPoint )
			local cap = wifiProxy.p.WirelessCapabilities
			if (cap&0x200) > 0 then	-- 2.4G support
				results.WIFI.WirelessCapabilities = "2.4G"
			end

			if (cap&0x400) > 0 then -- 5G support
				results.WIFI.WirelessCapabilities = results.WIFI.WirelessCapabilities and "2.4G/5G" or "5G"
			end
		end
	end

	if devType == NM_DEVICE_TYPE_MODEM and (checkAll or opts.MODEM ) then -- Check modem status
		local modemProxy = _M.dbus_session:GetProxy( obj, NM_BUS_NAME, NM_MODEM_INTERFACE )
		if modemProxy then
			results.MODEM = {}
			results.MODEM.ModemCapabilities = getModemCap(modemProxy.p.ModemCapabilities or 0)
			results.MODEM.CurrentCapabilities = getModemCap(modemProxy.p.CurrentCapabilities or 0)
		end
	end

	-- pt.dump( results )
	return results
end

-----------------------------------------------
-- List all connections, by:
-- filterTable is nil : All
-- filterTable can be:
--  interface = "ifname"  : Get only interface name is 'ifname'
--	mac = "mac-address" : Get only specical MAC address
--  type = "gsm/cdma/802-3-ethernet/802-11-wireless" : Get special type
--	uuid = "..." : By special UUID
-----------------------------------------------

_M.ListConnectionAndGetDetails = function( filterTable )
	local r, list = _M.nm_settings_proxy:ListConnections()
	if r and type(list) == "table" then
		local rtable = _M.GetConnectionDetails(list)
		-- pt.dump(rtable)
		if rtable then
			if type(filterTable) == "table" then	--Check filters
				local drops = {}

				local k,v
				for k,v in pairs(rtable) do
					local bye = false
					if type(v) == "table" then
						if type(v.connection) == "table" then
							if filterTable.interface and v.connection["interface-name"] ~= filterTable.interface then
								drops[k] = true
								bye = true
							end

							if not bye and filterTable.type and filterTable.type ~= v.connection["type"] then
								drops[k] = true
								bye = true
							end

							if not bye and filterTable.uuid and filterTable.uuid ~= v.connection["uuid"] then
								drops[k] = true
								bye = true
							end
						end

						if not bye and filterTable.mac then
							if type(v["802-11-wireless"]) == "table" and v["802-11-wireless"]["mac-address"] ~= filterTable.mac then
								drops[k] = true
								bye = true
							end
							
							if not bye and type(v["802-3-ethernet"]) == "table" and v["802-3-ethernet"]["mac-address"] ~= filterTable.mac then
								drops[k] = true
								bye = true
							end
						end
					end -- if v valid
				end --for

				for k,v in pairs(drops) do
					rtable[k] = nil		--Drop filter-out items.
				end
			end -- with filter table.

			return rtable
		else
			return {}
		end
	else
		return nil, "Could not get connections list"
	end
end

-- Return the ethernet configure.
function _M.GetEthernetConfig(devicename)
	if devicename and _M.ethcfg then
		return _M.ethcfg[devicename]
	else
		return nil
	end
end

-------------------------------------------
-- Get one/a set of connection details
-------------------------------------------
_M.GetConnectionDetails = function( obj_or_list,devname )
	local k,v
	local results = {}
	local olist,only4G

	if type(obj_or_list) == "string" then
		olist = {obj_or_list}
	elseif type(obj_or_list) == "table" then
		olist = obj_or_list
	else
		return nil, "Invalid connection object or object list"
	end

	for k,v in pairs(olist) do
		if type(v) == "string" then
			local conProxy = _M.dbus_session:GetProxy( v, NM_BUS_NAME, NM_CONNECTION_INTERFACE )
			if conProxy then
				local r, cfg = conProxy:GetSettings()
				if r and type(cfg) == "table" then
					--print( "CONN:", v, cfg["connection"]["uuid"] )
					--Fix the Dns showing because of LUA type conversion problem.
					if cfg["ipv4"] and cfg["ipv4"]["dns"] then
						local x
						for x=1,#cfg["ipv4"]["dns"],1 do
							cfg["ipv4"]["dns"][x] = nu.u32ToIp( cfg["ipv4"]["dns"][x] )
						end
					end

					results[v] = cfg
				end
			end
		end
	end

	if devname and _M.config[devname] and _M.config[devname]["connection"] then
		only4G = _M.config[devname]["connection"]["only4G"] or "no"
 	end 
	
	return results,only4G
end
-----------------------------------------------
-- Get Activate connection details.
-----------------------------------------------
_M.GetActivateConnectionDetails = function( obj_or_list )
	local k,v
	local results = {}
	local olist

	if type(obj_or_list) == "string" then
		olist = {obj_or_list}
	elseif type(obj_or_list) == "table" then
		olist = obj_or_list
	else
		return nil, "Invalid connection object or object list"
	end

	for k,v in pairs(olist) do
		if type(v) == "string" then
			local conProxy = _M.dbus_session:GetProxy( v, NM_BUS_NAME, NM_ACTIVE_CONNECTION_INTERFACE )
			if conProxy then
				local opts = {}
				opts.Connection = conProxy.p.Connection
				opts.State = conProxy.p.State
				opts.Default = conProxy.p.Default
				opts.Default6 = conProxy.p.Default6
				opts.Vpn = conProxy.p.Vpn
				opts.Id = conProxy.p.Id
				opts.Uuid = conProxy.p.Uuid
				opts.Type = conProxy.p.Type
				results[v] = opts

				--print( "**ActivateConnection", v, opts.Uuid )
			end
		end
	end

	return results
end


-----------------------------------------------
-- Remove a connection
-----------------------------------------------

local function removeconfig (obj)
	if not _M.config then 
		return
	end

	if obj.connection and obj.connection["interface-name"] then 
		local i
		for i=1, 4 do
			local devname =  "modem" .. tostring(i)
			local currentinterface = _M.GetModemInterface(devname)
			if currentinterface and currentinterface == obj.connection["interface-name"] then
				_dbg.Info( "NM_DAEMON", "removeconfig devname:",devname )
				_M.config[devname] = nil
				_M.config:__TRIGGER(true)
			end
		end		
	else
		_dbg.Error( "NM_DAEMON", "removeconfig error: connection or interface-name not exist!" )
	end
end

_M.DeleteConnection = function( obj , rmconfig )
	if rmconfig == nil then 
		rmconfig = true
	end

	--删除前判断连接设置是否存在，不存在直接return！
	local connections = _M.ListConnectionAndGetDetails()
	if not connections[obj] then
		return true
	else
		local conProxy = _M.dbus_session:GetProxy( obj, NM_BUS_NAME, NM_CONNECTION_INTERFACE )
		if not conProxy then
			return false, "Could not get the connection to remove"
		else
			if rmconfig then
				removeconfig(connections[obj])
			end
			conProxy:Delete()
			return true
		end
	end
end

function split( str,reps )
    local resultStrList = {}
    string.gsub(str,'[^'..reps..']+',function ( w )
        table.insert(resultStrList,w)
    end)
    return resultStrList
end

-----------------------------------------------
-- Connect device
-----------------------------------------------
_M.ConnectDevice = function( devObj, connObj )
	if not _M.nm_manager_proxy then
		return false, "Network manager module is not initialized yet"
	end

	if type(devObj) ~= "string" then
		return false, "Invalid network device specified"
	end

	if not connObj or connObj == "" or connObj == "*" then
		connObj = "/"
	end

	local r,o = _M.nm_manager_proxy:ActivateConnection( connObj, devObj, "/" )
	-- print("ActivateConnection", r,o)
	if r and o then
		return true
	else
		return false, o
	end
end

_M.DeactivateDevice = function(devObj)
	local devProxy = _M.dbus_session:GetProxy( devObj, NM_BUS_NAME, NM_DEVICE_INTERFACE )
	if devProxy then
		local ac = devProxy.p.ActiveConnection
		if not ac or ac == "" or ac == "/" then
			return true
		else
			_M.nm_manager_proxy:DeactivateConnection(ac)
			return true
		end
	else
		return false, "Invalid device"
	end
end

-----------------------------------------------
-- Need the parameters fixing
-----------------------------------------------
local NmConnectionTypeFixTable = {
	["connection"] = {
		["autoconnect"] = "b",
		["gateway-ping-timeout"] = "u",
		["read-only"] = "b",
		["timestamp"] = "t",
	},

	["802-3-ethernet"] = {
		["auto-negotiate"] = "b",
		["cloned-mac-address"] = "ay",
		["mac-address"] = "ay",
		["mtu"] = "u",
		["speed"] = "u",
		["wake-on-lan"] = "u",
	},

	["802-11-wireless"] = {
		["bssid"] = "ay",
		["channel"] = "u",
		["cloned-mac-address"] = "ay",
		["hidden"] = "b",
		["mac-address"] = "ay",
		["mtu"] = "u",
		["rate"] = "u",
		["ssid"] = "ay",
		["tx-power"] = "u"
	},

	["802-11-wireless-security"] = {
		["leap-password-flags"] = "u",
		["psk-flags"] = "u",
		["wep-key-flags"] = "u",
		["wep-key-type"] = "u",
		["wep-tx-keyidx"] = "u"
	},

	["serial"] = {
		["baud"] = "u",
		["bits"] = "u",
		["parity"] = "y",
		["send-delay"] = "t",
		["stopbits"] = "u"
	},

	["ppp"] = {
		["baud"] = "u",
		["crtscts"] = "b",
		["lcp-echo-failure"] = "u",
		["lcp-echo-interval"] = "u",
		["mppe-stateful"] = "b",
		["mru"] = "u",
		["mtu"] = "u",
		["no-vj-comp"] = "b",
		["noauth"] = "b",
		["nobsdcomp"] = "b",
		["nodeflate"] = "b",
		["refuse-chap"] = "b",
		["refuse-eap"] = "b",
		["refuse-mschap"] = "b",
		["refuse-mschapv2"] = "b",
		["refuse-pap"] = "b",
		["require-mppe"] = "b",
		["require-mppe-128"] = "b",
	},

	["ipv4"] = {
		["address-data"] = {
			{
				["prefix"] = "u"
			}
		},

		["addresses"] = "aau",

		["dhcp-send-hostname"] = "b",
		["dns"] = "au",
		["ignore-auto-dns"] = "b",
		["ignore-auto-routes"] = "b",
		["may-fail"] = "b",
		["never-default"] = "b",
		["route-data"] = {
			{
				["prefix"] = "u",
				["metric"] = "u"
			}
		},
		["route-metric"] = "x",
		["routes"] = "aau"
	},

	["gsm"] = {
		["home-only"] = "b",
		["password-flags"] = "u",
		["pin-flags"] = "u"
	},

	["cdma"] = {
		["password-flags"] = "u"
	}
}

-------------------------

local function typeCast( typeIdr, val )
	local t = typeIdr:sub(1,1)
	
	if t == "a" then	--This is array
		if type(val) ~= "table" then
			t = typeIdr:sub(2,2)	--Check some specical type, eg. "ay": string to bytes array
			if t == "y" and type(val) == "string" then
				local result = {}
				local i
				for i=1, val:len(), 1 do
					result[i] = l2dbus.DbusTypes.Byte.new( val:byte(i) )
				end
				return result
			else -- not support
				return val	--Not match, ignore
			end
		else
			local k,v
			for k,v in pairs(val) do
				val[k] = typeCast( typeIdr:sub(2), v )
			end

			return val
		end
	elseif t == "y" then
		return l2dbus.DbusTypes.Byte.new( tonumber(val) )
	elseif t == "b" then
		return l2dbus.DbusTypes.Boolean.new( val and true or false )
	elseif t == "n" then
		return l2dbus.DbusTypes.Int16.new( tonumber(val) )
	elseif t == "q" then
		return l2dbus.DbusTypes.Uint16.new( tonumber(val) )
	elseif t == "i" then
		return l2dbus.DbusTypes.Int32.new( tonumber(val) )
	elseif t == "u" then
		--Processing a special IP address style
		if type(val) == "string" then
			local t1,t2,t3,t4 = val:match("(%d+)%.(%d+)%.(%d+)%.(%d+)")
			if t1 and t2 and t3 and t4 then
				--print( "**** Convertion:", t1, t2, t3, t4, nu.ipToU32(val) )
				return l2dbus.DbusTypes.Uint32.new( nu.ipToU32(val) )
			end
		end

		return l2dbus.DbusTypes.Uint32.new( tonumber(val) )
	elseif t == "x" then
		return l2dbus.DbusTypes.Int64.new( tonumber(val) )
	elseif t == "t" then
		return l2dbus.DbusTypes.Uint64.new( tonumber(val) )
	elseif t == "d" then
		return l2dbus.DbusTypes.Double.new( tonumber(val) )
	elseif t == "s" then
		return val	--Ignore it
	elseif t == "o" then
		return val	--Ignore it
	else
		return val
	end
end

local function fixingTheSetParameters(sets, fixTable)
	local group, opt
	local idr, key

	if nil == fixTable then
		fixTable = NmConnectionTypeFixTable
	end

	for idr, group in pairs(sets) do
		local checkIdr = idr
		if type(idr) == "number" then	--Index, all matching
			checkIdr = 1
		end

		if fixTable[checkIdr] then	--With fixing options
			local fixing = fixTable[checkIdr]
	
			for key, opt in pairs(group) do
			
				if type(fixing[key]) == "string" then	--Directly type fix
					--print( "*** Fixed:", key, fixing[key] )
					group[key] = typeCast( fixing[key], opt )
				elseif type(fixing[key]) == "table" and type(opt) == "table" then	--With special members
					--print( "*** Special fixed:", key )
					fixingTheSetParameters( opt, fixing[key] )
				end
			end
	
		end
	end
end

-------------------------
-- For CoAF call.
_M.ConfigEthernet = function(devname, cfg)
	if devname and type(cfg) == "table" then
		local eth_cfg = {
			auto_neg = cfg.auto_neg or "on",
			speed = cfg.speed or "100",
			duplex = cfg.duplex or "full"
		}
		_M.ethcfg[devname] = eth_cfg
		_M.ethcfg:__TRIGGER(true)
		config_ethernet(devname, cfg)
		return true
	else
		return false, "invalid-arguments"
	end
end

-------------------------
--Change the interface settings.
_M.ChangeConnectionDetails = function( obj, sets )
	-- _dbg.Info( "NM_DAEMON", "ChangeConnectionDetails",obj)
	-- pt.dump (sets)
	if type(sets) == "table" and sets["physical"] then
		local physical_set = sets["physical"]
		sets["physical"] = nil --Clear it, since it's not the standard NetworkManager options.
		if _M.ethcfg and physical_set.deviceName then
			local eth_cfg = {
				auto_neg = physical_set.auto_neg or "on",
				speed = physical_set.speed or "100",
				duplex = physical_set.duplex or "full"
			}
			_M.ethcfg[physical_set.deviceName] = eth_cfg
			_M.ethcfg:__TRIGGER(true)
			config_ethernet(physical_set.deviceName, eth_cfg)
		end
	end
	
	local originsets = deepcopy(sets)
	local conProxy = _M.dbus_session:GetProxy( obj, NM_BUS_NAME, NM_CONNECTION_INTERFACE )
	if not conProxy then
		return false, "Could not get the connection to change/update"
	elseif type(sets) ~= "table" then
		return false, "Must special a valid table for changing connection"
	else
		--TODO: I MUST FIX SOME special DBUS TYPES.
		fixingTheSetParameters(sets)
		if sets.devname then 
			sets.devname = nil
		end
		-- conProxy:Update(sets)
		-- conProxy:Save()
		local r,o 
		if sets.connection and sets.connection.type then
			if sets.connection.type == "gsm" or sets.connection.type == "cdma" then
				sets.connection["autoconnect-retries"]=0
				sets.connection["autoconnect-priority"]=100
				r,o = conProxy:UpdateUnsaved (sets)
			else
				r,o = conProxy:Update(sets)
			end
		else
			return false, "connection or connection.type not exist!"
		end
		-- print( conProxy:Save() ) --save can save keyfiles
		-- print("****",r,o)
		if r then
			if sets.connection and sets.connection.type then
				if sets.connection.type == "gsm" or sets.connection.type == "cdma" then
					if originsets.connection.uuid then
	            		originsets.connection.uuid = nil
		   	 		end
					if sets.connection and sets.connection.only4G then
						local modemInfo= _M.GetModemInfo(originsets.devname)
						local revision
						if modemInfo and modemInfo["ModemInfo"] and modemInfo["ModemInfo"]["Revision"] then
							revision = modemInfo["ModemInfo"]["Revision"]
						else
							revision = nil
						end
						if revision and string.find(revision,"EC2")==1 then
							if sets.connection.only4G == "yes" then
								local ret,err = _M.ModemCommand(originsets.devname, [[at+qcfg="nwscanmode",3,1\r\n]])
							elseif sets.connection.only4G == "no" then
								local ret,err = _M.ModemCommand(originsets.devname, [[at+qcfg="nwscanmode",0,1\r\n]])		
							end
						end
					end
					_M.config[originsets.devname] = originsets
					_M.config:__TRIGGER(true)
				end
			end
			return true
		else
			return false, o
		end
	end

end

function _M.GetModemTTY(modemname)
	local modem_path = _M.modemcfg[modemname]
	if not modem_path then
		return nil
	end

	local lfs = require "lfs"
	local maxid = -1
	for item in lfs.dir("/sys/class/tty") do
			local id = item:match("ttyUSB(%d+)")
			if id then
					local f = io.popen("/usr/bin/readlink -f /sys/class/tty/" .. item, "r" )
					if f then
							local path = f:read("*l")
							f:close()
							if stringx.startswith(path, modem_path) then
									if tonumber(id) > maxid then
											maxid = tonumber(id)
									end
							end
					end
			end
	end

	if maxid ~= -1 then
		return "/dev/ttyUSB"..maxid
	else
		return nil
	end
end

_M.GetATPATH = function( modemname )
	if not modemname or not _M.modemdevices[modemname] then
		_dbg.Error( "NM_DAEMON", "no modem obj" )
		return nil
	end

	local modeminfo = _M.modemdevices[modemname]["org.freedesktop.ModemManager1.Modem"]
	if not modeminfo or type(modeminfo) ~= "table" or not modeminfo.PrimaryPort then
		_dbg.Error( "NM_DAEMON", "Get Modem PrimaryPort failed!" )
		return nil
	end	

	local AT_PORT = "/dev/ttyUSB"
	if stringx.startswith(modeminfo.PrimaryPort, "cdc-wdm") then
		return _M.GetModemTTY(modemname)
	elseif stringx.startswith(modeminfo.PrimaryPort, "ttyUSB") then
		AT_PORT = modeminfo.PrimaryPort
	end
	return AT_PORT
end

local function ppptalk(tty, cmd)
	local ppp_talk = "/usr/bin/ppp_talk"
	local args
	if type(cmd) == "string" then
		args = { "-p", tty, "--say", cmd, "-e", [[(.*)/\1/-]] }  --(.*)/\1/.
	elseif type(cmd) == "table" then
		args = { "-p", tty }
		local k, v
		for k, v in pairs(cmd) do
			table.insert(args, "--say")
			table.insert(args, v)
			table.insert(args, "-e")
			table.insert(args, [[(.*)/\1/-]])
		end
	else
		_dbg.Info( "NM_DAEMON", "ppptalk cmd error" )
		return false, "cmd error"
	end
	-- pt.dump(args)
	local pf = utils.xpopen( ppp_talk, args, "r", 10 )
    if not pf then
        return false, "exec-fail", 255
	end
	
	local ret = ""
	local line = pf:read("l")
	while line do
		ret = ret .. line
		-- _dbg.Info( "NM_DAEMON", "ppptalk line:" .. line )
		line = pf:read("l")
	end
	pf:close()
    return true, ret
end

-----------------------------------------------
-- Sent Modem Command
-----------------------------------------------
_M.ModemCommand = function( modemname, cmd )
	if not cmd then
		return false, "no cmd arg"
	end

	local AT_PORT = _M.GetATPATH(modemname)
	if AT_PORT then
		-- local runcmd = 'echo "'.. cmd ..'" > ' .. AT_PORT .. ' &'
		-- -- _dbg.Info( "NM_DAEMON", "run cmd", runcmd )
		-- os.execute( runcmd )
		-- print("ModemCommand", modemname, AT_PORT)
		-- pt.dump(cmd)
		local r,data = ppptalk(AT_PORT, cmd)
		return r,data
	else
		return false, "get at port error"
	end
end
-----------------------------------------------
-- Add a connection
-----------------------------------------------
_M.AddConnection = function( sets , unsave )
	if not unsave then
		unsave = false
	end
	
	if type(sets) ~= "table" or type(sets.connection) ~= "table" or not sets.connection.type or sets.connection.type == "" then
		return nil, "Invalid arguments"
	end
	local originsets = deepcopy(sets)
	fixingTheSetParameters(sets)	
	
	if sets.connection and sets.connection.type then
		if sets.connection.type == "gsm" or sets.connection.type == "cdma" then	
			sets.connection["autoconnect-retries"]=0
			sets.connection["autoconnect-priority"]=100
			
			local devname,setid,current,currentid
			setid = sets.connection.id
			devname = sets.devname
			_dbg.Info( "NM_DAEMON", "Process WWAN Connection", setid, devname )

			-- disable ipv6 解决获取到IPV6 nameserver时无法解析的问题
			sets["ipv6"] = {
				method = "ignore"
			}
			local currentinterface = _M.GetModemInterface(devname)
			if not currentinterface then
				_dbg.Error( "NM_DAEMON", "WWAN device not ready!", setid, devname )
				return nil,"WWAN device not ready!"
			end

			--处理接口改变情况
			if currentinterface ~= sets.connection["interface-name"] then
				_dbg.Info( "NM_DAEMON", "WWAN Interface Changed!")
				sets.connection["interface-name"] = currentinterface
				originsets.connection["interface-name"] = currentinterface
			end

			-- pt.dump(sets)
			local existconsetting = _M.ListConnectionAndGetDetails( { interface = currentinterface } )
			-- pt.dump(existconsetting)
			if existconsetting then
				local k, v
				for k, v in pairs(existconsetting) do
					_dbg.Info( "NM_DAEMON", "Process current connection:",k)
					-- local r, o = _M.DeleteConnection(k,false)
					-- if r then
					-- 	_dbg.Info( "NM_DAEMON", "DeleteConnection Success!")
					-- else
					-- 	_dbg.Info( "NM_DAEMON", "DeleteConnection Failed!")
					-- end

					sets.connection.uuid = v.connection.uuid
					local r, o = _M.ChangeConnectionDetails(k,sets)
					if r then
						_dbg.Info( "NM_DAEMON", "ChangeConnectionDetails Success!")
						_M.config[devname] = originsets
						_M.config:__TRIGGER(true)
						return true
					else
						_dbg.Info( "NM_DAEMON", "ChangeConnectionDetails Failed!")
						return nil, o
					end					
				end
			end

			if sets.devname then 
				sets.devname = nil
			end

			local canSetAPN = false
			if _M.modemdevices[devname] and _M.modemdevices[devname].EC2X then
				canSetAPN = true
			end

			-- local r, o = _M.nm_settings_proxy:AddConnectionUnsaved(sets)
			-- AddConnectionUnsaved: Add new connection but do not save it to disk immediately.
			if sets.gsm and sets.gsm.apn and canSetAPN then
				-- local ret,err = _M.ModemCommand(devname, [[AT+CGDCONT=1,\"IP\",\"]] .. sets.gsm.apn .. [[\"]])
				local ret,err = _M.ModemCommand(devname, [[AT+CGDCONT=1,"IP","]] .. sets.gsm.apn .. [["\r\n]])
				-- print(ret,err)
			end

			if sets.connection and sets.connection.only4G then
				local modemInfo= _M.GetModemInfo(devname)
				local revision
				if modemInfo and modemInfo["ModemInfo"] and modemInfo["ModemInfo"]["Revision"] then
					revision = modemInfo["ModemInfo"]["Revision"]
				else
					revision = nil
				end
				if revision and string.find(revision,"EC2")==1 then
					coaf.ScheduleDelayedTask(function()
						if sets.connection.only4G == "yes" then
							local ret,err = _M.ModemCommand(devname, [[at+qcfg="nwscanmode",3,1\r\n]])
						elseif sets.connection.only4G == "no" then
							local ret,err = _M.ModemCommand(devname, [[at+qcfg="nwscanmode",0,1\r\n]])
						end
					end,3)
				end
			end
			local r, o = _M.nm_settings_proxy:AddConnectionUnsaved(sets)	
			_dbg.Info( "NM_DAEMON", "AddConnectionUnsaved:", r, o)
			if r and o then
				_M.config[devname] = originsets
				_M.config:__TRIGGER(true)
				return o
			else
				return nil, o
			end			
		end
	end

	-- local r, o = _M.nm_settings_proxy:AddConnectionUnsaved(sets)
	-- AddConnectionUnsaved: Add new connection but do not save it to disk immediately.
	local r, o
	if unsave then
		r, o = _M.nm_settings_proxy:AddConnectionUnsaved(sets)
	else
		r, o = _M.nm_settings_proxy:AddConnection(sets)
	end
	-- print( "***",r,o)
	if r and o then
		return o
	else
		return nil, o
	end
end
-----------------------------------------------
-- WifiScan request
-----------------------------------------------

_M.WifiScan = function( obj )
	local WIFI = require "mod.wifiManager"
	return WIFI.WifiScan(obj)
end

--------------------
-- Get WIFI scan list
--------------------
_M.WifiScanList = function( obj )
	local WIFI = require "mod.wifiManager"
	return WIFI.WifiScanList(obj)
end

--------------------
-- Get WIFI scan count
--------------------

_M.WifiScanCount = function( obj )
	local WIFI = require "mod.wifiManager"
	return WIFI.WifiScanCount(obj)
end

--------------------

-- process usb/eth 
local function NewDevice( devicename, deviceinfo )
    if not deviceinfo then
        return nil, "no deviceinfo!"
	end

	--if no connection auto add connection*+
	if deviceinfo.Type == NM_DEVICE_TYPE_ETHERNET and deviceinfo.AvailableConnections and #deviceinfo.AvailableConnections == 0 then
		_dbg.Info( "NM_DAEMON", "add connection", deviceinfo.Name)
		local devName = deviceinfo.Name
		local connSettings = {
                ["connection"] = {},
                ["802-3-ethernet"] = {},
                ["ipv4"] = {}
		}
        connSettings["connection"]["type"] = "802-3-ethernet"
		connSettings["connection"]["autoconnect"] = true        --Force auto connection

		connSettings["connection"]["id"] = devName
		connSettings["connection"]["interface-name"] = devName
		connSettings["ipv4"]["method"] = "auto"
		local ret,err = _M.AddConnection(connSettings, true)
		if not ret then
			_dbg.Error( "NM_DAEMON", "Auto setup ETH error!", err)
			return nil, "Auto setup ETH error!"
		end
	end

	if _M.ethcfg and deviceinfo.Type == NM_DEVICE_TYPE_ETHERNET then
		local cfg = _M.ethcfg[devicename]
		config_ethernet(devicename, cfg)
	end

	_M.devices[devicename] = deviceinfo
	return true
end

_M.CheckSIM = function()
	local k,v
	for k, v in pairs(_M.modemdevices) do
		local getStated = false
		if v.EC2X then
			local AT_PORT = _M.GetATPATH(k)
			if AT_PORT then
				local r,data = ppptalk(AT_PORT, [[AT+QSIMSTAT?\r\n]])
				if r and data then
					-- print("CheckSIM", data)
					local _, q=string.find(data,"%+QSIMSTAT:%s*1,1")
					if q then
						if _M.simcardstat[k] and _M.simcardstat[k] == 0 then
							_dbg.Info( "NM_DAEMON", k, "Insert SIM")
							-- if _M.modemdevices[k] and _M.modemdevices[k].ModemObj  then
							-- 	_M.ResetModem(_M.modemdevices[k].ModemObj)
							-- end
						end
						_M.simcardstat[k] = 1
						getStated = true
					else
						_, q=string.find(data,"%+QSIMSTAT:%s*1,0")
						if q then
							-- _dbg.Error( "NM_DAEMON", k, "No SIM Detect!")
							_M.simcardstat[k] = 0
							getStated = true
						end
					end
				end
			end
		end

		-- pt.dump(v)
		if not getStated then
			if v["org.freedesktop.ModemManager1.Modem"] and v["org.freedesktop.ModemManager1.Modem"].Sim and v["org.freedesktop.ModemManager1.Modem"].Sim ~= "/" then
				_M.simcardstat[k] = 1
			else
				_M.simcardstat[k] = 0
			end
		end
	end
end

_M.Steps = function()
    local ret,err = _M.GetDevicesBy("*")
    if ret then
        local k,v
        local device_item
        local new_devices = {}
        for k,v in pairs(ret) do
            new_devices[v.Name] = v
        end
        local drop = {}
        local new = {}
        for k,v in pairs(_M.devices) do
            if not new_devices[k] then
                table.insert( drop, k )
            end
        end

        for k,v in pairs(new_devices) do
            if not _M.devices[k] then
                table.insert( new, k )
            end
        end
        table.sort(new)
        --process drop devices
        local i
        for i=1, #drop do
            --TODO:
            _M.devices[ drop[i] ] = nil
            collectgarbage()
        end

        --process new devices
        for i=1, #new do
            NewDevice( new[i], new_devices[ new[i] ] )
        end

        for k,v in pairs(_M.devices) do
			_M.UpdateDeviceStatus( k, new_devices[k] )
			if (v.TypeName=="ETHERNET" or v.TypeName=="MODEM" or v.TypeName=="WIFI") then
				static.Start(k,v.State)
			end
		end

    else
        _dbg.Error( "NM_DAEMON", "GetDevicesBy * error:", err )
	end
	_M.GetModemObject()
	_M.CheckSIM()
end



function _M.UpdateDeviceStatus(devicename, deviceinfo)
    -- update info
    if _M.devices[ devicename ] then
        _M.devices[ devicename ] = deviceinfo
    end
    return true
end

--------------------

local function isEC2X(path)
	local vid, pid
	local vf = io.open( path .. "/idVendor", "r" )
	local pf = io.open( path .. "/idProduct", "r" )
	if vf then
		vid = vf:read("l")
		vf:close()
	end

	if pf then
		pid = pf:read("l")
		pf:close()
	end

	if vid and pid and vid == "2c7c" and pid =="0125" then
		return true
	end
	return false
end

function _M.GetModemObject()
	if not _M.dbus_object_proxy then
		return nil, "Network manager bus object proxy is not initialized yet"
	end

	local r, objs = _M.dbus_object_proxy:GetManagedObjects()
	if not r or type(objs) ~= "table" then
		_dbg.Error( "NM_DAEMON", "Could not get dbus_object_proxy", objs )
		return nil, objs -- The objs as error message returns.
	end
	-- pt.dump(objs)
	local k, v
	for k, v in pairs(objs) do
		if v["org.freedesktop.ModemManager1.Modem"] and 
		v["org.freedesktop.ModemManager1.Modem"].Device == ( _M.modemcfg and _M.modemcfg["modem1"] or "/sys/devices/e0000000.ahb/e0018000.ehci/usb2/2-1/2-1.2" ) then
				_M.modemdevices['modem1'] = v
				_M.modemdevices['modem1'].ModemObj = k
				_M.modemdevices['modem1'].EC2X = isEC2X(_M.modemcfg["modem1"])
		end

		if v["org.freedesktop.ModemManager1.Modem"] and 
		v["org.freedesktop.ModemManager1.Modem"].Device == ( _M.modemcfg["modem2"] or "/sys/devices/e0000000.ahb/e0018000.ehci/usb2/2-1/2-1.4" ) then
				_M.modemdevices['modem2'] = v
				_M.modemdevices['modem2'].ModemObj = k
				_M.modemdevices['modem2'].EC2X = isEC2X(_M.modemcfg["modem2"])
		end
		
		if v["org.freedesktop.ModemManager1.Modem"] and 
		v["org.freedesktop.ModemManager1.Modem"].Device == ( _M.modemcfg["modem3"] or "/sys/devices/e0000000.ahb/e0018000.ehci/usb2/2-2" ) then
				_M.modemdevices['modem3'] = v
				_M.modemdevices['modem3'].ModemObj = k
				_M.modemdevices['modem3'].EC2X = isEC2X(_M.modemcfg["modem3"])
		end

		if v["org.freedesktop.ModemManager1.Modem"] and 
		v["org.freedesktop.ModemManager1.Modem"].Device == ( _M.modemcfg["modem4"] or "/sys/devices/e0000000.ahb/e0018000.ehci/usb2/2-1/2-1.3" ) then
				_M.modemdevices['modem4'] = v
				_M.modemdevices['modem4'].ModemObj = k
				_M.modemdevices['modem4'].EC2X = isEC2X(_M.modemcfg["modem4"])
		end
	end
	-- pt.dump(_M.modemdevices)
	return _M.modemdevices
end

function _M.GetDeviceBy( devicename )
	return _M.devices[devicename]
end

function _M.GetPPPbyName( modemName )	--根据modem名获取对应的ppp连接名称
	local ttyport = nil
	local modemdevice = _M.GetDeviceBy(modemName)
	if modemdevice then
		if modemdevice.Type ~= 8 then
			return modemName
		elseif string.sub(modemdevice.PrimaryPort,1,string.len("cdc-wdm"))=="cdc-wdm" then
			local modem = _M.GetModemInfo(modemName)
			if modem and modem.ModemInfo  then
				local _, v
				for _, v in pairs(modem.ModemInfo.Ports) do
					if v[1] and stringx.startswith(v[1], "wwan") then
						return v[1]
					end
				end
			end
		else
			local devinfo = _M.GetDeviceInfo(modemdevice.Object)
			if devinfo then
				if devinfo.IP4 and devinfo.IP4.Addresses and #devinfo.IP4.Addresses > 0 then
					local ip = devinfo.IP4.Addresses[1].address
					-- _dbg.Info("NM_DAEMON", "modemdevice ip:", ip)
					local ips = _M.GetLinksIp()
					if ips[ip] then
						return ips[ip]
					else
						return nil
					end
				end
			else
				_dbg.Error("NM_DAEMON", "GetPPPbyName GetDeviceInfo error!")			
			end
		end
	end
	return nil	
end

function _M.GetLinksIp1()
	local connections = _M.ListConnectionAndGetDetails() ---从连接中获取ip和ppp 刚启动时获取不到!!!
	local ret = {}
	local k, v
	for k, v in pairs(connections) do	
		if v.ipv4 and v.ipv4["address-data"] and #v.ipv4["address-data"] > 0 then			
			if v.ipv4["address-data"][1].address then
				_dbg.Info("NM_DAEMON", v.connection.id, "ip:",v.ipv4["address-data"][1].address)
				ret[v.ipv4["address-data"][1].address] = v.connection.id
			end
		end
	end	
	return ret
end

function _M.GetLinksIp()
	---ifconfig eth0 | grep "inet addr:" | awk '{print $2}' | cut -c 6-
	local ret = {}
	local intfs = funcs.EnumNetworks()   
	local i
	for i = 1, #intfs do
		local linkinfo = funcs.GetLinkDetail(intfs[i])
		-- pt.dump( linkinfo )
		if linkinfo.address and ( linkinfo.address ~= "0.0.0.0" ) then
			ret[linkinfo.address]=intfs[i]
		end
	end
	return ret
end

function _M.GetModemInfo( name , checkall )
	if checkall == nil then
		checkall = false
	end

	local modemdevices = _M.modemdevices
	if not modemdevices then
		return nil,"Get modem objects error"
	end

	local modeminfo = modemdevices[name]

	if not modeminfo then
		return nil,"Get modem object error name: "..name
	end

	local ret = {
		ModemObj = modeminfo.ModemObj,
		ModemInfo = modeminfo["org.freedesktop.ModemManager1.Modem"] or {},
		["3gppInfo"] = modeminfo["org.freedesktop.ModemManager1.Modem.Modem3gpp"] or {},
		SimStat = _M.simcardstat[name] or 0,
	}

	if not checkall then
		return ret
	end

	-- get Bearers
    if modeminfo["org.freedesktop.ModemManager1.Modem"] and modeminfo["org.freedesktop.ModemManager1.Modem"].Bearers
        and type(modeminfo["org.freedesktop.ModemManager1.Modem"].Bearers) == "table" then
        local bearers = {}
        local k,v
        for k, v in pairs(modeminfo["org.freedesktop.ModemManager1.Modem"].Bearers) do
            local bearer = _M.GetBearerInfo( v )
            bearers[v] = bearer
        end
        ret.ModemInfo.Bearers=bearers
	end
	
	-- get SIM
    if modeminfo["org.freedesktop.ModemManager1.Modem"] and modeminfo["org.freedesktop.ModemManager1.Modem"].Sim
        and modeminfo["org.freedesktop.ModemManager1.Modem"].Sim ~= "/" then
        -- _dbg.Error("MM", "GetSimInfo",modeminfo["org.freedesktop.ModemManager1.Modem"].Sim)
        local siminfo = _M.GetSimInfo( modeminfo["org.freedesktop.ModemManager1.Modem"].Sim )
        if siminfo then
            ret.ModemInfo.Sim = siminfo
        else
			-- _dbg.Error("MM", "GetSimInfo error!" ,modeminfo["org.freedesktop.ModemManager1.Modem"].Sim)
			ret.ModemInfo.Sim = {}
        end
    end

	return ret
end

--根据modem名获取usb接口 modem*->ttyUSB*
function _M.GetModemInterface( name ) 
	local modem = _M.GetModemInfo( name, false )
	-- pt.dump(modem)
	if modem then
		if modem.ModemInfo and modem.ModemInfo.PrimaryPort then
			return modem.ModemInfo.PrimaryPort
		end
	end
	return nil
end
   
function _M.GetBearerInfo( obj )
	-- print("GetBearerInfo",obj, MM_BUS_NAME, MM_BEARER_INTERFACE)
	if type(obj) ~= "string" then
		_dbg.Error("NM", "GetBearerInfo obj not a string!")
		return nil
	end
	local proxy = _M.dbus_session:GetProxy( obj, MM_BUS_NAME, MM_BEARER_INTERFACE )
	if not proxy then
		_dbg.Error("NM", "Could not get org.freedesktop.ModemManager1.Bearer interface proxy")
		return nil
	end
	local results = {}
	results.Interface = proxy.p.Interface
	results.Connected = proxy.p.Connected
	results.Suspended = proxy.p.Suspended
	results.Ip4Config = proxy.p.Ip4Config
	results.Ip6Config = proxy.p.Ip6Config
	results.IpTimeout = proxy.p.IpTimeout
	results.Properties = proxy.p.Properties
	return results 
end

function _M.ResetModem( modemname )
	if type(modemname) ~= "string" then
		_dbg.Error("NM", "ResetModem modemname not a string!")
		return false
	end

	if _M.modemdevices[modemname] and _M.modemdevices[modemname].ModemObj then
		local devinfo = _M.GetDeviceBy(modemname)
		-- pt.dump(devinfo)
		if devinfo and devinfo.Object then
			_M.DeactivateDevice(devinfo.Object)	--断开设备连接！

			--删除原来的连接配置
			local currentinterface = devinfo.PrimaryPort
			-- print("currentinterface", modemname, currentinterface )
			if currentinterface then
				local existconsetting = _M.ListConnectionAndGetDetails( { interface = currentinterface } )
				if existconsetting then
					local k, v
					for k, v in pairs(existconsetting) do
						local r, o = _M.DeleteConnection( k, false )
						-- print("ResetModem DeleteConnection:", r,o)
						if r then
							_dbg.Info( "NM_DAEMON", "DeleteConnection Success!")
						else
							_dbg.Info( "NM_DAEMON", "DeleteConnection Failed!")
						end
					end
				end
			end
		end

		local r, d = _M.ModemCommand(modemname, [[AT+CFUN=0\r\n]])
		if not r then
			_dbg.Error("NM", "ModemCommand CFUN=0 error!", d)
			return false
		else
			coaf.ScheduleDelayedTask( function()
				coaf.Sleep(2)
				r, d = _M.ModemCommand(modemname, [[AT+CFUN=1\r\n]])
				if not r then
					_dbg.Error("NM", "ModemCommand CFUN=1 error!", d)
				end
			end, 10 )
		end

		_M.modemdevices[modemname] = nil
		_M.checkThreads[modemname].inited = false
		return true
	else
		_dbg.Error("NM", "GetModem info error!")
		return false
	end
	-- local proxy = _M.dbus_session:GetProxy( obj, MM_BUS_NAME, MM_MODEM_INTERFACE )
	-- if not proxy then
	-- 	_dbg.Error("NM", "Could not get org.freedesktop.ModemManager1.Bearer interface proxy")
	-- 	return nil
	-- end
	-- local r = proxy:Reset()
	-- return r
end



function _M.GetSimInfo( obj )
	if type(obj) ~= "string" then
		-- _dbg.Error("NM", "GetSimInfo obj not a string!")
		return nil
	end

	if obj == "/" then
		return {}
	end

	local proxy = _M.dbus_session:GetProxy( obj, MM_BUS_NAME, MM_SIM_INTERFACE )
	if not proxy then
		_dbg.Error("NM", "Could not get org.freedesktop.ModemManager1.SIM interface proxy")
		return nil
	end
	local results = {}
	results.SimIdentifier = proxy.p.SimIdentifier				--基于IMSI或ICCID的模糊SIM标识符
	results.Imsi = proxy.p.Imsi									--SIM卡的IMSI
	results.OperatorIdentifier = proxy.p.OperatorIdentifier		--运营商标示符
	results.OperatorName = proxy.p.OperatorName					--网络运营商的名称

	return results
end

function _M:LoadConfigureFile(filename)
	local is_ok = os.execute("/bin/cp " .. filename .. " /etc/NetworkManager/system-connections/")
	local f = io.open(filename, "r")
	local line = f:read("l")
	local name 
	while line do
		name = line:match("interface-name=(.+)")
		if name then
			break
		end
		line = f:read("l")
	end
	f:close()

	if not name then
		return false
	end

	os.execute("/usr/bin/nmcli d connect " .."eth0")
	return is_ok
end

function _M:LoadConfigureFile(filename)
	local f = io.open(filename, "r")
	if not f then
		return false
	end
	os.execute("/bin/chmod 600 " .. filename)
	local is_ok = os.execute("/bin/cp " .. filename .. " /etc/NetworkManager/system-connections/")
	local line = f:read("l")
	local name 
	while line do
		name = line:match("interface-name=(.+)")
		if name then
			break
		end
		line = f:read("l")
	end
	f:close()

	if not name then
		return false
	end

	os.execute("/usr/bin/nmcli d connect " .. name)
	return is_ok
end

function _M.Destroy()
end
local BASE_MAC = "18:c3:f4:2"
local function Get_mac()
	local SERIAL_NO = "000000000"
	local mac
	local r,msg = pcall( dofile, "/usr/share/lua/5.3/sys_setting.lua" )
	if r then
		if type(_G.KV.SERIAL_HISTORIES) == "table" and #_G.KV.SERIAL_HISTORIES > 0 then
			SERIAL_NO = _G.KV.SERIAL_HISTORIES[ #_G.KV.SERIAL_HISTORIES ].serial
		end
	end
	if #SERIAL_NO == 13 then
		if tonumber("0x" .. string.sub(SERIAL_NO,9,13)) then
			mac = BASE_MAC .. string.sub(SERIAL_NO,9,9) .. ":" .. string.sub(SERIAL_NO,10,11) .. ":" .. string.sub(SERIAL_NO,12,13)
		end
	end
	return mac
end

function _M.resetMac()
	local mac = Get_mac()
	if not mac then
		return
	end
	print("=================mac============= ",mac)
	--TODO:
	local modemdevice = _M.GetDeviceBy("eth0")
	local devinfo = _M.GetDeviceInfo(modemdevice.Object)
	local specConn = devinfo.AvailableConnections[1]
	local conn = _M.GetConnectionDetails(specConn)
	conn = conn[specConn]
	local connSettings = {
		["connection"] = {},
		["802-3-ethernet"] = {},
		["ipv4"] = {}
	}
	connSettings["connection"]["type"] = "802-3-ethernet"
	connSettings["connection"]["autoconnect"] = true	--Force auto connection.
	connSettings["connection"]["uuid"] = conn["connection"]["uuid"]
	connSettings["connection"]["id"] = conn["connection"]["id"]
	connSettings["connection"]["interface-name"] = conn["connection"]["interface-name"]
	connSettings["802-3-ethernet"]["cloned-mac-address"] = nu.MACToArray(mac)
	connSettings["ipv4"]["method"] = conn["ipv4"]["method"]
	if connSettings["ipv4"]["method"] == "manual" then
		connSettings["ipv4"]["address-data"] = conn["ipv4"]["address-data"]
		connSettings["ipv4"]["gateway"] = conn["ipv4"]["gateway"]
		connSettings["ipv4"]["dns"] = conn["ipv4"]["dns"]
	end

	local r, o = _M.ChangeConnectionDetails(specConn,connSettings)
	if not r then
		print("resetMac error is:",o)
	end
	_M.ConnectDevice(modemdevice.Object)
end

-- 增加一个方法 专门为A1-plus获取网卡信息 只能获取一个模块的网卡信息
_M.getA1DeviceObject = function( modemPort, object )
	local has = nil
	for k, v in pairs(_M.modemdevices) do
		local AT_PORT = _M.GetATPATH(k)
		local SIM = v["org.freedesktop.ModemManager1.Modem"]["Sim"] or "/"
		if AT_PORT and SIM ~= "/" and modemPort == AT_PORT then
			has = true
			break
		end
	end
	if not has then
		return has
	end
	local ret = _M.GetDevicesBy()
	local ethObject = nil
	local i,v 
	for i,v in pairs(ret) do
		local r = _M.GetDeviceInfo(v.Object,{ WIFI = true, IP4 = true })
		if (r["PrimaryPort"] == "eth1" or r["PrimaryPort"] == "eth2") and r["IP4"] then
			ethObject = v.Object
			break
		end
	end
	if ethObject then
		local a = _M.GetDeviceInfo(ethObject,{ WIFI = true, IP4 = true })
		local b = _M.GetDeviceInfo(object,{ WIFI = true, IP4 = true })
		b["IP4"] = a["IP4"]
		b["State"] = a["State"]
		b["ActiveConnection"] = a["ActiveConnection"]
		return b
	end
	return ethObject
end

return _M
