local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local NM = require "mod.networkManager"

local _M = {
	lastDeviceCheckTime = nil,
	lastDeviceUpdateTime = nil,
	deviceUpdatePeriod = 300, -- 5 minutes update wifi interfaces.
	deviceCheckPeriod = 15, -- 15 seconds update.
	apRefreshInterval = 30,

	--Don't call NetworkManager too fast, it eat CPU!
	cached_device_objects = {},
	scanningTasks = {},
	scanResults = {}
}

function _M.Init()
end

local function wifiScanTask(thrObj, obj)
	--_dbg.Info( "NM_DAEMON", "Starting WiFi scanning task for:", obj )
	local wifiProxy = NM.GetWifiProxy(obj)
	if not wifiProxy then
		_dbg.Error( "NM_DAEMON", "Fail to get WiFi proxy for:", obj, "scanning task is terminated." )
	else

		--wifiProxy:RequestScan({})	--No options passed
		
		local r, list = wifiProxy:GetAllAccessPoints()
		if r and list then
			_M.scanResults[obj].count = #list
			local i=1
			local newer = {}
			while not thrObj.terminated and i <= #list do
				newer[ list[i] ] = true
				i = i + 1
			end

			--Loss APs first.
			local k,v
			for k,v in pairs(_M.scanResults[obj].APDetails) do
				if not newer[k] then
					--_dbg.Info( "NM_DAEMON", "Loss SSID:", v.Ssid )
					_M.scanResults[obj].APDetails[k] = nil --Drop old.
				end
			end

			--New APs, quick get details.
			i = 1
			local scanned = {}
			while not thrObj.terminated and i <= #list do
				local details = _M.scanResults[obj].APDetails[ list[i] ]
				if not details then
					details = NM.GetWifiAPInfo( list[i] )
					if details and details.Ssid and details.Ssid ~= "" then
						_M.scanResults[obj].APDetails[ list[i] ] = details
						_M.scanResults[obj].APDetails[ list[i] ].__update_time = os.time()
						--_dbg.Info( "NM_DAEMON", "New SSID scanned:", details.Ssid, details.Strength )
					end
				end
				scanned[i] = true
				i = i + 1
				coaf.Sleep(0.01) --10ms scan one. (Don't too fast, eat CPU)
			end

			--Update exists APs.
			i = 1
			while not thrObj.terminated and i <= #list do
				local details = _M.scanResults[obj].APDetails[ list[i] ]
				if details and not scanned[i] then
					local now = os.time()
					if not details.__update_time or now - details.__update_time < 0 or now - details.__update_time >= _M.apRefreshInterval then
						details = NM.GetWifiAPInfo( list[i] )
						if details and details.Ssid and details.Ssid ~= "" then
							details.__update_time = now
							_M.scanResults[obj].APDetails[ list[i] ] = details
							--_dbg.Info( "NM_DAEMON", "Update SSID:", details.Ssid, details.Strength )
						else
							_M.scanResults[obj].APDetails[ list[i] ] = nil
						end
					end
				end
				i = i + 1
				coaf.Sleep(0.1) --Yield task.(sleep 100ms) !!Don't too fast.
			end
		end
		wifiProxy = nil
	end

	--_dbg.Info( "NM_DAEMON", "WiFi scanning task for:", obj, "is finished." )
	_M.scanningTasks[obj] = nil
end

function _M.WifiScan(obj)
	if not obj or obj == "" then
		return false, "Invalid WiFi object"
	end
	local i, cached
	for i=1,#_M.cached_device_objects do
		if _M.cached_device_objects[i].Object == obj then
			cached = true
			break
		end
	end

	if not cached then --So should update devices list next check (New device found?)
		_M.lastDeviceUpdateTime = nil
	end

	if not _M.scanningTasks[obj] then
		local task = thread.New( wifiScanTask )
		if not task then
			_dbg.Error( "NM_DAEMON", "Fail to start WiFi scanning task for:", obj )
			return false, "Fail to create WiFi scanning task"
		end
		_M.scanningTasks[obj] = task
		if not _M.scanResults[obj] then
			_M.scanResults[obj] = {
				count = 0,
				APDetails = {},
				lastActiveTime = os.time()
			}
		else
			_M.scanResults[obj].lastActiveTime = os.time()
		end
		task:StartNow(obj)
	end
	return true
end

function _M.WifiScanList(obj)
	if not obj or obj == "" then
		return nil, "Invalid WiFi object"
	end

	if not _M.scanResults[obj] then
		if not _M.WifiScan(obj) then
			return nil, "Fail to create WiFi scanning task"
		end
	end

	local scanResult = _M.scanResults[obj]
	if not scanResult or not scanResult.APDetails then
		return {}
	end

	local results = {}
	local seen = {}
	local k,v
	for k,v in pairs(scanResult.APDetails) do
		--NOTE: bug fixed: ignore invalid utf8 ssid names.
		if v and v.LastSeen >= 0 and utf8.len(v.Ssid) and v.Strength then
			--NOTE: Bug fixed: Ignore invalid UTF-8 ssids.
			if not seen[v.Ssid] then
				table.insert( results, v )
				seen[v.Ssid] = { index=#results, strength=v.Strength }
			elseif v.Strength > seen[v.Ssid].strength then
				results[seen[v.Ssid].index] = v	--Replace new one.
				seen[v.Ssid].strength = v.Strength
			end
		end
	end

	table.sort( results, function(a,b)
		return a.Strength > b.Strength
	end )

	return results
end

function _M.WifiScanCount(obj)
	if not obj or obj == "" then
		return nil, "Invalid WiFi object"
	end

	if not _M.scanResults[obj] then
		if not _M.WifiScan(obj) then
			return nil, "Fail to create WiFi scanning task"
		end
	end

	local scanResult = _M.scanResults[obj]
	if not scanResult then
		return 0
	else
		return scanResult.count or 0
	end
end

function _M.UpdateScan()
	local now = os.time()
	if not _M.lastDeviceUpdateTime or (now - _M.lastDeviceCheckTime) >= _M.deviceUpdatePeriod or (now - _M.lastDeviceUpdateTime) < 0 then
		_M.lastDeviceUpdateTime = now
		local devices = NM.GetDevicesBy( "*", "WIFI" )
		if devices then
			_M.cached_device_objects = devices
		end
	end

	if not _M.lastDeviceCheckTime or (now - _M.lastDeviceCheckTime) >= _M.deviceCheckPeriod or (now - _M.lastDeviceCheckTime) < 0 then
		_M.lastDeviceCheckTime = now
		local i
		for i=1, #_M.cached_device_objects do
			_M.WifiScan( _M.cached_device_objects[i].Object )
		end
	end
end

return _M

