local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local NM = require "mod.networkManager"
local event = require "coaf.event"
--------------------------------

_G.g_event = nil

--------------------------------
local _M = {
	device_list_interval = 120,
	check_interval = 180,
	--Store each connection's status, status should be:
	--"checking', "sys-error", "gateway-unreachable", "internet-unreachable", "dns-resolve-error", "internet-ready"
	--each item will be: {
	--	Name = "interface-name",
	--	Type = "interface-type",
	--	Status = "status"
	--}
	connection_status = {
	},

	check_tasks = {
	},

	last_check_times = {
	},
	traceroute_text = {text="",status="finished"},
	traceroute_ticket = nil,
	traceroute_time = nil,
	not_actively_ping = false,
	click_detection = false
}

-----------------------------

local function onTracerouteNotify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.text = msg.text
end

-----------------------------

function _M.TracerouteStart( targetAddr )
	_M.traceroute_text.text = ""
	if not targetAddr or targetAddr == "" then
		return false
	end
	_M.TracerouteSet("stop")
	local ctx = _M.traceroute_text
	local topic = tostring(ctx)
	_M.traceroute_ticket = g_event:Subscribe( topic, onTracerouteNotify, ctx )
	local pid = coaf.Fork( function()
		g_event:NotifyBack( topic, { status="wait" } )
		local pfile = io.popen( "traceroute " .. targetAddr .. "", "r" )
		if not pfile then
			g_event:NotifyBack( topic, { status="fail" } )
		else
			local line = pfile:read("*l")
			local t = ""
			while line do
				if not line then
					break
				end
				t = t .. line .. "<br><br>"
				line = pfile:read("*l")
				g_event:NotifyBack( topic, { status="wait",text = t } )
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				g_event:NotifyBack( topic, { status="fail" } )
			else
				g_event:NotifyBack( topic, { status="finished", text = t } )
			end
		end
		os.exit(0)
	end)
	return true
end

-----------------------------

function _M.TracerouteSet( targetAddr )
	_M.traceroute_time = os.time()
	if not targetAddr or (targetAddr ~="get" and targetAddr ~= "stop") then
		return false
	end
	if targetAddr == "stop" then
		if _M.traceroute_text.status ~= "finished" then
			os.execute( "killall traceroute")
			_M.traceroute_text.status = "finished"
		end
		if _M.traceroute_ticket then
			g_event:Unsubscribe(_M.traceroute_ticket )
			_M.traceroute_ticket = nil
		end
		return true,_M.traceroute_text.text,"finished"
	else
		return true,_M.traceroute_text.text,_M.traceroute_text.status
	end
end

--------------------------------
local function onPingNotify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.code = msg.code
	ctx.lossRate = msg.lossRate
end

--------------------------------
local function Ping( targetAddr, intf )
	if not targetAddr or targetAddr == "" then
		return false, false, 100
	end

	local ctx = {
		status = "wait",
		lossRate = 100
	}

	local topic = tostring(ctx)
	local ticket = g_event:Subscribe( topic, onPingNotify, ctx )
	local pid = coaf.Fork( function()
		local pfile = io.popen( "/bin/ping " .. targetAddr .. " -W 1 -c 3 -q" .. (intf and (" -I " .. intf) or ""), "r" )
		if not pfile then
			g_event:NotifyBack( topic, { status="fail" } )
		else
			local line = pfile:read("*l")
			local pt, pr, lr
			while line do
				--2 packets transmitted, 2 packets received, 0% packet loss
				pt, pr, lr = line:match( "(%d+).*transmitted.*(%d+).*received.*(%d+)%%.*$" )
				if pt and pr and lr then
					break
				end
				line = pfile:read("*l")
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				g_event:NotifyBack( topic, { status="fail" } )
			else
				g_event:NotifyBack( topic, { status="finished", code = code, lossRate = tonumber(lr) } )
			end
		end
		os.exit(0)
	end)

	local r,flag = coaf.WaitPID( pid )
	if not r then
		if flag == "#CoAF-KILL#" then
			os.execute( "/bin/kill " .. pid )
		end
		g_event:Unsubscribe( ticket )
		return false, false, 100
	else
		if ctx.status == "wait" then
			coaf.Sleep(0.1)	--Wait a while
		end
		g_event:Unsubscribe( ticket )
		if ctx.status == "finished" then
			return true, ctx.code == 0, ctx.lossRate
		else
			return true, false, 100
		end
	end
end

--------------------------------
_M.Ping = Ping
--------------------------------
local function onNsLookupNotify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.code = msg.code
	ctx.addresses = msg.addresses
end

--------------------------------
local function NsLookup( targetHost )
	if not targetHost or targetHost == "" then
		return false, nil
	end

	local ctx = {
		status = "wait"
	}

	local topic = tostring(ctx)
	local ticket = g_event:Subscribe( topic, onNsLookupNotify, ctx )
	local pid = coaf.Fork( function()
		local pfile = io.popen( "/usr/bin/nslookup " .. targetHost )
		if not pfile then
			g_event:NotifyBack( topic, { status="fail" } )
		else
			local line = pfile:read("*l")
			local found_name
			local pfx, addr, addrs
			while line do
				if not found_name then
					found_name = line:match( "^[Nn]ame:%s+(.+)$" )
				else
					addr = line:match( "^[Aa]ddress.-:%s+(.+)$" )
					if addr then
						if not addrs then
							addrs = { addr }
						else
							table.insert( addrs, addr )
						end
					end
				end
				line = pfile:read("*l")
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				g_event:NotifyBack( topic, { status="fail" } )
			else
				g_event:NotifyBack( topic, { status="finished", code = code, addresses = addrs } )
			end
		end
		os.exit(0)
	end)

	local r,flag = coaf.WaitPID( pid )
	if not r then
		if flag == "#CoAF-KILL#" then
			os.execute( "/bin/kill " .. pid )
		end
		g_event:Unsubscribe( ticket )
		return false, nil
	else
		if ctx.status == "wait" then
			coaf.Sleep(0.1)	--Wait a while
		end
		g_event:Unsubscribe( ticket )
		if ctx.status == "finished" then
			return ctx.code == 0, ctx.addresses
		else
			return false, nil
		end
	end
end
--------------------------------
_M.NsLookup = NsLookup
--------------------------------

local function checkTask(thrObj, connectionObj, test_global_ips, test_global_domains)
	if not connectionObj or connectionObj == "" then
		_dbg.Error( "NET-STATUS", "Invalid connection object for checking:", connectionObj )
		_M.check_tasks[connectionObj] = nil
		return
	end

	local status_item = _M.connection_status[connectionObj]
	if not status_item then
		status_item = {
		}
		_M.connection_status[connectionObj] = status_item
	end

	local info,msg = NM.GetDeviceInfo( connectionObj, {IP4=true} )
	if info.Name == "usb0" then return end
	if not info or not info.IP4 then
		local i
		for i=1,3 do --Retry 3 times.
			coaf.Sleep(1)
			info,msg = NM.GetDeviceInfo( connectionObj, {IP4=true} )
			if info and info.IP4 then
				break
			end
		end
		if not info or not info.IP4 then
			_dbg.Error( "NET-STATUS", "Fail to get connection IP configurations, for:", connectionObj, ",With error:", msg )
			_M.check_tasks[connectionObj] = nil
			status_item.Status = "sys-error"
			return
		end
	end

	status_item.Name = info.Name
	status_item.Type = info.Type
	status_item.TypeName = NM.NM_DEVICE_TYPE_NAMES[info.Type] or "UNKNOWN"
	status_item.Status = "checking"

	local IP4 = info.IP4
	if not IP4.Gateway or IP4.Gateway == "" then
		_dbg.Warn( "NET-STATUS", "Invalid gateway configured for:", connectionObj, "(", info.Name, ")" )
		_M.check_tasks[connectionObj] = nil
		status_item.Status = "gateway-unreachable"
		return
	end

	local INTF = info.Name
	if info.Type and info.Type == 8 then --modem
		INTF = NM.GetPPPbyName(info.Name)
	end

	local done,reach,result = Ping( IP4.Gateway, INTF )
	if thrObj.terminated then 
		_M.check_tasks[connectionObj] = nil
		status_item.Status = "aborted"
		return
	end

	if not done or not reach or result >= 100 then
		_dbg.Warn( "NET-STATUS", "Gateway unreachable for:", connectionObj, "(", info.Name, ")" )
		_M.check_tasks[connectionObj] = nil
		status_item.Status = "gateway-unreachable"
		return
	end

	local i, check_ok
	for i=1, #test_global_ips do
	    done,reach,result = Ping( test_global_ips[i], info.Name )
	    if thrObj.terminated then 
		_M.check_tasks[connectionObj] = nil
		status_item.Status = "aborted"
		return
            end
	    if done and reach and result < 100 then
	 	_dbg.Info( "NET-STATUS", "Ping target:", test_global_ips[i], "OK with loss rate:", result, "for:", connectionObj, "(", info.Name, ")" )
	 	check_ok = true
	 	break
	    end
	end

	if not check_ok then
		_dbg.Warn( "NET-STATUS", "Internet unreachable for:", connectionObj, "(", info.Name, ")" )
		_M.check_tasks[connectionObj] = nil
		status_item.Status = "internet-unreachable"
		return
	end

	check_ok = false
	for i=1, #test_global_domains do
	    done,result = NsLookup( test_global_domains[i] )
	    if thrObj.terminated then 
		_M.check_tasks[connectionObj] = nil
		status_item.Status = "aborted"
		return
            end
	    if done and result then
	 	_dbg.Info( "NET-STATUS", "Check domain:", test_global_domains[i], "OK for:", connectionObj, "(", info.Name, ")" )
	 	check_ok = true
	 	break
	    end
	end
	if not check_ok then
		_dbg.Warn( "NET-STATUS", "DNS resolve fail for:", connectionObj, "(", info.Name, ")" )
		_M.check_tasks[connectionObj] = nil
		status_item.Status = "dns-resolve-error"
		return
	end

	_dbg.Info( "NET-STATUS", "Internet is reachable for:", connectionObj, "(", info.Name, ")" )
	_M.check_tasks[connectionObj] = nil
	status_item.Status = "internet-ready"
	return
end

--------------------------------
function _M.Init()
	if not _G.g_event then
		_G.g_event = event.createNew( "network_manager", true ) -- Don't start slave thread, since we never use it.
	end
end
--------------------------------
local check_counter = 0

function _M.Check(specObj, force)
	if force == true then
		_M.not_actively_ping = force
	end
	if _M.traceroute_text.status ~= "finished" and (os.time()-_M.traceroute_time)>30 then
		_M.TracerouteSet("stop")
	end
	if not specObj then
		local i,v
		if force then --Drop all current tasks.
		--	for i,v in pairs(_M.check_tasks) do
		--		v:Terminate()
		--	end
		--	_M.check_tasks = {}
		--	_M.connection_status = {}
			_M.last_check_times = {}
		else
			check_counter = check_counter + 1
			if check_counter > 1 then --Just update @1, period 'device_list_interval' seconds.
				if check_counter > _M.device_list_interval then
					check_counter = 0
				end
				return
			end
		end

		local devices = NM.devices
		if not devices then
			_dbg.Error( "NET-STATUS", "Fail to get network devices info" )
			return
		end
		local i,v
		for i,v in pairs(devices) do
			if v.State > 10 then --10 is unmanaged
				if not _M.check_tasks[ v.Object ] then --So create the task.
					local now = os.time()
					if not _M.last_check_times[v.Object] or (now - _M.last_check_times[v.Object]) < 0 or (now - _M.last_check_times[v.Object]) > _M.check_interval then
						local task = thread.New( checkTask )
						if task and _M.not_actively_ping then
							_M.check_tasks[v.Object] = task
							_M.last_check_times[v.Object] = now
							task:Start( v.Object, {"8.8.8.8"}, {"kiloview.com"})
							_M.click_detection = true
						end
					end
				end
			end
		end
	else
		if force and _M.check_tasks[specObj] then
		--	_M.check_tasks[specObj]:Terminate()
		--	_M.check_tasks[specObj] = nil
		--	_M.connection_status[specObj] = nil
			_M.last_check_times[specObj] = nil
		end

		if not _M.check_tasks[specObj] then
			local now = os.time()
			if not _M.last_check_times[specObj] or (now - _M.last_check_times[specObj]) < 0 or (now - _M.last_check_times[specObj]) > _M.check_interval then
				local task = thread.New( checkTask )
				if task and _M.not_actively_ping then
					_M.last_check_times[specObj] = now
					_M.check_tasks[specObj] = task
					task:Start( specObj, {"8.8.8.8"}, {"kiloview.com"})
					_M.click_detection = true
				end
			end
		end
	end
end

--------------------------------
function _M.GetCheckStatus(specObj)
	if not specObj or specObj == "" then
		return "idle", specObj, nil, nil
	end
	if _M.connection_status[specObj] then
		return _M.connection_status[specObj].Status, specObj, _M.connection_status[specObj].Name, _M.connection_status[specObj].Type
	else
		return "idle", specObj, nil, nil
	end
end

--------------------------------
function _M.GetSummaryStatus()
	local best_status
	local internet_ready = {
	}
	local k,v
	for k,v in pairs(_M.connection_status) do
		if v.Status == "internet-ready" then
			best_status = "internet-ready"
			table.insert( internet_ready, { Object=k, Name=v.Name, Type=v.Type, TypeName = v.TypeName } )
		elseif v.Status == "dns-resolve-error" then
			if best_status ~= "internet-ready" then
				best_status = "dns-resolve-error"
			end
		elseif v.Status == "gateway-unreachable" then
			if not best_status then
				best_status = "gateway-unreachable"
			end
		end
	end

	if best_status == "internet-ready" then
		table.sort( internet_ready, function(a,b)
			return a.Type < b.Type and a.Name < b.Name
		end)
		_M.not_actively_ping = false
		return best_status, internet_ready
	elseif _M.click_detection and _M.not_actively_ping then
		return "checking", nil
	else
		return  "", nil
	end
end

--------------------------------
return _M

