local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

require "kmCore"

local _M = {
	file_location = "/tmp/hls",
	pusher_groups = nil,
	created_sessions = {
	},
}

----------------------------------------------

local hls_meta = {}

function hls_meta:AddSession( idr, mediaSource, session_name, muxrate, options )

	if not mediaSource or not session_name or session_name == "" then
		_dbg.Error( "HLS", "Invalid HLS arguments (MediaSource/Session name ...). Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	if self.session then
		_dbg.Warn( "HLS", "The HLS session is already created. Stream:", self.stream_id, ", IDR:", idr )
		return self.session
	end

	if _M.created_sessions[ session_name ] then
		_dbg.Error( "HLS", "The session '" .. session_name .. "' is already created by others! Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	self.session = _M.pusher_groups:AddSession( "hls:" .. self.stream_id .. ":" .. idr, mediaSource, 4*1024*1024, 512*1024 )
	if not self.session then
		_dbg.Error( "HLS", "Fail to create HLS ts fragment session for Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end
	
	self.location = _M.file_location .. "/" .. session_name
	self.pushSession = self.location .. "/playlist.m3u8"
	self.segmentTime = options and options.segmentTime or 5
	self.segmentTime = math.ceil(self.segmentTime/1000) --To second(s)
	-- todo maxSegments default 4
	self.maxSegments = math.max( math.min(math.modf(400 / self.segmentTime / math.modf(muxrate / 1024 / 1024)), 4) , 2) --options and options.maxSegments or 4
	self.maxWrap = options and options.max_wrap or 0
	if self.maxWrap > 0 then
		self.segmentFilename = self.location .. "/SEG-%04d.ts"
	else
		self.segmentFilename = self.location .. "/SEG-%Y%m%d%H%M%S.ts"
	end

	os.execute( "/bin/mkdir -p " .. self.location )
	--os.execute( "/bin/rm -Rf " .. self.location .. "/*.ts" )
	os.execute( "/bin/rm -Rf " .. self.location .. "/*" )

	_M.pusher_groups:SetOption( self.session, {
		PushPath = self.pushSession,
		PushProtocolType = "hls", --NOTE: The new version of moon_streamer should set type to 'hls'(not 'ts') to correct the timestamp problem.
		hls_time = self.segmentTime, --NOTE: This parameter should in 'second' unit.
		hls_list_size = self.maxSegments,
		hls_wrap = self.maxWrap ,
		hls_segment_filename = self.segmentFilename,
	})

	_M.created_sessions[ session_name ] = {
		streamer = self,
	}

	_dbg.Info( "HLS", "Created HLS session, for Stream:", self.stream_id, ", IDR:", idr, ", Session:", session_name )
	self.started = false
	self.session_idr = idr
	self.session_name = session_name
	return self.session
end

function hls_meta:RemoveSession()

	if self.session then
		self:StopSession()
		_M.created_sessions[ self.session_name ] = nil
		_M.pusher_groups:RemoveSession( self.session )
		_dbg.Info( "HLS", "HLS session is destroied, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Session:", self.session_name )
		self.segmentTime = nil
		self.maxSegments = nil
		self.maxWrap = nil
		self.segmentFilename = nil
		self.pushSession = nil
		self.session = nil
		self.session_idr = nil
		self.session_name = nil
		collectgarbage()

		os.execute( "/bin/rm -Rf " .. self.location .. "/*" )
		self.location = nil
	end
	return true
end

function hls_meta:StartSession()
	if self.session and not self.started then
		_M.pusher_groups:StartSession( self.session )
		self.started = true
		_dbg.Info( "HLS", "HLS session is started, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Session:", self.session_name )
	end
	return self.started
end

function hls_meta:StopSession()
	if self.session and self.started then
		_M.pusher_groups:StopSession( self.session )
		self.started = false
		_dbg.Info( "HLS", "HLS session is stopped, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Session:", self.session_name )
	end
	return true
end

function hls_meta:Destroy()
	self:RemoveSession()
end

hls_meta.__index = hls_meta
----------------------------------------------

function _M.GetStreamer( stream_id )
	stream_id = tonumber(stream_id)
	stream_id = string.format( "%02d", stream_id )

	if not _M.pusher_groups then
		_M.pusher_groups = KM_STREAMING.NetPusher.createNew( "hls-ts-fragment" )
		if not _M.pusher_groups then
			_dbg.Error( "HLS", "Fail to create TS fragment net group!" )
			return nil
		end
	end

	local streamer = {
		stream_id = stream_id,
	}

	return setmetatable( streamer, hls_meta )
end
----------------------------------------------

return _M


