#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"

-- Change current dir to app location
coaf.ChDir(coaf.GetAppDir(arg))

local socket = require "cosocket.core"
local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
local service = require "coaf.service"
local event = require "coaf.event"

local roll = require "coaf.log.rollfile"

local withLog = false
local _x
for _x=1, #arg do
	if arg[_x] == "--log" then
		withLog = true
		break
	end
end

if withLog then
	-- ** initialize the log file **
	local roll = require "coaf.log.rollfile"
	os.execute( "/bin/mkdir -p /data/log/" )
	local logF = roll.Create("/data/log/streamer.log")
	if not logF or not logF:init() then
		_dbg.Error("STREAMER", "Fail to open log file: /data/log/streamer.log, so the log messages should be output to console.")
		withLog = false
	else
		_dbg.AddLogger("logfile", logF, nil, nil)
	end
	-- ** log **
end

local dbusCli = require "kiloview.dbusClients"
coaf.Init( "streamer_service" )

_G.print = function(...)
	_dbg.Info1( "STREAMER", ... )
end

if withLog then
	local stdio = require "coaf.log.stdio"
	stdio.SetDefaultFilterName( "STREAMER" )
	stdio.Start()
end

--NOTE: This is the CoAF event.
_G.streamer_event = event.createNew( "streamer", true ) --Don't start slave thread.

--This is an old version (for SIP-Endpoint) event driver.
require "levent"	--We need the lua-event module.
local levent = LuaEvent.instance()

--The global environments.
_G._ENV = {}

require "features"

local mo_server
local mo_applications
local mo_streams
local mo_audioEngine
local mo_junshi

--------

--TO setup the network global settings (network buffer size)
--65536 131072 262144 524288 1048576 2097152
os.execute( "/sbin/sysctl -w net.core.rmem_max=2097152" )
os.execute( "/sbin/sysctl -w net.core.wmem_max=2097152" )

os.execute( "/sbin/sysctl -w net.core.rmem_default=131072" )
os.execute( "/sbin/sysctl -w net.core.wmem_default=1048576" )

os.execute( [[/sbin/sysctl -w net.ipv4.tcp_rmem="131072 262144 2097152"]] )
os.execute( [[/sbin/sysctl -w net.ipv4.tcp_wmem="131072 1048576 2097152"]] )

os.execute( [[/sbin/sysctl -w net.ipv4.udp_mem="131072 1048576 2097152"]] )

os.execute( "/sbin/sysctl -w net.ipv4.udp_rmem_min=131072" )
os.execute( "/sbin/sysctl -w net.ipv4.udp_wmem_min=131072" )

--------

dbusCli.Init( "SYSTEM" )
service.InitSocket()
client.InitSocket( socket, adapter )

--------

local cosvc_Streamer
local fast_timer
local period_timer
local serial_timer
local web_url_timer

--------

local function loopSerial()
	mo_streams.check_serial()
end

local function loopUpdateWebUrl()
	mo_streams.check_web_url()
end

local function periodCheck()
	mo_server.Check()
	mo_applications.Check()
	mo_streams.Check()
end

local function fastCheck()
	if mo_junshi then
		mo_junshi.Step()
	end
	mo_streams.CheckSip()
	levent:Polling(1) --As quickly as possible. (1ms wait to save cpu)
end

local function Main()
	mo_server = require "server"
	mo_applications = require "applications"
	mo_streams = require "streams"
	mo_audioEngine = require "audio_engine"
	
	if _G.OEM_TYPE == "JUNS1" or _G.OEM_TYPE == "JUNS2" then
		mo_junshi = require "busobj.junshi"
	end

	levent:SetEnable(true)
	levent:SetWaitTimeout(3000)	--3secs call timeout
	print( "*** KILOVIEW SUNSHINE streaming server running ***" )

	cosvc_Streamer = service.New( "streamer", "unix:/var/run/streamer.sock" )
	if cosvc_Streamer then
		_dbg.Info( "STREAMER", "CoService of 'streamer' is created." )
		cosvc_Streamer:CreateCoServerProxy()
	else
		_dbg.Error( "STREAMER", "Fail created CoService!" )
	end

	mo_audioEngine.Init( cosvc_Streamer )
	mo_server.Init( cosvc_Streamer )
	mo_applications.Init( cosvc_Streamer )
	mo_streams.Init( cosvc_Streamer )

	cosvc_Streamer:ExposeObject( "/streams", mo_streams )

	os.execute("(/usr/local/bin/checktally.sh &)")
	
	if mo_junshi then
		mo_junshi.Init()
	end

	--占用CPU约5%
	fast_timer = thread.NewTimer( fastCheck, 0.01, 0.10 )
	period_timer = thread.NewTimer( periodCheck, 0.01, 1 )
	--占用CPU约2%
	if _G.FEATURE_NDI_HX then
		serial_timer = thread.NewTimer(loopSerial, 0.01, 3)
		web_url_timer = thread.NewTimer(loopUpdateWebUrl, 0.01, 3)
	end

	if fast_timer then
		fast_timer:start()
	end
	if period_timer then
		period_timer:start()
	end
	if serial_timer then
		serial_timer:start()
	end

	if web_url_timer then
		web_url_timer:start()
	end
end

--------

coaf.Start(Main)

--------
_dbg.Info( "STREAMER", "Streamer service is terminated." )

