local rpc = require "kiloview.ringRpc"
local nu = require "kiloview.netUtil"

local _M = {}

--Get default WWan device.
function _M.GetDefaultWWan(objId)
	local r, devices = rpc.Call( "NETWORK_MANAGER.GetDevicesBy", "*", "MODEM" )
	if not r or not devices or #devices < 1 then
		return nil
	elseif not objId or objId == "" or objId == "/" then
		return devices[1]
	else
		local k,v
		for k,v in pairs(devices) do
			if v.Object == objId then
				return v
			end
		end
		return nil
	end
end


--Get WWan connections (saved)
function _M.GetConnections(obj)
	--local objPath = _M.utils.url_encode(obj)

	local r, devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", obj, {GENERIC=true} )	--No options, so only get base informations.
	if not r or type(devInfo) ~= "table" then
		return nil
	end

	local activeUuid = ""
	if devInfo.ActiveConnection and devInfo.ActiveConnection ~= "" and devInfo.ActiveConnection ~= "/" then
		local activeConnInfo
		r, activeConnInfo = rpc.Call( "NETWORK_MANAGER.GetActivateConnectionDetails", devInfo.ActiveConnection )
		if r and activeConnInfo then
			activeUuid = activeConnInfo[devInfo.ActiveConnection].Uuid
		end
	end

	local connections = {}

	if type(devInfo.AvailableConnections) == "table" then
		local connDetails                           --GetConnectionDetails
		r, connDetails = rpc.Call( "NETWORK_MANAGER.GetConnectionDetails", devInfo.AvailableConnections )
		-- r, connDetails = rpc.Call( "NETWORK_MANAGER.ListConnectionAndGetDetails", { interface=devInfo.Name } )
		if r and connDetails then
			local k,v 
			for k,v in pairs(connDetails) do
				local title
				local connType
				local addrResolv
				local isCurrent = false
				local gsm_apn = ""
                local gsm_number = ""
                local gsm_user = ""
                local gsm_pass = ""
                local gsm_pin = ""
                local cdma_number = ""
                local cdma_user = ""
                local cdma_pass = ""

				title = v["connection"]["id"]
				if not title or title == "" then
					title = v["802-11-wireless"]["ssid"] or "UNKNOWN"
				end

				if v["connection"]["uuid"] and v["connection"]["uuid"] == activeUuid then
					isCurrent = true
				end

				if v["connection"]["type"] then
					if v["connection"]["type"] == "cdma" then
						connType = "cdma"
					else
						connType = "gsm"
					end
				else
					connType = "other"
				end

				if v["ipv4"] and v["ipv4"]["method"] then
					if v["ipv4"]["method"] == "auto" then
						addrResolv = "dhcp"
					else
						addrResolv = "manual"
					end
				end

                if connType == "gsm" and v["gsm"] then
                    gsm_apn = v["gsm"]["apn"] or ""
                    gsm_number = v["gsm"]["number"] or ""
                    gsm_user = v["gsm"]["username"] or ""
                    gsm_pass = v["gsm"]["password"] or ""
                    gsm_pin = v["gsm"]["pin"] or ""
                elseif connType == "cdma" and v["cdma"] then
                    cdma_number = v["cdma"]["number"] or ""
                    cdma_user = v["cdma"]["username"] or ""
                    cdma_pass = v["cdma"]["password"] or ""
                end

				table.insert( connections, {
					title = title,
					connType = connType,
					addrResolv = addrResolv,
					isCurrent = isCurrent,
					connObj = k,
					uuid = v["connection"]["uuid"] or "",
                    GSM_APN = gsm_apn,
                    GSM_Number = gsm_number,
                    GSM_User = gsm_user,
                    GSM_Password = gsm_pass,
                    GSM_Pin = gsm_pin,
                    CDMA_Number = cdma_number,
                    CDMA_User = cdma_user,
                    CDMA_Password = cdma_pass
				})
			end	--For
		end --If get details
	end -- If Avaliable connections got.
	
	return connections
end

--Set/new WWan connection
function _M.SetConnection( deviceObj, new, args )
    if not args or not args.Type then
        return false, "Invalid arguments"
    end

	local r, devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", deviceObj, {MODEM = true} )	--Get MODEM informations
	if not r or not devInfo then
		return false, "ERROR: Fail to get WWan device informations"
	end

	local devName = devInfo.Name

	local connSettings = {
		["connection"] = {},
		["ipv4"] = { method="auto" },
		devname = devName
	}

	connSettings["connection"]["type"] = args.Type
	connSettings["connection"]["autoconnect"] = ( args.AutoConnect == 1 or args.AutoConnect == "1" )
	connSettings["connection"]["interface-name"] = devInfo.PrimaryPort

	if new then	--New a connection.
		if not args.Title or args.Title == "" then
			args.Title = "WWAN-Connection"
		end

		connSettings["connection"]["id"] = args.Title
	else --Change a connection
		if args.Uuid and args.Uuid ~= "" then
			connSettings["connection"]["uuid"] = args.Uuid
			local r, detail = rpc.Call( "NETWORK_MANAGER.ListConnectionAndGetDetails", {uuid=args.Uuid} )
			if r and detail then
				local k,v
				local found
				for k,v in pairs(detail) do
					found = v
					break
				end
				
				if found then
					connSettings["connection"]["id"] = found["connection"]["id"]
				end
			end
		end
	end

	if args.Type == "gsm" then
		if not args.GSM_APN or args.GSM_APN == "" then
			return false, "Invalid GSM APN"
		end

		connSettings["gsm"] = {}
		connSettings["gsm"]["apn"] = args.GSM_APN or ""
		connSettings["gsm"]["number"] = args.GSM_Number or ""
		connSettings["gsm"]["username"] = type(args.GSM_User) and args.GSM_User ~= "" and args.GSM_User or nil
		
		local psk = args.GSM_Password
		if psk == "~dont~change" then
			psk = nil
		end

		connSettings["gsm"]["password"] = psk

		psk = args.GSM_Pin
		if psk == "~dont~change" then
			psk = nil
		end

		connSettings["gsm"]["pin"] = psk
	else
		connSettings["cdma"] = {}
		connSettings["cdma"]["number"] = args.CDMA_Number or ""
		connSettings["cdma"]["username"] = type(args.CDMA_User) and args.CDMA_User ~= "" and args.CDMA_User or nil
		
		local psk = args.CDMA_Password
		if psk == "~dont~change" then
			psk = nil
		end

		connSettings["cdma"]["password"] = psk

	end

	if new then
		local r, o = rpc.Call( "NETWORK_MANAGER.AddConnection", connSettings ) --addConnection
		if r and o then
			return true
		else
			return false, "ERROR: Add connection fail"
		end
	else
		local r, o = rpc.Call( "NETWORK_MANAGER.ChangeConnectionDetails", args.Object, connSettings )
		if r and o then
			return true
		else
			return false, "ERROR: Change connection fail"
		end
	end
end

--Forget/Delete a connection
function _M.DeleteConnection( conPath )
	if not conPath then
		return false, "ERROR: Invalid connection object"
	end

	local r, code = rpc.Call( "NETWORK_MANAGER.DeleteConnection", conPath )
	if r and code then
		return true
	else
		return false, "ERROR: Fault to delete WWan connection"
	end
end

--Connect a connection
function _M.ConnectConnection(obj, conPath )
	if not obj then
		return false, "ERROR: Invalid WWan device"
	end

	if not conPath then
		return false, "ERROR: Invalid connection object"
	end

	local r, code = rpc.Call( "NETWORK_MANAGER.ConnectDevice", obj, conPath )
	if r and code then
		return true
	else
		return false, "ERROR: Connect fail"
	end
end

--Disconnect WWan
function _M.DisconnectConnection( obj )
	if not obj then
		return false, "ERROR: Invalid WWan object"
	end

	local r, code = rpc.Call( "NETWORK_MANAGER.DeactivateDevice", obj )
	if r and code then
		return true
	else
		return false, "ERROR: Disconnect fail"
	end
end

--Get WWan detail
function _M.GetWWanDetails( device )
	if not device then
		return nil
	end

	local rslt = {}
	
	local r,devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", device.Object, { MODEM = true, IP4 = true } )
	if devInfo and devInfo.MODEM and devInfo.MODEM.CurrentCapabilities then
		rslt.Capabilities = devInfo.MODEM.CurrentCapabilities
	else
		rslt.Capabilities = ""
	end

	rslt.StateReasonCode = devInfo.StateReason
	
	if devInfo.StateReason ~= nil and devInfo.StateReason > 0 then
		local r,reason = rpc.Call("DEVICE_NAMES.GetDeviceStateReasonName", devInfo.StateReason, lang )
		if r and reason then
            	    rslt.StateReason = reason
	        else
        	    rslt.StateReason = ""
		end
        else
        	rslt.StateReason = ""
	end

	if devInfo.IP4 then
		if type(devInfo.IP4) == "table" and devInfo.IP4.Addresses and #devInfo.IP4.Addresses > 0 then
			rslt.IP = devInfo.IP4.Addresses[1].address
			rslt.Prefix = devInfo.IP4.Addresses[1].prefix
			rslt.Gateway = devInfo.IP4.Gateway
		else
			rslt.IP = ""
			rslt.Prefix = 0
			rslt.Gateway = ""
		end
	else
		rslt.IP = ""
		rslt.Prefix = 0
		rslt.Gateway = ""
	end

	return rslt
end

function _M.GetModemInfo ( device )                                                                                  
        if not device then                                                                                        
            return nil                                                     
        end                                                                  
                                                                                                             
        local rslt = {}                                      
                                                                     
        local r,devInfo = rpc.Call( "NETWORK_MANAGER.GetModemInfo", device )
        if devInfo then                                   
            return devInfo                                                
        else                                                          
            return nil
        end
end

function _M.ModemCommand ( device, command )
	if not device or not command then
		return false, "ERROR: miss args"
	end
	local r,done,info = rpc.Call( "NETWORK_MANAGER.ModemCommand", device, command )
	if r then
		return done, info
	else
		return false, "ERROR: send AT Command fail"
	end
end

--------------------------------------------------------

return _M


