------------------------------------------------------
-- KILOVIEW Device HTTP Api, v1.0
-- Copyright (c) 2016, KILOVIEW Electronics.
------------------------------------------------------

--[[

##ABSTRACT{G=RECORD}: 获得指定磁盘上的所有录像文件列表

##ARGUMENTS{GET,POST}:
	@@Device{STR,*}: 要获取录像文件列表的磁盘设备号（请参考 ^^REF{getDisks.lua}）
    @@Stream{STR}: 可选。 
        指定要获取录像文件列表的码流。如未指定，或值为空，表示获取所有码流的录像文件列表。
        %%main: 主码流
        %%sub: 子码流
		%%all: 全部
	~~

##RETURNS: 
	++Data{ARR}: 数组。每一项代表一个录像序列信息。这些信息格式如下：
	++{OBJ}:
		**Stream{STR}: 当前序列对应的码流
			%%main: 主码流
			%%sub: 子码流
		~~

		**RecordDevice{STR}: 录像磁盘的设备号
		**SequenceID{STR}: 录像序列号。
		>>{BOLD} 每一次录像称为一个“序列”，考虑到文件切割的情况，一个序列中可能有多个录像文件。SequenceID代表该序列的唯一编号。

		**RecordPath{STR}: 录像文件存放位置（目录）。该目录同时映射在HTTP的访问目录下。

		++Files{ARR}: 当前序列中的所有文件清单，数组。每一个元素为以下格式：
			++{OBJ}:
				**Filename{STR}: 文件名
				**Size{INT}: 文件大小(KB)
			~~
		~~
		~~

		**TotalSize{INT}: 当前序列所有文件的大小(KB)
		**TotalDuration{INT}: 当前序列的持续录像时间（秒）
		**StartTime{STR}: 整个序列的开始录像时间(yyyy-mm-dd hh:nn:ss)

		**Format{STR}: 录像文件格式
			%%ts: TS文件格式
			%%mp4: MP4文件格式（暂不支持）
		~~

		**IsCurrent{INT}: 是否为当前正在录像的序列。1表示True, 0表示False
	~~
	~~

##NOTES:

录像文件的完整访问地址是 "<b>RecordPath</b>/<b>Filename</b>"，例如 RecordPath = /media/usb/sda1/RECORD/M-20170611-150311-0001，对应文件名为 REC00000001.ts，
则访问地址为 /media/usb/sda1/RECORD/M-20170611-150311-0001/REC00000001.ts，您可以通过http直接得到（下载）该文件。

]]

local cgilua = require "cgilua"
local rpc = require "kiloview.ringRpc"

local SAPI = SAPI or cgilua.SAPI
SAPI.Response.contenttype( "application/json; charset=utf-8" )

local dev, filter
local args = require("M_Request").VerifyAndGetArgs()
if not args or not args.Device or args.Device == "" then
	cgilua.put([[
	{
		"Result":400,
		"Status":"ERROR: Invalid disk"
	}
]])
	return
end

dev = args.Device
filter = ""

if args.Stream == "main" then
	filter = "M.*"
elseif args.Stream == "sub" then
	filter = "S.*"
end

local r,cc,info = rpc.Call( "BUS.DiskRecordList", dev, filter )
if not r or not cc or type(info) ~= "table" then
	cgilua.put([[
	{
		"Result":400,
		"Status":"ERROR: Fail to get recording list"
	}
]])
	return
end

cgilua.put([[
{
	"Result":200,
	"Status":"OK",
	"Data":[
]])

local x

for x=1, #info do
	if x > 1 then
		cgilua.put( ",\n" )
	end

	cgilua.put([[
	{
		"Stream":"]] .. (info[x].STREAM or "") .. [[",
		"RecordDevice":"]] .. (info[x].DEVICE or "unknown") .. [[",
		"SequenceID":"]] .. (info[x].SEQUENCE or "") .. [[",
		"RecordPath":"]] .. (info[x].PATH or "") .. [[",
		"Files":[
]])

	if type(info[x].FILES) == "table" then
		local i
		for i=1, #info[x].FILES do
			if i>1 then
				cgilua.put( ",\n" )
			end
			local cFile = info[x].FILES[i].FILENAME
			local cSize = math.ceil(info[x].FILES[i].SIZE)
			cgilua.put( [[{ "Filename":"]] .. cFile .. [[", "Size":]] .. cSize .. [[}]] )
		end
	end

	cgilua.put([[
		],
]])

	local seqStartTime = "1970-01-01 00:00:00" 
	local seqDur = 0
	if info[x].START_TIME and info[x].START_TIME > 0 then
		if info[x].END_TIME and info[x].END_TIME > 0 then
			seqDur = info[x].END_TIME - info[x].START_TIME
		end
		local t = os.date("*t", math.floor(info[x].START_TIME))
		seqStartTime = string.format("%04d-%02d-%02d %02d:%02d:%02d", t.year, t.month, t.day, t.hour, t.min, t.sec)
	end
	
	cgilua.put([[
		"TotalSize":]] .. math.ceil(info[x].SIZE or 0) .. [[,
		"TotalDuration":]] .. seqDur .. [[,
		"StartTime":"]] .. seqStartTime .. [[",
		"Format":"]] .. (info[x].FORMAT or "ts") .. [["
	}
]])

end --For

cgilua.put([[
	]
}
]])

