------------------------------------------------------
-- KILOVIEW Device HTTP Api, v1.0
-- Copyright (c) 2016, KILOVIEW Electronics.
------------------------------------------------------

--[[

##ABSTRACT{G=RECORD}: 获得当前录像状态

##ARGUMENTS{GET,POST}:
    @@Stream{*}: 
        指定要获取录像状态的码流。
        %%main: 主码流
        %%sub: 子码流
	~~

##RETURNS: 
	++Data{OBJ}:
		**Stream{STR}: 当前码流
			%%main: 主码流
			%%sub: 子码流
		~~

		**Status{STR}: 当前码流的录像状态
			%%started: 正在录像 [注：仅当Status为started时，后续参数才有效]
			%%idle: 空闲，没有录像
			%%error: 错误（未知的）
			%%error.start.nodisk: 没有可用的存储器
			%%error.start.nospace: 存储空间不足
			%%error.recording.nospace: 因存储空间不足，强制停止录像
			%%error.start.not-start: 码流没有开启
			%%error.start.not-exist: 指定的流不存在
			%%error.start.no-source: 没有可用的媒体源
			%%error.start.start-fail: 启动录像失败
			%%error.status-error.update-record: 更新状态时发生错
		~~

		**RecordDevice{STR}: 录像磁盘的设备号
		**SequenceID{STR}: 当前录像的序列号。
		>>{BOLD} 每次开启的一次录像称为一个“序列”，考虑到文件切割的情况，一个序列中可能有多个录像文件。SequenceID代表该序列的唯一编号。

		**RecordPath{STR}: 录像文件存放位置（目录）。该目录同时映射在HTTP的访问目录下。

		++Files{ARR}: 当前序列中的所有文件清单，数组。每一个元素为以下格式：
			++{OBJ}:
				**Filename{STR}: 文件名
				**Size{INT}: 文件大小(KB)
			~~
		~~
		~~

		**CurrentFilename{STR}: 当前正在录制的文件名
		**CurrentSize{INT}: 当前文件大小(KB)
		**CurrentDuration{INT}: 当前文件的持续录像时间（秒）
		**CurrentStartTime{STR}: 当前文件的开始录像时间(yyyy-mm-dd hh:nn:ss)

		**TotalSize{INT}: 当前序列所有文件的大小(KB)
		**TotalDuration{INT}: 当前序列的持续录像时间（秒）
		**StartTime{STR}: 整个序列的开始录像时间(yyyy-mm-dd hh:nn:ss)

		**Format{STR}: 录像文件格式
			%%ts: TS文件格式
			%%mp4: MP4文件格式（暂不支持）
		~~
	~~

##NOTES:

录像文件的完整访问地址是 "<b>RecordPath</b>/<b>Filename</b>"，例如 RecordPath = /media/usb/sda1/RECORD/M-20170611-150311-0001，对应文件名为 REC00000001.ts，
则访问地址为 /media/usb/sda1/RECORD/M-20170611-150311-0001/REC00000001.ts，您可以通过http直接得到（下载）该文件。

]]

local cgilua = require "cgilua"
local rpc = require "kiloview.ringRpc"

local SAPI = SAPI or cgilua.SAPI
SAPI.Response.contenttype( "application/json; charset=utf-8" )

local stream
local args = require("M_Request").VerifyAndGetArgs()
if not args or not args.Stream or args.Stream == "" then
	cgilua.put([[
	{
		"Result":400,
		"Status":"ERROR: Invalid arguments"
	}
]])
	return
end

stream = args.Stream

local streamId = 0
if stream == "sub" then
	streamId = 1
else
	stream = "main"
end

local r,st,info = rpc.Call( "BUS.GetRecordStatus", streamId )
if not r or not st or type(info) ~= "table" then
	cgilua.put([[
	{
		"Result":400,
		"Status":"ERROR: Fail to get recording status"
	}
]])
	return
end

st = st:gsub(":",".")

cgilua.put([[
{
	"Result":200,
	"Status":"OK",
	"Data":{
		"Stream":"]] .. stream .. [[",
		"Status":"]] .. st .. [[",
		"RecordDevice":"]] .. (info.DEVICE or "unknown") .. [[",
		"SequenceID":"]] .. (info.SEQUENCE or "") .. [[",
		"RecordPath":"]] .. (info.PATH or "") .. [[",
		"Files":[
]])

local curSize = 0
local curFile = info.CURRENT_FILE or ""

if type(info.FILES) == "table" then
	local i
	for i=1, #info.FILES do
		if i>1 then
			cgilua.put( ",\n" )
		end
		local cFile = info.FILES[i].FILENAME
		local cSize = math.ceil(info.FILES[i].SIZE)
		if curFile == "" then
			curFile = cFile
		end
		if cFile == curFile then
			curSize = cSize
		end
		cgilua.put( [[{ "Filename":"]] .. cFile .. [[", "Size":]] .. cSize .. [[}]] )
	end
end

cgilua.put([[
		],
]])

local curStartTime = "1970-01-01 00:00:00" 
local curDur = 0
if info.CURRENT_START_TIME and info.CURRENT_START_TIME > 0 then
	--if info.CURRENT_END_TIME and info.CURRENT_END_TIME > 0 then
	--	curDur = info.CURRENT_END_TIME - info.CURRENT_START_TIME
	--end
	curDur = os.time() - info.CURRENT_START_TIME

	local t = os.date("*t", math.floor(info.CURRENT_START_TIME))
	curStartTime = string.format("%04d-%02d-%02d %02d:%02d:%02d", t.year, t.month, t.day, t.hour, t.min, t.sec)
end

local seqStartTime = "1970-01-01 00:00:00" 
local seqDur = 0
if info.START_TIME and info.START_TIME > 0 then
	--if info.END_TIME and info.END_TIME > 0 then
	--	seqDur = info.END_TIME - info.START_TIME
	--end
	seqDur = os.time() - info.START_TIME
	local t = os.date("*t", math.floor(info.START_TIME))
	seqStartTime = string.format("%04d-%02d-%02d %02d:%02d:%02d", t.year, t.month, t.day, t.hour, t.min, t.sec)
end

cgilua.put([[
		"CurrentFilename":"]] .. curFile ..[[",
		"CurrentSize":]] .. math.ceil(curSize) .. [[,
		"CurrentDuration":]] .. curDur .. [[,
		"CurrentStartTime":"]] .. curStartTime .. [[",
		"TotalSize":]] .. math.ceil(info.SIZE or 0) .. [[,
		"TotalDuration":]] .. seqDur .. [[,
		"StartTime":"]] .. seqStartTime .. [[",
		"Format":"]] .. (info.FORMAT or "ts") .. [["
	}
}
]])
