------------------------------------------------------
-- KILOVIEW Device HTTP Api, v1.0
-- Copyright (c) 2016, KILOVIEW Electronics.
------------------------------------------------------

--[[

##ABSTRACT{G=ENCODING}: 获取指定码流（主码流/子码流）的编码参数和/或状态

##ARGUMENTS{GET,POST}: 
    @@Stream{*}: 
        指定要获取的码流。
        %%main: 主码流
        %%sub: 子码流
    ~~

    @@Config{INT}: 可选。
        %%1: 指明需要获取编码参数配置。
        %%0: (或未指定，默认）不返回编码参数配置。
    ~~

    @@Status{INT}: 可选。
        %%1: 指明需要获取编码状态（默认）。
        %%0: 不返回编码状态。
    ~~
##RETURNS: 
	++Data{OBJ}:
        **Stream: 当前码流。
            %%main: 主码流
            %%sub: 子码流
        ~~

        **SnapUrl: http://...
        >> 用于获取预览视频图片的HTTP URL（注：本质上，无论哪个码流，SnapUrl获得的内容均相同，因为只有一个MotionJPEG压缩服务）
        
       
        ++Config{OBJ}: 编码配置参数 (请求参数中如果指定Config=1返回)
           **Scaling: 编码缩放大小。
                %%default: 表示不缩放，对主码流而言，输入视频是什么格式，编码大小就是多大。
                       >> 对子码流而言，默认大小是720x576。
                %%(w)x(h): (w)表示宽度，(h)表示高度（如1280x720）。指定的缩放大小。
            ~~
            
            **Chrome{INT}: 0|1, 0表示彩色，1表示灰度

            **Codec: H264|H265, 视频编码算法（目前仅支持H264）

            **Profile: base|main|high, 指定H.264编码的Profile

            **BRCtrl: cbr|vbr, Bitrate控制方式，cbr或vbr。

            **Bitrate{INT}: 编码码率，单位bps。如1000000表示1Mbps

            **FrameRateMode: 
             编码帧率模式
             %%default: 表示全帧率，即输入视频帧率为多少，编码帧率为多少
             %%half: 表示当输入视频的帧率 >=50 时，自动降半帧率
             %%custom: 表示手动指定帧率
            ~~

            **FrameRate{NUM}: 指定的编码帧率（如果FrameRateMode为default/half，则不必要指定FrameRate值）

            **GOPSize{INT}: H.264 GOP大小（决定I帧之间的间隔），以帧数为单位

            **RefFrame{INT}: 1|2, 参数帧。1表示单一参考帧，2表示多参考帧

            **BindAudioChannel{INT}: 绑定的音频编码引擎通道号, -1表示无音频绑定
        ~~
        ~~
       
        ++Status{OBJ}: 当前实时编码状态信息 (请求参数中如果指定Status=1返回)
            **Codec: H264|H265, 当前编码算法
            **Profile: base|main|high,当前编码的Profile
            **BindAudioChannel{INT}: 绑定的音频编码引擎通道号, -1表示无音频绑定（同Config中的同名参数）

            **Width{INT}: 实际编码的视频宽度
            **Height{INT}: 实际编码的视频高度
            **FrameRate{NUM}: 实际编码的帧率
            **Bitrate{INT}: 视频实际编码的实时码率(Kbps)
        ~~
	~~
##NOTES:
    注：音频如果需要获取实时状态，请根据BindAudioChannel参数的ID，另行获取音频的实时编码状态（参见^^REF{getAudioEngChannel.lua}）
]]

local cgilua = require "cgilua"
local rpc = require "kiloview.ringRpc"

local SAPI = SAPI or cgilua.SAPI
SAPI.Response.contenttype( "application/json; charset=utf-8" )

local args = require("M_Request").VerifyAndGetArgs()

if not args or not args.Stream or args.Stream == "" then
	cgilua.put([[
{
	"Result":400,
	"Status":"ERROR: Invalid arguments"
}
]])
	return
end

local stream = 0
local getConfig = false
local getStatus = true

if args.Stream:lower() == "sub" then
	stream = 1
end
if tonumber(args.Config) == 1 then
	getConfig = true
end
if tonumber( args.Status ) == 0 then
	getStatus = false
end

local host = cgilua.servervariable"SERVER_ADDR"
local port = 80
local r,cc,rets = rpc.Call( "BUS.Get", { ["webConfig.port"] = 80 } )
if r and rets then
	port = rets["webConfig.port"]
end

cgilua.put([[
{
	"Result":200,
	"Status":"OK",
	"Data":{
		"Stream":"]] .. (stream == 0 and "main" or "sub") .. [[",
		"SnapUrl":"http://]] .. host .. ":" .. port .. [[/actions/snap.lua?stream=]] .. stream .. [["
]])

if getConfig then
	local argPrefix = "encoder.stream[" .. stream .. "]"
	local reqParams = {
		["streamer.stream[" .. stream .."].audioChannel"] = -1,
		[argPrefix .. ".video.scale"] = "default",
		[argPrefix .. ".video.profile"] = "high",
		[argPrefix .. ".video.br_ctrl"] = "cbr",
		[argPrefix .. ".video.gop_size"] = 60,
		[argPrefix .. ".video.framerate_mode"] = "default",
		[argPrefix .. ".video.framerate"] = 0,
		[argPrefix .. ".video.quality"] = 80,
		[argPrefix .. ".video.ref_frames"] = 1,
		[argPrefix .. ".video.chrome"] = 1,
		[argPrefix .. ".video.bitrate"] = 1
	}
	local r, cc, gets = rpc.Call( "BUS.Get", reqParams )
	if r and gets then
		cgilua.put([[,
		"Config":{
			"Scaling":"]] .. gets[ argPrefix .. ".video.scale" ] .. [[",
			"Chrome":]] .. gets[argPrefix .. ".video.chrome"] .. [[,
			"Codec":"H264",
			"Profile":"]] .. gets[argPrefix .. ".video.profile"] .. [[",
			"BRCtrl":"]] .. gets[argPrefix .. ".video.br_ctrl"] .. [[",
			"Bitrate":]] .. gets[argPrefix .. ".video.bitrate"] ..[[,
			"FrameRateMode":"]] .. gets[argPrefix .. ".video.framerate_mode"] .. [[",
			"FrameRate":]] .. gets[argPrefix .. ".video.framerate"] .. [[,
			"GOPSize":]] .. gets[argPrefix .. ".video.gop_size"] .. [[,
			"RefFrame":]] .. gets[argPrefix .. ".video.ref_frames"] .. [[,
			"BindAudioChannel":]] .. gets["streamer.stream[" .. stream .."].audioChannel"] .. [[
		}]])
	end
end

if getStatus then
	local x, cfgs = rpc.Call( "BUS.GetStreamStatus", string.format("%02d",stream) )
	if x and cfgs then
		cgilua.put([[,
		"Status":{
			"Codec":"]] .. cfgs["video.codec"] .. [[",
			"Profile":"]] .. cfgs["video.profile"] .. [[",
			"Width":]] .. cfgs["video.width"] .. [[,
			"Height":]] .. cfgs["video.height"] .. [[,
			"FrameRate":]] .. cfgs["video.frameRate"] .. [[,
			"Bitrate":]] .. cfgs["video.bitrate"] .. [[
		}]])
	end
end

cgilua.put( [[
	}
}
]])
