#!/usr/bin/env lua
local coaf = require "coaf.core"
local thread = require "coaf.thread"
local socket = require "cosocket.core"
local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"

local function Usage(app)
	print("Usage:")
	print("\t" .. app .. " <command> [args...]" )
	print("")
	print("Supported commands:")
	print("\tlocations | L")
	print("\tservices  | S")
	print("\tdetail    | D")
	print("\tstart")
	print("\tstop")
	print("\tenable")
	print("\tdisable")
	print("\tsignal     | kill")
	print("\tsignal_all | killall")
	print("\tid")
	print("\tstart_disco")
	print("\tstop_disco")
	print("\tstart_pub")
	print("\tstop_pub")

	print("\tlink_token")
	print("\tadd_target")
	print("\tremove_target")

	print("\tblacklist")
	print("\tdebug")
	print("\tadd_f")
	print("\tclear_f")
	print("\trun")
	print("\thelp      | --help | -h")
	print("")
	print("You can use 'help <command>' to get command details.")
end

local cmd_tips = {
	["locations"] = {	"locations [name-filter] [status-filter]",
				"Get specified or configured/online locations.",
				"> name-filter   : can be '*'/'-', a specified name or regular expression",
				"> status-filter : can be '*'/'-', 'configured', 'online'"
			},

	["services"] = {	"services [name-filter] [status-filter]",
				"Get specified or configured/disabled/online services.",
				"> name-filter   : can be '*'/'-', a specified service name '<LOC>/<ID>' or regular expression",
				"> status-filter : can be '*'/'-', 'configured', 'disabled' or 'online'"
			},

	["detail"] = {	"detail <service-id>", "Get service detail information." },
	["start"] = { "start <service-id>", "Start specified service." },
	["stop"] = { "stop <service-id>", "Stop specified service." },
	["enable"] = { "enable <service-id> [save]", "Enable specified service (remove it from blacklist). 'save' option controls write blacklist on storage." },
	["disable"] = { "disable <service-id> [save]", "Disable specified service (add it to blacklist). 'save' option controls write blacklist on storage." },
	["signal"] = { "signal <service-id> [sig]", "Like the system command 'kill' to signal service. 'sig' can be SIGxxx (e.g. SIGINT), or signal number." },
	["signal_all"] = { "signal_all [sig]", "Like the system command 'killall' to signal all services. 'sig' can be SIGxxx (e.g. SIGINT), or signal number." },
	["blacklist"] = { "blacklist", "Get blacklist services." },
	["debug"] = { "debug [rx|tx|all] <on|off>", "Control to show Rx/Tx protocol data for debugging." },
	["add_f"] = { "add_f <debug_filters>", "Add debug filter(s), should be CMD=xxx LOC=xxx SVC=xxx OBJ=xxx separate them by space or ';' or ','. Multiple same filter is allowed, and you can use any Lua string pattern." },
	["clear_f"] = { "clear_f", "Clear all added debug filters" },
	["start_disco"] = { "start_disco [group-match]", "Start discovery, match specified 'group-match' ('*' means all, or lua string pattern)." },
	["stop_disco"] = { "stop_disco", "Stop discovery" },
	["start_pub"] = { "start_pub [group=<group name>] [link_token=<link author key>] [key=value ...]", "Start publish service, with optional key=value pairs" },
	["stop_pub"] = { "stop_pub", "Stop publish service" },
	["id"] = { "id <unique-id>", "Set discovery unique ID" },
	["link_token"] = { "link_token <TOKEN>", "Set the link token" },
	["add_target"] = { "add_target <location> <ip> <port> <TOKEN>", "Add a manual target" },
	["remove_target"] = { "remove_target <location>", "Remove a manual target" },
	["run"] = { "run <script-file>", "Run specified script file." }
}

local function HelpCommand( app, cmd )
	local tips = cmd_tips[cmd]
	if not tips then
		print("ERROR: Invalid command '" .. cmd .. "'." )
	else
		print("Usage:")
		print( "\t" .. app .. " " .. tips[1] )
		print("")
		local i
		for i=2, #tips do
			print( tips[i] )
		end
	end
end


coaf.Init("CoCtrl")
client.InitSocket(socket, adapter)

local function GetLocations(proxy, args, argIdx)
	local name_fil, type_fil
	if args[argIdx] then
		name_fil = args[argIdx]
		if name_fil == "-" then
			name_fil = "*"
		end
		if args[argIdx+1] then
			type_fil = args[argIdx+1]
			if type_fil == "-" then
				type_fil = "*"
			end
		end
	end
	local r, ret, addrs = proxy.GetLocations( name_fil, type_fil )
	if not r then
		print("Control Error:", ret)
	else
		local k, v
		local fmts = {}
		local maxw = 8
		for k,v in pairs(ret) do
			if #k > maxw then
				maxw = #k
			end
			table.insert(fmts, {k, v})
		end
		table.sort( fmts, function(a,b)
			return a[1] < b[1]
		end )
		local line = string.rep("-", maxw + 3 + 12 + 3 + 15)
		print(line)
		print(string.format(" %-" .. maxw .. "s | %-12s | %s", "Location", "Status", "Address"))
		print(line)
		for k,v in ipairs(fmts) do
			print(string.format(" %-" .. maxw .. "s | %-12s | %s", v[1], v[2], addrs[v[1]] or "-"))
		end
		print(line)
	end
end

local function GetServices(proxy, args, argIdx)
	local name_fil, type_fil
	if args[argIdx] then
		name_fil = args[argIdx]
		if name_fil == "-" then
			name_fil = "*"
		end
		if args[argIdx+1] then
			type_fil = args[argIdx+1]
			if type_fil == "-" then
				type_fil = "*"
			end
		end
	end
	local r, ret = proxy.GetServices( name_fil, type_fil )
	if not r then
		print("Control Error:", ret)
	else
		local k, v
		local fmts = {}
		local maxw = 7
		for k,v in pairs(ret) do
			if #k > maxw then
				maxw = #k
			end
			table.insert(fmts, {k, v})
		end
		table.sort( fmts, function(a,b)
			return a[1] < b[1]
		end )
		local line = string.rep("-", maxw + 3 + 12)
		print(line)
		print(string.format(" %-" .. maxw .. "s | %s", "Service", "Status"))
		print(line)
		for k,v in ipairs(fmts) do
			print(string.format(" %-" .. maxw .. "s | %s", v[1], v[2]))
		end
		print(line)
	end
end

local function GetDetail(proxy, args, argIdx)
	local svcId = args[argIdx]
	if type(svcId) ~= "string" then
		print("ERROR: Invalid service id!")
		return
	end

	local loc, id = svcId:match("(.*)/(.*)$")
	if not loc then
		loc = "local"
		id = svcId
	end

	local r, ret = proxy.GetServiceDetail(loc, id)
	if not r then
		print("Control error:", ret)
	else
		if not ret then
			print("ERROR: Fail to get service detail, perhaps specified invalid service ID?")
		else
			print("Service: " .. svcId)
			print("--------------------------")
			print("Exact service : " .. ret.exact_service)
			print("Status        : " .. ret.status)
			print("Address       : " .. (ret.address or "-"))
			print("Port          : " .. (ret.port or "-"))
			print("--------------------------")
		end
	end
end

local function StartService(proxy, args, argIdx)
	local svcId = args[argIdx]
	if type(svcId) ~= "string" then
		print("ERROR: Invalid service id!")
		return
	end

	local loc, id = svcId:match("(.*)/(.*)$")
	if not loc then
		loc = "local"
		id = svcId
	end

	local r, ret = proxy.StartService(loc, id)
	if not r then
		print("Control error:", ret)
	else
		if ret then
			print("Start service DONE:", svcId)
		else
			print("Start service FAIL:", svcId)
		end
	end
end

local function StopService(proxy, args, argIdx)
	local svcId = args[argIdx]
	if type(svcId) ~= "string" then
		print("ERROR: Invalid service id!")
		return
	end

	local loc, id = svcId:match("(.*)/(.*)$")
	if not loc then
		loc = "local"
		id = svcId
	end

	local r, ret = proxy.StopService(loc, id)
	if not r then
		print("Control error:", ret)
	else
		if ret then
			print("Stop service DONE:", svcId)
		else
			print("Stop service FAIL:", svcId)
		end
	end
end

local function SignalService(proxy, args, argIdx)
	local svcId = args[argIdx]
	if type(svcId) ~= "string" then
		print("ERROR: Invalid service id!")
		return
	end

	local loc, id = svcId:match("(.*)/(.*)$")
	if not loc then
		loc = "local"
		id = svcId
	end

	local sig
	if args[argIdx+1] then
		sig = tonumber(args[argIdx+1])
		if sig then
			if sig < 0 then --Consider '-1' like option.
				sig = -sig
			end
		else
			sig = args[argIdx+1]
		end
	end

	local r, ret = proxy.SignalService(loc, id, sig)
	if not r then
		print("Control error:", ret)
	else
		if ret then
			print("Signal service DONE:", svcId)
		else
			print("Signal service FAIL:", svcId)
		end
	end
end

local function SignalAll(proxy, args, argIdx)
	local sig
	if args[argIdx] then
		sig = tonumber(args[argIdx])
		if sig then
			if sig < 0 then --Consider '-1' like option.
				sig = -sig
			end
		else
			sig = args[argIdx]
		end
	end

	local r, ret = proxy.SignalAll(sig)
	if not r then
		print("Control error:", ret)
	else
		if ret then
			print("Signal all DONE!")
		else
			print("Signal all FAIL!")
		end
	end
end

local function EnableService(proxy, args, argIdx)
	local svcId = args[argIdx]
	if type(svcId) ~= "string" then
		print("ERROR: Invalid service id!")
		return
	end

	local loc, id = svcId:match("(.*)/(.*)$")
	if not loc then
		loc = "local"
		id = svcId
	end

	local save
	if args[argIdx+1] and args[argIdx+1] == "save" then
		save = true
	end

	local r, ret = proxy.EnableService(loc, id, save)
	if not r then
		print("Control error:", ret)
	else
		if ret then
			print("Enable service DONE:", svcId)
		else
			print("Enable service FAIL:", svcId)
		end
	end
end

local function DisableService(proxy, args, argIdx)
	local svcId = args[argIdx]
	if type(svcId) ~= "string" then
		print("ERROR: Invalid service id!")
		return
	end

	local loc, id = svcId:match("(.*)/(.*)$")
	if not loc then
		loc = "local"
		id = svcId
	end

	local save
	if args[argIdx+1] and args[argIdx+1] == "save" then
		save = true
	end

	local r, ret = proxy.DisableService(loc, id, save)
	if not r then
		print("Control error:", ret)
	else
		if ret then
			print("Disable service DONE:", svcId)
		else
			print("Disable service FAIL:", svcId)
		end
	end
end

local function DebugControl(proxy, args, argIdx)
	local opt1 = args[argIdx]
	local rx = true
	local tx = true
	if opt1 == "rx" or opt1 == "tx" or opt1 == "all" or opt1 == "-" then
		if opt1 == "rx" then
			tx = false
		elseif opt1 == "tx" then
			rx = false
		end

		opt1 = args[argIdx+1]
	end
	if not opt1 then
		HelpCommand( args[0], "debug" )
	else
		local onOff = (opt1 == "on")
		local r = true
		local ret
		if rx then
			r, ret = proxy.EnableDebugRxDetail(onOff)
		end
		if r and tx then
			r, ret = proxy.EnableDebugTxDetail(onOff)
		end
		if not r then
			print("Control error:", ret)
		else
			print("DONE.")
		end
	end
end

local function Run(proxy, args, argIdx)
	local file = args[argIdx]
	if not file then
		print("ERROR: Invalid script file specified.")
	end
	local r, ret, err = proxy.Run(file)
	if not r then
		print("Control error:", ret)
	else
		if not ret then
			print("Run fail with message:", err)
		else
			print("DONE.")
		end
	end
end

local function GetBlacklist(proxy, args, argIdx)
	local r, ret = proxy.GetBlacklist()
	if not r then
		print("Control error:", ret)
	elseif not ret then
		print("ERROR: Fail to get blacklist!")
	else
		print("Blacklist:")
		print("--------------------------------")
		local i, v
		for i,v in ipairs(ret) do
			print(v)
		end
		print("--------------------------------")
	end
end

local function Help(proxy, args, argIdx)
	if not args[argIdx] then
		Usage(args[0])
	else
		HelpCommand( args[0], args[argIdx] )
	end
end

local function AddDebugFilter(proxy, args, argIdx )
	local fil = args[argIdx]
	if not fil then
		print( "ERROR: Invalid filter string(s)" )
		return
	end

	local i
	for i=argIdx+1, #args do
		fil = fil .. "," .. args[i]
	end

	local r, ret = proxy.AddDebugFilter(fil)
	if not r then
		print("Control error:", ret)
	else
		print("Added debug filter:", fil)
	end
end

local function ClearDebugFilter(proxy, args, argIdx )
	local r, ret = proxy.ClearDebugFilter()
	if not r then
		print("Control error:", ret)
	else
		print("Debug filters are cleared.")
	end
end

local function StartDiscovery(proxy, args, argIdx )
	proxy:ChangeObject( "/discovery" )
	local r, ret = proxy.StartDiscovery(args[argIdx])
	if not r then
		print("Control error:", ret)
	else
		print("Discovery service is started.")
	end
end

local function StopDiscovery(proxy, args, argIdx )
	proxy:ChangeObject( "/discovery" )
	local r, ret = proxy.StopDiscovery()
	if not r then
		print("Control error:", ret)
	else
		print("Discovery service is stopped.")
	end
end

local function SetUniqueID(proxy, args, argIdx )
	if not args[argIdx] then
		print( "ERROR: Invalid unique id specified." )
		return
	end

	proxy:ChangeObject( "/discovery" )
	local r, ret = proxy.SetUniqueID(args[argIdx])
	if not r then
		print("Control error:", ret)
	else
		print("Unique ID set to:", args[argIdx])
	end
end

local function StartPublish(proxy, args, argIdx )
	local options = {}
	local i
	for i=argIdx, #args do
		local k,v = args[i]:match("(.-)=(.*)$")
		if k then
			options[k] = v
		end
	end

	proxy:ChangeObject( "/discovery" )
	local r, ret = proxy.PublishGlobal(options)
	if not r then
		print("Control error:", ret)
	else
		print("Service published.")
	end
end

local function StopPublish(proxy, args, argIdx )
	proxy:ChangeObject( "/discovery" )
	local r, ret = proxy.StopPublishGlobal()
	if not r then
		print("Control error:", ret)
	else
		print("Service unpublished.")
	end
end

local function SetLinkToken(proxy, args, argIdx )
	if not args[argIdx] then
		print( "ERROR: Invalid link token" )
		return
	end

	proxy:ChangeObject( "/discovery" )
	local r, ret = proxy.SetLinkToken(args[argIdx])
	if not r then
		print("Control error:", ret)
	else
		print("Set link token to:", args[argIdx])
	end
end

local function AddManualTarget(proxy, args, argIdx )
	local loc = args[argIdx]
	local ip = args[argIdx+1]
	local port = args[argIdx+2]
	local token = args[argIdx+3]
	if not loc or not ip or not port or not token then
		print( "ERROR: Invalid arguments, you can use 'help add_target' to get details." )
		return
	end

	proxy:ChangeObject( "/discovery" )
	local r, ret = proxy.AddManualTarget(loc, ip, port, token)
	if not r then
		print("Control error:", ret)
	else
		print("Manual target '" .. loc .. "' added." )
	end
end

local function RemoveManualTarget(proxy, args, argIdx )
	local loc = args[argIdx]
	if not loc then
		print( "ERROR: Invalid location" )
		return
	end

	proxy:ChangeObject( "/discovery" )
	local r, ret = proxy.RemoveManualTarget(loc)
	if not r then
		print("Control error:", ret)
	else
		print("Manual target '" .. loc .. "' removed." )
	end
end

local cmd_actions = {
	["locations"] = GetLocations,
	["L"] = GetLocations,
	["services"] = GetServices,
	["S"] = GetServices,
	["detail"] = GetDetail,
	["D"] = GetDetail,
	["start"] = StartService,
	["stop"] = StopService,
	["signal"] = SignalService,
	["signal_all"] = SignalAll,
	["kill"] = SignalService,
	["killall"] = SignalAll,
	["enable"] = EnableService,
	["disable"] = DisableService,
	["blacklist"] = GetBlacklist,
	["debug"] = DebugControl,
	["add_f"] = AddDebugFilter,
	["clear_f"] = ClearDebugFilter,

	["id"] = SetUniqueID,

	["start_disco"] = StartDiscovery,
	["stop_disco"] = StopDiscovery,

	["start_pub"] = StartPublish,
	["stop_pub"] = StopPublish,

	["link_token"] = SetLinkToken,
	["add_target"] = AddManualTarget,
	["remove_target"] = RemoveManualTarget,

	["run"] = Run,
	["help"] = Help,
	["--help"] = Help,
	["-h"] = Help
}

---- Main.
local function Main(args)
	if not args or not args[1] then
		Usage( args and args[0] or "coctrl")
		coaf.Exit(0)
		return
	end

	local action = cmd_actions[args[1]]
	if not action then
		Usage( args[0] or "coctrl" )
		coaf.Exit(0)
		return
	end

	if action == Help then
		action(nil, args, 2)
		coaf.Exit(0)
		return
	end

	local cli = client.New( "coserver", "local", "unix:/var/run/coaf.coserver.ctrl" )
	if not cli then
		print("ERROR: Fail to connect CoServer!" )
	else
		local proxy = cli:CreateProxy( "/" )
		if not proxy then
			print( "ERROR: Fail to get 'coserver' proxy!" )
		else
			action(proxy, args, 2)
			proxy:Destroy()
		end
		cli:Destroy()
	end
	coaf.Exit(0)
end

coaf.Start(Main, arg)
