#!/usr/bin/env lua
local coaf = require "coaf.core"
local thread = require "coaf.thread"
local socket = require "cosocket.core"
local co_server = require "coserver.core"
local service = require "coaf.service"
local _dbg = require "coaf.debug"

local _sup_disco, discovery = pcall(require,"coserver.discovery")
if not _sup_disco then
	discovery = nil
end

coaf.ChDir(coaf.GetAppDir(arg))

local env = require "environment"

local function Usage(app)
	print("Usage:")
	print("\t" .. app .. " [-c <configure-location>] [-e <env-file>]" )
	print("")
	print("The <configure-location> is default located at '/etc/coserver'" )
end

local env_file = nil

if arg then
	local i
	for i=1, #arg do
		if arg[i] == "-c" then
			i = i + 1
			if i > #arg then
				Usage(arg[0])
				os.exit(0)
			end
			co_server.config_location = arg[i]
			break
		elseif arg[i] == "-e" then
			i = i + 1
			if i > #arg then
				Usage(arg[0])
				os.exit(0)
			end
			env_file = arg[i]
		elseif arg[i] == "-h" or arg[i] == "--help" then
			Usage(arg[0])
			os.exit(0)
		end
	end
end

_G.CO_ENV = env.Load(env_file)

local gSVC
local gDiscoCastSVC

--To make socket default be 'close-on-exec'
socket.default_close_on_exec = true

coaf.Init("CoServer")
service.InitSocket()

co_server.on_TriggerProxyLinkEvent = function(name, action)
	if gSVC then
		gSVC:Notify("/", "ProxyLinkStatus", name, action )
	end
end

local function Main(args)

	coaf.CreatePIDFile( "/var/run/coserver.pid" )

	--Note: While init co_server, it will auto initialize 'discovery' if required.
	co_server.Init()

	local addr
	if type(co_server.configs.CONTROL_BIND) == "string" then
		addr = co_server.configs.CONTROL_BIND
		gSVC = service.New( "coserver", addr )
		if not gSVC then
			_dbg.Error("CoServer", "Fail to create control service while binding:", addr )
		else
			_dbg.Info( "CoServer", "Control service 'coserver' is bound at:", addr )
			
			--Create discovery service if enabled.
			if co_server.configs.ENABLE_DISCOVERY then
				gDiscoCastSVC = service.New( "discovery", addr )
				_dbg.Info( "CoServer", "Control service 'discovery' is bound at:", addr )
			end

			if addr:match( "unix:") then
				local path = addr:match("unix:(.+)$")
				os.execute( "/bin/chmod 777 " .. path )
			end
		end
	elseif type(co_server.configs.CONTROL_BIND) == "table" then
		local i, svc, disc_svc
		for i, addr in ipairs(co_server.configs.CONTROL_BIND) do
			svc = service.New( "coserver", addr )
			if not svc then
				_dbg.Error( "CoServer", "Fail to create control service while binding:", addr )
			else
				_dbg.Info( "CoServer", "Control service 'coserver' is bound at:", addr )

				--Create discovery service if enabled.
				if co_server.configs.ENABLE_DISCOVERY then
					disc_svc = service.New( "discovery", addr )
					_dbg.Info( "CoServer", "Control service 'discovery' is bound at:", addr )
				end

				if addr:match( "unix:") then
					local path = addr:match("unix:(.+)$")
					os.execute( "/bin/chmod 777 " .. path )
				end
				if not gSVC then
					gSVC = svc
				end

				if co_server.configs.ENABLE_DISCOVERY and not gDiscoCastSVC then
					gDiscoCastSVC = disc_svc
				end
			end
		end
	end

	if gSVC then
		gSVC:ExposeObject( "/", co_server )
		gSVC:ExposeObject( "/CO_ENV", _G.CO_ENV )

		--Expose discovery service if enabled.
		if co_server.configs.ENABLE_DISCOVERY and discovery then
			gSVC:ExposeObject( "/discovery", discovery )
		end
	end

	if gDiscoCastSVC and discovery then
		gDiscoCastSVC:ExposeObject( "/", discovery.CASTER )
	end

	--NOTE: Invoke to start.
	co_server.GoStart()
end

coaf.Start(Main)

if co_server.configs.ENABLE_DISCOVERY and discovery then
	--TO avoid some process remained.
	discovery.Cleanup()
end

co_server.StopAllServices()

_dbg.Print("CoServer", "*** Server is terminated.")


