local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local compa = require "coaf.compatible"
local ev = require "ev"
local unistd = compa.CheckPosix( "posix.unistd" )
local fcntl = compa.CheckPosix( "posix.fcntl" )
local errno = compa.CheckPosix( "posix.errno" )

--------------------------------------------
local PIPE_EVENT

local pp_event_meta = {}

function pp_event_meta:CloseSlaveRead(flag_cloexec,val)
	if self.slave_read_fd then
		if not flag_cloexec then
			unistd.close(self.slave_read_fd)
			self.slave_read_fd = nil
		else
			fcntl.fcntl( self.slave_read_fd, fcntl.F_SETFD, val and 1 or 0)
		end
	end
end

function pp_event_meta:CloseSlaveWrite(flag_cloexec,val)
	if self.slave_write_fd then
		if not flag_cloexec then
			unistd.close(self.slave_write_fd)
			self.slave_write_fd = nil
		else
			fcntl.fcntl( self.slave_write_fd, fcntl.F_SETFD, val and 1 or 0)
		end
	end
end

function pp_event_meta:CloseMasterRead(flag_cloexec,val)
	if self.master_read_fd then
		if not flag_cloexec then
			unistd.close(self.master_read_fd)
			self.master_read_fd = nil
		else
			fcntl.fcntl( self.master_read_fd, fcntl.F_SETFD, val and 1 or 0 )
		end
	end
end

function pp_event_meta:CloseMasterWrite(flag_cloexec,val)
	if self.master_write_fd then
		if not flag_cloexec then
			unistd.close(self.master_write_fd)
			self.master_write_fd = nil
		else
			fcntl.fcntl( self.master_write_fd, fcntl.F_SETFD, val and 1 or 0 )
		end
	end
end

function pp_event_meta:StartSlaveThread()
	--TODO: Nothing, just fake.
end

function pp_event_meta:NotifyMasterMessage(topic,msg,typ)
	if not self.master_write_fd then
		return nil
	end
	local cont = "\01" .. topic .. "\02" .. (msg or "nil") .. "\03" .. (typ or "text") .. "\04"
	local r = unistd.write( self.master_write_fd, cont )
	if not r then
		return -1
	else
		return r
	end
end

function pp_event_meta:NotifySlaveMessage(topic,msg,typ)
	if not self.slave_write_fd then
		return nil
	end
	local cont = "\01" .. topic .. "\02" .. (msg or "nil") .. "\03" .. (typ or "text") .. "\04"
	local r = unistd.write( self.slave_write_fd, cont )
	if not r then
		return -1
	else
		return r
	end
end

function pp_event_meta:ReadSlaveMessage()
	local i,blk
	local pre,nxt
	local cont
	local errcode

	repeat
		local bufsize = #self.slave_buffer
		for i=1, bufsize do
			blk = self.slave_buffer[1]
			pre,nxt = blk:match("(.-)\04(.*)$")
			if pre then
				cont = (self.slave_fetched or '') .. pre
				self.slave_fetched = nil
				if #nxt > 0 then
					self.slave_buffer[1] = nxt
				else
					table.remove(self.slave_buffer,1)
				end
				return cont:match("\01(.*)\02(.*)\03(.*)$")
			else
				if not self.slave_fetched then
					self.slave_fetched = blk
				else
					self.slave_fetched = self.slave_fetched .. blk
				end
				table.remove( self.slave_buffer, 1 )
			end
		end

		while true do
			local buf,msg,code = unistd.read( self.slave_read_fd, 8192 )
			if not buf then
				errcode = code
				break
			elseif #buf == 0 then
				errcode = -1
			end
			table.insert( self.slave_buffer, buf )
		end
	until #self.slave_buffer == 0;
	if errcode == errno.EAGAIN or errcode == errno.EWOULDBLOCK then
		return nil, "retry", nil
	else
		return nil, "error", errcode
	end
end

function pp_event_meta:GetFd()
	return self.slave_read_fd
end

function pp_event_meta:Destroy()
	if self.slave_read_fd then
		unistd.close(self.slave_read_fd)
		self.slave_read_fd = nil
	end
	if self.slave_write_fd then
		unistd.close(self.slave_write_fd)
		self.slave_write_fd = nil
	end
	if self.master_read_fd then
		unistd.close(self.master_read_fd)
		self.master_read_fd = nil
	end
	if self.master_write_fd then
		unistd.close(self.master_write_fd)
		self.master_write_fd = nil
	end
	PIPE_EVENT.all_event_objects[self.name] = nil
end

pp_event_meta.__index = pp_event_meta

--------------------------------------------

PIPE_EVENT = {
	all_event_objects = {}
}

local function noblocking_fd(fd)
	local flag = fcntl.fcntl( fd, fcntl.F_GETFL )
	if compa.lua_version >= 5.3 then
		flag = flag | fcntl.O_NONBLOCK
	elseif compa.bit then
		flag = compa.bit.bor( flag, fcntl.O_NONBLOCK )
	end
	fcntl.fcntl( fd, fcntl.F_SETFL, flag )
end

function PIPE_EVENT.createNew(name, mode)
	if PIPE_EVENT.all_event_objects[name] then
		return nil, "exist"
	end

	if not unistd then
		return nil, "not-support"
	end
	
	local master_read_fd, master_write_fd
	local slave_read_fd, slave_write_fd = unistd.pipe()
	if not slave_read_fd or not slave_write_fd then
		return nil, "pipe-error"
	end

	--Non-blocking  pipe.
	noblocking_fd( slave_read_fd )
	noblocking_fd( slave_write_fd )

	if mode == "bidi" or mode == "bidi-pipe" then
		master_read_fd, master_write_fd = unistd.pipe()
		if not master_read_fd or not master_write_fd then
			unistd.close(slave_read_fd)
			unistd.close(slave_write_fd)
			return nil, "pipe-error"
		end
	
		--Non-blocking  pipe.
		noblocking_fd( master_read_fd )
		noblocking_fd( master_write_fd )
	end

	local obj = {
		name = name,
		slave_read_fd = slave_read_fd,
		slave_write_fd = slave_write_fd,
		master_read_fd = master_read_fd,
		master_write_fd = master_write_fd,

		slave_fetched = nil,
		master_fetched = nil,

		slave_buffer = {},
		master_buffer = {},
	}
	
	PIPE_EVENT.all_event_objects[name] = obj
	return setmetatable(obj, pp_event_meta)
end

function PIPE_EVENT.free(name)
	local obj = PIPE_EVENT.all_event_objects[name]
	if obj then
		obj:Destroy()
	end
end

--------------------------------------------

local _M = {
	eventObjects = {}
}

--------------------------------------------

local event_meta = {}

function event_meta:CloseSlaveRead(flag,val)
	if self.eventType == "pipe" and self.eventObject then
		if not flag and self.watcher then
			self.watcher:destroy()
			self.watcher = nil
		end
		self.eventObject:CloseSlaveRead(flag,val)
	end
end

function event_meta:CloseSlaveWrite(flag,val)
	if self.eventType == "pipe" and self.eventObject then
		self.eventObject:CloseSlaveWrite(flag,val)
	end
end

function event_meta:CloseMasterRead(flag,val)
	if self.eventType == "pipe" and self.eventObject then
		self.eventObject:CloseMasterRead(flag,val)
	end
end

function event_meta:CloseMasterWrite(flag,val)
	if self.eventType == "pipe" and self.eventObject then
		self.eventObject:CloseMasterWrite(flag,val)
	end
end

function event_meta:StartSlaveThread()
	self.eventObject:StartSlaveThread()
end

function event_meta:Subscribe( topic, callback, obj )
	if not topic or not callback then
		_dbg.Error( "CoAF.Event", "Subscribe but without valid topic or callback function, for event object '" .. self.name .. "'" )
		return nil
	end
	local ticket = topic .. tostring(callback) .. tostring(obj)
	self.subscribers[ ticket ] = {
		topic = topic,
		callback = callback,
		obj = obj
	}
	return ticket
end

function event_meta:Unsubscribe( ticket )
	if ticket then
		self.subscribers[ ticket ] = nil
	end
end

function event_meta:Notify( topic, message )
	local r
	if type(message) == "string" then
		r = self.eventObject:NotifyMasterMessage( topic, message, "text" )
	elseif type(message) == "table" then
		local pretty = require "pl.pretty"
		local msg = pretty.write( message )
		r = self.eventObject:NotifyMasterMessage( topic, msg, "table" )
	else
		_dbg.Error( "CoAF.Event", "Invalid notify message for event object '" .. self.name .. "'" )
	end
	if r and r <= 0 then
		_dbg.Error( "CoAF.Event", "Fail to notify message for event object '" .. self.name .. "', topic=", topic )
	end
	return r
end

function event_meta:NotifyBack( topic, message )
	local r
	if type(message) == "string" then
		r = self.eventObject:NotifySlaveMessage( topic, message, "text" )
	elseif type(message) == "table" then
		local pretty = require "pl.pretty"
		local msg = pretty.write( message )
		r = self.eventObject:NotifySlaveMessage( topic, msg, "table" )
	else
		_dbg.Error( "CoAF.Event", "Invalid notify(self) message for event object '" .. self.name .. "'" )
	end
	if r and r <= 0 then
		_dbg.Error( "CoAF.Event", "Fail to notify(self) message for event object '" .. self.name .. "', topic=", topic )
	end
	return r
end

function event_meta:Destroy()
	if self.watcher then
		self.watcher:destroy()
		self.watcher = nil
	end

	if self.eventType == "pipe" then
		PIPE_EVENT.free( self.name )
	else
		if not CoAF_EVENT then
			require "coevent"
		end
		CoAF_EVENT.free( self.name )
	end
	self.eventObject = nil
	_M.eventObjects[ self.name ] = nil
end

function event_meta:_onMessage( topic, message, messageType )
	if message and messageType and messageType:sub(1,5) == "table" then
		local func,err = load( "return " .. message )
		if not func then
			_dbg.Error( "CoAF.Event", "Invalid table message while parsing it, error:", err )
			messageType = "bad-table"
		else
			local flag, ret = pcall(func)
			if flag then
				message = ret
			else
				_dbg.Error( "CoAF.Event", "Bad message table while parsing it, error:", ret )
				messageType = "bad-table"
			end
		end
	end

	local k, subObj 
	for k, subObj in pairs(self.subscribers) do
		if subObj.topic == topic and subObj.callback then
			if subObj.obj then
				subObj.callback( subObj.obj, topic, message, messageType )
			else
				subObj.callback( topic, message, messageType )
			end
		end
	end
end

event_meta.__index = event_meta

---------------------------------------------

function _M.createNew( name, noStartThread )
	if type(name) ~= "string" then
		_dbg.Error( "CoAF.Event", "Invalid event object name" )
		return nil
	end

	local obj = _M.get(name)
	if obj then
		return obj
	end

	local eventCObj 
	local eventType
	if type(noStartThread) == "string" then
		eventCObj = PIPE_EVENT.createNew(name, noStartThread) --AKA pipe type: 'pipe', 'bidi'/'bidi-pipe'
		eventType = "pipe"
	else
		if not CoAF_EVENT then
			require "coevent"
		end
		eventCObj = CoAF_EVENT.createNew(name, noStartThread)
		eventType = "coevent"
	end

	if not eventCObj then
		_dbg.Error( "CoAF.Event", "Fail to create event object with name '" .. name .. "'" )
		return nil
	end

	local obj = {
		eventObject = eventCObj,
		name = name,
		eventType = eventType,
		subscribers = {}
	}

	local watcher = coaf.NewIO(function(w, revents)
		while true do
			local topic, msg, msgType = obj.eventObject:ReadSlaveMessage()
			if not topic then
				if msg ~= "retry" then
					_dbg.Error( "CoAF.Event", "Event object read message error, closed. name=", obj.name )
					w:destroy()
					obj.watcher = nil
				end
				break
			else
				obj:_onMessage( topic, msg, msgType )
			end
		end
	end, eventCObj:GetFd(), ev.READ )

	if watcher then
		obj.watcher = watcher
		watcher:start()
	end

	local r = setmetatable( obj, event_meta )
	_M.eventObjects[ name ] = r
	return r
end

function _M.free( name )
	if type(name) ~= "string" then
		_dbg.Warn( "CoAF.Event", "Invalid event object name while free" )
		return false
	end
	local obj = _M.eventObjects[name]
	if obj then
		obj:Destroy()
		return true
	else
		return false
	end
end

function _M.get( name )
	if type(name) ~= "string" then
		_dbg.Warn( "CoAF.Event", "Invalid event object name while getting it" )
		return nil
	end
	return _M.eventObjects[name]
end

function _M.cleanup()
	local existObjects = _M.eventObjects
	_M.eventObjects = {}
	local k,v

	for k,v in pairs(existObjects) do
		v:Destroy()
	end
end

-----------------------

return _M

