--[==[

The Co-Application Framework (CoAF) RPC client module.
Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local _dbg = require "coaf.debug"
local protocol = require "coaf.rpc.protocol"
local DEF_COSERVER_ADDR = "unix:/var/run/coaf.coserver"

math.randomseed(os.time())

local _M = {
	socket = nil,
	adapter = nil,
	coserver_addr = DEF_COSERVER_ADDR,
	clientName = "CLI-" .. math.random(1, 1000000000),
}

-------------------------------------------------------
-- You must call 'InitSocket(socket_obj, adapter_obj)' first,
-- before using CoAF client.
-------------------------------------------------------
function _M.InitSocket(socket, adapter)
	if not socket then
		if ngx and ngx.socket then -- Check in openresty environment?
			_dbg.Info("CoAF.Client", "Init client socket: found ngnix-openresty socket")
			_M.socket = ngx.socket

			if not adapter then
				_M.adapter = require "coaf.cli_adapter.simple"
			else
				_M.adapter = adapter
			end
		else
			local status, socket = pcall(require, "cosocket.core")
			if status and socket then
				_dbg.Info("CoAF.Client", "Init client socket: found CoAF cosocket")
				_M.socket = socket
				if not adapter then
					_M.adapter = require "coaf.cli_adapter.simple"
				else
					_M.adapter = adapter
				end
			else
				_dbg.Error("CoAF.Client", "Init client socket: no socket libary found!")
			end
		end
	else
		_M.socket = socket
		if not adapter then
			_M.adapter = require "coaf.cli_adapter.simple"
		else
			_M.adapter = adapter
		end
	end
end

function _M.SetCoServerAddress(addr)
	_M.coserver_addr = addr or DEF_COSERVER_ADDR
end

-----------------------------------------------------------------
-- Object Proxy
-----------------------------------------------------------------
local proxy_meta = {}
local proxy_property_meta = {}
local proxy_notify_meta = {}
local proxy_stream_meta = require "coaf.client_stream_meta"

function proxy_meta:__index(func)
	local thisObj = self
	local flag_unprotect = self.__unprotect

	local caller = function(_this, ...)
		local args
		local callMethod
		if _this == thisObj then
			callMethod = "CALL"
			args = {...}
		else
			callMethod = "STATIC_CALL"
			args = {_this, ...}
		end

		local content = {
			CMD = callMethod,
			ARGS = func,
			--SN will be auto filled by Send(...)
			headers = {
				SVC = thisObj.__internal_client.serviceId,
				LOC = thisObj.__internal_client.location,
				OBJ = thisObj.__internal_objectPath,
				--VIA will be auto filled by Send(...)
			},
			parameters = args
		}

		if not thisObj.__internal_connection:Send(content) then
			_dbg.Error("CoAF.Client", "Proxy - Fail send message for calling:", func)
			if flag_unprotect then
				error( "#RPC-ERROR#: Fail send message" )
			end
			return false, "#RPC-ERROR#", "Fail send message"
		end

		local ack = thisObj.__internal_connection:RecvACK(content.SN)
		if not ack then
			_dbg.Error("CoAF.Client", "Proxy - Fail recv ACK for calling:", func)
			if flag_unprotect then
				error( "#RPC-ERROR#: Fail recv ACK" )
			end
			return false, "#RPC-ERROR#", "Fail recv ACK"
		end

		if not ack.headers or not ack.headers.RESULT then
			_dbg.Error("CoAF.Client", "Proxy - Invalid ACK message for calling:", func)
			if flag_unprotect then
				error( "#RPC-ERROR#: Invalid ACK message" )
			end
			return false, "#RPC-ERROR#", "Invalid ACK message"
		end

		if ack.headers.RESULT ~= "OK" then
			_dbg.Info("CoAF.Client", "Proxy - Call error:", ack.headers.RESULT, ack.headers.MSG )
			if flag_unprotect then
				error( ack.headers.RESULT .. ": " .. (ack.headers.MSG or "Unknown error") )
			end
			return false, ack.headers.RESULT, ack.headers.MSG or "Unknown error"
		else
			if type(ack.data) ~= "table" then
				return true
			else
				table.insert(ack.data, 1, "return " )
			end
			local flag, func
			func, flag = load( table.concat( ack.data ) )
			if not func then
				_dbg.Error("CoAF.Client", "Proxy - Bad response arguments format for calling:", flag)
				if flag_unprotect then
					error( "#RPC-ERROR#: Bad response arguments format" )
				end
				return false, "#RPC-ERROR#", "Bad response arguments format"
			end
			local rets
			flag, rets = pcall(func)
			if not flag then
				_dbg.Error("CoAF.Client", "Proxy - Bad response arguments for calling:", func)
				if flag_unprotect then
					error( "#RPC-ERROR#: Bad response arguments" )
				end
				return false, "#RPC-ERROR#", "Bad response arguments"
			end
			if not rets then
				if not flag_unprotect then
					return true
				end
			else
				if flag_unprotect then
					return table.unpack(rets)
				else
					return true, table.unpack(rets)
				end
			end
		end
	end

	self[func] = caller
	return caller
end

---------------------------------------------------------------------

local function proxy_Destroy(_self)
	_self.__internal_connection:Close()
	--_self.__internal_connection = nil
end

local function proxy_Subscribe(_self, ...)
	local thisObj = _self

	local content = {
		CMD = "SUBS",
		--SN will be auto filled by Send(...)
		headers = {
			SVC = thisObj.__internal_client.serviceId,
			LOC = thisObj.__internal_client.location,
			OBJ = thisObj.__internal_objectPath,
			FROM = thisObj.__from_name
			--VIA will be auto filled by Send(...)
		},
		parameters = {...}
	}

	if not thisObj.__internal_connection:Send(content) then
		_dbg.Error("CoAF.Client", "Proxy - fail send SUBS message")
		return false, "#RPC-ERROR#", "fail send message"
	end

	local ack = thisObj.__internal_connection:RecvACK(content.SN)
	if not ack then
		_dbg.Error("CoAF.Client", "Proxy - fail get ACK/SUBS message")
		return false, "#RPC-ERROR#", "fail get ACK message"
	end

	if not ack.headers or not ack.headers.RESULT then
		_dbg.Error("CoAF.Client", "Proxy - invalid ACK/SUBS message for:", key)
		return false, "#RPC-ERROR#", "invalid ACK message"
	end

	if ack.headers.RESULT ~= "OK" then
		_dbg.Info("CoAF.Client", "Proxy - Subscribe error:", ack.headers.RESULT, ack.headers.MSG )
		return false, ack.headers.RESULT, ack.headers.MSG or "Unknown error"
	end
	return true
end

local function proxy_Unsubscribe(_self, ...)
	local thisObj = _self

	local content = {
		CMD = "UNSUBS",
		--SN will be auto filled by Send(...)
		headers = {
			SVC = thisObj.__internal_client.serviceId,
			LOC = thisObj.__internal_client.location,
			OBJ = thisObj.__internal_objectPath,
			FROM = thisObj.__from_name
			--VIA will be auto filled by Send(...)
		},
		parameters = {...}
	}

	if not thisObj.__internal_connection:Send(content) then
		_dbg.Error("CoAF.Client", "Proxy - fail send UNSUBS message")
		return false, "#RPC-ERROR#", "fail send message"
	end

	local ack = thisObj.__internal_connection:RecvACK(content.SN)
	if not ack then
		_dbg.Error("CoAF.Client", "Proxy - fail get ACK/UNSUBS message")
		return false, "#RPC-ERROR#", "fail get ACK message"
	end

	if not ack.headers or not ack.headers.RESULT then
		_dbg.Error("CoAF.Client", "Proxy - invalid ACK/UNSUBS message for:", key)
		return false, "#RPC-ERROR#", "invalid ACK message"
	end

	if ack.headers.RESULT ~= "OK" then
		_dbg.Info("CoAF.Client", "Proxy - Unsubscribe error:", ack.headers.RESULT, ack.headers.MSG )
		return false, ack.headers.RESULT, ack.headers.MSG or "Unknown error"
	end
	return true
end

local function proxy_RegisterNotifier(_self, obj, objCall)
	if type(obj) == "table" then
		table.insert(_self.__internal_notifiers,{
			notifier = obj,
			objCall = objCall
		})
		return true
	else
		return false
	end
end

local function proxy_UnregisterNotifier(_self, obj)
	local i, v
	for i,v in ipairs(_self.__internal_notifiers) do
		if v.notifier == obj then
			table.remove( self.__internal_notifiers, i )
			break
		end
	end
end

local function proxy_DryRun(_self,timeout)
	if _self.__internal_connection then
		return _self.__internal_connection:RecvACK("dryrun",timeout)
	else
		return nil, "closed"
	end
end

local function proxy_ForkProxy(_self, objPath)
	return _self.__internal_client:CreateProxy(objPath)
end

local function proxy_ChangeObject(_self, objPath)
	--ping target object, to check object exist or not.
	local content = {
		CMD = "PING",
		headers = {
			SVC = _self.__internal_client.serviceId,
			LOC = _self.__internal_client.location,
			OBJ = objPath
		}
	}

	if not _self.__internal_connection:Send(content) then
		_dbg.Error("CoAF.Client", "ChangeObject - Fail to send PING message")
		return false
	end

	local ack = _self.__internal_connection:RecvACK(content.SN)
	if not ack or not ack.headers or not ack.headers.RESULT then
		_dbg.Error("CoAF.Client", "ChangeObject - Get ACK/PING fail or invalid")
		return false
	end

	if ack.headers.RESULT ~= "OK" then
		_dbg.Error("CoAF.Client", "ChangeObject - Check object ", objPath, "Fail:", ack.headers.RESULT, ack.headers.MSG)
		return false
	end

	_self.__internal_objectPath = objPath
	return true
end

local function proxy_ProcessNotifier(_self, content)
	if content.headers.OBJ and content.headers.OBJ ~= _self.__internal_objectPath then
		return false
	end

	local target = content.ARGS
	if not target or target == "" then
		target = "OnNotify"
	end
	local args
	if type(content.data) ~= "table" then
		args = {}
	else
		table.insert(content.data, 1, "return ")
		local func, flag = load(table.concat(content.data))
		if func then
			flag, args = pcall(func)
			if flag then
				if not args then
					args = {}
				end
			else
				_dbg.Error("CoAF.Client", "Invalid NOTIFY argument")
				return false
			end
		else
			_dbg.Error("CoAF.Client", "Invalid NOTIFY arguments format:", flag)
			return false
		end
	end

	local i,v
	local func
	for i,v in ipairs(_self.__internal_notifiers) do
		func = v.notifier[target]
		if type(func) == "function" then
			local done, err
			if v.objCall then
				done, err = pcall(func, v.notifier, table.unpack(args))
			else
				done, err = pcall(func, table.unpack(args))
			end
			if not done then
				_dbg.Error("CoAF.Client", "Notify error:", target, err )
			end
		end
	end
	return true
end

---------------------------------------------------------------------
function proxy_property_meta:__index(key)
	local thisObj = self.__owner
	local content = {
		CMD = "GET",
		--ARGS = key,
		--SN will be auto filled by Send(...)
		headers = {
			SVC = thisObj.__internal_client.serviceId,
			LOC = thisObj.__internal_client.location,
			OBJ = thisObj.__internal_objectPath,
			--VIA will be auto filled by Send(...)
		},
		parameters = {
			key
		}
	}

	if not thisObj.__internal_connection:Send(content) then
		_dbg.Error("CoAF.Client", "Proxy - fail send GET message for:", key)
		return nil
	end

	local ack = thisObj.__internal_connection:RecvACK(content.SN)
	if not ack then
		_dbg.Error("CoAF.Client", "Proxy - fail get ACK/GET message for:", key)
		return nil
	end

	if not ack.headers or not ack.headers.RESULT then
		_dbg.Error("CoAF.Client", "Proxy - invalid ACK/GET message for:", key)
		return nil
	end

	if ack.headers.RESULT ~= "OK" then
		_dbg.Info("CoAF.Client", "Proxy - Get property error:", ack.headers.RESULT, ack.headers.MSG )
		return nil
	else
		if type(ack.data) ~= "table" then
			return nil
		else
			table.insert(ack.data, 1, "return " )
		end
		local flag, func
		func, flag = load( table.concat( ack.data ) )
		if not func then
			_dbg.Error("CoAF.Client", "Proxy - Bad response for GET property:", key )
			return nil
		end
		local rets
		flag, rets = pcall(func)
		if not flag then
			_dbg.Error("CoAF.Client", "Proxy - Bad result for GET property:", key )
			return nil
		end
		if not rets then
			return nil
		else
			return table.unpack(rets)
		end
	end
end

function proxy_property_meta:__newindex(key,val)
	local thisObj = self.__owner

	local content = {
		CMD = "SET",
		--SN will be auto filled by Send(...)
		headers = {
			SVC = thisObj.__internal_client.serviceId,
			LOC = thisObj.__internal_client.location,
			OBJ = thisObj.__internal_objectPath,
			--VIA will be auto filled by Send(...)
		},
		parameters = {
			key, val
		}
	}

	if not thisObj.__internal_connection:Send(content) then
		_dbg.Error("CoAF.Client", "Proxy - fail send SET message for:", key)
		return
	end

	local ack = thisObj.__internal_connection:RecvACK(content.SN)
	if not ack then
		_dbg.Error("CoAF.Client", "Proxy - fail get ACK/SET message for:", key)
		return
	end

	if not ack.headers or not ack.headers.RESULT then
		_dbg.Error("CoAF.Client", "Proxy - invalid ACK/SET message for:", key)
		return
	end

	if ack.headers.RESULT ~= "OK" then
		_dbg.Info("CoAF.Client", "Proxy - Set property error:", ack.headers.RESULT, ack.headers.MSG )
		return
	end
end

function proxy_property_meta:__call(...)
	local args = {...}
	local thisObj = self.__owner

	if type(args[1]) == "string" then -- Get a sequence of properties.
		local content = {
			CMD = "GET",
			--SN will be auto filled by Send(...)
			headers = {
				SVC = thisObj.__internal_client.serviceId,
				LOC = thisObj.__internal_client.location,
				OBJ = thisObj.__internal_objectPath,
				--VIA will be auto filled by Send(...)
			},
			parameters = args
		}

		if not thisObj.__internal_connection:Send(content) then
			_dbg.Error("CoAF.Client", "Proxy - fail send GET message (multiple)")
			return nil
		end

		local ack = thisObj.__internal_connection:RecvACK(content.SN)
		if not ack then
			_dbg.Error("CoAF.Client", "Proxy - fail get ACK/GET message (multiple)")
			return nil
		end

		if not ack.headers or not ack.headers.RESULT then
			_dbg.Error("CoAF.Client", "Proxy - invalid ACK/GET message (multiple)")
			return nil
		end

		if ack.headers.RESULT ~= "OK" then
			_dbg.Info("CoAF.Client", "Proxy - Get properties error:", ack.headers.RESULT, ack.headers.MSG )
			return nil
		else
			if type(ack.data) ~= "table" then
				return nil
			else
				table.insert(ack.data, 1, "return " )
			end
			local flag, func
			func, flag = load( table.concat( ack.data ) )
			if not func then
				_dbg.Error("CoAF.Client", "Proxy - Bad response for GET multiple properties" )
				return nil
			end
			local rets
			flag, rets = pcall(func)
			if not flag then
				_dbg.Error("CoAF.Client", "Proxy - Bad result for GET multiple properties" )
				return nil
			end
			if not rets then
				return nil
			else
				return table.unpack(rets)
			end
		end
	elseif type(args[1]) == "table" then -- Set properties, in table: {key=value, key2=value2, ...}
		local content = {
			CMD = "SET",
			--SN will be auto filled by Send(...)
			headers = {
				SVC = thisObj.__internal_client.serviceId,
				LOC = thisObj.__internal_client.location,
				OBJ = thisObj.__internal_objectPath,
				--VIA will be auto filled by Send(...)
			},
			parameters = args
		}

		if not thisObj.__internal_connection:Send(content) then
			_dbg.Error("CoAF.Client", "Proxy - fail send SET message (multiple)")
			return
		end

		local ack = thisObj.__internal_connection:RecvACK(content.SN)
		if not ack then
			_dbg.Error("CoAF.Client", "Proxy - fail get ACK/SET message (multiple)")
			return
		end

		if not ack.headers or not ack.headers.RESULT then
			_dbg.Error("CoAF.Client", "Proxy - invalid ACK/SET message (multiple)")
			return
		end

		if ack.headers.RESULT ~= "OK" then
			_dbg.Info("CoAF.Client", "Proxy - Set multiple properties error:", ack.headers.RESULT, ack.headers.MSG )
			return
		end
	else
		_dbg.Error("CoAF.Client", "Proxy - Invalid property calling parameters")
		return
	end
end

--------------------------------------------------------------

function proxy_notify_meta:__index(func)
	local thisObj = self

	local caller = function(_this, ...)
		local args
		if _this == thisObj then
			args = {...}
		else
			args = {_this, ...}
		end

		local content = {
			CMD = "NOTIFY",
			ARGS = func,
			--SN will be auto filled by Send(...)
			headers = {
				SVC = thisObj.__owner.__internal_client.serviceId,
				LOC = thisObj.__owner.__internal_client.location,
				OBJ = thisObj.__owner.__internal_objectPath,
				--VIA will be auto filled by Send(...)
			},
			parameters = args
		}

		if not thisObj.__owner.__internal_connection:Send(content) then
			return false, "#RPC-ERROR#", "Fail send message"
		else
			return true
		end
	end

	self[func] = caller
	return caller
end

function proxy_notify_meta:__call(func, ...)
	local content = {
		CMD = "NOTIFY",
		ARGS = func,
		--SN will be auto filled by Send(...)
		headers = {
			SVC = thisObj.__internal_client.serviceId,
			LOC = thisObj.__internal_client.location,
			OBJ = thisObj.__internal_objectPath,
			--VIA will be auto filled by Send(...)
		},
		parameters = {...}
	}

	if not self.__owner.__internal_connection:Send(content) then
		return false, "#RPC-ERROR#", "Fail send message"
	else
		return true
	end
end

-----------------------------------------------------------------
-- Client object.
-----------------------------------------------------------------
local client_meta = {}

function client_meta:CreateProxy(objPath, exclusive)
	return self:CreateProxy2(objPath,exclusive,"1")
end

function client_meta:CreateProxy2(objPath, exclusive, compatible_version)
	if not self.connection then
		_dbg.Error("CoAF.Client", "Invalid connection while create proxy:", objPath)
		return nil
	end

	local newConn 
	
	if not exclusive then
		newConn = self.connection:ForkConnection()
		if not newConn then
			_dbg.Error("CoAF.Client", "Fail to fork connection while create proxy:", objPath)
			return nil
		end
	else
		newConn = self.connection
	end

	--ping target object, to check object exist or not.
	local content = {
		CMD = "PING",
		headers = {
			SVC = self.serviceId,
			LOC = self.location,
			OBJ = objPath
		}
	}

	if not newConn:Send(content) then
		_dbg.Error("CoAF.Client", "CreateProxy - Fail to send PING message")
		return nil
	end

	local ack = newConn:RecvACK(content.SN)
	if not ack or not ack.headers or not ack.headers.RESULT then
		_dbg.Error("CoAF.Client", "CreateProxy - Get ACK/PING fail or invalid")
		return nil
	end

	if not ngx or not ngx.socket then -- Check in openresty environment?
	--TODO: if not in openresty, then try authorization.
		if self.authKey and ack.headers.RESULT == "AUTH_REQ" and ack.headers.CHA then
			require "cocrypto"

			--_dbg.Info("CoAF.Client", "[AUTH] CreateProxy - Remote require authorization, challenge is:", ack.headers.CHA)

			local tip = CoAF_CRYPTO.get_challenge()
			local dig = CoAF_CRYPTO.finger_print( ack.headers.CHA .. tip .. self.authKey )

			content = {
				CMD = "PING",
				headers = {
					SVC = self.serviceId,
					LOC = self.location,
					UNIQ = self.authUniq,
					OBJ = objPath,
					CHA = ack.headers.CHA,
					TIP = tip,
					DIG = dig
				}
			}
			if not newConn:Send(content) then
				_dbg.Error("CoAF.Client", "[AUTH] CreateProxy - Fail to send PING message")
				return false
			end
		
			ack = newConn:RecvACK(content.SN)
			if not ack or not ack.headers or not ack.headers.RESULT then
				_dbg.Error("CoAF.Client", "[AUTH] CreateProxy - Get ACK/PING fail or invalid")
				return nil
			end
		end
	end
	
	if ack.headers.RESULT ~= "OK" then
		_dbg.Error("CoAF.Client", "CreateProxy - Check object ", objPath, "Fail:", ack.headers.RESULT, ack.headers.MSG)
		return nil
	end

	local proxy = {
		__unprotect = false,
		__internal_client = self,
		__internal_connection = newConn,
		__internal_objectPath = objPath,
		__internal_notifiers = {},

		__ProcessNotifier = proxy_ProcessNotifier,

		proxy_Destroy = proxy_Destroy,
		proxy_Subscribe = proxy_Subscribe,
		proxy_Unsubscribe = proxy_Unsubscribe,
		proxy_RegisterNotifier = proxy_RegisterNotifier,
		proxy_UnregisterNotifier = proxy_UnregisterNotifier,
		proxy_ForkProxy = proxy_ForkProxy,
		proxy_ChangeObject = proxy_ChangeObject,
		proxy_DryRun = proxy_DryRun,

		_P = {
		},

		_NOTIFY = {
		},

		_STREAM = {
			data = {
			}
		}
	}

	--Compatible with old version (1)
	if compatible_version == "1" then
		proxy.Destroy = proxy_Destroy
		proxy.Subscribe = proxy_Subscribe
		proxy.Unsubscribe = proxy_Unsubscribe
		proxy.RegisterNotifier = proxy_RegisterNotifier
		proxy.UnregisterNotifier = proxy_UnregisterNotifier
		proxy.ForkProxy = proxy_ForkProxy
		proxy.ChangeObject = proxy_ChangeObject
		proxy.DryRun = proxy_DryRun
	end

	--__from_name is used for "SUBS/UNSUBS"
	proxy.__from_name = tostring(proxy):match(".*:%s*(.*)$") .. "/" .. _M.clientName
	proxy._P.__owner = proxy
	proxy._NOTIFY.__owner = proxy
	proxy._STREAM.__owner = proxy

	setmetatable( proxy._P, proxy_property_meta)
	setmetatable( proxy._NOTIFY, proxy_notify_meta)
	setmetatable( proxy._STREAM, proxy_stream_meta)
	setmetatable( proxy, proxy_meta)

	newConn:BindProxy(proxy)

	--table.insert(self.allProxies, proxy)
	return proxy
end

function client_meta:Destroy()
	--To destroy all proxies.
	--[[
	local k,v
	for k,v in pairs(self.allProxies) do
		v:Destroy()
	end
	self.allProxies = {}
	]]

	if self.connection then
		self.connection:Close()
		self.connection = nil
	end
	collectgarbage()
end

client_meta.__index = client_meta

-----------------------------------------------------------------
-- Create new client
-----------------------------------------------------------------
function _M.New(serviceId, loc, addr, tmout, authKey, authUniq)
	if not _M.socket or not _M.adapter then
		_dbg.Error( "CoAF.Client", "Invalid socket/adapter. Please initialize first.")
		return nil
	end

	if type(addr) == "number" and not tmout then
		tmout = addr
		addr = nil
	else
		tmout = tonumber(tmout) or 30
	end

	local conn, err
	local startTime = os.time()
	local now
	while true do
		conn, err = _M.adapter.New(_M.socket, addr or _M.coserver_addr, tmout)
		if not conn then
			if err:match("refused") or err:match("[Nn]o such") --[[or err:match("in progress")]] then
				now = os.time()
				now = now - startTime
				if now < 0 or now >= tmout then
					_dbg.Error("CoAF.Client", "Fail to connect service socket address:", addr or _M.coserver_addr)
					return nil
				end

				if _M.socket.sleep then
					_M.socket.sleep(1) --Try sleep 1 second.
				elseif ngx and ngx.sleep then
					ngx.sleep(1)
				else
					_dbg.Warn("CoAF.Client", "Fail to wait some time to try connect service again!")
				end
			else
				_dbg.Error("CoAF.Client", "Fail to connect service socket address:", addr or _M.coserver_addr)
				return nil
			end
		else
			break
		end
	end

	if not conn:ConnectService( serviceId, loc, authKey, authUniq ) then
		conn:Close()
		_dbg.Error("CoAF.Client", "Fail to connect service of:", (loc or "local") .. "/" .. (serviceId or "nil"))
		return nil
	end

	return setmetatable({
		connection = conn,
		serviceId = serviceId,
		location = loc or "local",
		authKey = authKey,
		authUniq = authUniq,
		allProxies = {}
	},client_meta)
end

return _M
