local coaf = require "coaf.core"
local ev = require "ev"
local _dbg = require "coaf.debug"
local compa = require "coaf.compatible"

local _M = {
	defaultFilterName = "STDIO",
	defaultFunc = _dbg.Print,
	matches = {},
	ready = false
}

function _M.SetDefaultFilterName(filName)
	_M.defaultFilterName = filName or "STDIO"
end

function _M.SetDefaultLogFunction(func)
	_M.defaultFunc = func
end

-- ** NOTE: **
-- 'matchPattern' is a standard Lua string.match(...) pattern,
-- you can also use '(...)' to match and save. This is used 
-- to match lower std-out content, and if matched, do output
-- to debug logger according to matchTable_or_func.
--
-- matchTable_or_func: has 2 cases:
-- 1) if matchTable_or_func is a function, that, if matched
--    by 'matchPattern', call this function (generally, it
--    will be _dbg.Info/_dbg.Error/_dbg.Warn/...). The output
--    filter name is specified by 'filterName'. If not specified 
--    'filterName', it uses the module default filter name.
--
-- 2) More complex, if 'matchPattern' used '(...)' matching,
--    matching will returns 1 or more matched results. The
--    concepts is all returned results stored into a array:
--    { match1, match2, match3, ... }.
--
--    so, you can specify a matching table (matchTable_or_func),
--    to match the returned array items, and do specific action.
--
--    The each matching table item format is:
--    {
--        match = { value1, value2, value3, ... },
--        filterName = "filter name", -- optional, can be nil
--        action = <output function>, -- gernally, be _dbg.Info, Error, ...
--    }
--
--    multiple matching item can be specified. Like:
--    {
--    	{ match={"ERR"}, filterName="LOWER", action=_dbg.Error },
--    	{ match={"INFO","1"}, filterName="LOWER", action=_dbg.Info1 },
--    	{ match={"INFO","3"}, filterName="LOWER", action=_dbg.Info3 },
--		{ match={ nil, nil }, filterName="LOWER", action=_dbg.Info } --Match anything else.
--    }
function _M.SetFilter( matchPattern, matchTable_or_func, filterName )
	if not matchPattern then
		return false, "Invalid matching pattern"
	end
	
	if type(matchTable_or_func) == "function" then
		--TODO: Check duplication.
		table.insert( _M.matches, {
			pattern = matchPattern,
			filter = {
				{ match={}, filterName = filterName, action = matchTable_or_func }
			}
		})
	elseif type(matchTable_or_func) == "table" then
		if type(matchTable_or_func[1]) == "table" then	--Is array
			table.insert( _M.matches, {
				pattern = matchPattern,
				filter = matchTable_or_func
			})
		else --Just one option
			table.insert( _M.matches, {
				pattern = matchPattern,
				filter = { matchTable_or_func }
			})
		end
	else
		return false, "Invalid matching table or function"
	end
	return true
end

function _M.OutputLine(line)
	if #_M.matches == 0 then
		_M.defaultFunc( _M.defaultFilterName, line )
	else
		local i, matchItem
		for i, matchItem in ipairs(_M.matches) do
			local mr = { line:match( matchItem.pattern ) }
			if #mr > 0 then
				local j, filterItem
				for j, filterItem in ipairs(matchItem.filter) do
					local x
					local match = true
					for x=1, #mr do
						if filterItem.match[x] and not mr[x]:match(filterItem.match[x]) then
							match = false
							break
						end
					end
					if match then
						filterItem.action( filterItem.filterName or _M.defaultFilterName, line )
						return
					end
				end
			end
		end
		_M.defaultFunc( _M.defaultFilterName, line )
	end
end

function _M.Start()
	if _M.ready then return true end
	local unistd = require "posix.unistd"
	local fcntl = require "posix.fcntl"
	local errno = require "posix.errno"

	local read_fd, write_fd = unistd.pipe()
	if not read_fd or not write_fd then
		return false, "Fail to create IO pipe for stdio monitor"
	end

	local ver = _VERSION:match("Lua (.+)$")
	ver = tonumber(ver) or 5.1

	--Non-blocking read pipe.
	local flag = fcntl.fcntl( read_fd, fcntl.F_GETFL )
	if compa.lua_version >= 5.3 then
		flag = flag | fcntl.O_NONBLOCK
	elseif compa.bit then
		flag = compa.bit.bor(flag, fcntl.O_NONBLOCK)
	end
	fcntl.fcntl( read_fd, fcntl.F_SETFL, flag )
	--TODO: This be not very good.
	--      read side: sub-process don't inherit from
	fcntl.fcntl( read_fd, fcntl.F_SETFD, 1 )

	--Non-blocking write pipe.
	flag = fcntl.fcntl( write_fd, fcntl.F_GETFL )
	if compa.lua_version >= 5.3 then
		flag = flag | fcntl.O_NONBLOCK
	elseif compa.bit then
		flag = compa.bit.bor(flag, fcntl.O_NONBLOCK)
	end
	fcntl.fcntl( write_fd, fcntl.F_SETFL, flag )

	_M.lineBuffer = ''

	local task = coaf.NewIO( function(task_obj)
		while true do
			local buf,e,errcode = unistd.read( read_fd, 256 )
			if not buf then
				if errcode ~= errno.EAGAIN and errcode ~= errno.EWOULDBLOCK then
					task_obj:destroy()
				end
				break
			elseif #buf == 0 then
				task_obj:destroy()
				break
			end

			local pre,nx = buf:match("(.-)[\r]?[\n](.*)$")
			if pre then
				buf = _M.lineBuffer .. pre	--A whole line
				_M.OutputLine(buf)		
				buf = nx
				pre,nx = buf:match("(.-)[\r]?[\n](.*)$")
				while pre do
					_M.OutputLine(pre)
					buf = nx
					pre,nx = buf:match("(.-)[\r]?[\n](.*)$")
				end
				_M.lineBuffer = buf
			else
				_M.lineBuffer = _M.lineBuffer .. buf
				if #_M.lineBuffer >= 4096 then --Too long.
					_M.OutputLine(_M.lineBuffer)
					_M.lineBuffer = ''
				end
			end
		end
	end, read_fd, ev.READ )
	if not task then
		return false, "Fail to monitor stdio"
	end
	
	task:start()
	_dbg.DisableConsole() --Don't output to console, because I will watch the console.
	unistd.dup2( write_fd, 1 )	-- stdout
	unistd.dup2( write_fd, 2 )	-- stderr
	unistd.close( write_fd )
	return true
end

return _M

