--[==[

The Co-Application Framework (CoAF) RPC service module.

Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local compa = require "coaf.compatible"
local _dbg = require "coaf.debug"
local protocol = require "coaf.rpc.protocol"
local connection = require "coaf.rpc.connection"
local daemon = require "coaf.rpc.daemon"

local _ready, coaf = pcall( require, "coaf.core" )
if not _ready then
	coaf = nil
end

local DEF_COSERVER_ADDR = "unix:/var/run/coaf.coserver"

local _M = {
	socket = nil,
	sleep = nil,
	coserver_addr = DEF_COSERVER_ADDR,
	services = {},
	socketDaemons = {},
}

-------------------------------------------------------
-- You must call 'InitSocket(socket_obj)' first, before
-- using CoAF service.
-------------------------------------------------------
function _M.InitSocket(socket)
	if not socket then
		if ngx and ngx.socket then -- Check in openresty environment?
			_dbg.Info("CoAF.Service", "Init service socket: found ngnix-openresty socket")
			_M.socket = ngx.socket
			_M.sleep = ngx.sleep
		else
			local status, socket = pcall(require, "cosocket.core")
			if status and socket then
				_dbg.Info("CoAF.Service", "Init service socket: found CoAF cosocket")
				_M.socket = socket
				_M.sleep = socket.sleep
			else
				_dbg.Error("CoAF.Service", "Init service socket: no socket libary found!")
			end
		end
	else
		_M.socket = socket
	end
end

function _M.SetCoServerAddress(addr)
	_M.coserver_addr = addr or DF_COSERVER_ADDR
end

------------------------------------------------------
local service_meta = {}

-- NOTE:
-- Simplely, the object path is '/path/to/object'.
-- if an object path is '/path/to/object:[sub1[.sub2[.sub3 ...]].]subobject', that
-- try check the sub object of registered object '/path/to/object',
-- the sub object is obj.sub1.sub2.sub3.subobject
-- e.g.:
-- /configure:video
-- /configure:video.source.format

function service_meta:GetObject(objPath)
	if not objPath then
		return nil
	end
	local obj = self.exposedObjects[objPath]
	if obj then return obj end --First: try full match.

	local token, subs = objPath:match("(.+):(.*)$")
	if not token then
		return nil
	else
		obj = self.exposedObjects[token]
		if not obj then
			return nil
		end
		subs = subs .. "." --terminator.
		local maybeNum, nextObj
		for token in subs:gmatch("(.-)%.") do
			if token == '' then --Invalid.
				return nil
			end
			maybeNum = tonumber(token)
			nextObj = obj[token]
			if not nextObj and maybeNum then
				nextObj = obj[maybeNum]
			end
			if type(nextObj) ~= "table" then
				return nil
			else
				obj = nextObj
			end
		end
		return nextObj
	end
end

function service_meta:Destroy()
	self:DisconnectCoServer()
	local i, daemon
	for i, daemon in ipairs(self.daemons) do
		if daemon then
			daemon:RemoveService( self.serviceId )
		end
	end
	self.daemons = {}
	_M.services[self.serviceId] = nil
	collectgarbage()
end

function service_meta:ExposeObject(objPath, objTable)
	self.exposedObjects[objPath] = objTable
end

function service_meta:RemoveObject(objPath)
	self.exposedObjects[objPath] = nil
end

function service_meta:Subscribe(connObj, objPath, from, ...)
	if not self:GetObject(objPath) then
		return nil, "Object not exist"
	end

	-- coserver proxy link, always send notify messages.
	if connObj == self.coserver_connection then
		return true
	end

	if not connObj.subscribes then
		connObj.subscribes = {}
	end

	local svcSubs = connObj.subscribes[ self.serviceId ]
	if not svcSubs then
		svcSubs = {}
		connObj.subscribes[ self.serviceId ] = svcSubs
	end

	local objSubs = svcSubs[ objPath ]
	if not objSubs then
		objSubs = {}
		svcSubs[ objPath ] = objSubs
	end

	local topics = {...}
	local i, j, v
	for i=1, #topics do
		if not objSubs[ topics[i] ] then
			objSubs[ topics[i] ] = {from}
		else
			local exists = false
			for j, v in ipairs( objSubs[topics[i]] ) do
				if v == from then --Already exist
					exists = true
					break
				end
			end
			if not exists then
				table.insert( objSubs[ topics[i] ], from )
			end
		end
	end
	return true
end

function service_meta:Unsubscribe(connObj, objPath, from, ...)
	if not self:GetObject(objPath) then
		return nil, "Object not exist"
	end

	-- coserver proxy link, always send notify messages.
	if connObj == self.coserver_connection then
		return true
	end

	if not connObj.subscribes then
		return true
	end

	local svcSubs = connObj.subscribes[ self.serviceId ]
	if not svcSubs then
		return true
	end

	local objSubs = svcSubs[ objPath ]
	if not objSubs then
		return true
	end

	local topics = {...}
	local i, j, v
	for i=1, #topics do
		if objSubs[ topics[i] ] then
			for j, v in ipairs( objSubs[ topics[i] ] ) do
				if v == from then
					table.remove( objSubs[ topics[i] ], j )
					break
				end
			end
			if #objSubs[ topics[i] ] == 0 then
				objSubs[ topics[i] ] = nil --Drop subscribe topic.
			end
		end
	end
	return true
end

function service_meta:Notify(objPath, topic, ...)
	local i,daemon 
	local notifyCnt = 0
	
	local msg

	for i, daemon in ipairs(self.daemons) do
		if daemon and daemon.peers then
			local k,conn
			for k, conn in pairs(daemon.peers) do
				if conn and (conn == self.coserver_connection or conn.subscribes) then
					local svcSubs
					if conn ~= self.coserver_connection then
						svcSubs = conn.subscribes[ self.serviceId ]
					end

					if conn == self.coserver_connection or svcSubs then
						local objSubs 
						if conn ~= self.coserver_connection then
							objSubs = svcSubs[objPath]
						end

						if conn == self.coserver_connection or (objSubs and objSubs[topic]) then
							-- Generate notify message here.
							if not msg then
								local req = {
									SN = self:NewSN(),
									CMD = "NOTIFY",
									ARGS = topic,
									headers = {
										SVC = self.serviceId,
										OBJ = objPath
									},
									parameters = {...}
								}
								msg = table.concat( protocol.FormatContent(req) )
							end
							-- Generated.

							conn:Send(msg)
							notifyCnt = notifyCnt + 1
						end
					end -- 
				end -- With subscribes.
			end -- Each peer connections.
		end -- Each daemon
	end
	return notifyCnt
end

local function CoServerProxyLinkCoro( service, addr, port, kaTimeout )
	local coro = coroutine.running()
	while true do
		local done = service:__InternalCoServerProxyLink(coro,addr,port, kaTimeout)
		if done then
			local r, err = coroutine.yield( coro )
			if not r then
				break
			else
				service.coserver_connection = nil
				_dbg.Info( "CoAF.Service", "Co-server proxy link is shutdown and reconnecting, for service:", service.serviceId )
			end
		else
			_M.sleep(3)
			_dbg.Info( "CoAF.Service", "Co-server proxy link is trying reconnect, for service:", service.serviceId )
		end
	end
	if service.coserver_connection then
		service.coserver_connection:Destroy()
		service.coserver_connection = nil
	end
end

function service_meta:__InternalCoServerProxyLink(coro, addr,port, kaTimeout)
	local sock = _M.socket.tcp()

	if not sock then
		_dbg.Error("CoAF.Service", "Fail to create CoServer proxy connection! For service:", self.serviceId, " To:", addr, port)
		return false
	end

	local status, err = sock:connect(addr, port)
	if not status then
		_dbg.Error("CoAF.Service", "Fail to connect CoServer. For service:", self.serviceId, " To:", addr, port)
		sock:close()
		return false
	end

	if kaTimeout and kaTimeout <= 0 then
		kaTimeout = 60
	end

	local req = {
		SN = self:NewSN(),
		CMD = "PROXY",
		headers = {
			SVC = self.serviceId,
			LOC = loc,
			KA = kaTimeout
		}
	}

	status, err = sock:send( table.concat( protocol.FormatContent(req) ) )
	if not status then
		sock:close()
		_dbg.Error("CoAF.Service", "Fail to establish connection to CoServer (Send error). For service:", self.serviceId, " To:", addr, port)
		return false
	end

	local parser = protocol.NewParser()
	local data, partical
	local ack
	while true do
		data, err, partical = sock:receive( parser.next_fetch )

		if not data then
			if err ~= "timeout" and err ~= "busy" then
				sock:close()
				_dbg.Error("CoAF.Service", "Fail to establish connection to CoServer (I/O error). For service:", self.serviceId, " To:", addr, port)
				return false
			end
			if partial and #partial > 0 then
				data = partial
			elseif err == "timeout" then
				sock:close()
				_dbg.Error("CoAF.Service", "Fail to establish connection to CoServer (timeout). For service:", self.serviceId, " To:", addr, port)
				return false
			else  --NOTE: here! Very important: (BUG found)
			-- In coaf environment, 'busy' means the socket is using by others.
			-- In this case, if this task (coroutine, or main task) does not yield
			-- out, it will fall in a dead loop: Other task will never be waked up
			-- and this loop will always continue ('busy' is always get).
				if coaf then
					coaf.Sleep(0.01)
				end
			end
		end

		if data then
			status, err = parser:Push(data)
			if status and err == "complete" then
				local content = parser.content
				parser:Clear()

				if content.CMD == "ACK/PROXY" and tonumber(content.SN) == req.SN then
					ack = content
					break
				end
			end
		end
	end

	if ack.headers.RESULT == "OK" then
		--TODO: Check which daemon is stand-alone?
		local daemon = self.daemons[1]
		if daemon then
			self.coserver_connection = daemon:AddSpecialConnection( sock, nil, function()
				coroutine.resume( coro, true, "reconnect" )
			end)
		end
		if not self.coserver_connection then
			_dbg.Error("CoAF.Service", "Fail to add CoServer proxy connection to daemon, For service:", self.serviceId, " To:", addr, port)
			sock:close()
			return false
		else
			_dbg.Info("CoAF.Service", "CoServer proxy established. For service:", self.serviceId )
			if kaTimeout then
				self.coserver_connection:SetKeepAlive( {keepAlive=true, keepAliveTimeout=kaTimeout} )
			end
			return true
		end
	else
		_dbg.Error("CoAF.Service", "Fail to establish proxy connection to CoServer, reason:", ack.headers.MSG, "For service:", self.serviceId, " To:", addr, port)
		sock:close()
		return false
	end
end

function service_meta:CreateCoServerProxy(addr, loc, kaTimeout)
	if self.coserver_thread then
		_dbg.Warn("CoAF.Service", "Service to CoServer connection is already created. Service:", self.serviceId )
		return true
	end

	if not _M.socket then
		_dbg.Error("CoAF.Service", "Invalid socket library initialized. ConnectToServer for service:", self.serviceId)
		return false
	end

	if not addr then
		addr = _M.coserver_addr
	end

	if not loc then
		loc = "local"
	end

	local port
	if not addr:match("^unix:") then
		local ip
		ip, port = addr:match("(.*):([0-9]+)")
		if ip then
			if ip == '' or ip == '*' then
				addr = "127.0.0.1"
			else
				addr = ip
			end
			port = tonumber(port) or 14000
		else
			port = 14000
		end
		port = tonumber(port) or 14000
	else
		port = 0 --NOTE: openresty must specify a port number.
	end 

	self.coserver_thread = coroutine.create( CoServerProxyLinkCoro )
	coroutine.resume( self.coserver_thread, self, addr, port, kaTimeout )
	return true
end

function service_meta:ReleaseCoServerProxy()
	if self.coserver_thread then
		coroutine.resume( self.coserver_thread, false, "#CoAF-KILL#" )
		self.coserver_thread = nil
	end
end

function service_meta:NewSN()
	local sn = self.internal_sn or 0
	self.internal_sn = sn + 1
	if self.internal_sn > 99999999 then
		self.internal_sn = 0
	end
	return sn
end

function service_meta:GetInvokingConnection()
	return self.__last_invoke_connection
end

service_meta.__index = service_meta

------------------------------------------------------

function _M.New(service_id, sock_addr)

	if not service_id then
		_dbg.Error("CoAF.Service", "Create service with invalid service id.")
		return nil
	end

	local svcObj = _M.services[service_id]
	--Note here: a service should be bound at multiple addresses.
	if svcObj then
		_dbg.Info("CoAF.Service", "Pick up the exist service because it is already created while create new service:", service_id )
	end

	if not sock_addr then --No socket address, so create a stand-alone service.
		if svcObj then --Pick up the exist object.
			return svcObj
		end

		local dm = daemon.New( nil, "service:" .. service_id )
		if not dm then
			_dbg.Error("CoAF.Service", "Fail to create daemon for service:", service_id)
			return nil
		end
		svcObj = {
			daemons = { dm },
			internal_sn = 0,
			exposedObjects = {},
			coserver_connection = nil,
			coserver_thread = nil,
			serviceId = service_id,
			__last_invoke_connection = nil,
		}
		setmetatable(svcObj, service_meta)
		dm:RegisterService(service_id, svcObj)
		_M.services[service_id] = svcObj
		dm:Start()
		return svcObj
	else
		if not _M.socket then
			_dbg.Error("CoAF.Service", "Invalid socket library initialized. Create service:", service_id)
			return false
		end

		local addr, port
		if not sock_addr:match("^unix:") then
			addr, port = sock_addr:match("(.*):([0-9]+)")
			if not addr then
				_dbg.Error("CoAF.Service", "Invalid service binding socket address of:", sock_addr, " while creating service:", service_id )
				return nil
			end
			if addr == '' then
				addr = '*'
			end
			port = tonumber(port)
			sock_addr = addr .. ":" .. port  -- Formal format.
		else
			addr = sock_addr
			port = 0 --NOTE: openresty must accept a number port.
		end

		local dm = _M.socketDaemons[sock_addr]
		if dm then
			_dbg.Info("CoAF.Service", "Create service:", service_id, " using exists daemon of address:", sock_addr )
		else
			local unix, path = sock_addr:match("(unix:)(.*)")
			if unix then
				os.remove(path)	-- Try remove exist unix domain file.
			end
			local sock = _M.socket.tcp()
			if not sock then
				_dbg.Error("CoAF.Service", "Fail to create service socket for: ", service_id )
				return nil
			end

			local x, err = sock:bind(addr, port)
			if not x then
				_dbg.Error("CoAF.Service", "Fail to bind address at ", sock_addr, " for service:", service_id )
				sock:close()
				return nil
			end

			sock:listen()
			dm = daemon.New(sock, sock_addr)
			if not dm then
				_dbg.Error("CoAF.Service", "Fail to create daemon for service:", service_id)
				sock:close()
				return nil
			end
			_M.socketDaemons[sock_addr] = dm
		end

		if not svcObj then
			svcObj = {
				daemons = { dm },
				internal_sn = 0,
				exposedObjects = {},
				coserver_connection = nil,
				coserver_thread = nil,
				serviceId = service_id,
				__last_invoke_connection = nil,
			}
			setmetatable(svcObj, service_meta)
			_M.services[service_id] = svcObj
		else
			local i, v
			local fnd
			for i, v in ipairs(svcObj.daemons) do
				if v == dm then
					fnd = true
					break
				end
			end
			if not fnd then
				table.insert( svcObj.daemons, dm )
			end
		end

		dm:RegisterService(service_id, svcObj)
		dm:Start()
		return svcObj
	end
end

function _M.GetService(service_id)
	return _M.services[service_id]
end

--Notify subscribed topic.
function _M.Notify(service_id, objPath, topic, ...)
	local svc = _M.services[service_id]
	if not svc then
		return
	end
	return svc:Notify(objPath, topic, ...)
end

----------------------------------------------------------

return _M


