---------------------------------------------------------
-- CoServer services manager.
---------------------------------------------------------
local lfs = require "lfs"
local _dbg = require "coaf.debug"
local coaf = require "coaf.core"
local pretty = require "pl.pretty"
local thread = require "coaf.thread"

local _M = {
	config_location = "/etc/coserver",
	blacklist_file = "blacklist.conf",
	expose_file = "expose.conf",
	services = {},
	auto_starts = {},
	blacklist = {},
	remote_expose = {},
	running = {}
}

local LOCAL = "local"
local ANY = "ANY"
local PATH_DELIM = "/"

if coaf.OSType() == "Windows" then
	_M.config_location = ".\\config"
	PATH_DELIM = "\\"
end

---------------------------------------------------------
-- Service configure file:
--
--		ACTION = EXEC(prog[,args_table])  [EXECP]
--		or
--		ACTION = ACCEPT_PROXY(["bind-address-pattern"[,"LOC", "SERVICE_ID"]])
--		or
--		ACTION = REQUEST_PROXY("address"[, "LOC", "SERVICE_ID"] )
--
--		TIMEOUT = <wait-timeout-in-seconds>
--		KEEP_ALIVE = <false|true|interleave-value>
--
--		ALT_SOCKETS = {"alternative-conn-address1", "alternative-conn-address2", ...}
--		AUTO_START = <seq_no>|true (means any)
--
-- bellow arguments only for 'EXEC':
--
--		PIDFILE = <false|"pid-file-name">
--		NOT_COAF = <true|false>
--		PROXY = <"none"|"any"|"accept-bind-address-pattern">
--
---------------------------------------------------------

function _M.ParseServiceConfig(cfgStr, loc, service_id)
	local funcGen = {
[[
	local args={...}
	local actions = require "coserver.actions"
	local EXEC = actions.EXEC
	local EXECP = actions.EXECP
	local ACCEPT_PROXY = actions.ACCEPT_PROXY
	local REQUEST_PROXY = actions.REQUEST_PROXY
	
	local ACTION
	local TIMEOUT
	local KEEP_ALIVE
	local ALT_SOCKETS
	local AUTO_START
	local PIDFILE
	local PROXY
	local KILL_SIG
	local AUTHOR_KEY
	local NOT_COAF

	local service = {
		LOC = args[1] or "ANY",
		ID = args[2] or "ANY"
	}
]]
	}

	table.insert( funcGen, cfgStr )
	table.insert( funcGen, [[
	
	service.ACTION = ACTION or service.ACTION
	service.TIMEOUT = TIMEOUT or service.TIMEOUT
	service.KEEP_ALIVE = KEEP_ALIVE or service.KEEP_ALIVE
	service.ALT_SOCKETS = ALT_SOCKETS or service.ALT_SOCKETS
	service.AUTO_START = AUTO_START or service.AUTO_START
	service.PIDFILE = PIDFILE or service.PIDFILE
	service.PROXY = PROXY or service.PROXY
	service.KILL_SIG = KILL_SIG or service.KILL_SIG
	service.AUTHOR_KEY = AUTHOR_KEY or service.AUTHOR_KEY
	service.NOT_COAF = NOT_COAF or service.NOT_COAF

	return service
]])
	local func, err = load(table.concat( funcGen ))
	if not func then
		_dbg.Error("CoServer", "Fail to parse service configure string, of " .. (loc or "ANY") .. "/" .. (service_id or "ANY") .. ":", err)
		return nil
	end
	local flag, ret = pcall(func, loc, service_id)
	if not flag then
		_dbg.Error("CoServer", "Fail to load service configure string, of " .. (loc or "ANY") .. "/" .. (service_id or "ANY") .. ":", ret)
		return nil
	end
	return ret
end

-----------------------------------------------------------

local function CheckServiceConfig(file, loc, service_id)
	local f = io.open(file, "r")
	if not f then
		_dbg.Error("CoServer", "Fail to open service configure file: ", file, "For service: " .. (loc or "ANY") .. "/" .. (service_id or "ANY") )
		return nil
	end
	local cfg = f:read("*a")
	if not cfg then
		_dbg.Error("CoServer", "Fail to read service configure file: ", file, "For service: " .. (loc or "ANY") .. "/" .. (service_id or "ANY") )
		f:close()
		return nil
	end
	f:close()
	local ret = _M.ParseServiceConfig(cfg, loc, service_id)
	if ret then
		ret.CONFIG_FILE = file
	end
	return ret
end

local function CheckServices(dir, loc)
	local attrs, item, file
	local svc_config
	local rets = {}

	for item in lfs.dir(dir) do
		if item ~= "." and item ~= ".." then
			file = dir .. PATH_DELIM .. item
			attrs = lfs.attributes(file)
			if attrs and attrs.mode == "file" then
				svc_config = CheckServiceConfig(file, loc, item)
				if svc_config then
					table.insert(rets, svc_config)
				end
			end
		end
	end
	return rets
end

-------------------------------------------------------

function _M.MatchServiceID(from, to)
	if from == to then
		return true
	end

	local from_loc, from_svc
	local to_loc, to_svc

	if type(from) == "table" then
		from_loc = from[1] or "ANY"
		from_svc = from[2] or "ANY"
	else
		from_loc, from_svc = from:match("(.*)/(.*)$")
		if not from_loc then
			from_loc = from
			from_svc = "ANY"
		end
	end

	if type(to) == "table" then
		to_loc = to[1] or "ANY"
		to_svc = to[2] or "ANY"
	else
		to_loc, to_svc = to:match("(.*)/(.*)$")
		if not to_loc then
			to_loc = to
			to_svc = "ANY"
		end
	end

	if (from_loc == to_loc or from_loc == "ANY" or to_loc == "ANY") and
	   (from_svc == to_svc or from_svc == "ANY" or to_svc == "ANY") then
		return true
	else
		return false
	end
end

-------------------------------------------------------

function _M.LoadServices()
	local attrs = lfs.attributes( _M.config_location)
	if not attrs or attrs.mode ~= "directory" then
		_dbg.Error("CoServer", "Invalid configure directory of:", _M.config_location)
		return
	end

	local newList = {}		-- To save new loaded services.

	local item
	local svcs, dir
	local i, v
	for item in lfs.dir(_M.config_location) do
		if item ~= "." and item ~= ".." then
			dir = _M.config_location .. PATH_DELIM .. item
			attrs = lfs.attributes(dir)
			if attrs and attrs.mode == "directory" then
				svcs = CheckServices(dir, item)
				for i,v in ipairs(svcs) do
					_dbg.Info("CoServer", "Loaded service of:", v.LOC .. "/" .. v.ID )
					newList[v.LOC .. "/" .. v.ID] = v
				end
			end
		end
	end

	--Merge manual added services, if reloaded services.
	for i,v in pairs(_M.services) do
		if not newList[i] and v.MANUAL then
			newList[i] = v
		end
	end

	_M.services = newList

	--Generate auto start list
	newList = {}
	for i,v in pairs(_M.services) do
		if v.AUTO_START then
			table.insert(newList, {seq= (tonumber(v.AUTO_START) or 0xffffffff), service_id = i} )	--Append to table first, then sort.
		end
	end

	table.sort(newList, function(a,b) 
		if a == b then
			return false
		elseif not a.seq or not b.seq then
			return false
		else
			return a.seq < b.seq
		end
	end)

	_M.auto_starts = newList
end

-------------
function _M.GetExactService(loc,svcId)
	loc = loc or "ANY"
	svcId = svcId or "ANY"
	local svc, key
	key = loc .. "/" .. svcId
	svc = _M.services[key]
	if svc then
		return key, svc
	end

	if svcId ~= "ANY" then
		key = loc .. "/ANY"
		svc = _M.services[key]
		if svc then
			return key, svc
		end
	end

	if loc ~= "ANY" then
		key = "ANY/" .. svcId
		svc = _M.services[key]
		if svc then
			return key, svc
		end

		if svcId ~= "ANY" then
			key = "ANY/ANY"
			svc = _M.services[key]
			if svc then
				return key, svc
			end
		end
	end

	return nil
end
-------------

function _M.LoadBlacklist()
	local f = io.open( _M.config_location .. PATH_DELIM .. _M.blacklist_file, "r")
	if not f then
		return false
	end

	local trunkCode = "return " .. f:read("*a")
	f:close()

	local func,err = load(trunkCode)
	if not func then
		_dbg.Error("CoServer", "LoadBlacklist - Fail to load blacklist configure file:", err)
		return false
	end

	local sta, ret = pcall(func)
	if not sta then
		_dbg.Error("CoServer", "LoadBlacklist - Invalid blacklist configure file:", ret)
		return false
	end

	if type(ret) == "table" then
		_M.blacklist = ret
	else
		_dbg.Warn("CoServer", "LoadBlacklist - Bad blacklist content.")
	end
	return true
end

-------------
-- Remote expose:	expose.conf
-- {
--		-- REMOTE-ID				LOCAL-ID
-- 		["LOC/service_id"] = "local_LOC/local_service_id"
--
-- 		* or this (below) means auto set location as requested location:
--
-- 		["ANY/service_id"] = "local_LOC/local_service_id" 
-- 		
-- 		[...] = "..."
--
-- 		** New features (after 1.1 version):
--
-- 		[REMOTE-ID] = {
-- 			local_service = "local_LOC/local_service_id",
--			allow_addresses = { "ip-pattern", "ip-pattern2", "*", "unix", "local", "ip", "remote", ... },
--			allow_objects = { "/", "/path/to/object", ... }
-- 		}
-- }

function _M.LoadRemoteExposeList()
	local f = io.open( _M.config_location .. PATH_DELIM .. _M.expose_file, "r")
	if not f then
		return false
	end

	local trunkCode = "return " .. f:read("*a")
	f:close()

	local func,err = load(trunkCode)
	if not func then
		_dbg.Error("CoServer", "LoadRemoteExposeList - Fail to load expose configure file:", err)
		return false
	end

	local sta, ret = pcall(func)
	if not sta then
		_dbg.Error("CoServer", "LoadRemoteExposeList - Invalid expose configure file:", ret)
		return false
	end

	if type(ret) == "table" then
		_M.remote_expose = ret
	else
		_dbg.Warn("CoServer", "LoadRemoteExposeList - Bad expose content.")
	end
	return true
end

-------------

local function CheckAdvancedTranslateService(svc, req_obj, ipaddr)
	if type(svc) ~= "table" or not svc.local_service then
		return nil
	end
	local svcStr = svc.local_service
	if req_obj and svc.allow_objects then
		local objs
		if type(svc.allow_objects) ~= "table" then
			objs = { svc.allow_objects }
		else
			objs = svc.allow_objects
		end
		if #objs > 0 then
			local i, obj, fnd
			for i,obj in ipairs(objs) do
				if obj == req_obj then
					fnd = true
					break
				end
			end
			if not fnd then
				return nil
			end
		end
	end

	if ipaddr and svc.allow_addresses then
		local addrs
		if type(svc.allow_addresses) ~= "table" then
			addrs = { svc.allow_addresses }
		else
			addrs = svc.allow_addresses
			if #addrs == 0 then
				return svcStr
			end
		end
		local i, pat
		for i,pat in ipairs(addrs) do
			if pat == "*" or ((pat == "unix" or pat == "local") and ipaddr:match("^unix")) or
				((pat == "ip" or pat == "remote") and not ipaddr:match("^unix")) or
				ipaddr:match(pat) then
				return svcStr
			end
		end
	else
		return svcStr
	end

	return nil
end

function _M.TranslateRemoteService(loc,svcId,transTable,req_obj,ipaddr)
	loc = loc or "ANY"
	svcId = svcId or "ANY"
	local svc, key
	key = loc .. "/" .. svcId
	if not transTable then
		transTable = _M.remote_expose
	end

	local t_loc, t_svc
	svc = transTable[key]
	if type(svc) == "table" then
		svc = CheckAdvancedTranslateService(svc, req_obj, ipaddr)
	end
	if svc then
		t_loc, t_svc = svc:match("(.*)/(.*)$")
		return svc, t_loc == "ANY" and loc or t_loc, t_svc == "ANY" and svcId or t_svc
	end

	if svcId ~= "ANY" then
		key = loc .. "/ANY"
		svc = transTable[key]
		if type(svc) == "table" then
			svc = CheckAdvancedTranslateService(svc, req_obj, ipaddr)
		end
		if svc then
			t_loc, t_svc = svc:match("(.*)/(.*)$")
			return svc, t_loc == "ANY" and loc or t_loc, t_svc == "ANY" and svcId or t_svc
		end
	end

	if loc ~= "ANY" then
		key = "ANY/" .. svcId
		svc = transTable[key]
		if type(svc) == "table" then
			svc = CheckAdvancedTranslateService(svc, req_obj, ipaddr)
		end
		if svc then
			t_loc, t_svc = svc:match("(.*)/(.*)$")
			return svc, t_loc == "ANY" and loc or t_loc, t_svc == "ANY" and svcId or t_svc
		end

		if svcId ~= "ANY" then
			key = "ANY/ANY"
			svc = transTable[key]
			if type(svc) == "table" then
				svc = CheckAdvancedTranslateService(svc, req_obj, ipaddr)
			end
			if svc then
				t_loc, t_svc = svc:match("(.*)/(.*)$")
				return svc, t_loc == "ANY" and loc or t_loc, t_svc == "ANY" and svcId or t_svc
			end
		end
	end

	return nil
end

function _M.MatchTranslatedServices(loc,svcId,transTable,ipaddr)
	local rslt = {}

	loc = loc or "ANY"
	svcId = svcId or "ANY"
	local key, svc
	if not transTable then
		transTable = _M.remote_expose
	end

	local exp_svc, loc_svc, skip
	for exp_svc, loc_svc in pairs(transTable) do
		skip = false
		key, svc = exp_svc:match("(.*)/(.*)$")
		if not key then
			skip = true
		elseif loc == "ANY" then
			if key == "ANY" then --All is any, can't confirm what location it is, so skip this bad case.
				skip = true
			--else use the key.
			end
		else
			if key == "ANY" then
				key = loc
			elseif loc ~= key then
				skip = true
			end
		end

		if not skip then --Try match service-id.
			if svcId == "ANY" then
				if svc == "ANY" then --Can't confirm what service-id it is, skip this bad case
					skip = true
				--else use the svc.
				end
			else
				if svc == "ANY" then
					svc = svcId
				elseif svc ~= svcId then
					skip = true
				end
			end
		end

		if not skip then --Check the permission.
			if type(loc_svc) == "table" and not CheckAdvancedTranslateService(loc_svc, nil, ipaddr) then
				skip = true
			end
		end

		if not skip then
			table.insert( rslt, key .. "/" .. svc )
		end
	end
	if #rslt == 0 then
		return nil
	else
		return rslt
	end
end

-------------

function _M.TranslateLocalService(loc,svcId,transTable)
	loc = loc or "ANY"
	svcId = svcId or "ANY"
	local svc, key
	if not transTable then
		transTable = _M.remote_expose
	end

	local matchs = {
		loc .. "/" .. svcId,
		loc .. "/ANY",
		"ANY/" .. svcId,
		"ANY/ANY"
	}

	local matchRslt = {}

	local k, v, i
	for k,v in pairs(transTable) do
		if type(v) == "table" then
			v = v.local_service  --Advanced case.
		end

		for i=1, #matchs do
			if v == matchs[i] then
				if i == 1 then -- Exact matched, so return directly.
					return k, k:match("(.*)/(.*)$")
				else
					matchRslt[i] = k
				end
			end
		end
	end

	local m_loc, m_svc
	for i=2,#matchs do
		if matchRslt[i] then
			m_loc, m_svc = matchRslt[i]:match("(.*)/(.*)$")
			if m_loc == "ANY" then
				m_loc = loc
			end

			if m_svc == "ANY" then
				m_svc = svc
			end
			return matchRslt[i], m_loc, m_svc
		end
	end
	return nil
end

-------------

function _M.AddService(loc, service_id, cfg, save)
	if not cfg then
		_dbg.Error("CoServer", "AddService - Invalid service configuration")
		return nil
	end

	loc = loc or "ANY"
	service_id = service_id or "ANY"

	if type(cfg) == "table" then
		if type(cfg.ACTION) == "string" then
			cfg.LOC = loc
			cfg.ID = service_id
			local func, err = load([[
local args={...}
local service = args[1]
local actions = require "coserver.actions"
local EXEC = actions.EXEC
local EXECP = actions.EXECP
local ACCEPT_PROXY = actions.ACCEPT_PROXY
local REQUEST_PROXY = actions.REQUEST_PROXY

]] .. "service.ACTION = " .. cfg.ACTION .. [[

return true
]])
			if not func then
				_dbg.Error("CoServer", "AddService - Bad ACTION argument:", err)
				return nil
			end
			local status
			status, err = pcall(func, cfg)
			if not status then
				_dbg.Error("CoServer", "AddService - Fail parse ACTION argument:", err)
				return nil
			end
		end
	elseif type(cfg) == "string" then
		cfg = _M.ParseServiceConfig(cfg, loc, service_id)
		if not cfg then
			_dbg.Error("CoServer", "AddService - Fail to parse service configuration")
			return nil
		end
	else
		_dbg.Error("CoServer", "AddService - Invalid service configuration content")
		return nil
	end

	cfg.MANUAL = true
	local key = cfg.LOC .. "/" .. cfg.ID
	if _M.services[key] then
		_dbg.Warn("CoServer", "AddService - The service of " .. key .. " is already exist, so replaced old one.")
	end

	_M.services[key] = cfg
	if cfg.AUTO_START then
		local seq = tonumber(cfg.AUTO_START) or 0xffffffff
		local i, v
		local ins_pos
		for i, v in ipairs(_M.auto_starts) do
			if not ins_pos and v.seq > seq then
				ins_pos = i
			end
			if v.service_id == key then --already exists
				ins_pos = -1
				break
			end
		end
		if not ins_pos then
			table.insert(_M.auto_start, {
				seq = seq,
				service_id = key
			})

			_M.StartService(loc, service_id) --Since auto start, try start it.
		elseif ins_pos > 0 then
			table.insert(_M.auto_start, ins_pos, {
				seq = seq,
				service_id = key
			})

			_M.StartService(loc, service_id) --Since auto start, try start it.
		end
	end

	if save then
		local loc = _M.config_location .. PATH_DELIM .. cfg.LOC
		local r,err = lfs.mkdir(loc)
		if not r and err ~= "File exists" then
			_dbg.Warn("CoServer", "AddService - Fail to create service saving location:", loc)
			return cfg
		end

		local f = io.open( loc .. PATH_DELIM .. cfg.ID, "w")
		if not f then
			_dbg.Warn("CoServer", "AddService - Fail to create service file:", loc .. PATH_DELIM .. cfg.ID )
			return cfg
		end

		local k,v
		for k,v in pairs(cfg) do
			if k ~= "MANUAL" and k ~= "CONFIG_FILE" and k ~= "LOC" and k ~= "ID" then
				f:write( k .. " = ")
				if type(v) == "string" then
					f:write("\"")
					f:write(v)
					f:write("\"\r\n")
				elseif type(v) == "table" and not v.IS_COSERVER_ACTION then
					f:write( pretty.write(v,""))
					f:write("\r\n")
				else
					f:write(tostring(v))
					f:write("\r\n")
				end
			end
		end

		f:close()
		cfg.CONFIG_FILE = loc .. PATH_DELIM .. cfg.ID
	end
	return cfg
end

-------------

function _M.RemoveService(loc, service_id, save)
	local key = (loc or "ANY") .. "/" .. (service_id or "ANY")
	local file
	if _M.services[key] then
		file = _M.services[key].CONFIG_FILE
		_M.StopService(loc, service_id) --Stop first.
		local i,v
		for i,v in ipairs(_M.auto_starts) do
			if v.service_id == key then
				table.remove( _M.auto_starts,i )
				break
			end
		end
		_M.services[key] = nil
	end
	if file and save then
		os.remove(file)
	end
	return true
end

-------------

function _M.EnableService(loc, service_id, save)
	local key = (loc or "ANY") .. "/" .. (service_id or "ANY")
	local i,v
	local fnd
	for i,v in ipairs(_M.blacklist) do
		if v == key then
			table.remove( _M.blacklist, i )
			fnd = true
			break
		end
	end
	if fnd then
		fnd = false
		for i,v in ipairs(_M.auto_starts) do
			if v.service_id == key then
				fnd = true
				break
			end
		end

		if fnd then -- Auto start required, so start this service.
			_M.StartService(loc, service_id)
		end

		if save then
			local f = io.open( _M.config_location .. PATH_DELIM .. _M.blacklist_file, "w")
			if f then
				f:write(pretty.write(_M.blacklist, "\t"))
				f:close()
			else
				_dbg.Warn("CoServer", "EnableService - Fail to write blacklist file.")
			end
		end
		return true
	else
		return false
	end
end

-------------

function _M.DisableService(loc, service_id, save)
	local key = (loc or "ANY") .. "/" .. (service_id or "ANY")
	local i,v
	local fnd
	for i,v in ipairs(_M.blacklist) do
		if v == key then
			fnd = true
			break
		end
	end
	if not fnd then
		table.insert(_M.blacklist, key)
	end

	_M.StopService(loc, service_id) --Stop service, if it's started.

	if save then
		local f = io.open( _M.config_location .. PATH_DELIM .. _M.blacklist_file, "w")
		if f then
			f:write(pretty.write(_M.blacklist, "\t"))
			f:close()
		else
			_dbg.Warn("CoServer", "DisableService - Fail to write blacklist file.")
		end
	end
	return true
end

-------------

function _M.ServiceIsDisabled(loc, svcId)
	loc = loc or "ANY"
	svcId = svcId or "ANY"

	local matchs = {
		"ANY/ANY", loc .. "/ANY", "ANY/" .. svcId, loc .. "/" .. svcId
	}
	local i,v
	local j
	for i,v in pairs(_M.blacklist) do
		for j = 1,4 do
			if v == matchs[j] then
				return true
			end
		end
	end
	return false 
end

-------------

local function serviceThreadTask(threadObj, svcId, svcCfg)
	_dbg.Info("CoServer", "Service '" .. svcId .. "' task is running...")
	while not threadObj.terminated do
		if not svcCfg.ACTION( svcCfg, "start" ) then --pass options to ACTION
			break
		end
		if threadObj.terminated or not svcCfg.KEEP_ALIVE then
			break
		else
			local interval = tonumber(svcCfg.KEEP_ALIVE)
			if interval and interval > 0 then
				coaf.Sleep(interval)
			end
		end
	end
	svcCfg.ACTION( svcCfg, "stop" )
	_M.running[svcId] = nil
	_dbg.Info("CoSerer", "Service '" .. svcId .. "' task is terminated.")
end

-------------

function _M.StartService(loc, service_id)
	local key = (loc or "ANY") .. "/" .. (service_id or "ANY")
	if _M.running[key] and _M.running[key] ~= "ONESHOT" then
		--_dbg.Warn("CoServer", "StartService - Service is already started:", key)
		return true
	end

	local svcCfg = _M.services[key]
	if not svcCfg then
		_dbg.Error("CoService", "StartService - Service not found:", key)
		return false
	end

	if _M.ServiceIsDisabled(loc, service_id) then
		_dbg.Info("CoServer", "StartService - But service is disabled:", key )
		return false
	end

	if not svcCfg.ACTION then
		_dbg.Error("CoServer", "StartService - Invalid service action specified. For:", key)
		return false
	end

	if svcCfg.ACTION.IS_ONESHOT then
		svcCfg.ACTION(svcCfg, "start")
		_M.running[key] = "ONESHOT"
	else
		local thr = thread.New(serviceThreadTask)
		_M.running[key] = thr
		thr:Start(key, svcCfg)
	end

	return true
end

function _M.StopService(loc, service_id)
	local key = (loc or "ANY") .. "/" .. (service_id or "ANY")
	local thr = _M.running[key]
	if thr then
		if type(thr) == "string" then
			local svcCfg = _M.services[key]
			if svcCfg and svcCfg.ACTION then
				svcCfg.ACTION( svcCfg, "stop" )
			end
			_M.running[key] = nil
		else
			thr:Terminate()
			_M.running[key] = nil
		end
	end
	return true
end

function _M.SignalService(loc, service_id, sig)
	local key = (loc or "ANY") .. "/" .. (service_id or "ANY")
	local thr = _M.running[key]
	if thr then
		local svcCfg = _M.services[key]
		if svcCfg and svcCfg.ACTION then
			return svcCfg.ACTION( svcCfg, "signal", sig )
		else
			return false
		end
	else
		return false
	end
end

function _M.SignalAll(sig)
	local key, thr
	for key, thr in pairs(_M.running) do
		if thr then
			local svcCfg = _M.services[key]
			if svcCfg and svcCfg.ACTION then
				svcCfg.ACTION( svcCfg, "signal", sig )
			end
		end
	end
	return true
end

function _M.StopAllServices()
	local key, thr
	for key, thr in pairs(_M.running) do
		if thr then
			if type(thr) == "string" then
				local svcCfg = _M.services[key]
				if svcCfg and svcCfg.ACTION then
					svcCfg.ACTION( svcCfg, "stop" )
				end
				_M.running[key] = nil
			else
				thr:Terminate()
				_M.running[key] = nil
			end
		end
	end
	_M.running = {}
end

-----------------

return _M
