local ProxyController = require("l2dbus.proxyctrl")
local l2dbus = require("l2dbus")
local ev = require("ev")

local _M = {
	blockingMode = true
}

--------------------------------------------------------
-- ClientSession object methods
--------------------------------------------------------

local session_methods = {}
session_methods.__index = session_methods

--------------------------------------------------------

--------------------------------------------------------

function session_methods:GetProxy(path, busname, ifname)
	if type(path) ~= "string" or type(busname) ~= "string" or type(ifname) ~= "string" then
		return nil, "Invalid arguments"
	end
	local idx = path.."@"..busname.."::"..ifname
	if self.proxyObjects[idx] then --Already created, reuse it.
		return self.proxyObjects[idx]
	end

	--Otherwise, try create new one.
	local proxyCtrl, msg = _M.LookupProxyController( busname, path )
	if not proxyCtrl then
		return nil, msg
	end
	local proxy, msg = _M.LookupInterfaceProxy( busname, path, ifname )
	if not proxy then
		return nil, msg
	end

	local proxy_obj = {
		interface = ifname,
		proxyCtrl = proxyCtrl,
		proxy = proxy,
		p = {
			proxy = proxy,
			allprops = {}
		},	--All properties

		signals = {}
	}
	
	local methods_meta = {}
	local properties_meta = {}

	methods_meta.__index = methods_meta

	function methods_meta:ConnectSignal( signame, callback )
		if self.signals[signame] then --exists
			self.proxyCtrl:disconnectSignal( self.signals[signame] )
			self.signals[signame] = nil
		end
		
		local h = self.proxyCtrl:connectSignal( self.interface, signame, callback )
		--print("***", self.interface, signame, h )
		self.signals[signame] = h
	end

	function methods_meta:DisconnectSignal( signame )
		if self.signals[signame] then --exists
			self.proxyCtrl:disconnectSignal( self.signals[signame] )
			self.signals[signame] = nil
		end
	end

	function methods_meta:__Free()
		local k,v
		for k,v in pairs(self.signals) do
			self.proxyCtrl:disconnectSignal(v)
		end

		self.signals = {}
	end

	function properties_meta:__index(key)
		if self.allprops[key] ~= nil and self.allprops[key]:find('g') then --Exists, and get property
			local r, got = _M.ActualCall( self.proxy, "g", key )
			if not r then
				return nil
			else
				return got
			end
		else
			return nil
		end
	end

	function properties_meta:__newindex(key,val)
		if self.allprops[key] ~= nil and self.allprops[key]:find('s') then --Exists, and set property
			_M.ActualCall( self.proxy, "s", key, val )
		end
	end

	local function proxy_build_method( name )
		return function(_self, ...)
			return _M.ActualCall(_self.proxy, "m", name,...)
		end
	end

	local k,v
	for k,v in pairs(proxy.m) do
		--print( "***Method:", k, type(v) )
		if type(v) == "function" then --Is a method
			methods_meta[k] = proxy_build_method(k)
		end
	end

	for k,v in pairs(proxy.p.get) do
		--print( "***Property Get:", k, type(v) )
		if type(v) == "function" then --Is a get property
			proxy_obj.p.allprops[k] = "g"
		end
	end

	for k,v in pairs(proxy.p.set) do
		--print( "***Property Set:", k, type(v) )
		if type(v) == "function" then --Is a set property
			proxy_obj.p.allprops[k] = proxy_obj.p.allprops[k] and "gs" or "s"
		end
	end

	setmetatable( proxy_obj.p, properties_meta )
	setmetatable( proxy_obj, methods_meta )

	self.proxyObjects[idx] = proxy_obj
	return proxy_obj
end

function session_methods:GetMultiProxies(path, busname, ifaces)
	if type(path) ~= "string" or type(busname) ~= "string" then
		return nil, "Invalid arguments"
	end

	if type(ifaces) == "string" then --Only one interface, same as the GetProxy
		return self:GetProxy(path,busname,ifaces)
	elseif type(ifaces) == "table" then --A map of interfaces
		local obj = {}
		local k,v
		for k,v in pairs(ifaces) do
			if type(v) == "string" then
				local pobj = self:GetProxy(path,busname,v)
				if pobj then
					obj[k] = pobj
				end
			end
		end
		return obj
	else
		return nil, "Invalid interfaces specified"
	end
end

--[[
function session_methods:GetDefaultMasterProxy()
	return self:GetMultiProxies( "/com/kiloview/MOONLIGHT/System","com.kiloview.MOONLIGHT.System",
									{
										base = "com.kiloview.MOONLIGHT.System.iModule",
										master = "com.kiloview.MOONLIGHT.System.iSystemMaster"
									} )
end

function session_methods:GetLogProxy(path,busname)
	local obj = self:GetMultiProxies( "/com/kiloview/MOONLIGHT/System/Log","com.kiloview.MOONLIGHT.System",
								 		{
											base = "com.kiloview.MOONLIGHT.System.iModule",
											log = "com.kiloview.MOONLIGHT.System.iModuleLog"
								 		} )

end
--]]

function session_methods:Free()
	local k,v 
	for k,v in pairs(self.proxyObjects) do
		v:__Free()	--Free all created proxies.
	end

	self.proxyObjects = nil	--Drop all.
end

--------------------------------------------------------
-- Global methods
--------------------------------------------------------
_M.CreateClientSession = function()
	--Return a client session
	return setmetatable({ proxyObjects = {} }, session_methods)
end

_M.LookupProxyController = function( busname, path )
	if _M.gConnection and type(busname) == "string" and type(path) == "string" then
		local ctrlr = ProxyController.new( _M.gConnection, busname, path )
		if ctrlr then
			if ctrlr:bind() then
				ctrlr:setBlockingMode(_M.blockingMode)
				return ctrlr
			else
				return nil, "Bind to " .. path .. " @ " .. busname .. " fail!"
			end
		else
			return nil, "Fail to open proxy controller of " .. path .. " @ " .. busname
		end
	else
		return nil, "moon1Clients not initialized, or invalid bus name/object path passed."
	end
end

--local pretty = require("pl.pretty")

_M.LookupInterfaceProxy = function( busname, path, ifname )
	if type(ifname) ~= "string" then
		return nil, "Invalid interface name"
	end

	local ctrlr, msg = _M.LookupProxyController( busname, path )
	if not ctrlr then
		return nil, msg
	else
		--TEST TEST
		--local data = ctrlr:getIntrospectionData()
		--pretty.dump(data)
		local proxy = ctrlr:getProxy( ifname )
		if proxy then
			return proxy
		else
			return nil, "Could not get proxy of interface '" .. ifname .. "'"
		end
	end
end

_M.Init = function(bustype, main_loop, disp)
	l2dbus.Trace.setFlags(l2dbus.Trace.ALL)
	_M.gMainLoop = main_loop or ev.Loop.default
	_M.gMainLoop:now()
	_M.gDispatcher = disp or l2dbus.Dispatcher.new(require("l2dbus_ev").MainLoop.new(_M.gMainLoop))
	assert( nil ~= _M.gDispatcher )

	local bus = l2dbus.Dbus.BUS_SYSTEM
	if nil ~= bustype then
		if bustype == "SESSION" then
			bus = l2dbus.Dbus.BUS_SESSION
		end	--TODO: Other types
	end

	_M.gConnection = l2dbus.Connection.openStandard(_M.gDispatcher, bus)
	assert( nil ~= _M.gConnection )

	_M.gProxyControllers = {}
	_M.gProxies = {}
end

_M.Cleanup = function()
	local k,v
	_M.gProxies = nil	--Clear all items.
	for k,v in pairs(_M.gProxyControllers) do
		v:disconnectAllSignals()
		v:unbind()
	end
	_M.gProxyControllers = nil --Clear all items.
    _M.gDispatcher = nil
	_M.gMainLoop = nil
end

_M.RegisterIdleCall = function( callback )
	local idle = ev.Idle.new( callback )
	idle:start(_M.gMainLoop)
end

_M.RegisterTimerCall = function( callback, after, period )
	local timer = ev.Timer.new( callback, after, period )
	timer:start(_M.gMainLoop)
end

_M.Run = function()
	if _M.gMainLoop then
		_M.gMainLoop:loop()
	end
--[[ Same function as above.
	if _M.gDispatcher then
		_M.gDispatcher:run(l2dbus.Dispatcher.DISPATCH_WAIT)
	end
--]]
end

_M.Stop = function()
	if _M.gMainLoop then
		_M.gMainLoop:unloop()
	end
end

_M.Sleep = function(ms)
  local periphery = require( "periphery" )
  periphery.sleep_ms(ms)
end

_M.SetModuleName = function(name)
  default_module_name = name
end

_M.SetBlockingMode = function(be)
  _M.blockingMode = be
  local k,v
  for k,v in pairs(_M.gProxyControllers) do
    if v then
      v:setBlockingMode(be)
    end
  end
end

_M.ActualCall = function( proxy, callt, method, ... )
	local invokeFunc
	if callt == "m" or callt == "M" then
		invokeFunc = proxy.m[method]
	elseif callt == "g" or callt == "G" then --Get property
		invokeFunc = proxy.p.get[method]
	elseif callt == "s" or callt == "S" then --Set property
		invokeFunc = proxy.p.set[method]
	end

	if type(invokeFunc) == "function" then
		local results = { pcall( invokeFunc, ... ) }

		if not results[1] then --pcall fail
			return results[1], results[2]	--results[2] is error message
		elseif not results[2] then --DBus call fail
			return false, results[4]	--results[4] is call error message
		else --got reply or results
			if _M.blockingMode then --Blocking mode, directly return the results
				table.remove( results, 1 )	--Remove the pcall() status at first
				return table.unpack(results)
			else --wait reply
				if "userdata" == type(results[3]) then
					--TODO: Here is a bug, but how can I get the proxy controller?
					--      Currently, this is never used.
					local reply, errName, errMsg = proxyCtrl:waitForReply(results[3])
					if not reply then
						return false, tostring(errName) .. " : " .. tostring(errMsg)
					else
						local status, results = pcall(reply.getArgsAsArray, reply)
						if status then
							return status, table.unpack(results)
						else
							return status, results
						end
					end
				else
					return true, results[3]
				end
			end
		end
	else
		return false, "The interface does not provided '" .. method .. "' method or property"
	end
end

_M.CallEntry = function( objpath, busname, iface, method, ... )
	if type(method) ~= "string" then
		return false, "Invalid method name"
	end
	local proxyCtrl, msg = _M.LookupProxyController( busname, objpath )
	if not proxyCtrl then
		return false, msg
	end

	local proxy, msg = _M.LookupInterfaceProxy( busname, objpath, iface )
	if not proxy then
		return false, msg
	end

	return _M.ActualCall( proxy, "m", method, ... )
end

_M.PropertyEntry = function( objpath, busname, iface, act, propName, ... )
	if type(propName) ~= "string" then
		return false, "Invalid property name"
	end
	local proxyCtrl, msg = _M.LookupProxyController( busname, objpath )
	if not proxyCtrl then
		return false, msg
	end

	local proxy, msg = _M.LookupInterfaceProxy( busname, objpath, iface )
	if not proxy then
		return false, msg
	end

	return _M.ActualCall( proxy, act, propName, ... )
end

return _M


