local ProxyController = require("l2dbus.proxyctrl")
local l2dbus = require("l2dbus")
local svc = require("l2dbus.service")
local ev = require("ev")

local SERVICE_ERROR_NAME = "com.kiloview.Error"

local _M = {
	blockingMode = true,
	busType = l2dbus.Dbus.BUS_SYSTEM
}


--------------------------------------------------------
-- Service object methods
--------------------------------------------------------

local svc_methods = {}
svc_methods.__index = svc_methods

--------------------------------------------------------

function svc_methods:AddInterfaceWithXmlMeta(ifacename, xmlMeta)
	local meta = svc.convertXmlToIntfMeta( ifacename, xmlMeta )
	if not meta then
		return false, "Fail to parse the XML meta of interface " .. ifacename
	else
		return self:AddInterfaceWithMeta(ifacename,meta)
	end
end

function svc_methods:AddInterfaceWithMeta(ifacename, meta)
	if self.allInterfaces[ifacename] then	--Already registered
		return true	
	end

	local result = self.service:addInterface(ifacename, meta)
	if result then
		if not self.withProperties then
			self:__buildPropertiesAccess()
			self.withProperties = true
		end

		self.allInterfaces[ifacename] = meta	--Save it's meta
	end
	return result
end

function svc_methods:RemoveInterface(ifacename)
	if self.allInterfaces[ifacename] then
		self.service:removeInterface( ifacename )
		self.allInterfaces[ifacename] = nil
		return true
	else
		return false, "No registered interface " .. ifacename
	end
end

function svc_methods:__buildPropertiesAccess()
	local serviceObj = self

	local function propertiesGet( ctx, intfName, propName )
		local pObj = serviceObj.allProperties[intfName]
		if type(pObj) ~= "table" then
			ctx:error( SERVICE_ERROR_NAME, "Unknown interface '"..intfName.."' or its property '" ..propName .. "'" )
		else
			local pType = type(pObj[propName])
			if pType == "number" or pType == "string" or pType == "boolean" or pType == "table" then
				ctx:reply(pObj[propName])
			else
				if type( pObj["get_"..propName] ) == "function" then
					ctx:reply( pObj["get_"..propName](pObj) )
				else
					ctx:error( SERVICE_ERROR_NAME, "Unknown property '" .. propName .. "' getting from '"..intfName.."'" )
				end
			end
		end
	end

	local function propertiesSet( ctx, intfName, propName, value )
		local pObj = serviceObj.allProperties[intfName]
		if type(pObj) ~= "table" then
			ctx:error( SERVICE_ERROR_NAME, "Unknown interface '"..intfName.."' or its property '" ..propName .. "'" )
		else
			local pType = type(pObj[propName])
			if pType == "number" or pType == "string" or pType == "boolean" or pType == "table" then
				pObj[propName] = value
				ctx:reply()
			else
				if type( pObj["set_"..propName] ) == "function" then
					pObj["set_"..propName](pObj,value)
					ctx:reply()
				else
					ctx:error( SERVICE_ERROR_NAME, "Unknown property '" .. propName .. "' setting to '"..intfName.."'" )
				end
			end
		end
	end

	local function propertiesGetAll( ctx, intfName )
		local meta = serviceObj.allInterfaces[intfName]
		if type(meta) ~= "table" or type(meta.properties) ~= "table" then
			ctx:error(SERVICE_ERROR_NAME, "Unknown interface '"..intfName.."' or its property '" ..propName .. "'")
		else
			local pObj = serviceObj.allProperties[intfName]
			if type(pObj) ~= "table" then
				ctx:error( SERVICE_ERROR_NAME, "Unknown interface '"..intfName.."' or its property '" ..propName .. "'" )
				return
			end

			local results = {}

			local k,v 
			for k,v in pairs(meta.properties) do
				
				local pType = type(pObj[v.name])
				if pType == "number" or pType == "string" or pType == "boolean" or pType == "table" then
					results[v.name] = pObj[v.name]
				else
					if type( pObj["get_"..propName] ) == "function" then
						results[v.name] = pObj["get_"..propName](pObj)
					end
				end

			end

			ctx:reply(results)
		end
	end

    self.service:registerMethodHandler( l2dbus.Dbus.INTERFACE_PROPERTIES,
    									"Set",
    									propertiesSet )

    self.service:registerMethodHandler( l2dbus.Dbus.INTERFACE_PROPERTIES,
    									"Get",
    									propertiesGet )

    self.service:registerMethodHandler( l2dbus.Dbus.INTERFACE_PROPERTIES,
    									"GetAll",
    									propertiesGetAll )
	
end

function svc_methods:BindInterfaceObject(ifacename, obj)

	local function buildObjDbusMethod( bindObj, funcName )
		return function( ctx, ... )
			local results = { pcall(bindObj[funcName], bindObj, ...) }
			if results[1] then
				table.remove( results, 1 )
				ctx:reply( table.unpack(results) )
			else
				ctx:error( SERVICE_ERROR_NAME, results[2] or "Call method error" )
			end
		end
	end	

	if self.allInterfaces[ifacename] then
		local k,v

		if type(self.allInterfaces[ifacename].methods) == "table" then
			for k,v in pairs(self.allInterfaces[ifacename].methods) do
				if type(v) == "table" and v.name then
					--if type(obj[v.name]) == "function" then	--Bind this function
						self.service:registerMethodHandler( ifacename, v.name, buildObjDbusMethod( obj, v.name ) )
					--end
				end
			end
		end

		if type(self.allInterfaces[ifacename].properties) == "table" then
			self.allProperties[ifacename] = obj
		end
		
		return true
	else
		return false
	end
end

function svc_methods:EmitSignal(ifacename, sigName, ...)
	if self.allInterfaces[ifacename] then
		self.service:emit( self.serviceConn.conn, ifacename, sigName, ... )
	end
end

function svc_methods:Online()
	if not self.attached then
		self.service:attach( self.serviceConn.conn )
		self.attached = true
	end
end

function svc_methods:Offline()
	if self.attached then
		self.service:detach( self.serviceConn.conn )
		self.attached = false
	end
end

function svc_methods:Free()
	self:Offline()
	self.allProperties = {}

	local k, v
	for k,v in pairs(self.allInterfaces) do
		self.service:removeInterface(k)
	end

	self.allInterfaces = {}
end

--------------------------------------------------------
-- ServiceConnection object methods
--------------------------------------------------------

local svcconn_methods = {}
svcconn_methods.__index = svcconn_methods

--------------------------------------------------------

local defaultServiceHandler = function( ctx, intfName, member, args )
	ctx:error( SERVICE_ERROR_NAME, "unhandled request for " ..
			   ((intfName == nil) and "" or (intfName .. ".")) .. member)
end

function svcconn_methods:CreateService(svcName)
	if self.services[svcName] then --Exists, direct return
		return self.services[svcName]
	end

	local service = svc.new(svcName, true, defaultServiceHandler)
	if service then
		local svc = setmetatable( {
			serviceConn = self,
			service = service,
			withProperties = false,
			allInterfaces = {},
			allProperties = {}
		}, svc_methods )
		self.services[svcName] = svc
		return svc
	else
		return nil, "Could not create dbus service of " .. svcName 
	end
end

function svcconn_methods:ListServices()
	return self.services
end

function svcconn_methods:RemoveService(svcName)
	if self.services[svcName] then
		self.services[svcName]:Free()
		self.services[svcName] = nil
	end
end

function svcconn_methods:Free()
	local k,v
	for k,v in pairs(self.services) do
		v:Free()
	end

    self.dbusProxyCtrl:setBlockingMode(true)
    self.dbusProxyCtrl:disconnectAllSignals()

    local status,result = self.dbusProxy.m.ReleaseName(self.name)
    if l2dbus.Dbus.RELEASE_NAME_REPLY_RELEASED ~= result then
    	io.stderr:write("Failed to release bus name " .. self.name .. "!\n")
    end
	self.dbusProxy = nil
	self.dbusProxyCtrl = nil
	self.conn:flush()
	self.conn = nil
	self.services = {}
end

--------------------------------------------------------
-- Global methods
--------------------------------------------------------
_M.CreateServiceConnection = function(busname)
	if _M.gServiceConnections[busname] then
		return _M.gServiceConnections[busname]
	end

	--Create a connection first.
	local conn = l2dbus.Connection.openStandard(_M.gDispatcher, _M.busType)
	if not conn then
		return nil, "Could not create dbus connection"
	end

    local dbusProxyCtrl = ProxyController.new(conn, l2dbus.Dbus.SERVICE_DBUS, l2dbus.Dbus.PATH_DBUS)
    dbusProxyCtrl:bind(true)
    dbusProxyCtrl:setBlockingMode(true)
	local dbusProxy = dbusProxyCtrl:getProxy(l2dbus.Dbus.INTERFACE_DBUS)
	
	local status,result = dbusProxy.m.NameHasOwner(busname)
    if result then
		dbusProxy = nil
		dbusProxyCtrl = nil
		conn = nil
		return nil, "Bus name " .. busname .. " is already claimed!"
    end
	
	--Try register the connection with name
    status,result = dbusProxy.m.RequestName(busname, l2dbus.Dbus.NAME_FLAG_DO_NOT_QUEUE)
	--print( "****", status, result )
    if result ~= l2dbus.Dbus.REQUEST_NAME_REPLY_PRIMARY_OWNER then
		dbusProxy = nil
		dbusProxyCtrl = nil
		conn = nil
		return nil, "Require bus name of " .. busname .. " fail!"
	end

	--Return a service connection object
	local svcconn = setmetatable({
		name = busname,
		conn = conn,
		dbusProxyCtrl = dbusProxyCtrl,
		dbusProxy = dbusProxy,
		services = {}
	}, svcconn_methods)

	_M.gServiceConnections[busname] = svcconn
	return svcconn
end

_M.RemoveServiceConnection = function(busname)
	if _M.gServiceConnections[busname] then
		local obj = _M.gServiceConnections[busname]
		obj:Free()
		_M.gServiceConnections[busname] = nil
	end
end


_M.Init = function(bustype, main_loop, disp)
	--l2dbus.Trace.setFlags(l2dbus.Trace.ALL)
	_M.gMainLoop = main_loop or ev.Loop.default
	_M.gMainLoop:now()
	_M.gDispatcher = disp or l2dbus.Dispatcher.new(require("l2dbus_ev").MainLoop.new(_M.gMainLoop))
	assert( nil ~= _M.gDispatcher )

	--_M.gConnection = l2dbus.Connection.openStandard(_M.gDispatcher, l2dbus.Dbus.BUS_SYSTEM)
	--assert( nil ~= _M.gConnection )

	_M.busType = l2dbus.Dbus.BUS_SYSTEM
	if nil ~= bustype then
		if bustype == "SESSION" then
			_M.busType = l2dbus.Dbus.BUS_SESSION
		end	--TODO: Other types
	end

	_M.gServiceConnections = {}
end

_M.Cleanup = function()
	local k,v
	for k,v in pairs(_M.gServiceConnections) do
		v:Free()
	end
	_M.gServiceConnections = {}
    _M.gDispatcher = nil
	_M.gMainLoop = nil
end

_M.RegisterIdleCall = function( callback )
	local idle = ev.Idle.new( callback )
	idle:start(_M.gMainLoop)
end

_M.RegisterTimerCall = function( callback, after, period )
	local timer = ev.Timer.new( callback, after, period )
	timer:start(_M.gMainLoop)
end

_M.Run = function()
	if _M.gMainLoop then
		_M.gMainLoop:loop()
	end
--[[ Same function as above.
	if _M.gDispatcher then
		_M.gDispatcher:run(l2dbus.Dispatcher.DISPATCH_WAIT)
	end
--]]
end

_M.Stop = function()
	if _M.gMainLoop then
		_M.gMainLoop:unloop()
	end
end

_M.Sleep = function(ms)
  local periphery = require( "periphery" )
  periphery.sleep_ms(ms)
end

return _M


