local _M = {}

local lfs = require "lfs"
local presetStoDir = "/data/presets"
local sysPresetStoDir = "/etc/sys-preset"
local tmd = require "kiloview.systemd-timedate"

local function file_exist(path)
	local f = io.open(path, "r")
	if f then
		f:close()
		return true
	else
		return false
	end
end


--------------------
-- Preset storage:
-- /data/presets
--      |__ <id>
--      |    |__cfg.info - Preset information.
--      |...

local function OpenPresetDB( spec_id, savePos )
	savePos = savePos or presetStoDir
	local attrs = lfs.attributes( savePos )
	if not attrs or attrs.mode ~= "directory" then
		return nil, nil
	end

	local presetTable = {}
	local counter = 0

	local item
	for item in lfs.dir( savePos ) do
		if item ~= "." and item ~= ".." and (not spec_id or spec_id == item) then
			attrs = lfs.attributes( savePos .. "/" .. item )
			if attrs and attrs.mode == "directory" then
				local cfgFile = savePos .. "/" .. item .. "/cfg.info"
				local x, ret = pcall( dofile, cfgFile )
				if x and type(ret) == "table" then
					counter=counter + 1
					ret.id = item
					table.insert( presetTable, ret )
				end
			end
		end
	end
	
	table.sort(presetTable, function(a,b)
		if not a or not b or a==b then
			return false
		else
			return a.name < b.name
		end
	end)

	return counter, presetTable
end

--------------------

function _M.GetPresetList()
	local counter, presetTable = OpenPresetDB()
	if not counter or not presetTable then
		return nil
	else
		return presetTable
	end
end

--------------------
function _M.GetPresetDetail( id, savePos )
	savePos = savePos or presetStoDir
	local counter, presetTable = OpenPresetDB(id, savePos)
	if not counter or not presetTable then
		return nil
	else
		local result = presetTable[1]
		
		if type(result) == "table" then
			
			result.configs = {}
			local attrs = lfs.attributes( savePos .. "/" .. (result.id or "") .. "/configs" )
			if attrs and attrs.mode == "directory" then
				for item in lfs.dir( savePos .. "/" .. result.id .. "/configs" ) do
					if item ~= "." and item ~= ".." then
						result.configs[ item ].file = savePos .. "/" .. result.id .. "/configs/" .. item
					end
				end
			end
			
			return result
		else
			return nil
		end	
	end
end

--------------------
local function GetTelnetStatus()
	local x = io.popen( "/usr/bin/systemctl status telnet.socket", "r" )
	if not x then
		return nil
	else
		local status = "false"
		local ln = x:read("*l")
		while ln do
			local flag = ln:match( "^%s*Active:%s*(%w*)%s*.*" )
			if flag then
				if flag == "active" then
					status = "true"
				else
					status = "false"
				end
				break
			end
			ln = x:read("*l")
		end
		x:close()
		return status
	end
end

function _M.AddPreset( name, args, saveTo, saveSysSettings, saveNetSettings)
	local id
	local f = io.popen( "/usr/bin/uuidgen", "r" )
	if f then
		id = f:read("l")
		f:close()
	end

	if not id then
		print( "ERROR: Fail to generate preset unique id." )
		return nil, "uuidgen"
	end

	saveTo = saveTo or presetStoDir

	--If empty name, using UUID instead.
	if name == "" then
		name = nil
	end

	--Avoid name conflict.
	if name then
		local counter, items = OpenPresetDB(nil, saveTo)
		local instead_name = name
		local instead_cc = 1
		if counter and #items > 0 then
			for _i, item in ipairs(items) do
				if item.name == instead_name then
					instead_name = name .. "_" .. instead_cc
					instead_cc = instead_cc + 1
				end
			end
			name = instead_name
		end
	end

	os.execute("/bin/rm -Rf " .. saveTo .. "/" .. id )
	os.execute( "/bin/mkdir -p " .. saveTo .. "/" .. id )
	local cfgFile = io.open( saveTo .. "/" .. id .. "/cfg.info", "w" )
	if not cfgFile then
		print( "ERROR: Fail to create the preset configure file!\n" )
		lfs.rmdir( saveTo .. "/" .. id )
		return nil, "create"
	end
	
	os.execute( "/bin/mkdir -p " .. saveTo .. "/" .. id .. "/configs" )

	local desc
	local saveTelnetSettings
	local saveOverlay = false
	local patch_cmds = {}
	local touch_files = {}

	if type(args) == "table" then
		if args.hostname then
			os.execute( "/bin/mkdir -p " .. saveTo .. "/" .. id .. "/etc" )
			os.execute("/bin/cp -Rf /etc/hostname " .. saveTo .. "/" .. id .. "/etc/")
		end

		if args.time then
			saveSysSettings = true
		end

		if args.maintenance then
			os.execute("/bin/cp -Rf /data/configs/timer.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/timer.conf")
		end

		if args.adjust then
			os.execute("/bin/cp -Rf /data/configs/avio.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/avio.conf")
		end

		if args.audio then
			os.execute("/bin/cp -Rf /data/configs/audioEngine.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/audioEngine.conf")
		end

		if args.code then
			os.execute("/bin/cp -Rf /data/configs/encoder.conf " .. saveTo .. "/" .. id .. "/configs/")
			os.execute("/bin/cp -Rf /data/configs/streamer.conf " .. saveTo .. "/" .. id .. "/configs/")
			os.execute("/bin/cp -Rf /data/configs/media_users.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/encoder.conf")
			table.insert( touch_files, "/data/configs/streamer.conf")
			table.insert( touch_files, "/data/configs/media_users.conf")
		end

		if args.overlay then
			os.execute("/bin/cp -Rf /data/configs/osd.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/osd.conf")
			saveOverlay = true
			--Only save images to tempoarary storage.
			if saveTo:match("^/tmp/") then
				local attrs = lfs.attributes( "/data/osd_images" )
				if attrs and attrs.mode == "directory" then
					os.execute("/bin/cp -Rf /data/osd_images " .. saveTo .. "/" .. id .. "/" )
				end
			end
		end

		if args.web then
			os.execute("/bin/cp /data/configs/web.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/web.conf")
		end

		if args.onvif then
			os.execute("/bin/cp /data/configs/onvif.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/onvif.conf")
		end

		if args.easyman then
			os.execute("/bin/cp /data/configs/hostapm.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/hostapm.conf")
		end

		if args.network then
			saveNetSettings = true
			saveTelnetSettings = true
		end

		if args.ptz then
			os.execute("/bin/cp -Rf /data/configs/ptz.conf " .. saveTo .. "/" .. id .. "/configs/")
			table.insert( touch_files, "/data/configs/ptz.conf")
		end

	elseif type(args) == "string" then
		desc = args
		os.execute( "/bin/cp -Rf /data/configs/*.conf " .. saveTo .. "/" .. id .. "/configs/" )
		table.insert( touch_files, "/data/configs/*.conf")
		--os.execute( "/bin/rm -f " .. saveTo .. "/" .. id .. "/configs/web.conf" ) --Don't save web configs.
		saveSysSettings = true
		--saveNetSettings = ?
	end
	
	if saveSysSettings then
		os.execute( "/bin/mkdir -p " .. saveTo .. "/" .. id .. "/etc/systemd" )
		os.execute("/bin/cp -Rf /etc/localtime " .. saveTo .. "/" .. id .. "/etc/")
		os.execute("/bin/cp -Rf /etc/systemd/timesyncd.conf " .. saveTo .. "/" .. id .. "/etc/systemd/" )
		if file_exist("/etc/systemd/system/multi-user.target.wants/systemd-timesyncd.service") then
			os.execute(
				"/bin/mkdir -p " .. saveTo .. "/" .. id .. "/etc/systemd/system/multi-user.target.wants;" ..
				"cp -Rf /etc/systemd/system/multi-user.target.wants/systemd-timesyncd.service " ..
				saveTo .. "/" .. id .. "/etc/systemd/system/multi-user.target.wants/"
			)
		else
			table.insert(patch_cmds, "/bin/rm -Rf /etc/systemd/system/multi-user.target.wants/systemd-timesyncd.service")
		end
	end

	if saveNetSettings then
		os.execute( "/bin/mkdir -p " .. saveTo .. "/" .. id .. "/etc/NetworkManager/system-connections" )
		os.execute( "/bin/cp -Rf /etc/NetworkManager/system-connections/* " .. saveTo .. "/" .. id .. "/etc/NetworkManager/system-connections/" )
		os.execute( "/bin/cp -Rf /data/configs/ethernet.conf " .. saveTo .. "/" .. id .. "/configs/" )
		table.insert( touch_files, "/data/configs/ethernet.conf")
		table.insert(patch_cmds, "/bin/chmod 600 /etc/NetworkManager/system-connections/*")
	end

	if saveTelnetSettings then
		if file_exist("/usr/lib/systemd/system/sockets.target.wants/telnet.socket") then
			table.insert(patch_cmds, "/bin/ln -sf /usr/lib/systemd/system/telnet.socket /usr/lib/systemd/system/sockets.target.wants/telnet.socket")
		else
			table.insert(patch_cmds, "/bin/rm -Rf /usr/lib/systemd/system/sockets.target.wants/telnet.socket")
		end
	end

	cfgFile:write( [==[
return {
]==] )
	
	cfgFile:write( 'name = [==[' .. (name or id) .. ']==],\n' )
	cfgFile:write( 'desc = [==[' .. (desc or "") .. ']==],\n' )
	cfgFile:write( 'withOverlay = ' .. (saveOverlay and "true" or "false") .. ',\n' )
	cfgFile:write( [==[
}
]==] )
	cfgFile:close()

	cfgFile = io.open(saveTo .. "/" .. id .. "/prepare.cmd", "w")
	if cfgFile then
		cfgFile:write("#!/bin/sh\n")
		for _x, file in ipairs(touch_files) do
			cfgFile:write("/bin/rm -Rf " .. file .. "\n")
		end
		cfgFile:close()
	end

	cfgFile = io.open(saveTo .. "/" .. id .. "/patch.cmd", "w")
	if cfgFile then
		cfgFile:write("#!/bin/sh\n")
		for _x, cmd in ipairs(patch_cmds) do
			cfgFile:write(cmd)
			cfgFile:write("\n")
		end
		cfgFile:close()
	end
	
	return id
end

--------------------------------------
function _M.PresetAsRestoreDefault( id )
	return true
	--[[ Not used.
	local attrs = lfs.attributes( presetStoDir .. "/" .. id )
	if attrs and attrs.mode == "directory" then
		lfs.mkdir( sysPresetStoDir )
		os.execute( "/bin/ln -sf " .. presetStoDir .. "/" .. id .. " " .. sysPresetStoDir .. "/user" )
		return true
	else
		return false, "Invalid preset"
	end
	--]]
end
--------------------------------------

function _M.DeletePreset( id, savePos )
	savePos = savePos or presetStoDir
	local attrs = lfs.attributes( savePos .. "/" .. id )
	if attrs and attrs.mode == "directory" then
		os.execute( "/bin/rm -Rf " .. savePos .. "/" .. id )
		return true
	else
		return false, "Invalid preset"
	end
end

function _M.ExportPreset( id, savePos )
	savePos = savePos or presetStoDir
	local counter, presetTable = OpenPresetDB(id, savePos)
	if not counter or not presetTable or #presetTable == 0 then
		return false, "Preset is not existed"
	end

	local presetName, presetItem
	presetItem = presetTable[1]
	presetName = presetItem.name or presetItem.id

	presetName = presetName:gsub("%s+", "")
	presetName = presetName:gsub("%p+", "")

	os.execute("/bin/rm -Rf /tmp/preset/exports/" .. id)
	os.execute("/bin/mkdir -p /tmp/preset/exports/" .. id)
	--Yes sure, '/tmp/preset/exports/<id>/<id>'
	os.execute("/bin/cp -Rf " .. savePos .."/" .. id .. " /tmp/preset/exports/" .. id .. "/" )

	if presetItem.withOverlay and not savePos:match("^/tmp/") then
		local attrs = lfs.attributes( "/data/osd_images" )
		if attrs and attrs.mode == "directory" then
			os.execute("/bin/cp -Rf /data/osd_images " .. "/tmp/preset/exports/" .. id .. "/" .. id .. "/" )
		end
	end

	--Zip with:
	--  <id>/...any-files...
	os.execute("/bin/rm -Rf /tmp/preset/" .. presetName .. ".zip")
	local _ = os.execute(  "cd /tmp/preset/exports/" .. id .. "; /usr/bin/zip -b . -r -y /tmp/preset/" .. presetName .. ".zip " .. id .. 
						   "; /bin/chmod 644 /tmp/preset/" .. presetName .. ".zip" )
	os.execute("/bin/rm -Rf /tmp/preset/exports/" .. id)
	return true, "/preset/" .. presetName .. ".zip"
end

-----------------------------------------
-- Import 'filepath' zip file
-- If 'tmp_store' is true, only save into /tmp/...

function _M.ImportPreset( filepath, tmp_store )
	local f = io.open(filepath, "r")
	if not f then
		return  false, "Invalid preset file"
	end

	local tmppath = "/tmp/preset/imports"
	local is_ok = os.execute("/bin/mkdir -p " .. tmppath)

	if not is_ok then
		os.execute("/bin/rm " .. filepath)
		return false, "Import preset failed"
	end

	os.execute("/bin/rm -Rf " .. tmppath .. "/*" )
	is_ok = os.execute("/usr/bin/unzip '" .. filepath .. "' -d " .. tmppath)
	--Now, it will generate "/tmp/preset/imports/<id>/...any-files.."

	os.execute("/bin/rm " .. filepath)

	if not is_ok then
		os.execute("/bin/rm -Rf " .. tmppath .. "/*")
		return false, "Extract preset file failed"
	end

	local id
	for item in lfs.dir( tmppath ) do
		if item ~= "." and item ~= ".." then
			local attrs = lfs.attributes(tmppath .. "/" .. item)
			if attrs and attrs.mode == "directory" then
				id = item
				break
			end
		end
	end

	if not id then
		os.execute("/bin/rm -Rf " .. tmppath .. "/*")
		return false, "Bad preset format"
	end

	if tmp_store then
		return true, id, tmppath
	end

	--Copy back to preset storage.
	os.execute("/bin/rm -Rf " .. presetStoDir .. "/" .. id)
	os.execute("/bin/cp -Rf " .. tmp_store .. "/" .. id .. " " .. presetStoDir)
	os.execute("/bin/rm -Rf " .. tmppath .. "/*")

	-- 3rd returns is nil, means use the system default preset location
	return true, id, nil
end

-----------------------------------------

function _M.LoadPreset( id, savePos )
	savePos = savePos or presetStoDir
	local attrs = lfs.attributes(savePos .. "/" .. id)
	if not attrs or attrs.mode ~= "directory" then
		return false, "Preset is not existed"
	end

	local cfgFile = savePos .. "/" .. id .. "/cfg.info"
	local x, ret = pcall( dofile, cfgFile )

	if not x or type(ret) ~= "table" then
		return false, "Bad preset format"
	else
		ret.id = id
	end

	if file_exist(savePos .. "/" .. id .. "/prepare.cmd") then
		os.execute( "source " .. savePos .. "/" .. id .. "/prepare.cmd")
	end

	os.execute("/bin/cp -Rf " .. savePos .. "/" .. id .. "/configs/* /data/configs")
	os.execute("/bin/cp -Rf " .. savePos .. "/" .. id .. "/etc/* /etc/")
	
	local attrs = lfs.attributes( savePos .. "/" .. id .. "/osd_images" )
	if attrs and attrs.mode == "directory" then
		os.execute( "/bin/mkdir -p /data/osd_images; /bin/rm -Rf /data/osd_images/*")
		os.execute( "/bin/cp -Rf " .. savePos .. "/" .. id .. "/osd_images/* /data/osd_images/")
	end

	if file_exist(savePos .. "/" .. id .. "/patch.cmd") then
		os.execute( "source " .. savePos .. "/" .. id .. "/patch.cmd")
	end

	return true
end


return _M
