local periphery = require "periphery"
local SERIAL = periphery.Serial

local _PTZ = {}

--============================================================--
function _PTZ.CreateNew(dev,baud)
--============================================================--

local _M = {
	protocol = nil,
	serial_port = nil,
	address = 0,
	speed_options = {
		pan = 0.5,
		tilt = 0.5,
		zoom = 0.5,
		focus = 0.5,
		iris = 0.5
	}
}

-----

local status, err
if type(dev) == "table" then
	status, err = pcall(function () _M.serial_port = SERIAL(dev) end)
else
	status, err = pcall(function () _M.serial_port = SERIAL(dev,baud) end)
end
if not status then
	--print( "ERROR: Fail to open serial port!" )
	return nil
end

----

function _M:Close()
	if nil ~= self.serial_port then
		self.serial_port:close()
		self.serial_port = nil
	end
	collectgarbage()
end

----
function _M:GetSerialObject()
	return self.serial_port
end
----

function _M:GetSerialObject()                                                 
	return self.serial_port                                               
end                                                                           

----


function _M:LoadProtocol( protocolName )
	if protocolName == "null" then
		self.protocol = nil
	else
		local status
		status = pcall( function() self.protocol = require("protocols." .. protocolName) end )
		if not status then
			print( "ERROR: Fail to load PTZ protocol '" .. protocolName .. "'." )
			self.protocol = nil
		end
	end

	collectgarbage()
end

----

function _M:SetAddress( addr )
	self.address = addr
end

--Options(opts):
--  pan   = ...,
--  tilt  = ...,
--  zoom  = ...,
--  focus = ...,
--  iris  = ...
--  ...     (etc)
function _M:SetSpeed( opts )
	if type(opts) == "table" then
		self.speed_options.pan = opts.pan or self.speed_options.pan
		self.speed_options.tilt = opts.tilt or self.speed_options.tilt
		self.speed_options.zoom = opts.zoom or self.speed_options.zoom
		self.speed_options.focus = opts.focus or self.speed_options.focus
		self.speed_options.iris = opts.iris or self.speed_options.iris
	end
end

----
function _M:DoAction( action, ... )
	if self.serial_port and self.protocol then
		local produce = self.protocol.supports[action]
		if produce then
			local data_arr = produce( self.address, ... )
			if type(data_arr) == "table" then
				local i
				for i=1, #data_arr do
					if type(data_arr[i]) == "table" then
						self.serial_port:write( string.char( table.unpack(data_arr[i]) ) )
						self.serial_port:flush()
					elseif type(data_arr[i]) == "string" then
						self.serial_port:write( data_arr[i] )
						self.serial_port:flush()
					elseif type(data_arr[i]) == "number" then
						periphery.sleep_ms( data_arr[i] )
					end

					--if self.protocol.AckFilter then
						--TODO: Process reading.
					--end
				end
				return true
			else
				return false, "protocol-error"
			end
		else
			return false, "not-support"
		end
	else
		return false, "not-ready"
	end
end

function _M:GetActionSpeedArgs(action,arg1,arg2)
	if action == "up" or action == "down" then
		return arg1 or self.speed_options.tilt
	elseif action == "left" or action == "right" then
		return arg1 or self.speed_options.pan
	elseif action == "left-up" or action == "left-down" or action == "right-up" or action == "right-down" then
		return arg1 or self.speed_options.pan, arg2 or self.speed_options.tilt
	elseif action == "zoom-in" or action == "zoom-out" then
		return arg1 or self.speed_options.zoom
	elseif action == "focus-near" or action == "focus-far" then
		return arg1 or self.speed_options.focus
	elseif action == "iris-up" or action == "iris-down" then
		return arg1 or self.speed_options.iris
	else
		return arg1,arg2
	end
end

function _M:MoveUp(speed)
	return self:DoAction( "up", speed or self.speed_options.tilt )
end

function _M:MoveDown(speed)
	return self:DoAction( "down", speed or self.speed_options.tilt )
end

function _M:MoveLeft(speed)
	return self:DoAction( "left", speed or self.speed_options.pan )
end

function _M:MoveRight(speed)
	return self:DoAction( "right", speed or self.speed_options.pan )
end

function _M:MoveLeftUp(pan_speed, tilt_speed)
	return self:DoAction( "left-up", pan_speed or self.speed_options.pan, tilt_speed or self.speed_options.tilt )
end

function _M:MoveRightUp(pan_speed, tilt_speed)
	return self:DoAction( "right-up", pan_speed or self.speed_options.pan, tilt_speed or self.speed_options.tilt )
end

function _M:MoveLeftDown(pan_speed, tilt_speed)
	return self:DoAction( "left-down", pan_speed or self.speed_options.pan, tilt_speed or self.speed_options.tilt )
end

function _M:MoveRightDown(pan_speed, tilt_speed)
	return self:DoAction( "right-down", pan_speed or self.speed_options.pan, tilt_speed or self.speed_options.tilt )
end

function _M:MoveHome()
	return self:DoAction( "home" )
end

function _M:StopMove()
	return self:DoAction( "stop-move" )
end

function _M:ZoomIn(speed)
	return self:DoAction( "zoom-in", speed or self.speed_options.zoom )
end

function _M:ZoomOut(speed)
	return self:DoAction( "zoom-out", speed or self.speed_options.zoom )
end

function _M:StopZoom()
	return self:DoAction( "stop-zoom" )
end

function _M:FocusNear(speed)
	return self:DoAction( "focus-near", speed or self.speed_options.focus )
end

function _M:FocusFar(speed)
	return self:DoAction( "focus-far", speed or self.speed_options.focus )
end

function _M:StopFocus()
	return self:DoAction( "stop-focus" )
end

function _M:IrisUp(speed)
	return self:DoAction( "iris-up", speed or self.speed_options.iris )
end

function _M:IrisDown(speed)
	return self:DoAction( "iris-down", speed or self.speed_options.iris )
end

function _M:StopIris()
	return self:DoAction( "stop-iris" )
end

function _M:StopAll()
	return self:DoAction( "stop-all" )
end

function _M:SavePreset(id)
	return self:DoAction( "save-preset", id )
end

function _M:RemovePreset(id)
	return self:DoAction( "remove-preset", id )
end

function _M:LoadPreset(id)
	return self:DoAction( "load-preset", id )
end


return _M

--=====================================================--
end
--=====================================================--

return _PTZ

