--[==[
The super-table declaration helpers.
This is a part of Co-Application Framework (CoAF).

(C) KILOVIEW Electronics Co., Ltd.
]==]

local _M = {
}

---------------------------------------

local type_meta_table = {
}

function type_meta_table:__get( cb )
	self["get"] = cb
	return self
end

function type_meta_table:__set( cb )
	self["set"] = cb
	return self
end

function type_meta_table:__changed(cb)
	self["changed"] = cb
	return self
end

function type_meta_table:__nosave(b)
	self["nosave"] = b
	return self
end

function type_meta_table:__range(from, to)
	self["range_from"] = to and from or 1
	self["range_to"] = to and to or from
	return self
end

type_meta_table.__index = type_meta_table

--------------------
-- Type convensions.
--------------------
local convensionFuncs = {
	["INT"] = function(__self,x)
		local num = tonumber(x)
		if num then
			return math.floor(num)
		else 
			return nil
		end
	end,

	["STR"] = function(__self,x)
		if type(x) == "string" then
			return x
		elseif type(x) == "boolean" then
			return x and "true" or "false"
		elseif type(x) == "number" then
			return '' .. x
		else 
			return nil
		end
	end,

	["NUM"] = function(__self,x)
		local num = tonumber(x)
		if num then
			return num
		else 
			return nil
		end
	end,

	["BOOL"] = function(__self,x)
		if type(x) == "boolean" then
			return x
		elseif type(x) == "number" then
			return x ~= 0
		elseif type(x) == "string" then
			x = x:lower()
			if x == "true" or x == "yes" then
				return true
			else
				return false
			end
		else
			return nil
		end
	end
}

-------------------------------------------------------

local function ConstructTypeTable( dataType, default_decl )
	local typeTable = {
		["type"] = dataType,
		["default_decl"] = default_decl
	}
	
	typeTable["convert"] = convensionFuncs[dataType]
	return setmetatable(typeTable, type_meta_table)
end

--[[++++++++++++++++++++++++++++++++++++++++++++++

Simple type declaration:

	x = INT(<default>) : __get( func(...) ) : __set( func(...) ) : __changed( func(...) ) : __nosave(true)
	x = NUM(<default>) : ...
	x = STR(<default>) : ...
	x = BOOL() : ...

	e.g.

	x = INT(100) : __noSave(true) : 
	__get( function(...)
	end) : 
	__changed( function(...)
	end)

Object declaration:

	x = OBJ{
		<members declaration>
	} : __noSave(...) : __changed(...)
)

Array declaration:

	x = ARR( INT/NUM/STR/OBJ/ARR/SET ) : __range(from, to) : ...

Set declaration:

	x = SET( INT/NUM/STR/OBJ/ARR/SET ) : ...

>>> So the properties can be:
	__get( callback-func ) : Property get
	__set( callback-func ) : Property set
	__changed( callback-func ) : Changed
	__nosave( boolean ) : Not save to persitation
	__range( from, to ) : Array range
+++++++++++++++++++++++++++++++++++++++++++++++++]]

-- Declare an interger property.
function _M.INT( decl )
	return ConstructTypeTable("INT", decl)
end

-- Alias: INTEGER = INT
_M.INTEGER = _M.INT

-- Declare a numeric
function _M.NUM( decl )
	return ConstructTypeTable("NUM", decl)
end

-- Alias: NUMBER/FLOAT/DOUBLE = NUM
_M.NUMBER = _M.NUM
_M.FLOAT = _M.NUM
_M.DOUBLE = _M.NUM

-- Declare a string.
function _M.STR( decl )
	return ConstructTypeTable("STR", decl)
end

-- Alias: STRING = STR
_M.STRING = _M.STR

-- Declare a boolean.
function _M.BOOL( decl )
	return ConstructTypeTable( "BOOL", decl )
end

-- Alias: BOOLEAN = BOOL
_M.BOOLEAN = _M.BOOL

-- Declare any type:
function _M.ANY(decl)
	return ConstructTypeTable( "ANY", decl )
end

-- Declare an object
function _M.OBJ( decl )
	return ConstructTypeTable( "OBJ", decl )
end

-- Alias: OBJECT = OBJ
_M.OBJECT = _M.OBJ

-- Declare a set
function _M.SET( decl )
	return ConstructTypeTable( "SET", decl )
end

-- Declare an array
function _M.ARR( decl )
	return ConstructTypeTable( "ARR", decl )
end

-- Alias: ARRAY = ARR
_M.ARRAY = _M.ARR

--==================

return _M
